// created by jay 0.7 (c) 1998 Axel.Schreiner@informatik.uni-osnabrueck.de

#line 2 "cs-parser.jay"
//
// cs-parser.jay: The Parser for the C# compiler
//
// Authors: Miguel de Icaza (miguel@gnome.org)
//          Ravi Pratap     (ravi@ximian.com)
//          Marek Safar	    (marek.safar@gmail.com)
//
// Dual Licensed under the terms of the GNU GPL and the MIT X11 license
//
// (C) 2001 Ximian, Inc (http://www.ximian.com)
// (C) 2004-2011 Novell, Inc
// Copyright 2011-2012 Xamarin Inc.
//

using System.Text;
using System.IO;
using System;
using System.Collections.Generic;

namespace Mono.CSharp
{
	/// <summary>
	///    The C# Parser
	/// </summary>
	public class CSharpParser
	{
		[Flags]
		enum ParameterModifierType
		{
			Ref		= 1 << 1,
			Out		= 1 << 2,
			This	= 1 << 3,
			Params	= 1 << 4,
			Arglist	= 1 << 5,
			DefaultValue = 1 << 6,
			
			All = Ref | Out | This | Params | Arglist | DefaultValue,
			PrimaryConstructor = Ref | Out | Params | DefaultValue
		}
		
		static readonly object ModifierNone = 0;
	
		NamespaceContainer current_namespace;
		TypeContainer current_container;
		TypeDefinition current_type;
		PropertyBase current_property;
		EventProperty current_event;
		EventField current_event_field;
		FieldBase current_field;
	
		/// <summary>
		///   Current block is used to add statements as we find
		///   them.  
		/// </summary>
		Block      current_block;
		
		BlockVariable current_variable;

		Delegate   current_delegate;
		
		AnonymousMethodExpression current_anonymous_method;

		/// <summary>
		///   This is used by the unary_expression code to resolve
		///   a name against a parameter.  
		/// </summary>
		
		// FIXME: This is very ugly and it's very hard to reset it correctly
		// on all places, especially when some parameters are autogenerated.
		ParametersCompiled current_local_parameters;

		bool parsing_anonymous_method;
		
		bool async_block;

		///
		/// An out-of-band stack.
		///
		Stack<object> oob_stack;

		///
		/// Controls the verbosity of the errors produced by the parser
		///
		int yacc_verbose_flag;

		/// 
		/// Used by the interactive shell, flags whether EOF was reached
		/// and an error was produced
		///
		public bool UnexpectedEOF;

		///
		/// The current file.
		///
		readonly CompilationSourceFile file;

		///
		/// Temporary Xml documentation cache.
		/// For enum types, we need one more temporary store.
		///
		string tmpComment;
		string enumTypeComment;
	       		
		/// Current attribute target
		string current_attr_target;
		
		ParameterModifierType valid_param_mod;
		
		bool default_parameter_used;

		/// When using the interactive parser, this holds the
		/// resulting expression
		public Class InteractiveResult;

		//
		// Keeps track of global data changes to undo on parser error
		//
		public Undo undo;

		bool? interactive_async;
		
		Stack<Linq.QueryBlock> linq_clause_blocks;

		ModuleContainer module;
		
		readonly CompilerContext compiler;
		readonly LanguageVersion lang_version;
		readonly bool doc_support;
		readonly CompilerSettings settings;
		readonly Report report;
		
		//
		// Instead of allocating carrier array everytime we
		// share the bucket for very common constructs which can never
		// be recursive
		//
		List<Parameter> parameters_bucket;
		
		//
		// Full AST support members
		//
		LocationsBag lbag;
		List<Tuple<Modifiers, Location>> mod_locations;
		Stack<Location> location_stack;
#line default

  /** error output stream.
      It should be changeable.
    */
  public System.IO.TextWriter ErrorOutput = System.Console.Out;

  /** simplified error message.
      @see <a href="#yyerror(java.lang.String, java.lang.String[])">yyerror</a>
    */
  public void yyerror (string message) {
    yyerror(message, null);
  }
#pragma warning disable 649
  /* An EOF token */
  public int eof_token;
#pragma warning restore 649
  /** (syntax) error message.
      Can be overwritten to control message format.
      @param message text to be displayed.
      @param expected vector of acceptable tokens, if available.
    */
  public void yyerror (string message, string[] expected) {
    if ((yacc_verbose_flag > 0) && (expected != null) && (expected.Length  > 0)) {
      ErrorOutput.Write (message+", expecting");
      for (int n = 0; n < expected.Length; ++ n)
        ErrorOutput.Write (" "+expected[n]);
        ErrorOutput.WriteLine ();
    } else
      ErrorOutput.WriteLine (message);
  }

  /** debugging support, requires the package jay.yydebug.
      Set to null to suppress debugging messages.
    */
//t  internal yydebug.yyDebug debug;

  protected const int yyFinal = 7;
//t // Put this array into a separate class so it is only initialized if debugging is actually used
//t // Use MarshalByRefObject to disable inlining
//t class YYRules : MarshalByRefObject {
//t  public static readonly string [] yyRule = {
//t    "$accept : compilation_unit",
//t    "compilation_unit : outer_declaration opt_EOF",
//t    "$$1 :",
//t    "compilation_unit : interactive_parsing $$1 opt_EOF",
//t    "compilation_unit : documentation_parsing",
//t    "outer_declaration : opt_extern_alias_directives opt_using_directives",
//t    "outer_declaration : opt_extern_alias_directives opt_using_directives namespace_or_type_declarations opt_attributes",
//t    "outer_declaration : opt_extern_alias_directives opt_using_directives attribute_sections",
//t    "outer_declaration : error",
//t    "opt_EOF :",
//t    "opt_EOF : EOF",
//t    "extern_alias_directives : extern_alias_directive",
//t    "extern_alias_directives : extern_alias_directives extern_alias_directive",
//t    "extern_alias_directive : EXTERN_ALIAS IDENTIFIER IDENTIFIER SEMICOLON",
//t    "extern_alias_directive : EXTERN_ALIAS error",
//t    "using_directives : using_directive",
//t    "using_directives : using_directives using_directive",
//t    "using_directive : using_namespace",
//t    "using_namespace : USING namespace_or_type_expr SEMICOLON",
//t    "using_namespace : USING IDENTIFIER ASSIGN namespace_or_type_expr SEMICOLON",
//t    "using_namespace : USING error",
//t    "$$2 :",
//t    "$$3 :",
//t    "namespace_declaration : opt_attributes NAMESPACE namespace_name $$2 OPEN_BRACE $$3 opt_extern_alias_directives opt_using_directives opt_namespace_or_type_declarations CLOSE_BRACE opt_semicolon_error",
//t    "namespace_declaration : opt_attributes NAMESPACE namespace_name",
//t    "opt_semicolon_error :",
//t    "opt_semicolon_error : SEMICOLON",
//t    "opt_semicolon_error : error",
//t    "namespace_name : IDENTIFIER",
//t    "namespace_name : namespace_name DOT IDENTIFIER",
//t    "namespace_name : error",
//t    "opt_semicolon :",
//t    "opt_semicolon : SEMICOLON",
//t    "opt_comma :",
//t    "opt_comma : COMMA",
//t    "opt_using_directives :",
//t    "opt_using_directives : using_directives",
//t    "opt_extern_alias_directives :",
//t    "opt_extern_alias_directives : extern_alias_directives",
//t    "opt_namespace_or_type_declarations :",
//t    "opt_namespace_or_type_declarations : namespace_or_type_declarations",
//t    "namespace_or_type_declarations : namespace_or_type_declaration",
//t    "namespace_or_type_declarations : namespace_or_type_declarations namespace_or_type_declaration",
//t    "namespace_or_type_declaration : type_declaration",
//t    "namespace_or_type_declaration : namespace_declaration",
//t    "namespace_or_type_declaration : attribute_sections CLOSE_BRACE",
//t    "type_declaration : class_declaration",
//t    "type_declaration : struct_declaration",
//t    "type_declaration : interface_declaration",
//t    "type_declaration : enum_declaration",
//t    "type_declaration : delegate_declaration",
//t    "opt_attributes :",
//t    "opt_attributes : attribute_sections",
//t    "attribute_sections : attribute_section",
//t    "attribute_sections : attribute_sections attribute_section",
//t    "$$4 :",
//t    "attribute_section : OPEN_BRACKET $$4 attribute_section_cont",
//t    "$$5 :",
//t    "attribute_section_cont : attribute_target COLON $$5 attribute_list opt_comma CLOSE_BRACKET",
//t    "attribute_section_cont : attribute_list opt_comma CLOSE_BRACKET",
//t    "attribute_section_cont : IDENTIFIER error",
//t    "attribute_section_cont : error",
//t    "attribute_target : IDENTIFIER",
//t    "attribute_target : EVENT",
//t    "attribute_target : RETURN",
//t    "attribute_list : attribute",
//t    "attribute_list : attribute_list COMMA attribute",
//t    "$$6 :",
//t    "attribute : attribute_name $$6 opt_attribute_arguments",
//t    "attribute_name : namespace_or_type_expr",
//t    "opt_attribute_arguments :",
//t    "opt_attribute_arguments : OPEN_PARENS attribute_arguments CLOSE_PARENS",
//t    "attribute_arguments :",
//t    "attribute_arguments : positional_or_named_argument",
//t    "attribute_arguments : named_attribute_argument",
//t    "attribute_arguments : attribute_arguments COMMA positional_or_named_argument",
//t    "attribute_arguments : attribute_arguments COMMA named_attribute_argument",
//t    "positional_or_named_argument : expression",
//t    "positional_or_named_argument : named_argument",
//t    "positional_or_named_argument : error",
//t    "$$7 :",
//t    "named_attribute_argument : IDENTIFIER ASSIGN $$7 expression",
//t    "named_argument : identifier_inside_body COLON opt_named_modifier named_argument_expr",
//t    "named_argument_expr : expression_or_error",
//t    "opt_named_modifier :",
//t    "opt_named_modifier : REF",
//t    "opt_named_modifier : OUT",
//t    "opt_class_member_declarations :",
//t    "opt_class_member_declarations : class_member_declarations",
//t    "class_member_declarations : class_member_declaration",
//t    "class_member_declarations : class_member_declarations class_member_declaration",
//t    "class_member_declaration : constant_declaration",
//t    "class_member_declaration : field_declaration",
//t    "class_member_declaration : method_declaration",
//t    "class_member_declaration : property_declaration",
//t    "class_member_declaration : event_declaration",
//t    "class_member_declaration : indexer_declaration",
//t    "class_member_declaration : operator_declaration",
//t    "class_member_declaration : constructor_declaration",
//t    "class_member_declaration : primary_constructor_body",
//t    "class_member_declaration : destructor_declaration",
//t    "class_member_declaration : type_declaration",
//t    "class_member_declaration : attributes_without_members",
//t    "class_member_declaration : incomplete_member",
//t    "class_member_declaration : error",
//t    "$$8 :",
//t    "primary_constructor_body : OPEN_BRACE $$8 opt_statement_list block_end",
//t    "$$9 :",
//t    "$$10 :",
//t    "$$11 :",
//t    "$$12 :",
//t    "$$13 :",
//t    "struct_declaration : opt_attributes opt_modifiers opt_partial STRUCT $$9 type_declaration_name $$10 opt_primary_parameters opt_class_base opt_type_parameter_constraints_clauses $$11 OPEN_BRACE $$12 opt_class_member_declarations CLOSE_BRACE $$13 opt_semicolon",
//t    "struct_declaration : opt_attributes opt_modifiers opt_partial STRUCT error",
//t    "$$14 :",
//t    "constant_declaration : opt_attributes opt_modifiers CONST type IDENTIFIER $$14 constant_initializer opt_constant_declarators SEMICOLON",
//t    "constant_declaration : opt_attributes opt_modifiers CONST type error",
//t    "opt_constant_declarators :",
//t    "opt_constant_declarators : constant_declarators",
//t    "constant_declarators : constant_declarator",
//t    "constant_declarators : constant_declarators constant_declarator",
//t    "constant_declarator : COMMA IDENTIFIER constant_initializer",
//t    "$$15 :",
//t    "constant_initializer : ASSIGN $$15 constant_initializer_expr",
//t    "constant_initializer : error",
//t    "constant_initializer_expr : constant_expression",
//t    "constant_initializer_expr : array_initializer",
//t    "$$16 :",
//t    "field_declaration : opt_attributes opt_modifiers member_type IDENTIFIER $$16 opt_field_initializer opt_field_declarators SEMICOLON",
//t    "$$17 :",
//t    "field_declaration : opt_attributes opt_modifiers FIXED simple_type IDENTIFIER $$17 fixed_field_size opt_fixed_field_declarators SEMICOLON",
//t    "field_declaration : opt_attributes opt_modifiers FIXED simple_type error SEMICOLON",
//t    "opt_field_initializer :",
//t    "$$18 :",
//t    "opt_field_initializer : ASSIGN $$18 variable_initializer",
//t    "opt_field_declarators :",
//t    "opt_field_declarators : field_declarators",
//t    "field_declarators : field_declarator",
//t    "field_declarators : field_declarators field_declarator",
//t    "field_declarator : COMMA IDENTIFIER",
//t    "$$19 :",
//t    "field_declarator : COMMA IDENTIFIER ASSIGN $$19 variable_initializer",
//t    "opt_fixed_field_declarators :",
//t    "opt_fixed_field_declarators : fixed_field_declarators",
//t    "fixed_field_declarators : fixed_field_declarator",
//t    "fixed_field_declarators : fixed_field_declarators fixed_field_declarator",
//t    "fixed_field_declarator : COMMA IDENTIFIER fixed_field_size",
//t    "$$20 :",
//t    "fixed_field_size : OPEN_BRACKET $$20 expression CLOSE_BRACKET",
//t    "fixed_field_size : OPEN_BRACKET error",
//t    "variable_initializer : expression",
//t    "variable_initializer : array_initializer",
//t    "variable_initializer : error",
//t    "$$21 :",
//t    "method_declaration : method_header $$21 method_body_expression_block",
//t    "$$22 :",
//t    "$$23 :",
//t    "method_header : opt_attributes opt_modifiers member_type method_declaration_name OPEN_PARENS $$22 opt_formal_parameter_list CLOSE_PARENS $$23 opt_type_parameter_constraints_clauses",
//t    "$$24 :",
//t    "$$25 :",
//t    "$$26 :",
//t    "method_header : opt_attributes opt_modifiers PARTIAL VOID $$24 method_declaration_name OPEN_PARENS $$25 opt_formal_parameter_list CLOSE_PARENS $$26 opt_type_parameter_constraints_clauses",
//t    "method_header : opt_attributes opt_modifiers member_type modifiers method_declaration_name OPEN_PARENS opt_formal_parameter_list CLOSE_PARENS",
//t    "method_header : opt_attributes opt_modifiers member_type method_declaration_name error",
//t    "method_body_expression_block : method_body",
//t    "method_body_expression_block : expression_block",
//t    "method_body : block",
//t    "method_body : SEMICOLON",
//t    "$$27 :",
//t    "expression_block : ARROW $$27 expression SEMICOLON",
//t    "opt_formal_parameter_list :",
//t    "opt_formal_parameter_list : formal_parameter_list",
//t    "formal_parameter_list : fixed_parameters",
//t    "formal_parameter_list : fixed_parameters COMMA parameter_array",
//t    "formal_parameter_list : fixed_parameters COMMA arglist_modifier",
//t    "formal_parameter_list : parameter_array COMMA error",
//t    "formal_parameter_list : fixed_parameters COMMA parameter_array COMMA error",
//t    "formal_parameter_list : arglist_modifier COMMA error",
//t    "formal_parameter_list : fixed_parameters COMMA ARGLIST COMMA error",
//t    "formal_parameter_list : parameter_array",
//t    "formal_parameter_list : arglist_modifier",
//t    "formal_parameter_list : error",
//t    "fixed_parameters : fixed_parameter",
//t    "fixed_parameters : fixed_parameters COMMA fixed_parameter",
//t    "fixed_parameter : opt_attributes opt_parameter_modifier parameter_type identifier_inside_body",
//t    "fixed_parameter : opt_attributes opt_parameter_modifier parameter_type identifier_inside_body OPEN_BRACKET CLOSE_BRACKET",
//t    "fixed_parameter : attribute_sections error",
//t    "fixed_parameter : opt_attributes opt_parameter_modifier parameter_type error",
//t    "$$28 :",
//t    "fixed_parameter : opt_attributes opt_parameter_modifier parameter_type identifier_inside_body ASSIGN $$28 constant_expression",
//t    "opt_parameter_modifier :",
//t    "opt_parameter_modifier : parameter_modifiers",
//t    "parameter_modifiers : parameter_modifier",
//t    "parameter_modifiers : parameter_modifiers parameter_modifier",
//t    "parameter_modifier : REF",
//t    "parameter_modifier : OUT",
//t    "parameter_modifier : THIS",
//t    "parameter_array : opt_attributes params_modifier type IDENTIFIER",
//t    "parameter_array : opt_attributes params_modifier type IDENTIFIER ASSIGN constant_expression",
//t    "parameter_array : opt_attributes params_modifier type error",
//t    "params_modifier : PARAMS",
//t    "params_modifier : PARAMS parameter_modifier",
//t    "params_modifier : PARAMS params_modifier",
//t    "arglist_modifier : ARGLIST",
//t    "$$29 :",
//t    "$$30 :",
//t    "$$31 :",
//t    "$$32 :",
//t    "property_declaration : opt_attributes opt_modifiers member_type member_declaration_name $$29 OPEN_BRACE $$30 accessor_declarations $$31 CLOSE_BRACE $$32 opt_property_initializer",
//t    "$$33 :",
//t    "property_declaration : opt_attributes opt_modifiers member_type member_declaration_name $$33 expression_block",
//t    "opt_property_initializer :",
//t    "$$34 :",
//t    "opt_property_initializer : ASSIGN $$34 property_initializer SEMICOLON",
//t    "property_initializer : expression",
//t    "property_initializer : array_initializer",
//t    "$$35 :",
//t    "$$36 :",
//t    "indexer_declaration : opt_attributes opt_modifiers member_type indexer_declaration_name OPEN_BRACKET $$35 opt_formal_parameter_list CLOSE_BRACKET $$36 indexer_body",
//t    "indexer_body : OPEN_BRACE accessor_declarations CLOSE_BRACE",
//t    "indexer_body : expression_block",
//t    "accessor_declarations : get_accessor_declaration",
//t    "accessor_declarations : get_accessor_declaration accessor_declarations",
//t    "accessor_declarations : set_accessor_declaration",
//t    "accessor_declarations : set_accessor_declaration accessor_declarations",
//t    "accessor_declarations : error",
//t    "$$37 :",
//t    "get_accessor_declaration : opt_attributes opt_modifiers GET $$37 accessor_body",
//t    "$$38 :",
//t    "set_accessor_declaration : opt_attributes opt_modifiers SET $$38 accessor_body",
//t    "accessor_body : block",
//t    "accessor_body : SEMICOLON",
//t    "accessor_body : error",
//t    "$$39 :",
//t    "$$40 :",
//t    "$$41 :",
//t    "$$42 :",
//t    "interface_declaration : opt_attributes opt_modifiers opt_partial INTERFACE $$39 type_declaration_name $$40 opt_class_base opt_type_parameter_constraints_clauses $$41 OPEN_BRACE opt_interface_member_declarations CLOSE_BRACE $$42 opt_semicolon",
//t    "interface_declaration : opt_attributes opt_modifiers opt_partial INTERFACE error",
//t    "opt_interface_member_declarations :",
//t    "opt_interface_member_declarations : interface_member_declarations",
//t    "interface_member_declarations : interface_member_declaration",
//t    "interface_member_declarations : interface_member_declarations interface_member_declaration",
//t    "interface_member_declaration : constant_declaration",
//t    "interface_member_declaration : field_declaration",
//t    "interface_member_declaration : method_declaration",
//t    "interface_member_declaration : property_declaration",
//t    "interface_member_declaration : event_declaration",
//t    "interface_member_declaration : indexer_declaration",
//t    "interface_member_declaration : operator_declaration",
//t    "interface_member_declaration : constructor_declaration",
//t    "interface_member_declaration : type_declaration",
//t    "$$43 :",
//t    "operator_declaration : opt_attributes opt_modifiers operator_declarator $$43 method_body_expression_block",
//t    "operator_type : type_expression_or_array",
//t    "operator_type : VOID",
//t    "$$44 :",
//t    "operator_declarator : operator_type OPERATOR overloadable_operator OPEN_PARENS $$44 opt_formal_parameter_list CLOSE_PARENS",
//t    "operator_declarator : conversion_operator_declarator",
//t    "overloadable_operator : BANG",
//t    "overloadable_operator : TILDE",
//t    "overloadable_operator : OP_INC",
//t    "overloadable_operator : OP_DEC",
//t    "overloadable_operator : TRUE",
//t    "overloadable_operator : FALSE",
//t    "overloadable_operator : PLUS",
//t    "overloadable_operator : MINUS",
//t    "overloadable_operator : STAR",
//t    "overloadable_operator : DIV",
//t    "overloadable_operator : PERCENT",
//t    "overloadable_operator : BITWISE_AND",
//t    "overloadable_operator : BITWISE_OR",
//t    "overloadable_operator : CARRET",
//t    "overloadable_operator : OP_SHIFT_LEFT",
//t    "overloadable_operator : OP_SHIFT_RIGHT",
//t    "overloadable_operator : OP_EQ",
//t    "overloadable_operator : OP_NE",
//t    "overloadable_operator : OP_GT",
//t    "overloadable_operator : OP_LT",
//t    "overloadable_operator : OP_GE",
//t    "overloadable_operator : OP_LE",
//t    "overloadable_operator : IS",
//t    "$$45 :",
//t    "conversion_operator_declarator : IMPLICIT OPERATOR type OPEN_PARENS $$45 opt_formal_parameter_list CLOSE_PARENS",
//t    "$$46 :",
//t    "conversion_operator_declarator : EXPLICIT OPERATOR type OPEN_PARENS $$46 opt_formal_parameter_list CLOSE_PARENS",
//t    "conversion_operator_declarator : IMPLICIT error",
//t    "conversion_operator_declarator : EXPLICIT error",
//t    "constructor_declaration : constructor_declarator constructor_body",
//t    "$$47 :",
//t    "$$48 :",
//t    "constructor_declarator : opt_attributes opt_modifiers IDENTIFIER $$47 OPEN_PARENS opt_formal_parameter_list CLOSE_PARENS $$48 opt_constructor_initializer",
//t    "constructor_body : block_prepared",
//t    "constructor_body : SEMICOLON",
//t    "opt_constructor_initializer :",
//t    "opt_constructor_initializer : constructor_initializer",
//t    "$$49 :",
//t    "constructor_initializer : COLON BASE OPEN_PARENS $$49 opt_argument_list CLOSE_PARENS",
//t    "$$50 :",
//t    "constructor_initializer : COLON THIS OPEN_PARENS $$50 opt_argument_list CLOSE_PARENS",
//t    "constructor_initializer : COLON error",
//t    "constructor_initializer : error",
//t    "$$51 :",
//t    "destructor_declaration : opt_attributes opt_modifiers TILDE $$51 IDENTIFIER OPEN_PARENS CLOSE_PARENS method_body",
//t    "$$52 :",
//t    "event_declaration : opt_attributes opt_modifiers EVENT type member_declaration_name $$52 opt_event_initializer opt_event_declarators SEMICOLON",
//t    "$$53 :",
//t    "$$54 :",
//t    "event_declaration : opt_attributes opt_modifiers EVENT type member_declaration_name OPEN_BRACE $$53 event_accessor_declarations $$54 CLOSE_BRACE",
//t    "event_declaration : opt_attributes opt_modifiers EVENT type error",
//t    "opt_event_initializer :",
//t    "$$55 :",
//t    "opt_event_initializer : ASSIGN $$55 event_variable_initializer",
//t    "opt_event_declarators :",
//t    "opt_event_declarators : event_declarators",
//t    "event_declarators : event_declarator",
//t    "event_declarators : event_declarators event_declarator",
//t    "event_declarator : COMMA IDENTIFIER",
//t    "$$56 :",
//t    "event_declarator : COMMA IDENTIFIER ASSIGN $$56 event_variable_initializer",
//t    "$$57 :",
//t    "event_variable_initializer : $$57 variable_initializer",
//t    "event_accessor_declarations : add_accessor_declaration remove_accessor_declaration",
//t    "event_accessor_declarations : remove_accessor_declaration add_accessor_declaration",
//t    "event_accessor_declarations : add_accessor_declaration",
//t    "event_accessor_declarations : remove_accessor_declaration",
//t    "event_accessor_declarations : error",
//t    "$$58 :",
//t    "add_accessor_declaration : opt_attributes opt_modifiers ADD $$58 event_accessor_block",
//t    "$$59 :",
//t    "remove_accessor_declaration : opt_attributes opt_modifiers REMOVE $$59 event_accessor_block",
//t    "event_accessor_block : opt_semicolon",
//t    "event_accessor_block : block",
//t    "attributes_without_members : attribute_sections CLOSE_BRACE",
//t    "incomplete_member : opt_attributes opt_modifiers member_type CLOSE_BRACE",
//t    "$$60 :",
//t    "$$61 :",
//t    "$$62 :",
//t    "enum_declaration : opt_attributes opt_modifiers ENUM type_declaration_name opt_enum_base $$60 OPEN_BRACE $$61 opt_enum_member_declarations $$62 CLOSE_BRACE opt_semicolon",
//t    "opt_enum_base :",
//t    "opt_enum_base : COLON type",
//t    "opt_enum_base : COLON error",
//t    "opt_enum_member_declarations :",
//t    "opt_enum_member_declarations : enum_member_declarations",
//t    "opt_enum_member_declarations : enum_member_declarations COMMA",
//t    "enum_member_declarations : enum_member_declaration",
//t    "enum_member_declarations : enum_member_declarations COMMA enum_member_declaration",
//t    "enum_member_declaration : opt_attributes IDENTIFIER",
//t    "$$63 :",
//t    "enum_member_declaration : opt_attributes IDENTIFIER $$63 ASSIGN constant_expression",
//t    "enum_member_declaration : opt_attributes IDENTIFIER error",
//t    "enum_member_declaration : attributes_without_members",
//t    "$$64 :",
//t    "$$65 :",
//t    "$$66 :",
//t    "delegate_declaration : opt_attributes opt_modifiers DELEGATE member_type type_declaration_name OPEN_PARENS $$64 opt_formal_parameter_list CLOSE_PARENS $$65 opt_type_parameter_constraints_clauses $$66 SEMICOLON",
//t    "opt_nullable :",
//t    "opt_nullable : INTERR_NULLABLE",
//t    "namespace_or_type_expr : member_name",
//t    "namespace_or_type_expr : qualified_alias_member IDENTIFIER opt_type_argument_list",
//t    "namespace_or_type_expr : qualified_alias_member IDENTIFIER generic_dimension",
//t    "member_name : simple_name_expr",
//t    "member_name : namespace_or_type_expr DOT IDENTIFIER opt_type_argument_list",
//t    "member_name : namespace_or_type_expr DOT IDENTIFIER generic_dimension",
//t    "simple_name_expr : IDENTIFIER opt_type_argument_list",
//t    "simple_name_expr : IDENTIFIER generic_dimension",
//t    "opt_type_argument_list :",
//t    "opt_type_argument_list : OP_GENERICS_LT type_arguments OP_GENERICS_GT",
//t    "opt_type_argument_list : OP_GENERICS_LT error",
//t    "type_arguments : type",
//t    "type_arguments : type_arguments COMMA type",
//t    "$$67 :",
//t    "type_declaration_name : IDENTIFIER $$67 opt_type_parameter_list",
//t    "member_declaration_name : method_declaration_name",
//t    "method_declaration_name : type_declaration_name",
//t    "method_declaration_name : explicit_interface IDENTIFIER opt_type_parameter_list",
//t    "indexer_declaration_name : THIS",
//t    "indexer_declaration_name : explicit_interface THIS",
//t    "explicit_interface : IDENTIFIER opt_type_argument_list DOT",
//t    "explicit_interface : qualified_alias_member IDENTIFIER opt_type_argument_list DOT",
//t    "explicit_interface : explicit_interface IDENTIFIER opt_type_argument_list DOT",
//t    "opt_type_parameter_list :",
//t    "opt_type_parameter_list : OP_GENERICS_LT_DECL type_parameters OP_GENERICS_GT",
//t    "type_parameters : type_parameter",
//t    "type_parameters : type_parameters COMMA type_parameter",
//t    "type_parameter : opt_attributes opt_type_parameter_variance IDENTIFIER",
//t    "type_parameter : error",
//t    "type_and_void : type_expression_or_array",
//t    "type_and_void : VOID",
//t    "member_type : type_and_void",
//t    "type : type_expression_or_array",
//t    "type : void_invalid",
//t    "simple_type : type_expression",
//t    "simple_type : void_invalid",
//t    "parameter_type : type_expression_or_array",
//t    "parameter_type : VOID",
//t    "type_expression_or_array : type_expression",
//t    "type_expression_or_array : type_expression rank_specifiers",
//t    "type_expression : namespace_or_type_expr opt_nullable",
//t    "type_expression : namespace_or_type_expr pointer_stars",
//t    "type_expression : builtin_type_expression",
//t    "void_invalid : VOID",
//t    "builtin_type_expression : builtin_types opt_nullable",
//t    "builtin_type_expression : builtin_types pointer_stars",
//t    "builtin_type_expression : VOID pointer_stars",
//t    "type_list : base_type_name",
//t    "type_list : type_list COMMA base_type_name",
//t    "base_type_name : type",
//t    "builtin_types : OBJECT",
//t    "builtin_types : STRING",
//t    "builtin_types : BOOL",
//t    "builtin_types : DECIMAL",
//t    "builtin_types : FLOAT",
//t    "builtin_types : DOUBLE",
//t    "builtin_types : integral_type",
//t    "integral_type : SBYTE",
//t    "integral_type : BYTE",
//t    "integral_type : SHORT",
//t    "integral_type : USHORT",
//t    "integral_type : INT",
//t    "integral_type : UINT",
//t    "integral_type : LONG",
//t    "integral_type : ULONG",
//t    "integral_type : CHAR",
//t    "primary_expression : type_name_expression",
//t    "primary_expression : literal",
//t    "primary_expression : array_creation_expression",
//t    "primary_expression : parenthesized_expression",
//t    "primary_expression : default_value_expression",
//t    "primary_expression : invocation_expression",
//t    "primary_expression : element_access",
//t    "primary_expression : this_access",
//t    "primary_expression : base_access",
//t    "primary_expression : post_increment_expression",
//t    "primary_expression : post_decrement_expression",
//t    "primary_expression : object_or_delegate_creation_expression",
//t    "primary_expression : anonymous_type_expression",
//t    "primary_expression : typeof_expression",
//t    "primary_expression : sizeof_expression",
//t    "primary_expression : checked_expression",
//t    "primary_expression : unchecked_expression",
//t    "primary_expression : pointer_member_access",
//t    "primary_expression : anonymous_method_expression",
//t    "primary_expression : undocumented_expressions",
//t    "type_name_expression : simple_name_expr",
//t    "type_name_expression : IDENTIFIER GENERATE_COMPLETION",
//t    "type_name_expression : member_access",
//t    "literal : boolean_literal",
//t    "literal : LITERAL",
//t    "literal : NULL",
//t    "boolean_literal : TRUE",
//t    "boolean_literal : FALSE",
//t    "open_parens_any : OPEN_PARENS",
//t    "open_parens_any : OPEN_PARENS_CAST",
//t    "close_parens : CLOSE_PARENS",
//t    "close_parens : COMPLETE_COMPLETION",
//t    "parenthesized_expression : OPEN_PARENS expression CLOSE_PARENS",
//t    "parenthesized_expression : OPEN_PARENS expression COMPLETE_COMPLETION",
//t    "member_access : primary_expression DOT identifier_inside_body opt_type_argument_list",
//t    "member_access : primary_expression DOT identifier_inside_body generic_dimension",
//t    "member_access : primary_expression INTERR_OPERATOR DOT identifier_inside_body opt_type_argument_list",
//t    "member_access : builtin_types DOT identifier_inside_body opt_type_argument_list",
//t    "member_access : BASE DOT identifier_inside_body opt_type_argument_list",
//t    "member_access : AWAIT DOT identifier_inside_body opt_type_argument_list",
//t    "member_access : qualified_alias_member identifier_inside_body opt_type_argument_list",
//t    "member_access : qualified_alias_member identifier_inside_body generic_dimension",
//t    "member_access : primary_expression DOT GENERATE_COMPLETION",
//t    "member_access : primary_expression DOT IDENTIFIER GENERATE_COMPLETION",
//t    "member_access : builtin_types DOT GENERATE_COMPLETION",
//t    "member_access : builtin_types DOT IDENTIFIER GENERATE_COMPLETION",
//t    "invocation_expression : primary_expression open_parens_any opt_argument_list close_parens",
//t    "invocation_expression : primary_expression open_parens_any argument_list error",
//t    "invocation_expression : primary_expression open_parens_any error",
//t    "opt_object_or_collection_initializer :",
//t    "opt_object_or_collection_initializer : object_or_collection_initializer",
//t    "object_or_collection_initializer : OPEN_BRACE opt_member_initializer_list close_brace_or_complete_completion",
//t    "object_or_collection_initializer : OPEN_BRACE member_initializer_list COMMA CLOSE_BRACE",
//t    "opt_member_initializer_list :",
//t    "opt_member_initializer_list : member_initializer_list",
//t    "member_initializer_list : member_initializer",
//t    "member_initializer_list : member_initializer_list COMMA member_initializer",
//t    "member_initializer_list : member_initializer_list error",
//t    "member_initializer : IDENTIFIER ASSIGN initializer_value",
//t    "member_initializer : AWAIT ASSIGN initializer_value",
//t    "member_initializer : GENERATE_COMPLETION",
//t    "member_initializer : non_assignment_expression opt_COMPLETE_COMPLETION",
//t    "member_initializer : OPEN_BRACE expression_list CLOSE_BRACE",
//t    "member_initializer : OPEN_BRACKET_EXPR expression_list CLOSE_BRACKET ASSIGN initializer_value",
//t    "member_initializer : OPEN_BRACE CLOSE_BRACE",
//t    "initializer_value : expression",
//t    "initializer_value : object_or_collection_initializer",
//t    "opt_argument_list :",
//t    "opt_argument_list : argument_list",
//t    "argument_list : argument_or_named_argument",
//t    "argument_list : argument_list COMMA argument",
//t    "argument_list : argument_list COMMA named_argument",
//t    "argument_list : argument_list COMMA error",
//t    "argument_list : COMMA error",
//t    "argument : expression",
//t    "argument : non_simple_argument",
//t    "argument_or_named_argument : argument",
//t    "argument_or_named_argument : named_argument",
//t    "non_simple_argument : REF variable_reference",
//t    "non_simple_argument : REF declaration_expression",
//t    "non_simple_argument : OUT variable_reference",
//t    "non_simple_argument : OUT declaration_expression",
//t    "non_simple_argument : ARGLIST OPEN_PARENS argument_list CLOSE_PARENS",
//t    "non_simple_argument : ARGLIST OPEN_PARENS CLOSE_PARENS",
//t    "declaration_expression : OPEN_PARENS declaration_expression CLOSE_PARENS",
//t    "declaration_expression : variable_type identifier_inside_body",
//t    "declaration_expression : variable_type identifier_inside_body ASSIGN expression",
//t    "variable_reference : expression",
//t    "element_access : primary_expression OPEN_BRACKET_EXPR expression_list_arguments CLOSE_BRACKET",
//t    "element_access : primary_expression INTERR_OPERATOR OPEN_BRACKET_EXPR expression_list_arguments CLOSE_BRACKET",
//t    "element_access : primary_expression OPEN_BRACKET_EXPR expression_list_arguments error",
//t    "element_access : primary_expression OPEN_BRACKET_EXPR error",
//t    "expression_list : expression_or_error",
//t    "expression_list : expression_list COMMA expression_or_error",
//t    "expression_list_arguments : expression_list_argument",
//t    "expression_list_arguments : expression_list_arguments COMMA expression_list_argument",
//t    "expression_list_argument : expression",
//t    "expression_list_argument : named_argument",
//t    "this_access : THIS",
//t    "base_access : BASE OPEN_BRACKET_EXPR expression_list_arguments CLOSE_BRACKET",
//t    "base_access : BASE OPEN_BRACKET error",
//t    "post_increment_expression : primary_expression OP_INC",
//t    "post_decrement_expression : primary_expression OP_DEC",
//t    "object_or_delegate_creation_expression : NEW new_expr_type open_parens_any opt_argument_list CLOSE_PARENS opt_object_or_collection_initializer",
//t    "object_or_delegate_creation_expression : NEW new_expr_type object_or_collection_initializer",
//t    "array_creation_expression : NEW new_expr_type OPEN_BRACKET_EXPR expression_list CLOSE_BRACKET opt_rank_specifier opt_array_initializer",
//t    "array_creation_expression : NEW new_expr_type rank_specifiers opt_array_initializer",
//t    "array_creation_expression : NEW rank_specifier array_initializer",
//t    "array_creation_expression : NEW new_expr_type OPEN_BRACKET CLOSE_BRACKET OPEN_BRACKET_EXPR error CLOSE_BRACKET",
//t    "array_creation_expression : NEW new_expr_type error",
//t    "$$68 :",
//t    "new_expr_type : $$68 simple_type",
//t    "anonymous_type_expression : NEW OPEN_BRACE anonymous_type_parameters_opt_comma CLOSE_BRACE",
//t    "anonymous_type_expression : NEW OPEN_BRACE GENERATE_COMPLETION",
//t    "anonymous_type_parameters_opt_comma : anonymous_type_parameters_opt",
//t    "anonymous_type_parameters_opt_comma : anonymous_type_parameters COMMA",
//t    "anonymous_type_parameters_opt :",
//t    "anonymous_type_parameters_opt : anonymous_type_parameters",
//t    "anonymous_type_parameters : anonymous_type_parameter",
//t    "anonymous_type_parameters : anonymous_type_parameters COMMA anonymous_type_parameter",
//t    "anonymous_type_parameters : COMPLETE_COMPLETION",
//t    "anonymous_type_parameters : anonymous_type_parameter COMPLETE_COMPLETION",
//t    "anonymous_type_parameter : identifier_inside_body ASSIGN variable_initializer",
//t    "anonymous_type_parameter : identifier_inside_body",
//t    "anonymous_type_parameter : member_access",
//t    "anonymous_type_parameter : error",
//t    "opt_rank_specifier :",
//t    "opt_rank_specifier : rank_specifiers",
//t    "rank_specifiers : rank_specifier",
//t    "rank_specifiers : rank_specifier rank_specifiers",
//t    "rank_specifier : OPEN_BRACKET CLOSE_BRACKET",
//t    "rank_specifier : OPEN_BRACKET dim_separators CLOSE_BRACKET",
//t    "dim_separators : COMMA",
//t    "dim_separators : dim_separators COMMA",
//t    "opt_array_initializer :",
//t    "opt_array_initializer : array_initializer",
//t    "array_initializer : OPEN_BRACE CLOSE_BRACE",
//t    "array_initializer : OPEN_BRACE variable_initializer_list opt_comma CLOSE_BRACE",
//t    "variable_initializer_list : variable_initializer",
//t    "variable_initializer_list : variable_initializer_list COMMA variable_initializer",
//t    "typeof_expression : TYPEOF open_parens_any typeof_type_expression CLOSE_PARENS",
//t    "typeof_type_expression : type_and_void",
//t    "typeof_type_expression : error",
//t    "generic_dimension : GENERIC_DIMENSION",
//t    "qualified_alias_member : IDENTIFIER DOUBLE_COLON",
//t    "sizeof_expression : SIZEOF open_parens_any type CLOSE_PARENS",
//t    "sizeof_expression : SIZEOF open_parens_any type error",
//t    "checked_expression : CHECKED open_parens_any expression CLOSE_PARENS",
//t    "checked_expression : CHECKED error",
//t    "unchecked_expression : UNCHECKED open_parens_any expression CLOSE_PARENS",
//t    "unchecked_expression : UNCHECKED error",
//t    "pointer_member_access : primary_expression OP_PTR IDENTIFIER opt_type_argument_list",
//t    "$$69 :",
//t    "anonymous_method_expression : DELEGATE opt_anonymous_method_signature $$69 block",
//t    "$$70 :",
//t    "anonymous_method_expression : ASYNC DELEGATE opt_anonymous_method_signature $$70 block",
//t    "opt_anonymous_method_signature :",
//t    "opt_anonymous_method_signature : anonymous_method_signature",
//t    "$$71 :",
//t    "anonymous_method_signature : OPEN_PARENS $$71 opt_formal_parameter_list CLOSE_PARENS",
//t    "default_value_expression : DEFAULT open_parens_any type CLOSE_PARENS",
//t    "unary_expression : primary_expression",
//t    "unary_expression : BANG prefixed_unary_expression",
//t    "unary_expression : TILDE prefixed_unary_expression",
//t    "unary_expression : OPEN_PARENS_CAST type CLOSE_PARENS prefixed_unary_expression",
//t    "unary_expression : AWAIT prefixed_unary_expression",
//t    "unary_expression : BANG error",
//t    "unary_expression : TILDE error",
//t    "unary_expression : OPEN_PARENS_CAST type CLOSE_PARENS error",
//t    "unary_expression : AWAIT error",
//t    "prefixed_unary_expression : unary_expression",
//t    "prefixed_unary_expression : PLUS prefixed_unary_expression",
//t    "prefixed_unary_expression : MINUS prefixed_unary_expression",
//t    "prefixed_unary_expression : OP_INC prefixed_unary_expression",
//t    "prefixed_unary_expression : OP_DEC prefixed_unary_expression",
//t    "prefixed_unary_expression : STAR prefixed_unary_expression",
//t    "prefixed_unary_expression : BITWISE_AND prefixed_unary_expression",
//t    "prefixed_unary_expression : PLUS error",
//t    "prefixed_unary_expression : MINUS error",
//t    "prefixed_unary_expression : OP_INC error",
//t    "prefixed_unary_expression : OP_DEC error",
//t    "prefixed_unary_expression : STAR error",
//t    "prefixed_unary_expression : BITWISE_AND error",
//t    "multiplicative_expression : prefixed_unary_expression",
//t    "multiplicative_expression : multiplicative_expression STAR prefixed_unary_expression",
//t    "multiplicative_expression : multiplicative_expression DIV prefixed_unary_expression",
//t    "multiplicative_expression : multiplicative_expression PERCENT prefixed_unary_expression",
//t    "multiplicative_expression : multiplicative_expression STAR error",
//t    "multiplicative_expression : multiplicative_expression DIV error",
//t    "multiplicative_expression : multiplicative_expression PERCENT error",
//t    "additive_expression : multiplicative_expression",
//t    "additive_expression : additive_expression PLUS multiplicative_expression",
//t    "additive_expression : additive_expression MINUS multiplicative_expression",
//t    "additive_expression : additive_expression PLUS error",
//t    "additive_expression : additive_expression MINUS error",
//t    "additive_expression : additive_expression AS type",
//t    "additive_expression : additive_expression IS pattern_type_expr opt_identifier",
//t    "additive_expression : additive_expression IS pattern_expr",
//t    "additive_expression : additive_expression AS error",
//t    "additive_expression : additive_expression IS error",
//t    "additive_expression : AWAIT IS type",
//t    "additive_expression : AWAIT AS type",
//t    "pattern_type_expr : variable_type",
//t    "pattern_expr : literal",
//t    "pattern_expr : PLUS prefixed_unary_expression",
//t    "pattern_expr : MINUS prefixed_unary_expression",
//t    "pattern_expr : sizeof_expression",
//t    "pattern_expr : default_value_expression",
//t    "pattern_expr : OPEN_PARENS_CAST type CLOSE_PARENS prefixed_unary_expression",
//t    "pattern_expr : STAR",
//t    "pattern_expr : pattern_expr_invocation",
//t    "pattern_expr : pattern_property",
//t    "pattern_expr_invocation : type_name_expression OPEN_PARENS opt_pattern_list CLOSE_PARENS",
//t    "pattern_property : type_name_expression OPEN_BRACE pattern_property_list CLOSE_BRACE",
//t    "pattern_property_list : pattern_property_entry",
//t    "pattern_property_list : pattern_property_list COMMA pattern_property_entry",
//t    "pattern_property_entry : identifier_inside_body IS pattern",
//t    "pattern : pattern_expr",
//t    "pattern : pattern_type_expr opt_identifier",
//t    "opt_pattern_list :",
//t    "opt_pattern_list : pattern_list",
//t    "pattern_list : pattern_argument",
//t    "pattern_list : pattern_list COMMA pattern_argument",
//t    "pattern_argument : pattern",
//t    "pattern_argument : IDENTIFIER COLON pattern",
//t    "shift_expression : additive_expression",
//t    "shift_expression : shift_expression OP_SHIFT_LEFT additive_expression",
//t    "shift_expression : shift_expression OP_SHIFT_RIGHT additive_expression",
//t    "shift_expression : shift_expression OP_SHIFT_LEFT error",
//t    "shift_expression : shift_expression OP_SHIFT_RIGHT error",
//t    "relational_expression : shift_expression",
//t    "relational_expression : relational_expression OP_LT shift_expression",
//t    "relational_expression : relational_expression OP_GT shift_expression",
//t    "relational_expression : relational_expression OP_LE shift_expression",
//t    "relational_expression : relational_expression OP_GE shift_expression",
//t    "relational_expression : relational_expression OP_LT error",
//t    "relational_expression : relational_expression OP_GT error",
//t    "relational_expression : relational_expression OP_LE error",
//t    "relational_expression : relational_expression OP_GE error",
//t    "equality_expression : relational_expression",
//t    "equality_expression : equality_expression OP_EQ relational_expression",
//t    "equality_expression : equality_expression OP_NE relational_expression",
//t    "equality_expression : equality_expression OP_EQ error",
//t    "equality_expression : equality_expression OP_NE error",
//t    "and_expression : equality_expression",
//t    "and_expression : and_expression BITWISE_AND equality_expression",
//t    "and_expression : and_expression BITWISE_AND error",
//t    "exclusive_or_expression : and_expression",
//t    "exclusive_or_expression : exclusive_or_expression CARRET and_expression",
//t    "exclusive_or_expression : exclusive_or_expression CARRET error",
//t    "inclusive_or_expression : exclusive_or_expression",
//t    "inclusive_or_expression : inclusive_or_expression BITWISE_OR exclusive_or_expression",
//t    "inclusive_or_expression : inclusive_or_expression BITWISE_OR error",
//t    "conditional_and_expression : inclusive_or_expression",
//t    "conditional_and_expression : conditional_and_expression OP_AND inclusive_or_expression",
//t    "conditional_and_expression : conditional_and_expression OP_AND error",
//t    "conditional_or_expression : conditional_and_expression",
//t    "conditional_or_expression : conditional_or_expression OP_OR conditional_and_expression",
//t    "conditional_or_expression : conditional_or_expression OP_OR error",
//t    "null_coalescing_expression : conditional_or_expression",
//t    "null_coalescing_expression : conditional_or_expression OP_COALESCING null_coalescing_expression",
//t    "conditional_expression : null_coalescing_expression",
//t    "conditional_expression : null_coalescing_expression INTERR expression COLON expression",
//t    "conditional_expression : null_coalescing_expression INTERR expression error",
//t    "conditional_expression : null_coalescing_expression INTERR expression COLON error",
//t    "conditional_expression : null_coalescing_expression INTERR expression COLON CLOSE_BRACE",
//t    "assignment_expression : prefixed_unary_expression ASSIGN expression",
//t    "assignment_expression : prefixed_unary_expression OP_MULT_ASSIGN expression",
//t    "assignment_expression : prefixed_unary_expression OP_DIV_ASSIGN expression",
//t    "assignment_expression : prefixed_unary_expression OP_MOD_ASSIGN expression",
//t    "assignment_expression : prefixed_unary_expression OP_ADD_ASSIGN expression",
//t    "assignment_expression : prefixed_unary_expression OP_SUB_ASSIGN expression",
//t    "assignment_expression : prefixed_unary_expression OP_SHIFT_LEFT_ASSIGN expression",
//t    "assignment_expression : prefixed_unary_expression OP_SHIFT_RIGHT_ASSIGN expression",
//t    "assignment_expression : prefixed_unary_expression OP_AND_ASSIGN expression",
//t    "assignment_expression : prefixed_unary_expression OP_OR_ASSIGN expression",
//t    "assignment_expression : prefixed_unary_expression OP_XOR_ASSIGN expression",
//t    "lambda_parameter_list : lambda_parameter",
//t    "lambda_parameter_list : lambda_parameter_list COMMA lambda_parameter",
//t    "lambda_parameter : parameter_modifier parameter_type identifier_inside_body",
//t    "lambda_parameter : parameter_type identifier_inside_body",
//t    "lambda_parameter : IDENTIFIER",
//t    "lambda_parameter : AWAIT",
//t    "opt_lambda_parameter_list :",
//t    "opt_lambda_parameter_list : lambda_parameter_list",
//t    "$$72 :",
//t    "lambda_expression_body : $$72 expression",
//t    "lambda_expression_body : block",
//t    "lambda_expression_body : error",
//t    "expression_or_error : expression",
//t    "expression_or_error : error",
//t    "$$73 :",
//t    "lambda_expression : IDENTIFIER ARROW $$73 lambda_expression_body",
//t    "$$74 :",
//t    "lambda_expression : AWAIT ARROW $$74 lambda_expression_body",
//t    "$$75 :",
//t    "lambda_expression : ASYNC identifier_inside_body ARROW $$75 lambda_expression_body",
//t    "$$76 :",
//t    "$$77 :",
//t    "lambda_expression : OPEN_PARENS_LAMBDA $$76 opt_lambda_parameter_list CLOSE_PARENS ARROW $$77 lambda_expression_body",
//t    "$$78 :",
//t    "$$79 :",
//t    "lambda_expression : ASYNC OPEN_PARENS_LAMBDA $$78 opt_lambda_parameter_list CLOSE_PARENS ARROW $$79 lambda_expression_body",
//t    "expression : assignment_expression",
//t    "expression : non_assignment_expression",
//t    "non_assignment_expression : conditional_expression",
//t    "non_assignment_expression : lambda_expression",
//t    "non_assignment_expression : query_expression",
//t    "non_assignment_expression : ARGLIST",
//t    "undocumented_expressions : REFVALUE OPEN_PARENS non_assignment_expression COMMA type CLOSE_PARENS",
//t    "undocumented_expressions : REFTYPE open_parens_any expression CLOSE_PARENS",
//t    "undocumented_expressions : MAKEREF open_parens_any expression CLOSE_PARENS",
//t    "constant_expression : expression",
//t    "boolean_expression : expression",
//t    "opt_primary_parameters :",
//t    "opt_primary_parameters : primary_parameters",
//t    "primary_parameters : OPEN_PARENS opt_formal_parameter_list CLOSE_PARENS",
//t    "opt_primary_parameters_with_class_base :",
//t    "opt_primary_parameters_with_class_base : class_base",
//t    "opt_primary_parameters_with_class_base : primary_parameters",
//t    "opt_primary_parameters_with_class_base : primary_parameters class_base",
//t    "$$80 :",
//t    "opt_primary_parameters_with_class_base : primary_parameters class_base OPEN_PARENS $$80 opt_argument_list CLOSE_PARENS",
//t    "$$81 :",
//t    "$$82 :",
//t    "$$83 :",
//t    "$$84 :",
//t    "class_declaration : opt_attributes opt_modifiers opt_partial CLASS $$81 type_declaration_name $$82 opt_primary_parameters_with_class_base opt_type_parameter_constraints_clauses $$83 OPEN_BRACE opt_class_member_declarations CLOSE_BRACE $$84 opt_semicolon",
//t    "opt_partial :",
//t    "opt_partial : PARTIAL",
//t    "opt_modifiers :",
//t    "opt_modifiers : modifiers",
//t    "modifiers : modifier",
//t    "modifiers : modifiers modifier",
//t    "modifier : NEW",
//t    "modifier : PUBLIC",
//t    "modifier : PROTECTED",
//t    "modifier : INTERNAL",
//t    "modifier : PRIVATE",
//t    "modifier : ABSTRACT",
//t    "modifier : SEALED",
//t    "modifier : STATIC",
//t    "modifier : READONLY",
//t    "modifier : VIRTUAL",
//t    "modifier : OVERRIDE",
//t    "modifier : EXTERN",
//t    "modifier : VOLATILE",
//t    "modifier : UNSAFE",
//t    "modifier : ASYNC",
//t    "opt_class_base :",
//t    "opt_class_base : class_base",
//t    "class_base : COLON type_list",
//t    "class_base : COLON type_list error",
//t    "opt_type_parameter_constraints_clauses :",
//t    "opt_type_parameter_constraints_clauses : type_parameter_constraints_clauses",
//t    "type_parameter_constraints_clauses : type_parameter_constraints_clause",
//t    "type_parameter_constraints_clauses : type_parameter_constraints_clauses type_parameter_constraints_clause",
//t    "type_parameter_constraints_clause : WHERE IDENTIFIER COLON type_parameter_constraints",
//t    "type_parameter_constraints_clause : WHERE IDENTIFIER error",
//t    "type_parameter_constraints : type_parameter_constraint",
//t    "type_parameter_constraints : type_parameter_constraints COMMA type_parameter_constraint",
//t    "type_parameter_constraint : type",
//t    "type_parameter_constraint : NEW OPEN_PARENS CLOSE_PARENS",
//t    "type_parameter_constraint : CLASS",
//t    "type_parameter_constraint : STRUCT",
//t    "opt_type_parameter_variance :",
//t    "opt_type_parameter_variance : type_parameter_variance",
//t    "type_parameter_variance : OUT",
//t    "type_parameter_variance : IN",
//t    "$$85 :",
//t    "block : OPEN_BRACE $$85 opt_statement_list block_end",
//t    "block_end : CLOSE_BRACE",
//t    "block_end : COMPLETE_COMPLETION",
//t    "$$86 :",
//t    "block_prepared : OPEN_BRACE $$86 opt_statement_list CLOSE_BRACE",
//t    "opt_statement_list :",
//t    "opt_statement_list : statement_list",
//t    "statement_list : statement",
//t    "statement_list : statement_list statement",
//t    "statement : block_variable_declaration",
//t    "statement : valid_declaration_statement",
//t    "statement : labeled_statement",
//t    "statement : error",
//t    "interactive_statement_list : interactive_statement",
//t    "interactive_statement_list : interactive_statement_list interactive_statement",
//t    "interactive_statement : block_variable_declaration",
//t    "interactive_statement : interactive_valid_declaration_statement",
//t    "interactive_statement : labeled_statement",
//t    "valid_declaration_statement : block",
//t    "valid_declaration_statement : empty_statement",
//t    "valid_declaration_statement : expression_statement",
//t    "valid_declaration_statement : selection_statement",
//t    "valid_declaration_statement : iteration_statement",
//t    "valid_declaration_statement : jump_statement",
//t    "valid_declaration_statement : try_statement",
//t    "valid_declaration_statement : checked_statement",
//t    "valid_declaration_statement : unchecked_statement",
//t    "valid_declaration_statement : lock_statement",
//t    "valid_declaration_statement : using_statement",
//t    "valid_declaration_statement : unsafe_statement",
//t    "valid_declaration_statement : fixed_statement",
//t    "interactive_valid_declaration_statement : block",
//t    "interactive_valid_declaration_statement : empty_statement",
//t    "interactive_valid_declaration_statement : interactive_expression_statement",
//t    "interactive_valid_declaration_statement : selection_statement",
//t    "interactive_valid_declaration_statement : iteration_statement",
//t    "interactive_valid_declaration_statement : jump_statement",
//t    "interactive_valid_declaration_statement : try_statement",
//t    "interactive_valid_declaration_statement : checked_statement",
//t    "interactive_valid_declaration_statement : unchecked_statement",
//t    "interactive_valid_declaration_statement : lock_statement",
//t    "interactive_valid_declaration_statement : using_statement",
//t    "interactive_valid_declaration_statement : unsafe_statement",
//t    "interactive_valid_declaration_statement : fixed_statement",
//t    "embedded_statement : valid_declaration_statement",
//t    "embedded_statement : block_variable_declaration",
//t    "embedded_statement : labeled_statement",
//t    "embedded_statement : error",
//t    "empty_statement : SEMICOLON",
//t    "$$87 :",
//t    "labeled_statement : identifier_inside_body COLON $$87 statement",
//t    "variable_type : variable_type_simple",
//t    "variable_type : variable_type_simple rank_specifiers",
//t    "variable_type_simple : type_name_expression opt_nullable",
//t    "variable_type_simple : type_name_expression pointer_stars",
//t    "variable_type_simple : builtin_type_expression",
//t    "variable_type_simple : void_invalid",
//t    "pointer_stars : pointer_star",
//t    "pointer_stars : pointer_star pointer_stars",
//t    "pointer_star : STAR",
//t    "identifier_inside_body : IDENTIFIER",
//t    "identifier_inside_body : AWAIT",
//t    "$$88 :",
//t    "block_variable_declaration : variable_type identifier_inside_body $$88 opt_local_variable_initializer opt_variable_declarators SEMICOLON",
//t    "$$89 :",
//t    "block_variable_declaration : CONST variable_type identifier_inside_body $$89 const_variable_initializer opt_const_declarators SEMICOLON",
//t    "opt_local_variable_initializer :",
//t    "opt_local_variable_initializer : ASSIGN block_variable_initializer",
//t    "opt_local_variable_initializer : error",
//t    "opt_variable_declarators :",
//t    "opt_variable_declarators : variable_declarators",
//t    "opt_using_or_fixed_variable_declarators :",
//t    "opt_using_or_fixed_variable_declarators : variable_declarators",
//t    "variable_declarators : variable_declarator",
//t    "variable_declarators : variable_declarators variable_declarator",
//t    "variable_declarator : COMMA identifier_inside_body",
//t    "variable_declarator : COMMA identifier_inside_body ASSIGN block_variable_initializer",
//t    "const_variable_initializer :",
//t    "const_variable_initializer : ASSIGN constant_initializer_expr",
//t    "opt_const_declarators :",
//t    "opt_const_declarators : const_declarators",
//t    "const_declarators : const_declarator",
//t    "const_declarators : const_declarators const_declarator",
//t    "const_declarator : COMMA identifier_inside_body ASSIGN constant_initializer_expr",
//t    "block_variable_initializer : variable_initializer",
//t    "block_variable_initializer : STACKALLOC simple_type OPEN_BRACKET_EXPR expression CLOSE_BRACKET",
//t    "block_variable_initializer : STACKALLOC simple_type",
//t    "expression_statement : statement_expression SEMICOLON",
//t    "expression_statement : statement_expression COMPLETE_COMPLETION",
//t    "expression_statement : statement_expression CLOSE_BRACE",
//t    "interactive_expression_statement : interactive_statement_expression SEMICOLON",
//t    "interactive_expression_statement : interactive_statement_expression COMPLETE_COMPLETION",
//t    "statement_expression : expression",
//t    "interactive_statement_expression : expression",
//t    "interactive_statement_expression : error",
//t    "selection_statement : if_statement",
//t    "selection_statement : switch_statement",
//t    "if_statement : IF open_parens_any boolean_expression CLOSE_PARENS embedded_statement",
//t    "if_statement : IF open_parens_any boolean_expression CLOSE_PARENS embedded_statement ELSE embedded_statement",
//t    "if_statement : IF open_parens_any boolean_expression error",
//t    "$$90 :",
//t    "switch_statement : SWITCH open_parens_any expression CLOSE_PARENS OPEN_BRACE $$90 opt_switch_sections CLOSE_BRACE",
//t    "switch_statement : SWITCH open_parens_any expression error",
//t    "opt_switch_sections :",
//t    "opt_switch_sections : switch_sections",
//t    "switch_sections : switch_section",
//t    "switch_sections : switch_sections switch_section",
//t    "switch_sections : error",
//t    "switch_section : switch_labels statement_list",
//t    "switch_labels : switch_label",
//t    "switch_labels : switch_labels switch_label",
//t    "switch_label : CASE constant_expression COLON",
//t    "switch_label : CASE constant_expression error",
//t    "switch_label : CASE pattern_expr_invocation COLON",
//t    "switch_label : DEFAULT_COLON",
//t    "iteration_statement : while_statement",
//t    "iteration_statement : do_statement",
//t    "iteration_statement : for_statement",
//t    "iteration_statement : foreach_statement",
//t    "while_statement : WHILE open_parens_any boolean_expression CLOSE_PARENS embedded_statement",
//t    "while_statement : WHILE open_parens_any boolean_expression error",
//t    "do_statement : DO embedded_statement WHILE open_parens_any boolean_expression CLOSE_PARENS SEMICOLON",
//t    "do_statement : DO embedded_statement error",
//t    "do_statement : DO embedded_statement WHILE open_parens_any boolean_expression error",
//t    "$$91 :",
//t    "for_statement : FOR open_parens_any $$91 for_statement_cont",
//t    "$$92 :",
//t    "for_statement_cont : opt_for_initializer SEMICOLON $$92 for_condition_and_iterator_part embedded_statement",
//t    "for_statement_cont : error",
//t    "$$93 :",
//t    "for_condition_and_iterator_part : opt_for_condition SEMICOLON $$93 for_iterator_part",
//t    "for_condition_and_iterator_part : opt_for_condition close_parens_close_brace",
//t    "for_iterator_part : opt_for_iterator CLOSE_PARENS",
//t    "for_iterator_part : opt_for_iterator CLOSE_BRACE",
//t    "close_parens_close_brace : CLOSE_PARENS",
//t    "close_parens_close_brace : CLOSE_BRACE",
//t    "opt_for_initializer :",
//t    "opt_for_initializer : for_initializer",
//t    "$$94 :",
//t    "for_initializer : variable_type identifier_inside_body $$94 opt_local_variable_initializer opt_variable_declarators",
//t    "for_initializer : statement_expression_list",
//t    "opt_for_condition :",
//t    "opt_for_condition : boolean_expression",
//t    "opt_for_iterator :",
//t    "opt_for_iterator : for_iterator",
//t    "for_iterator : statement_expression_list",
//t    "statement_expression_list : statement_expression",
//t    "statement_expression_list : statement_expression_list COMMA statement_expression",
//t    "foreach_statement : FOREACH open_parens_any type error",
//t    "foreach_statement : FOREACH open_parens_any type identifier_inside_body error",
//t    "$$95 :",
//t    "foreach_statement : FOREACH open_parens_any type identifier_inside_body IN expression CLOSE_PARENS $$95 embedded_statement",
//t    "jump_statement : break_statement",
//t    "jump_statement : continue_statement",
//t    "jump_statement : goto_statement",
//t    "jump_statement : return_statement",
//t    "jump_statement : throw_statement",
//t    "jump_statement : yield_statement",
//t    "break_statement : BREAK SEMICOLON",
//t    "continue_statement : CONTINUE SEMICOLON",
//t    "continue_statement : CONTINUE error",
//t    "goto_statement : GOTO identifier_inside_body SEMICOLON",
//t    "goto_statement : GOTO CASE constant_expression SEMICOLON",
//t    "goto_statement : GOTO DEFAULT SEMICOLON",
//t    "return_statement : RETURN opt_expression SEMICOLON",
//t    "return_statement : RETURN expression error",
//t    "return_statement : RETURN error",
//t    "throw_statement : THROW opt_expression SEMICOLON",
//t    "throw_statement : THROW expression error",
//t    "throw_statement : THROW error",
//t    "yield_statement : identifier_inside_body RETURN opt_expression SEMICOLON",
//t    "yield_statement : identifier_inside_body RETURN expression error",
//t    "yield_statement : identifier_inside_body BREAK SEMICOLON",
//t    "opt_expression :",
//t    "opt_expression : expression",
//t    "try_statement : TRY block catch_clauses",
//t    "try_statement : TRY block FINALLY block",
//t    "try_statement : TRY block catch_clauses FINALLY block",
//t    "try_statement : TRY block error",
//t    "catch_clauses : catch_clause",
//t    "catch_clauses : catch_clauses catch_clause",
//t    "opt_identifier :",
//t    "opt_identifier : identifier_inside_body",
//t    "catch_clause : CATCH opt_catch_filter block",
//t    "$$96 :",
//t    "catch_clause : CATCH open_parens_any type opt_identifier CLOSE_PARENS $$96 opt_catch_filter block_prepared",
//t    "catch_clause : CATCH open_parens_any error",
//t    "catch_clause : CATCH open_parens_any type opt_identifier CLOSE_PARENS error",
//t    "opt_catch_filter :",
//t    "opt_catch_filter : IF open_parens_any expression CLOSE_PARENS",
//t    "checked_statement : CHECKED block",
//t    "unchecked_statement : UNCHECKED block",
//t    "$$97 :",
//t    "unsafe_statement : UNSAFE $$97 block",
//t    "lock_statement : LOCK open_parens_any expression CLOSE_PARENS embedded_statement",
//t    "lock_statement : LOCK open_parens_any expression error",
//t    "$$98 :",
//t    "$$99 :",
//t    "fixed_statement : FIXED open_parens_any variable_type identifier_inside_body $$98 using_or_fixed_variable_initializer opt_using_or_fixed_variable_declarators CLOSE_PARENS $$99 embedded_statement",
//t    "$$100 :",
//t    "$$101 :",
//t    "using_statement : USING open_parens_any variable_type identifier_inside_body $$100 using_initialization CLOSE_PARENS $$101 embedded_statement",
//t    "using_statement : USING open_parens_any expression CLOSE_PARENS embedded_statement",
//t    "using_statement : USING open_parens_any expression error",
//t    "using_initialization : using_or_fixed_variable_initializer opt_using_or_fixed_variable_declarators",
//t    "using_initialization : error",
//t    "using_or_fixed_variable_initializer :",
//t    "using_or_fixed_variable_initializer : ASSIGN variable_initializer",
//t    "query_expression : first_from_clause query_body",
//t    "query_expression : nested_from_clause query_body",
//t    "query_expression : first_from_clause COMPLETE_COMPLETION",
//t    "query_expression : nested_from_clause COMPLETE_COMPLETION",
//t    "first_from_clause : FROM_FIRST identifier_inside_body IN expression",
//t    "first_from_clause : FROM_FIRST type identifier_inside_body IN expression",
//t    "nested_from_clause : FROM identifier_inside_body IN expression",
//t    "nested_from_clause : FROM type identifier_inside_body IN expression",
//t    "$$102 :",
//t    "from_clause : FROM identifier_inside_body IN $$102 expression_or_error",
//t    "$$103 :",
//t    "from_clause : FROM type identifier_inside_body IN $$103 expression_or_error",
//t    "query_body : query_body_clauses select_or_group_clause opt_query_continuation",
//t    "query_body : select_or_group_clause opt_query_continuation",
//t    "query_body : query_body_clauses COMPLETE_COMPLETION",
//t    "query_body : query_body_clauses error",
//t    "query_body : error",
//t    "$$104 :",
//t    "select_or_group_clause : SELECT $$104 expression_or_error",
//t    "$$105 :",
//t    "$$106 :",
//t    "select_or_group_clause : GROUP $$105 expression_or_error $$106 by_expression",
//t    "by_expression : BY expression_or_error",
//t    "by_expression : error",
//t    "query_body_clauses : query_body_clause",
//t    "query_body_clauses : query_body_clauses query_body_clause",
//t    "query_body_clause : from_clause",
//t    "query_body_clause : let_clause",
//t    "query_body_clause : where_clause",
//t    "query_body_clause : join_clause",
//t    "query_body_clause : orderby_clause",
//t    "$$107 :",
//t    "let_clause : LET identifier_inside_body ASSIGN $$107 expression_or_error",
//t    "$$108 :",
//t    "where_clause : WHERE $$108 expression_or_error",
//t    "$$109 :",
//t    "$$110 :",
//t    "$$111 :",
//t    "join_clause : JOIN identifier_inside_body IN $$109 expression_or_error ON $$110 expression_or_error EQUALS $$111 expression_or_error opt_join_into",
//t    "$$112 :",
//t    "$$113 :",
//t    "$$114 :",
//t    "join_clause : JOIN type identifier_inside_body IN $$112 expression_or_error ON $$113 expression_or_error EQUALS $$114 expression_or_error opt_join_into",
//t    "opt_join_into :",
//t    "opt_join_into : INTO identifier_inside_body",
//t    "$$115 :",
//t    "orderby_clause : ORDERBY $$115 orderings",
//t    "orderings : order_by",
//t    "$$116 :",
//t    "orderings : order_by COMMA $$116 orderings_then_by",
//t    "orderings_then_by : then_by",
//t    "$$117 :",
//t    "orderings_then_by : orderings_then_by COMMA $$117 then_by",
//t    "order_by : expression",
//t    "order_by : expression ASCENDING",
//t    "order_by : expression DESCENDING",
//t    "then_by : expression",
//t    "then_by : expression ASCENDING",
//t    "then_by : expression DESCENDING",
//t    "opt_query_continuation :",
//t    "$$118 :",
//t    "opt_query_continuation : INTO identifier_inside_body $$118 query_body",
//t    "interactive_parsing : EVAL_STATEMENT_PARSER EOF",
//t    "interactive_parsing : EVAL_USING_DECLARATIONS_UNIT_PARSER using_directives opt_COMPLETE_COMPLETION",
//t    "$$119 :",
//t    "interactive_parsing : EVAL_STATEMENT_PARSER $$119 interactive_statement_list opt_COMPLETE_COMPLETION",
//t    "interactive_parsing : EVAL_COMPILATION_UNIT_PARSER interactive_compilation_unit",
//t    "interactive_compilation_unit : opt_extern_alias_directives opt_using_directives",
//t    "interactive_compilation_unit : opt_extern_alias_directives opt_using_directives namespace_or_type_declarations",
//t    "opt_COMPLETE_COMPLETION :",
//t    "opt_COMPLETE_COMPLETION : COMPLETE_COMPLETION",
//t    "close_brace_or_complete_completion : CLOSE_BRACE",
//t    "close_brace_or_complete_completion : COMPLETE_COMPLETION",
//t    "documentation_parsing : DOC_SEE doc_cref",
//t    "doc_cref : doc_type_declaration_name opt_doc_method_sig",
//t    "doc_cref : builtin_types opt_doc_method_sig",
//t    "doc_cref : VOID opt_doc_method_sig",
//t    "doc_cref : builtin_types DOT IDENTIFIER opt_doc_method_sig",
//t    "doc_cref : doc_type_declaration_name DOT THIS",
//t    "$$120 :",
//t    "doc_cref : doc_type_declaration_name DOT THIS OPEN_BRACKET $$120 opt_doc_parameters CLOSE_BRACKET",
//t    "doc_cref : EXPLICIT OPERATOR type opt_doc_method_sig",
//t    "doc_cref : IMPLICIT OPERATOR type opt_doc_method_sig",
//t    "doc_cref : OPERATOR overloadable_operator opt_doc_method_sig",
//t    "doc_type_declaration_name : type_declaration_name",
//t    "doc_type_declaration_name : doc_type_declaration_name DOT type_declaration_name",
//t    "opt_doc_method_sig :",
//t    "$$121 :",
//t    "opt_doc_method_sig : OPEN_PARENS $$121 opt_doc_parameters CLOSE_PARENS",
//t    "opt_doc_parameters :",
//t    "opt_doc_parameters : doc_parameters",
//t    "doc_parameters : doc_parameter",
//t    "doc_parameters : doc_parameters COMMA doc_parameter",
//t    "doc_parameter : opt_parameter_modifier parameter_type",
//t  };
//t public static string getRule (int index) {
//t    return yyRule [index];
//t }
//t}
  protected static readonly string [] yyNames = {    
    "end-of-file",null,null,null,null,null,null,null,null,null,null,null,
    null,null,null,null,null,null,null,null,null,null,null,null,null,null,
    null,null,null,null,null,null,null,null,null,null,null,null,null,null,
    null,null,null,null,null,null,null,null,null,null,null,null,null,null,
    null,null,null,null,null,null,null,null,null,null,null,null,null,null,
    null,null,null,null,null,null,null,null,null,null,null,null,null,null,
    null,null,null,null,null,null,null,null,null,null,null,null,null,null,
    null,null,null,null,null,null,null,null,null,null,null,null,null,null,
    null,null,null,null,null,null,null,null,null,null,null,null,null,null,
    null,null,null,null,null,null,null,null,null,null,null,null,null,null,
    null,null,null,null,null,null,null,null,null,null,null,null,null,null,
    null,null,null,null,null,null,null,null,null,null,null,null,null,null,
    null,null,null,null,null,null,null,null,null,null,null,null,null,null,
    null,null,null,null,null,null,null,null,null,null,null,null,null,null,
    null,null,null,null,null,null,null,null,null,null,null,null,null,null,
    null,null,null,null,null,null,null,null,null,null,null,null,null,null,
    null,null,null,null,null,null,null,null,null,null,null,null,null,null,
    null,null,null,null,null,null,null,null,null,null,null,null,null,null,
    null,null,null,null,null,null,null,"EOF","NONE","ERROR",
    "FIRST_KEYWORD","ABSTRACT","AS","ADD","BASE","BOOL","BREAK","BYTE",
    "CASE","CATCH","CHAR","CHECKED","CLASS","CONST","CONTINUE","DECIMAL",
    "DEFAULT","DELEGATE","DO","DOUBLE","ELSE","ENUM","EVENT","EXPLICIT",
    "EXTERN","FALSE","FINALLY","FIXED","FLOAT","FOR","FOREACH","GOTO",
    "IF","IMPLICIT","IN","INT","INTERFACE","INTERNAL","IS","LOCK","LONG",
    "NAMESPACE","NEW","NULL","OBJECT","OPERATOR","OUT","OVERRIDE",
    "PARAMS","PRIVATE","PROTECTED","PUBLIC","READONLY","REF","RETURN",
    "REMOVE","SBYTE","SEALED","SHORT","SIZEOF","STACKALLOC","STATIC",
    "STRING","STRUCT","SWITCH","THIS","THROW","TRUE","TRY","TYPEOF",
    "UINT","ULONG","UNCHECKED","UNSAFE","USHORT","USING","VIRTUAL","VOID",
    "VOLATILE","WHERE","WHILE","ARGLIST","PARTIAL","ARROW","FROM",
    "FROM_FIRST","JOIN","ON","EQUALS","SELECT","GROUP","BY","LET",
    "ORDERBY","ASCENDING","DESCENDING","INTO","INTERR_NULLABLE",
    "EXTERN_ALIAS","REFVALUE","REFTYPE","MAKEREF","ASYNC","AWAIT",
    "INTERR_OPERATOR","GET","SET","LAST_KEYWORD","OPEN_BRACE",
    "CLOSE_BRACE","OPEN_BRACKET","CLOSE_BRACKET","OPEN_PARENS",
    "CLOSE_PARENS","DOT","COMMA","COLON","SEMICOLON","TILDE","PLUS",
    "MINUS","BANG","ASSIGN","OP_LT","OP_GT","BITWISE_AND","BITWISE_OR",
    "STAR","PERCENT","DIV","CARRET","INTERR","DOUBLE_COLON","OP_INC",
    "OP_DEC","OP_SHIFT_LEFT","OP_SHIFT_RIGHT","OP_LE","OP_GE","OP_EQ",
    "OP_NE","OP_AND","OP_OR","OP_MULT_ASSIGN","OP_DIV_ASSIGN",
    "OP_MOD_ASSIGN","OP_ADD_ASSIGN","OP_SUB_ASSIGN",
    "OP_SHIFT_LEFT_ASSIGN","OP_SHIFT_RIGHT_ASSIGN","OP_AND_ASSIGN",
    "OP_XOR_ASSIGN","OP_OR_ASSIGN","OP_PTR","OP_COALESCING",
    "OP_GENERICS_LT","OP_GENERICS_LT_DECL","OP_GENERICS_GT","LITERAL",
    "IDENTIFIER","OPEN_PARENS_LAMBDA","OPEN_PARENS_CAST",
    "GENERIC_DIMENSION","DEFAULT_COLON","OPEN_BRACKET_EXPR",
    "EVAL_STATEMENT_PARSER","EVAL_COMPILATION_UNIT_PARSER",
    "EVAL_USING_DECLARATIONS_UNIT_PARSER","DOC_SEE","GENERATE_COMPLETION",
    "COMPLETE_COMPLETION","UMINUS",
  };

  /** index-checked interface to yyNames[].
      @param token single character or %token value.
      @return token name or [illegal] or [unknown].
    */
//t  public static string yyname (int token) {
//t    if ((token < 0) || (token > yyNames.Length)) return "[illegal]";
//t    string name;
//t    if ((name = yyNames[token]) != null) return name;
//t    return "[unknown]";
//t  }

#pragma warning disable 414
  int yyExpectingState;
#pragma warning restore 414
  /** computes list of expected tokens on error by tracing the tables.
      @param state for which to compute the list.
      @return list of token names.
    */
  protected int [] yyExpectingTokens (int state){
    int token, n, len = 0;
    bool[] ok = new bool[yyNames.Length];
    if ((n = yySindex[state]) != 0)
      for (token = n < 0 ? -n : 0;
           (token < yyNames.Length) && (n+token < yyTable.Length); ++ token)
        if (yyCheck[n+token] == token && !ok[token] && yyNames[token] != null) {
          ++ len;
          ok[token] = true;
        }
    if ((n = yyRindex[state]) != 0)
      for (token = n < 0 ? -n : 0;
           (token < yyNames.Length) && (n+token < yyTable.Length); ++ token)
        if (yyCheck[n+token] == token && !ok[token] && yyNames[token] != null) {
          ++ len;
          ok[token] = true;
        }
    int [] result = new int [len];
    for (n = token = 0; n < len;  ++ token)
      if (ok[token]) result[n++] = token;
    return result;
  }
  protected string[] yyExpecting (int state) {
    int [] tokens = yyExpectingTokens (state);
    string [] result = new string[tokens.Length];
    for (int n = 0; n < tokens.Length;  n++)
      result[n++] = yyNames[tokens [n]];
    return result;
  }

  /** the generated parser, with debugging messages.
      Maintains a state and a value stack, currently with fixed maximum size.
      @param yyLex scanner.
      @param yydebug debug message writer implementing yyDebug, or null.
      @return result of the last reduction, if any.
      @throws yyException on irrecoverable parse error.
    */
  internal Object yyparse (yyParser.yyInput yyLex, Object yyd)
				 {
//t    this.debug = (yydebug.yyDebug)yyd;
    return yyparse(yyLex);
  }

  /** initial size and increment of the state/value stack [default 256].
      This is not final so that it can be overwritten outside of invocations
      of yyparse().
    */
  protected int yyMax;

  /** executed at the beginning of a reduce action.
      Used as $$ = yyDefault($1), prior to the user-specified action, if any.
      Can be overwritten to provide deep copy, etc.
      @param first value for $1, or null.
      @return first.
    */
  protected Object yyDefault (Object first) {
    return first;
  }

	static int[] global_yyStates;
	static object[] global_yyVals;
#pragma warning disable 649
	protected bool use_global_stacks;
#pragma warning restore 649
	object[] yyVals;					// value stack
	object yyVal;						// value stack ptr
	int yyToken;						// current input
	int yyTop;

  /** the generated parser.
      Maintains a state and a value stack, currently with fixed maximum size.
      @param yyLex scanner.
      @return result of the last reduction, if any.
      @throws yyException on irrecoverable parse error.
    */
  internal Object yyparse (yyParser.yyInput yyLex)
  {
    if (yyMax <= 0) yyMax = 256;		// initial size
    int yyState = 0;                   // state stack ptr
    int [] yyStates;               	// state stack 
    yyVal = null;
    yyToken = -1;
    int yyErrorFlag = 0;				// #tks to shift
	if (use_global_stacks && global_yyStates != null) {
		yyVals = global_yyVals;
		yyStates = global_yyStates;
   } else {
		yyVals = new object [yyMax];
		yyStates = new int [yyMax];
		if (use_global_stacks) {
			global_yyVals = yyVals;
			global_yyStates = yyStates;
		}
	}

    /*yyLoop:*/ for (yyTop = 0;; ++ yyTop) {
      if (yyTop >= yyStates.Length) {			// dynamically increase
        global::System.Array.Resize (ref yyStates, yyStates.Length+yyMax);
        global::System.Array.Resize (ref yyVals, yyVals.Length+yyMax);
      }
      yyStates[yyTop] = yyState;
      yyVals[yyTop] = yyVal;
//t      if (debug != null) debug.push(yyState, yyVal);

      /*yyDiscarded:*/ while (true) {	// discarding a token does not change stack
        int yyN;
        if ((yyN = yyDefRed[yyState]) == 0) {	// else [default] reduce (yyN)
          if (yyToken < 0) {
            yyToken = yyLex.advance() ? yyLex.token() : 0;
//t            if (debug != null)
//t              debug.lex(yyState, yyToken, yyname(yyToken), yyLex.value());
          }
          if ((yyN = yySindex[yyState]) != 0 && ((yyN += yyToken) >= 0)
              && (yyN < yyTable.Length) && (yyCheck[yyN] == yyToken)) {
//t            if (debug != null)
//t              debug.shift(yyState, yyTable[yyN], yyErrorFlag-1);
            yyState = yyTable[yyN];		// shift to yyN
            yyVal = yyLex.value();
            yyToken = -1;
            if (yyErrorFlag > 0) -- yyErrorFlag;
            goto continue_yyLoop;
          }
          if ((yyN = yyRindex[yyState]) != 0 && (yyN += yyToken) >= 0
              && yyN < yyTable.Length && yyCheck[yyN] == yyToken)
            yyN = yyTable[yyN];			// reduce (yyN)
          else
            switch (yyErrorFlag) {
  
            case 0:
              yyExpectingState = yyState;
              // yyerror(String.Format ("syntax error, got token `{0}'", yyname (yyToken)), yyExpecting(yyState));
//t              if (debug != null) debug.error("syntax error");
              if (yyToken == 0 /*eof*/ || yyToken == eof_token) throw new yyParser.yyUnexpectedEof ();
              goto case 1;
            case 1: case 2:
              yyErrorFlag = 3;
              do {
                if ((yyN = yySindex[yyStates[yyTop]]) != 0
                    && (yyN += Token.yyErrorCode) >= 0 && yyN < yyTable.Length
                    && yyCheck[yyN] == Token.yyErrorCode) {
//t                  if (debug != null)
//t                    debug.shift(yyStates[yyTop], yyTable[yyN], 3);
                  yyState = yyTable[yyN];
                  yyVal = yyLex.value();
                  goto continue_yyLoop;
                }
//t                if (debug != null) debug.pop(yyStates[yyTop]);
              } while (-- yyTop >= 0);
//t              if (debug != null) debug.reject();
              throw new yyParser.yyException("irrecoverable syntax error");
  
            case 3:
              if (yyToken == 0) {
//t                if (debug != null) debug.reject();
                throw new yyParser.yyException("irrecoverable syntax error at end-of-file");
              }
//t              if (debug != null)
//t                debug.discard(yyState, yyToken, yyname(yyToken),
//t  							yyLex.value());
              yyToken = -1;
              goto continue_yyDiscarded;		// leave stack alone
            }
        }
        int yyV = yyTop + 1-yyLen[yyN];
//t        if (debug != null)
//t          debug.reduce(yyState, yyStates[yyV-1], yyN, YYRules.getRule (yyN), yyLen[yyN]);
        yyVal = yyV > yyTop ? null : yyVals[yyV]; // yyVal = yyDefault(yyV > yyTop ? null : yyVals[yyV]);
        switch (yyN) {
case 1:
#line 381 "cs-parser.jay"
  {
		Lexer.check_incorrect_doc_comment ();
	  }
  break;
case 2:
#line 382 "cs-parser.jay"
  { Lexer.CompleteOnEOF = false; }
  break;
case 6:
  case_6();
  break;
case 7:
#line 401 "cs-parser.jay"
  {
		module.AddAttributes ((Attributes) yyVals[0+yyTop], current_namespace);
	  }
  break;
case 8:
  case_8();
  break;
case 13:
  case_13();
  break;
case 14:
#line 446 "cs-parser.jay"
  {
		Error_SyntaxError (yyToken);
	  }
  break;
case 17:
  case_17();
  break;
case 18:
  case_18();
  break;
case 19:
  case_19();
  break;
case 20:
  case_20();
  break;
case 21:
  case_21();
  break;
case 22:
  case_22();
  break;
case 23:
  case_23();
  break;
case 24:
  case_24();
  break;
case 27:
  case_27();
  break;
case 28:
  case_28();
  break;
case 29:
  case_29();
  break;
case 30:
  case_30();
  break;
case 43:
  case_43();
  break;
case 44:
#line 631 "cs-parser.jay"
  {
		current_namespace.DeclarationFound = true;
	  }
  break;
case 45:
  case_45();
  break;
case 53:
  case_53();
  break;
case 54:
  case_54();
  break;
case 55:
  case_55();
  break;
case 56:
  case_56();
  break;
case 57:
  case_57();
  break;
case 58:
  case_58();
  break;
case 59:
  case_59();
  break;
case 60:
  case_60();
  break;
case 61:
  case_61();
  break;
case 62:
  case_62();
  break;
case 63:
#line 755 "cs-parser.jay"
  { yyVal = "event"; }
  break;
case 64:
#line 756 "cs-parser.jay"
  { yyVal = "return"; }
  break;
case 65:
#line 763 "cs-parser.jay"
  {
		yyVal = new List<Attribute> (4) { (Attribute) yyVals[0+yyTop] };
	  }
  break;
case 66:
  case_66();
  break;
case 67:
#line 780 "cs-parser.jay"
  {
		++lexer.parsing_block;
	  }
  break;
case 68:
  case_68();
  break;
case 70:
#line 799 "cs-parser.jay"
  { yyVal = null; }
  break;
case 71:
#line 803 "cs-parser.jay"
  {
		yyVal = yyVals[-1+yyTop];
	  }
  break;
case 72:
#line 808 "cs-parser.jay"
  { yyVal = null; }
  break;
case 73:
  case_73();
  break;
case 74:
  case_74();
  break;
case 75:
  case_75();
  break;
case 76:
  case_76();
  break;
case 77:
#line 850 "cs-parser.jay"
  {
	  	yyVal = new Argument ((Expression) yyVals[0+yyTop]);
	  }
  break;
case 79:
  case_79();
  break;
case 80:
#line 863 "cs-parser.jay"
  {
		++lexer.parsing_block;
	  }
  break;
case 81:
  case_81();
  break;
case 82:
  case_82();
  break;
case 84:
#line 894 "cs-parser.jay"
  { yyVal = null; }
  break;
case 85:
#line 898 "cs-parser.jay"
  { 
		yyVal = Argument.AType.Ref;
	  }
  break;
case 86:
#line 902 "cs-parser.jay"
  { 
		yyVal = Argument.AType.Out;
	  }
  break;
case 89:
  case_89();
  break;
case 90:
  case_90();
  break;
case 104:
  case_104();
  break;
case 105:
  case_105();
  break;
case 106:
  case_106();
  break;
case 107:
#line 979 "cs-parser.jay"
  {
	  }
  break;
case 108:
  case_108();
  break;
case 109:
  case_109();
  break;
case 110:
  case_110();
  break;
case 111:
  case_111();
  break;
case 112:
  case_112();
  break;
case 113:
#line 1029 "cs-parser.jay"
  {
		Error_SyntaxError (yyToken);
	  }
  break;
case 114:
  case_114();
  break;
case 115:
  case_115();
  break;
case 116:
  case_116();
  break;
case 119:
#line 1078 "cs-parser.jay"
  {
		current_field.AddDeclarator ((FieldDeclarator) yyVals[0+yyTop]);
	  }
  break;
case 120:
#line 1082 "cs-parser.jay"
  {
		current_field.AddDeclarator ((FieldDeclarator) yyVals[0+yyTop]);
	  }
  break;
case 121:
  case_121();
  break;
case 122:
#line 1098 "cs-parser.jay"
  {
		++lexer.parsing_block;
	  }
  break;
case 123:
  case_123();
  break;
case 124:
  case_124();
  break;
case 127:
  case_127();
  break;
case 128:
  case_128();
  break;
case 129:
  case_129();
  break;
case 130:
  case_130();
  break;
case 131:
#line 1176 "cs-parser.jay"
  {
		report.Error (1641, GetLocation (yyVals[-1+yyTop]), "A fixed size buffer field must have the array size specifier after the field name");
	  }
  break;
case 133:
  case_133();
  break;
case 134:
  case_134();
  break;
case 137:
#line 1206 "cs-parser.jay"
  {
		current_field.AddDeclarator ((FieldDeclarator) yyVals[0+yyTop]);
	  }
  break;
case 138:
#line 1210 "cs-parser.jay"
  {
		current_field.AddDeclarator ((FieldDeclarator) yyVals[0+yyTop]);
	  }
  break;
case 139:
  case_139();
  break;
case 140:
#line 1223 "cs-parser.jay"
  {
		++lexer.parsing_block;
	  }
  break;
case 141:
  case_141();
  break;
case 144:
#line 1242 "cs-parser.jay"
  {
		current_field.AddDeclarator ((FieldDeclarator) yyVals[0+yyTop]);
	  }
  break;
case 145:
#line 1246 "cs-parser.jay"
  {
		current_field.AddDeclarator ((FieldDeclarator) yyVals[0+yyTop]);
	  }
  break;
case 146:
  case_146();
  break;
case 147:
#line 1262 "cs-parser.jay"
  {
		++lexer.parsing_block;
	  }
  break;
case 148:
  case_148();
  break;
case 149:
  case_149();
  break;
case 152:
  case_152();
  break;
case 153:
  case_153();
  break;
case 154:
  case_154();
  break;
case 155:
#line 1329 "cs-parser.jay"
  {
		valid_param_mod = ParameterModifierType.All;
	  }
  break;
case 156:
  case_156();
  break;
case 157:
  case_157();
  break;
case 158:
#line 1369 "cs-parser.jay"
  {
		lexer.parsing_generic_declaration = true;
	  }
  break;
case 159:
  case_159();
  break;
case 160:
#line 1379 "cs-parser.jay"
  {
		lexer.ConstraintsParsing = true;
	  }
  break;
case 161:
  case_161();
  break;
case 162:
  case_162();
  break;
case 163:
  case_163();
  break;
case 167:
#line 1457 "cs-parser.jay"
  { yyVal = null; }
  break;
case 168:
  case_168();
  break;
case 169:
  case_169();
  break;
case 170:
#line 1481 "cs-parser.jay"
  { yyVal = ParametersCompiled.EmptyReadOnlyParameters; }
  break;
case 172:
  case_172();
  break;
case 173:
  case_173();
  break;
case 174:
  case_174();
  break;
case 175:
  case_175();
  break;
case 176:
  case_176();
  break;
case 177:
  case_177();
  break;
case 178:
  case_178();
  break;
case 179:
#line 1539 "cs-parser.jay"
  {
		yyVal = new ParametersCompiled (new Parameter[] { (Parameter) yyVals[0+yyTop] } );
	  }
  break;
case 180:
#line 1543 "cs-parser.jay"
  {
		yyVal = new ParametersCompiled (new Parameter [] { new ArglistParameter (GetLocation (yyVals[0+yyTop])) }, true);
	  }
  break;
case 181:
  case_181();
  break;
case 182:
  case_182();
  break;
case 183:
  case_183();
  break;
case 184:
  case_184();
  break;
case 185:
  case_185();
  break;
case 186:
  case_186();
  break;
case 187:
  case_187();
  break;
case 188:
#line 1621 "cs-parser.jay"
  {
	  	++lexer.parsing_block;
	  }
  break;
case 189:
  case_189();
  break;
case 190:
#line 1662 "cs-parser.jay"
  { yyVal = Parameter.Modifier.NONE; }
  break;
case 192:
#line 1670 "cs-parser.jay"
  {
		yyVal = yyVals[0+yyTop];
	  }
  break;
case 193:
  case_193();
  break;
case 194:
  case_194();
  break;
case 195:
  case_195();
  break;
case 196:
  case_196();
  break;
case 197:
  case_197();
  break;
case 198:
  case_198();
  break;
case 199:
  case_199();
  break;
case 200:
  case_200();
  break;
case 201:
  case_201();
  break;
case 202:
#line 1760 "cs-parser.jay"
  {
		Error_DuplicateParameterModifier (GetLocation (yyVals[-1+yyTop]), Parameter.Modifier.PARAMS);
	  }
  break;
case 203:
  case_203();
  break;
case 204:
  case_204();
  break;
case 205:
  case_205();
  break;
case 206:
  case_206();
  break;
case 207:
  case_207();
  break;
case 208:
#line 1810 "cs-parser.jay"
  {
		current_property = null;
	  }
  break;
case 209:
  case_209();
  break;
case 210:
  case_210();
  break;
case 212:
  case_212();
  break;
case 213:
  case_213();
  break;
case 216:
#line 1872 "cs-parser.jay"
  {
	  	valid_param_mod = ParameterModifierType.Params | ParameterModifierType.DefaultValue;
	  }
  break;
case 217:
  case_217();
  break;
case 218:
  case_218();
  break;
case 219:
#line 1918 "cs-parser.jay"
  {
		lbag.AppendToMember (current_property, GetLocation (yyVals[-2+yyTop]), GetLocation (yyVals[0+yyTop]));
	  }
  break;
case 220:
  case_220();
  break;
case 225:
  case_225();
  break;
case 226:
  case_226();
  break;
case 227:
  case_227();
  break;
case 228:
  case_228();
  break;
case 229:
  case_229();
  break;
case 231:
  case_231();
  break;
case 232:
  case_232();
  break;
case 233:
#line 2055 "cs-parser.jay"
  {
	  }
  break;
case 234:
  case_234();
  break;
case 235:
  case_235();
  break;
case 236:
  case_236();
  break;
case 237:
  case_237();
  break;
case 238:
#line 2095 "cs-parser.jay"
  {
		Error_SyntaxError (yyToken);	  
	  }
  break;
case 241:
  case_241();
  break;
case 242:
  case_242();
  break;
case 243:
#line 2120 "cs-parser.jay"
  {
		report.Error (525, GetLocation (yyVals[0+yyTop]), "Interfaces cannot contain fields or constants");
	  }
  break;
case 244:
#line 2124 "cs-parser.jay"
  {
		report.Error (525, GetLocation (yyVals[0+yyTop]), "Interfaces cannot contain fields or constants");
	  }
  break;
case 249:
#line 2132 "cs-parser.jay"
  {
	  	report.Error (567, GetLocation (yyVals[0+yyTop]), "Interfaces cannot contain operators");
	  }
  break;
case 250:
#line 2136 "cs-parser.jay"
  {
	  	report.Error (526, GetLocation (yyVals[0+yyTop]), "Interfaces cannot contain contructors");
	  }
  break;
case 251:
#line 2140 "cs-parser.jay"
  {
	  	report.Error (524, GetLocation (yyVals[0+yyTop]), "Interfaces cannot declare classes, structs, interfaces, delegates, or enumerations");
	  }
  break;
case 252:
#line 2146 "cs-parser.jay"
  {
	  }
  break;
case 253:
  case_253();
  break;
case 255:
  case_255();
  break;
case 256:
  case_256();
  break;
case 257:
  case_257();
  break;
case 259:
#line 2237 "cs-parser.jay"
  { yyVal = Operator.OpType.LogicalNot; }
  break;
case 260:
#line 2238 "cs-parser.jay"
  { yyVal = Operator.OpType.OnesComplement; }
  break;
case 261:
#line 2239 "cs-parser.jay"
  { yyVal = Operator.OpType.Increment; }
  break;
case 262:
#line 2240 "cs-parser.jay"
  { yyVal = Operator.OpType.Decrement; }
  break;
case 263:
#line 2241 "cs-parser.jay"
  { yyVal = Operator.OpType.True; }
  break;
case 264:
#line 2242 "cs-parser.jay"
  { yyVal = Operator.OpType.False; }
  break;
case 265:
#line 2244 "cs-parser.jay"
  { yyVal = Operator.OpType.Addition; }
  break;
case 266:
#line 2245 "cs-parser.jay"
  { yyVal = Operator.OpType.Subtraction; }
  break;
case 267:
#line 2247 "cs-parser.jay"
  { yyVal = Operator.OpType.Multiply; }
  break;
case 268:
#line 2248 "cs-parser.jay"
  {  yyVal = Operator.OpType.Division; }
  break;
case 269:
#line 2249 "cs-parser.jay"
  { yyVal = Operator.OpType.Modulus; }
  break;
case 270:
#line 2250 "cs-parser.jay"
  { yyVal = Operator.OpType.BitwiseAnd; }
  break;
case 271:
#line 2251 "cs-parser.jay"
  { yyVal = Operator.OpType.BitwiseOr; }
  break;
case 272:
#line 2252 "cs-parser.jay"
  { yyVal = Operator.OpType.ExclusiveOr; }
  break;
case 273:
#line 2253 "cs-parser.jay"
  { yyVal = Operator.OpType.LeftShift; }
  break;
case 274:
#line 2254 "cs-parser.jay"
  { yyVal = Operator.OpType.RightShift; }
  break;
case 275:
#line 2255 "cs-parser.jay"
  { yyVal = Operator.OpType.Equality; }
  break;
case 276:
#line 2256 "cs-parser.jay"
  { yyVal = Operator.OpType.Inequality; }
  break;
case 277:
#line 2257 "cs-parser.jay"
  { yyVal = Operator.OpType.GreaterThan; }
  break;
case 278:
#line 2258 "cs-parser.jay"
  { yyVal = Operator.OpType.LessThan; }
  break;
case 279:
#line 2259 "cs-parser.jay"
  { yyVal = Operator.OpType.GreaterThanOrEqual; }
  break;
case 280:
#line 2260 "cs-parser.jay"
  { yyVal = Operator.OpType.LessThanOrEqual; }
  break;
case 281:
  case_281();
  break;
case 282:
#line 2274 "cs-parser.jay"
  {
		valid_param_mod = ParameterModifierType.DefaultValue;
	  }
  break;
case 283:
  case_283();
  break;
case 284:
#line 2297 "cs-parser.jay"
  {
		valid_param_mod = ParameterModifierType.DefaultValue;
	  }
  break;
case 285:
  case_285();
  break;
case 286:
  case_286();
  break;
case 287:
  case_287();
  break;
case 288:
  case_288();
  break;
case 289:
  case_289();
  break;
case 290:
  case_290();
  break;
case 291:
  case_291();
  break;
case 293:
#line 2421 "cs-parser.jay"
  { current_block = null; yyVal = null; }
  break;
case 296:
#line 2433 "cs-parser.jay"
  {
		++lexer.parsing_block;
	  }
  break;
case 297:
  case_297();
  break;
case 298:
#line 2443 "cs-parser.jay"
  {
		++lexer.parsing_block;
	  }
  break;
case 299:
  case_299();
  break;
case 300:
  case_300();
  break;
case 301:
  case_301();
  break;
case 302:
  case_302();
  break;
case 303:
  case_303();
  break;
case 304:
  case_304();
  break;
case 305:
  case_305();
  break;
case 306:
  case_306();
  break;
case 307:
  case_307();
  break;
case 308:
  case_308();
  break;
case 309:
  case_309();
  break;
case 311:
#line 2566 "cs-parser.jay"
  {
	  	++lexer.parsing_block;
	  }
  break;
case 312:
  case_312();
  break;
case 315:
#line 2583 "cs-parser.jay"
  {
		current_event_field.AddDeclarator ((FieldDeclarator) yyVals[0+yyTop]);
	  }
  break;
case 316:
#line 2587 "cs-parser.jay"
  {
		current_event_field.AddDeclarator ((FieldDeclarator) yyVals[0+yyTop]);
	  }
  break;
case 317:
  case_317();
  break;
case 318:
#line 2600 "cs-parser.jay"
  {
		++lexer.parsing_block;
	  }
  break;
case 319:
  case_319();
  break;
case 320:
  case_320();
  break;
case 321:
#line 2625 "cs-parser.jay"
  {
		yyVal = yyVals[0+yyTop];
	  }
  break;
case 324:
  case_324();
  break;
case 325:
  case_325();
  break;
case 326:
  case_326();
  break;
case 327:
  case_327();
  break;
case 328:
  case_328();
  break;
case 329:
  case_329();
  break;
case 330:
  case_330();
  break;
case 331:
  case_331();
  break;
case 333:
  case_333();
  break;
case 334:
  case_334();
  break;
case 335:
  case_335();
  break;
case 336:
  case_336();
  break;
case 337:
  case_337();
  break;
case 338:
  case_338();
  break;
case 340:
#line 2788 "cs-parser.jay"
  {
		yyVal = yyVals[0+yyTop];
	 }
  break;
case 341:
  case_341();
  break;
case 344:
#line 2802 "cs-parser.jay"
  {
	  	lbag.AddLocation (yyVals[-1+yyTop], GetLocation (yyVals[0+yyTop]));
	  }
  break;
case 346:
  case_346();
  break;
case 347:
  case_347();
  break;
case 348:
  case_348();
  break;
case 349:
  case_349();
  break;
case 350:
  case_350();
  break;
case 352:
#line 2876 "cs-parser.jay"
  {
		valid_param_mod = ParameterModifierType.Ref | ParameterModifierType.Out | ParameterModifierType.Params | ParameterModifierType.DefaultValue;
	  }
  break;
case 353:
  case_353();
  break;
case 354:
#line 2895 "cs-parser.jay"
  {
		lexer.ConstraintsParsing = false;
	  }
  break;
case 355:
  case_355();
  break;
case 357:
  case_357();
  break;
case 359:
  case_359();
  break;
case 360:
  case_360();
  break;
case 362:
  case_362();
  break;
case 363:
  case_363();
  break;
case 364:
  case_364();
  break;
case 365:
  case_365();
  break;
case 367:
  case_367();
  break;
case 368:
  case_368();
  break;
case 369:
  case_369();
  break;
case 370:
  case_370();
  break;
case 371:
#line 3014 "cs-parser.jay"
  {
		lexer.parsing_generic_declaration = true;
	  }
  break;
case 372:
  case_372();
  break;
case 373:
  case_373();
  break;
case 375:
  case_375();
  break;
case 376:
  case_376();
  break;
case 377:
  case_377();
  break;
case 378:
  case_378();
  break;
case 379:
  case_379();
  break;
case 380:
  case_380();
  break;
case 382:
  case_382();
  break;
case 383:
  case_383();
  break;
case 384:
  case_384();
  break;
case 385:
  case_385();
  break;
case 386:
  case_386();
  break;
case 388:
#line 3132 "cs-parser.jay"
  {
		yyVal = new TypeExpression (compiler.BuiltinTypes.Void, GetLocation (yyVals[0+yyTop]));
	  }
  break;
case 389:
#line 3139 "cs-parser.jay"
  {
		lexer.parsing_generic_declaration = true;
	  }
  break;
case 395:
  case_395();
  break;
case 397:
#line 3169 "cs-parser.jay"
  {
		yyVal = new ComposedCast ((FullNamedExpression) yyVals[-1+yyTop], (ComposedTypeSpecifier) yyVals[0+yyTop]);
	  }
  break;
case 398:
  case_398();
  break;
case 399:
#line 3188 "cs-parser.jay"
  {
		yyVal = new ComposedCast ((ATypeNameExpression) yyVals[-1+yyTop], (ComposedTypeSpecifier) yyVals[0+yyTop]);
	  }
  break;
case 401:
  case_401();
  break;
case 402:
  case_402();
  break;
case 403:
#line 3209 "cs-parser.jay"
  {
		yyVal = new ComposedCast ((FullNamedExpression) yyVals[-1+yyTop], (ComposedTypeSpecifier) yyVals[0+yyTop]);
	  }
  break;
case 404:
#line 3213 "cs-parser.jay"
  {
		yyVal = new ComposedCast (new TypeExpression (compiler.BuiltinTypes.Void, GetLocation (yyVals[-1+yyTop])), (ComposedTypeSpecifier) yyVals[0+yyTop]);
	  }
  break;
case 405:
  case_405();
  break;
case 406:
  case_406();
  break;
case 407:
  case_407();
  break;
case 408:
#line 3246 "cs-parser.jay"
  { yyVal = new TypeExpression (compiler.BuiltinTypes.Object, GetLocation (yyVals[0+yyTop])); }
  break;
case 409:
#line 3247 "cs-parser.jay"
  { yyVal = new TypeExpression (compiler.BuiltinTypes.String, GetLocation (yyVals[0+yyTop])); }
  break;
case 410:
#line 3248 "cs-parser.jay"
  { yyVal = new TypeExpression (compiler.BuiltinTypes.Bool, GetLocation (yyVals[0+yyTop])); }
  break;
case 411:
#line 3249 "cs-parser.jay"
  { yyVal = new TypeExpression (compiler.BuiltinTypes.Decimal, GetLocation (yyVals[0+yyTop])); }
  break;
case 412:
#line 3250 "cs-parser.jay"
  { yyVal = new TypeExpression (compiler.BuiltinTypes.Float, GetLocation (yyVals[0+yyTop])); }
  break;
case 413:
#line 3251 "cs-parser.jay"
  { yyVal = new TypeExpression (compiler.BuiltinTypes.Double, GetLocation (yyVals[0+yyTop])); }
  break;
case 415:
#line 3256 "cs-parser.jay"
  { yyVal = new TypeExpression (compiler.BuiltinTypes.SByte, GetLocation (yyVals[0+yyTop])); }
  break;
case 416:
#line 3257 "cs-parser.jay"
  { yyVal = new TypeExpression (compiler.BuiltinTypes.Byte, GetLocation (yyVals[0+yyTop])); }
  break;
case 417:
#line 3258 "cs-parser.jay"
  { yyVal = new TypeExpression (compiler.BuiltinTypes.Short, GetLocation (yyVals[0+yyTop])); }
  break;
case 418:
#line 3259 "cs-parser.jay"
  { yyVal = new TypeExpression (compiler.BuiltinTypes.UShort, GetLocation (yyVals[0+yyTop])); }
  break;
case 419:
#line 3260 "cs-parser.jay"
  { yyVal = new TypeExpression (compiler.BuiltinTypes.Int, GetLocation (yyVals[0+yyTop])); }
  break;
case 420:
#line 3261 "cs-parser.jay"
  { yyVal = new TypeExpression (compiler.BuiltinTypes.UInt, GetLocation (yyVals[0+yyTop])); }
  break;
case 421:
#line 3262 "cs-parser.jay"
  { yyVal = new TypeExpression (compiler.BuiltinTypes.Long, GetLocation (yyVals[0+yyTop])); }
  break;
case 422:
#line 3263 "cs-parser.jay"
  { yyVal = new TypeExpression (compiler.BuiltinTypes.ULong, GetLocation (yyVals[0+yyTop])); }
  break;
case 423:
#line 3264 "cs-parser.jay"
  { yyVal = new TypeExpression (compiler.BuiltinTypes.Char, GetLocation (yyVals[0+yyTop])); }
  break;
case 445:
  case_445();
  break;
case 449:
#line 3307 "cs-parser.jay"
  { yyVal = new NullLiteral (GetLocation (yyVals[0+yyTop])); }
  break;
case 450:
#line 3311 "cs-parser.jay"
  { yyVal = new BoolLiteral (compiler.BuiltinTypes, true, GetLocation (yyVals[0+yyTop])); }
  break;
case 451:
#line 3312 "cs-parser.jay"
  { yyVal = new BoolLiteral (compiler.BuiltinTypes, false, GetLocation (yyVals[0+yyTop])); }
  break;
case 456:
  case_456();
  break;
case 457:
#line 3345 "cs-parser.jay"
  {
		yyVal = new ParenthesizedExpression ((Expression) yyVals[-1+yyTop], GetLocation (yyVals[-2+yyTop]));
	  }
  break;
case 458:
  case_458();
  break;
case 459:
  case_459();
  break;
case 460:
  case_460();
  break;
case 461:
  case_461();
  break;
case 462:
  case_462();
  break;
case 463:
  case_463();
  break;
case 464:
  case_464();
  break;
case 465:
  case_465();
  break;
case 466:
#line 3406 "cs-parser.jay"
  {
		yyVal = new CompletionMemberAccess ((Expression) yyVals[-2+yyTop], null,GetLocation (yyVals[0+yyTop]));
	  }
  break;
case 467:
  case_467();
  break;
case 468:
#line 3414 "cs-parser.jay"
  {
		yyVal = new CompletionMemberAccess ((Expression) yyVals[-2+yyTop], null, lexer.Location);
	  }
  break;
case 469:
  case_469();
  break;
case 470:
  case_470();
  break;
case 471:
  case_471();
  break;
case 472:
  case_472();
  break;
case 473:
#line 3444 "cs-parser.jay"
  { yyVal = null; }
  break;
case 475:
  case_475();
  break;
case 476:
  case_476();
  break;
case 477:
#line 3466 "cs-parser.jay"
  { yyVal = null; }
  break;
case 478:
#line 3470 "cs-parser.jay"
  {
		yyVal = yyVals[0+yyTop];
	}
  break;
case 479:
  case_479();
  break;
case 480:
  case_480();
  break;
case 481:
  case_481();
  break;
case 482:
  case_482();
  break;
case 483:
  case_483();
  break;
case 484:
#line 3508 "cs-parser.jay"
  {
		yyVal = new CompletionElementInitializer (null, GetLocation (yyVals[0+yyTop]));
	  }
  break;
case 485:
  case_485();
  break;
case 486:
  case_486();
  break;
case 487:
  case_487();
  break;
case 488:
  case_488();
  break;
case 491:
#line 3547 "cs-parser.jay"
  { yyVal = null; }
  break;
case 493:
  case_493();
  break;
case 494:
  case_494();
  break;
case 495:
  case_495();
  break;
case 496:
  case_496();
  break;
case 497:
  case_497();
  break;
case 498:
#line 3599 "cs-parser.jay"
  {
		yyVal = new Argument ((Expression) yyVals[0+yyTop]);
	  }
  break;
case 502:
  case_502();
  break;
case 503:
#line 3617 "cs-parser.jay"
  {
		yyVal = new Argument ((Expression) yyVals[0+yyTop], Argument.AType.Ref);
	  }
  break;
case 504:
  case_504();
  break;
case 505:
#line 3626 "cs-parser.jay"
  {
		yyVal = new Argument ((Expression) yyVals[0+yyTop], Argument.AType.Out);
	  }
  break;
case 506:
  case_506();
  break;
case 507:
  case_507();
  break;
case 508:
  case_508();
  break;
case 509:
  case_509();
  break;
case 510:
  case_510();
  break;
case 512:
  case_512();
  break;
case 513:
  case_513();
  break;
case 514:
  case_514();
  break;
case 515:
  case_515();
  break;
case 516:
  case_516();
  break;
case 517:
  case_517();
  break;
case 518:
  case_518();
  break;
case 519:
  case_519();
  break;
case 520:
#line 3751 "cs-parser.jay"
  {
	  	yyVal = new Argument ((Expression) yyVals[0+yyTop]);
	  }
  break;
case 522:
#line 3759 "cs-parser.jay"
  {
		yyVal = new This (GetLocation (yyVals[0+yyTop]));
	  }
  break;
case 523:
  case_523();
  break;
case 524:
  case_524();
  break;
case 525:
#line 3779 "cs-parser.jay"
  {
		yyVal = new UnaryMutator (UnaryMutator.Mode.PostIncrement, (Expression) yyVals[-1+yyTop], GetLocation (yyVals[0+yyTop]));
	  }
  break;
case 526:
#line 3786 "cs-parser.jay"
  {
		yyVal = new UnaryMutator (UnaryMutator.Mode.PostDecrement, (Expression) yyVals[-1+yyTop], GetLocation (yyVals[0+yyTop]));
	  }
  break;
case 527:
  case_527();
  break;
case 528:
  case_528();
  break;
case 529:
  case_529();
  break;
case 530:
  case_530();
  break;
case 531:
  case_531();
  break;
case 532:
  case_532();
  break;
case 533:
  case_533();
  break;
case 534:
#line 3853 "cs-parser.jay"
  {
		++lexer.parsing_type;
	  }
  break;
case 535:
  case_535();
  break;
case 536:
  case_536();
  break;
case 537:
#line 3875 "cs-parser.jay"
  {
		yyVal = new EmptyCompletion ();
	  }
  break;
case 540:
#line 3884 "cs-parser.jay"
  { yyVal = null; }
  break;
case 542:
  case_542();
  break;
case 543:
  case_543();
  break;
case 544:
#line 3904 "cs-parser.jay"
  {
		yyVal = new EmptyCompletion ();
	  }
  break;
case 545:
#line 3908 "cs-parser.jay"
  {
	  	yyVal = yyVals[-1+yyTop];
	  }
  break;
case 546:
  case_546();
  break;
case 547:
  case_547();
  break;
case 548:
  case_548();
  break;
case 549:
  case_549();
  break;
case 553:
  case_553();
  break;
case 554:
  case_554();
  break;
case 555:
  case_555();
  break;
case 556:
#line 3968 "cs-parser.jay"
  {
		yyVal = 2;
	  }
  break;
case 557:
#line 3972 "cs-parser.jay"
  {
		yyVal = ((int) yyVals[-1+yyTop]) + 1;
	  }
  break;
case 558:
#line 3979 "cs-parser.jay"
  {
		yyVal = null;
	  }
  break;
case 559:
#line 3983 "cs-parser.jay"
  {
		yyVal = yyVals[0+yyTop];
	  }
  break;
case 560:
  case_560();
  break;
case 561:
  case_561();
  break;
case 562:
  case_562();
  break;
case 563:
  case_563();
  break;
case 564:
  case_564();
  break;
case 566:
  case_566();
  break;
case 567:
  case_567();
  break;
case 568:
  case_568();
  break;
case 569:
  case_569();
  break;
case 570:
  case_570();
  break;
case 571:
  case_571();
  break;
case 572:
  case_572();
  break;
case 573:
  case_573();
  break;
case 574:
  case_574();
  break;
case 575:
  case_575();
  break;
case 576:
#line 4115 "cs-parser.jay"
  {
		start_anonymous (false, (ParametersCompiled) yyVals[0+yyTop], false, GetLocation (yyVals[-1+yyTop]));
	  }
  break;
case 577:
#line 4119 "cs-parser.jay"
  {
		yyVal = end_anonymous ((ParametersBlock) yyVals[0+yyTop]);
	  }
  break;
case 578:
#line 4123 "cs-parser.jay"
  {
		start_anonymous (false, (ParametersCompiled) yyVals[0+yyTop], true, GetLocation (yyVals[-2+yyTop]));
	  }
  break;
case 579:
#line 4127 "cs-parser.jay"
  {
		yyVal = end_anonymous ((ParametersBlock) yyVals[0+yyTop]);
	  }
  break;
case 580:
#line 4134 "cs-parser.jay"
  {
		yyVal = ParametersCompiled.Undefined;
	  }
  break;
case 582:
#line 4142 "cs-parser.jay"
  {
	  	valid_param_mod = ParameterModifierType.Ref | ParameterModifierType.Out;
	  }
  break;
case 583:
  case_583();
  break;
case 584:
  case_584();
  break;
case 586:
#line 4166 "cs-parser.jay"
  {
		yyVal = new Unary (Unary.Operator.LogicalNot, (Expression) yyVals[0+yyTop], GetLocation (yyVals[-1+yyTop]));
	  }
  break;
case 587:
#line 4170 "cs-parser.jay"
  {
		yyVal = new Unary (Unary.Operator.OnesComplement, (Expression) yyVals[0+yyTop], GetLocation (yyVals[-1+yyTop]));
	  }
  break;
case 588:
  case_588();
  break;
case 589:
  case_589();
  break;
case 590:
  case_590();
  break;
case 591:
  case_591();
  break;
case 592:
  case_592();
  break;
case 593:
  case_593();
  break;
case 595:
#line 4234 "cs-parser.jay"
  { 
	  	yyVal = new Unary (Unary.Operator.UnaryPlus, (Expression) yyVals[0+yyTop], GetLocation (yyVals[-1+yyTop]));
	  }
  break;
case 596:
#line 4238 "cs-parser.jay"
  { 
		yyVal = new Unary (Unary.Operator.UnaryNegation, (Expression) yyVals[0+yyTop], GetLocation (yyVals[-1+yyTop]));
	  }
  break;
case 597:
#line 4242 "cs-parser.jay"
  {
		yyVal = new UnaryMutator (UnaryMutator.Mode.PreIncrement, (Expression) yyVals[0+yyTop], GetLocation (yyVals[-1+yyTop]));
	  }
  break;
case 598:
#line 4246 "cs-parser.jay"
  {
		yyVal = new UnaryMutator (UnaryMutator.Mode.PreDecrement, (Expression) yyVals[0+yyTop], GetLocation (yyVals[-1+yyTop]));
	  }
  break;
case 599:
#line 4250 "cs-parser.jay"
  {
		yyVal = new Indirection ((Expression) yyVals[0+yyTop], GetLocation (yyVals[-1+yyTop]));
	  }
  break;
case 600:
#line 4254 "cs-parser.jay"
  {
		yyVal = new Unary (Unary.Operator.AddressOf, (Expression) yyVals[0+yyTop], GetLocation (yyVals[-1+yyTop]));
	  }
  break;
case 601:
  case_601();
  break;
case 602:
  case_602();
  break;
case 603:
  case_603();
  break;
case 604:
  case_604();
  break;
case 605:
  case_605();
  break;
case 606:
  case_606();
  break;
case 608:
  case_608();
  break;
case 609:
  case_609();
  break;
case 610:
  case_610();
  break;
case 611:
  case_611();
  break;
case 612:
  case_612();
  break;
case 613:
  case_613();
  break;
case 615:
  case_615();
  break;
case 616:
  case_616();
  break;
case 617:
  case_617();
  break;
case 618:
  case_618();
  break;
case 619:
#line 4362 "cs-parser.jay"
  {
		yyVal = new As ((Expression) yyVals[-2+yyTop], (Expression) yyVals[0+yyTop], GetLocation (yyVals[-1+yyTop]));
	  }
  break;
case 620:
  case_620();
  break;
case 621:
  case_621();
  break;
case 622:
  case_622();
  break;
case 623:
  case_623();
  break;
case 624:
  case_624();
  break;
case 625:
  case_625();
  break;
case 628:
#line 4418 "cs-parser.jay"
  {
		yyVal = new Unary (Unary.Operator.UnaryPlus, (Expression) yyVals[0+yyTop], GetLocation (yyVals[-1+yyTop]));
	  }
  break;
case 629:
#line 4422 "cs-parser.jay"
  {
		yyVal = new Unary (Unary.Operator.UnaryNegation, (Expression) yyVals[0+yyTop], GetLocation (yyVals[-1+yyTop]));
	  }
  break;
case 632:
  case_632();
  break;
case 633:
#line 4433 "cs-parser.jay"
  {
		yyVal = new WildcardPattern (GetLocation (yyVals[0+yyTop]));
	  }
  break;
case 636:
#line 4442 "cs-parser.jay"
  {
		yyVal = new RecursivePattern ((ATypeNameExpression) yyVals[-3+yyTop], (Arguments) yyVals[-1+yyTop], GetLocation (yyVals[-2+yyTop]));
	  }
  break;
case 637:
#line 4449 "cs-parser.jay"
  {
		yyVal = new PropertyPattern ((ATypeNameExpression) yyVals[-3+yyTop], (List<PropertyPatternMember>) yyVals[-1+yyTop], GetLocation (yyVals[-2+yyTop]));
	  }
  break;
case 638:
  case_638();
  break;
case 639:
  case_639();
  break;
case 640:
  case_640();
  break;
case 642:
  case_642();
  break;
case 643:
#line 4491 "cs-parser.jay"
  {
		yyVal = new Arguments (0);
	  }
  break;
case 645:
  case_645();
  break;
case 646:
  case_646();
  break;
case 647:
#line 4517 "cs-parser.jay"
  {
		yyVal = new Argument ((Expression) yyVals[0+yyTop]);
	  }
  break;
case 648:
  case_648();
  break;
case 650:
  case_650();
  break;
case 651:
  case_651();
  break;
case 652:
  case_652();
  break;
case 653:
  case_653();
  break;
case 655:
  case_655();
  break;
case 656:
  case_656();
  break;
case 657:
  case_657();
  break;
case 658:
  case_658();
  break;
case 659:
  case_659();
  break;
case 660:
  case_660();
  break;
case 661:
  case_661();
  break;
case 662:
  case_662();
  break;
case 664:
  case_664();
  break;
case 665:
  case_665();
  break;
case 666:
  case_666();
  break;
case 667:
  case_667();
  break;
case 669:
  case_669();
  break;
case 670:
  case_670();
  break;
case 672:
  case_672();
  break;
case 673:
  case_673();
  break;
case 675:
  case_675();
  break;
case 676:
  case_676();
  break;
case 678:
  case_678();
  break;
case 679:
  case_679();
  break;
case 681:
  case_681();
  break;
case 682:
  case_682();
  break;
case 684:
  case_684();
  break;
case 686:
  case_686();
  break;
case 687:
  case_687();
  break;
case 688:
  case_688();
  break;
case 689:
  case_689();
  break;
case 690:
  case_690();
  break;
case 691:
  case_691();
  break;
case 692:
  case_692();
  break;
case 693:
  case_693();
  break;
case 694:
  case_694();
  break;
case 695:
  case_695();
  break;
case 696:
  case_696();
  break;
case 697:
  case_697();
  break;
case 698:
  case_698();
  break;
case 699:
  case_699();
  break;
case 700:
  case_700();
  break;
case 701:
  case_701();
  break;
case 702:
  case_702();
  break;
case 703:
  case_703();
  break;
case 704:
  case_704();
  break;
case 705:
  case_705();
  break;
case 706:
  case_706();
  break;
case 707:
#line 4860 "cs-parser.jay"
  { yyVal = ParametersCompiled.EmptyReadOnlyParameters; }
  break;
case 708:
  case_708();
  break;
case 709:
#line 4870 "cs-parser.jay"
  {
		start_block (Location.Null);
	  }
  break;
case 710:
  case_710();
  break;
case 712:
  case_712();
  break;
case 714:
  case_714();
  break;
case 715:
  case_715();
  break;
case 716:
  case_716();
  break;
case 717:
  case_717();
  break;
case 718:
  case_718();
  break;
case 719:
  case_719();
  break;
case 720:
  case_720();
  break;
case 721:
#line 4937 "cs-parser.jay"
  {
	  	valid_param_mod = ParameterModifierType.Ref | ParameterModifierType.Out;
	  }
  break;
case 722:
  case_722();
  break;
case 723:
  case_723();
  break;
case 724:
#line 4951 "cs-parser.jay"
  {
	  	valid_param_mod = ParameterModifierType.Ref | ParameterModifierType.Out;	  
	  }
  break;
case 725:
  case_725();
  break;
case 726:
  case_726();
  break;
case 732:
#line 4976 "cs-parser.jay"
  {
		yyVal = new ArglistAccess (GetLocation (yyVals[0+yyTop]));
	  }
  break;
case 733:
  case_733();
  break;
case 734:
  case_734();
  break;
case 735:
  case_735();
  break;
case 737:
#line 5005 "cs-parser.jay"
  {
		yyVal = new BooleanExpression ((Expression) yyVals[0+yyTop]);
	  }
  break;
case 738:
#line 5012 "cs-parser.jay"
  {
	  	yyVal = null;
	  }
  break;
case 740:
  case_740();
  break;
case 741:
#line 5033 "cs-parser.jay"
  {
	  	yyVal = null;
	  }
  break;
case 742:
#line 5037 "cs-parser.jay"
  {
		yyVal = null;
	  }
  break;
case 743:
#line 5041 "cs-parser.jay"
  {
		yyVal = yyVals[0+yyTop];
	  }
  break;
case 744:
#line 5045 "cs-parser.jay"
  {
		yyVal = yyVals[-1+yyTop];
	  }
  break;
case 745:
  case_745();
  break;
case 746:
  case_746();
  break;
case 747:
#line 5070 "cs-parser.jay"
  {
	  }
  break;
case 748:
  case_748();
  break;
case 749:
  case_749();
  break;
case 750:
  case_750();
  break;
case 751:
  case_751();
  break;
case 752:
#line 5122 "cs-parser.jay"
  { yyVal = null; }
  break;
case 753:
#line 5124 "cs-parser.jay"
  { yyVal = yyVals[0+yyTop]; }
  break;
case 754:
  case_754();
  break;
case 755:
#line 5137 "cs-parser.jay"
  {
		lexer.parsing_modifiers = false;		
	  }
  break;
case 757:
  case_757();
  break;
case 758:
  case_758();
  break;
case 759:
  case_759();
  break;
case 760:
  case_760();
  break;
case 761:
  case_761();
  break;
case 762:
  case_762();
  break;
case 763:
  case_763();
  break;
case 764:
  case_764();
  break;
case 765:
  case_765();
  break;
case 766:
  case_766();
  break;
case 767:
  case_767();
  break;
case 768:
  case_768();
  break;
case 769:
  case_769();
  break;
case 770:
  case_770();
  break;
case 771:
  case_771();
  break;
case 772:
  case_772();
  break;
case 775:
#line 5252 "cs-parser.jay"
  {
		current_type.SetBaseTypes ((List<FullNamedExpression>) yyVals[0+yyTop]);
	 }
  break;
case 776:
  case_776();
  break;
case 778:
#line 5266 "cs-parser.jay"
  {
		yyVal = yyVals[0+yyTop];
	  }
  break;
case 779:
  case_779();
  break;
case 780:
  case_780();
  break;
case 781:
  case_781();
  break;
case 782:
  case_782();
  break;
case 783:
  case_783();
  break;
case 784:
  case_784();
  break;
case 785:
  case_785();
  break;
case 786:
  case_786();
  break;
case 787:
#line 5358 "cs-parser.jay"
  {
		yyVal = new SpecialContraintExpr (SpecialConstraint.Class, GetLocation (yyVals[0+yyTop]));
	  }
  break;
case 788:
#line 5362 "cs-parser.jay"
  {
		yyVal = new SpecialContraintExpr (SpecialConstraint.Struct, GetLocation (yyVals[0+yyTop]));
	  }
  break;
case 789:
#line 5369 "cs-parser.jay"
  {
		yyVal = null;
	  }
  break;
case 790:
  case_790();
  break;
case 791:
#line 5383 "cs-parser.jay"
  {
		yyVal = new VarianceDecl (Variance.Covariant, GetLocation (yyVals[0+yyTop]));
	  }
  break;
case 792:
#line 5387 "cs-parser.jay"
  {
		yyVal = new VarianceDecl (Variance.Contravariant, GetLocation (yyVals[0+yyTop]));
	  }
  break;
case 793:
  case_793();
  break;
case 794:
#line 5412 "cs-parser.jay"
  {
		yyVal = yyVals[0+yyTop];
	  }
  break;
case 795:
  case_795();
  break;
case 796:
  case_796();
  break;
case 797:
  case_797();
  break;
case 798:
  case_798();
  break;
case 803:
#line 5456 "cs-parser.jay"
  {
		current_block.AddStatement ((Statement) yyVals[0+yyTop]);
	  }
  break;
case 804:
#line 5460 "cs-parser.jay"
  {
		current_block.AddStatement ((Statement) yyVals[0+yyTop]);
	  }
  break;
case 806:
  case_806();
  break;
case 809:
#line 5484 "cs-parser.jay"
  {
		current_block.AddStatement ((Statement) yyVals[0+yyTop]);
	  }
  break;
case 810:
#line 5488 "cs-parser.jay"
  {
		current_block.AddStatement ((Statement) yyVals[0+yyTop]);
	  }
  break;
case 839:
  case_839();
  break;
case 840:
  case_840();
  break;
case 841:
  case_841();
  break;
case 842:
  case_842();
  break;
case 843:
  case_843();
  break;
case 846:
  case_846();
  break;
case 847:
  case_847();
  break;
case 848:
  case_848();
  break;
case 852:
  case_852();
  break;
case 853:
#line 5629 "cs-parser.jay"
  {
		yyVal = ComposedTypeSpecifier.CreatePointer (GetLocation (yyVals[0+yyTop]));
	  }
  break;
case 855:
#line 5637 "cs-parser.jay"
  {
	  	yyVal = Error_AwaitAsIdentifier (yyVals[0+yyTop]);
	  }
  break;
case 856:
  case_856();
  break;
case 857:
  case_857();
  break;
case 858:
  case_858();
  break;
case 859:
  case_859();
  break;
case 861:
  case_861();
  break;
case 862:
  case_862();
  break;
case 866:
  case_866();
  break;
case 869:
  case_869();
  break;
case 870:
  case_870();
  break;
case 871:
#line 5737 "cs-parser.jay"
  {
		report.Error (145, lexer.Location, "A const field requires a value to be provided");
	  }
  break;
case 872:
#line 5741 "cs-parser.jay"
  {
		current_variable.Initializer = (Expression) yyVals[0+yyTop];
	  }
  break;
case 877:
  case_877();
  break;
case 879:
  case_879();
  break;
case 880:
  case_880();
  break;
case 881:
  case_881();
  break;
case 882:
#line 5786 "cs-parser.jay"
  { yyVal = yyVals[-1+yyTop]; }
  break;
case 883:
  case_883();
  break;
case 884:
#line 5796 "cs-parser.jay"
  { yyVal = yyVals[-1+yyTop]; }
  break;
case 885:
#line 5797 "cs-parser.jay"
  { yyVal = yyVals[-1+yyTop]; }
  break;
case 886:
  case_886();
  break;
case 887:
  case_887();
  break;
case 888:
  case_888();
  break;
case 891:
  case_891();
  break;
case 892:
  case_892();
  break;
case 893:
  case_893();
  break;
case 894:
#line 5869 "cs-parser.jay"
  {
		start_block (GetLocation (yyVals[0+yyTop]));
	  }
  break;
case 895:
  case_895();
  break;
case 896:
  case_896();
  break;
case 897:
#line 5889 "cs-parser.jay"
  {
		report.Warning (1522, 1, current_block.StartLocation, "Empty switch block"); 
	  }
  break;
case 901:
#line 5899 "cs-parser.jay"
  {
		Error_SyntaxError (yyToken);
	  }
  break;
case 903:
  case_903();
  break;
case 904:
#line 5916 "cs-parser.jay"
  {
		current_block.AddStatement ((Statement) yyVals[0+yyTop]);
	  }
  break;
case 905:
  case_905();
  break;
case 906:
  case_906();
  break;
case 907:
  case_907();
  break;
case 908:
#line 5943 "cs-parser.jay"
  {
		yyVal = new SwitchLabel (null, GetLocation (yyVals[0+yyTop]));
	  }
  break;
case 913:
  case_913();
  break;
case 914:
  case_914();
  break;
case 915:
  case_915();
  break;
case 916:
  case_916();
  break;
case 917:
  case_917();
  break;
case 918:
  case_918();
  break;
case 919:
#line 6003 "cs-parser.jay"
  {
		yyVal = yyVals[0+yyTop];
	  }
  break;
case 920:
  case_920();
  break;
case 921:
  case_921();
  break;
case 922:
  case_922();
  break;
case 923:
  case_923();
  break;
case 924:
#line 6044 "cs-parser.jay"
  {
		yyVal = new Tuple<Location,Location> (GetLocation (yyVals[-2+yyTop]), (Location) yyVals[0+yyTop]);
	  }
  break;
case 925:
  case_925();
  break;
case 926:
  case_926();
  break;
case 927:
  case_927();
  break;
case 929:
#line 6072 "cs-parser.jay"
  { lexer.putback ('}'); }
  break;
case 930:
#line 6076 "cs-parser.jay"
  { yyVal = new EmptyStatement (lexer.Location); }
  break;
case 932:
  case_932();
  break;
case 933:
  case_933();
  break;
case 935:
#line 6100 "cs-parser.jay"
  { yyVal = null; }
  break;
case 937:
#line 6105 "cs-parser.jay"
  { yyVal = new EmptyStatement (lexer.Location); }
  break;
case 941:
  case_941();
  break;
case 942:
  case_942();
  break;
case 943:
  case_943();
  break;
case 944:
  case_944();
  break;
case 945:
  case_945();
  break;
case 952:
  case_952();
  break;
case 953:
  case_953();
  break;
case 954:
  case_954();
  break;
case 955:
  case_955();
  break;
case 956:
  case_956();
  break;
case 957:
  case_957();
  break;
case 958:
  case_958();
  break;
case 959:
  case_959();
  break;
case 960:
  case_960();
  break;
case 961:
  case_961();
  break;
case 962:
  case_962();
  break;
case 963:
  case_963();
  break;
case 964:
  case_964();
  break;
case 965:
  case_965();
  break;
case 966:
  case_966();
  break;
case 969:
#line 6329 "cs-parser.jay"
  {
		yyVal = new TryCatch ((Block) yyVals[-1+yyTop], (List<Catch>) yyVals[0+yyTop], GetLocation (yyVals[-2+yyTop]), false);
	  }
  break;
case 970:
  case_970();
  break;
case 971:
  case_971();
  break;
case 972:
  case_972();
  break;
case 973:
  case_973();
  break;
case 974:
  case_974();
  break;
case 977:
  case_977();
  break;
case 978:
  case_978();
  break;
case 979:
  case_979();
  break;
case 980:
  case_980();
  break;
case 981:
  case_981();
  break;
case 983:
  case_983();
  break;
case 984:
#line 6449 "cs-parser.jay"
  {
		yyVal = new Checked ((Block) yyVals[0+yyTop], GetLocation (yyVals[-1+yyTop]));
	  }
  break;
case 985:
#line 6456 "cs-parser.jay"
  {
		yyVal = new Unchecked ((Block) yyVals[0+yyTop], GetLocation (yyVals[-1+yyTop]));
	  }
  break;
case 986:
  case_986();
  break;
case 987:
#line 6466 "cs-parser.jay"
  {
		yyVal = new Unsafe ((Block) yyVals[0+yyTop], GetLocation (yyVals[-2+yyTop]));
	  }
  break;
case 988:
  case_988();
  break;
case 989:
  case_989();
  break;
case 990:
  case_990();
  break;
case 991:
  case_991();
  break;
case 992:
  case_992();
  break;
case 993:
  case_993();
  break;
case 994:
  case_994();
  break;
case 995:
  case_995();
  break;
case 996:
  case_996();
  break;
case 997:
  case_997();
  break;
case 999:
  case_999();
  break;
case 1000:
#line 6570 "cs-parser.jay"
  {
		Error_MissingInitializer (lexer.Location);
	  }
  break;
case 1001:
  case_1001();
  break;
case 1002:
  case_1002();
  break;
case 1003:
  case_1003();
  break;
case 1004:
  case_1004();
  break;
case 1005:
  case_1005();
  break;
case 1006:
  case_1006();
  break;
case 1007:
  case_1007();
  break;
case 1008:
  case_1008();
  break;
case 1009:
  case_1009();
  break;
case 1010:
#line 6674 "cs-parser.jay"
  {
		current_block = new Linq.QueryBlock (current_block, lexer.Location);
	  }
  break;
case 1011:
  case_1011();
  break;
case 1012:
#line 6690 "cs-parser.jay"
  {
		current_block = new Linq.QueryBlock (current_block, lexer.Location);
	  }
  break;
case 1013:
  case_1013();
  break;
case 1014:
  case_1014();
  break;
case 1015:
  case_1015();
  break;
case 1017:
  case_1017();
  break;
case 1018:
  case_1018();
  break;
case 1019:
#line 6754 "cs-parser.jay"
  {
	  	current_block = new Linq.QueryBlock (current_block, lexer.Location);
	  }
  break;
case 1020:
  case_1020();
  break;
case 1021:
  case_1021();
  break;
case 1022:
  case_1022();
  break;
case 1023:
  case_1023();
  break;
case 1024:
#line 6793 "cs-parser.jay"
  {
	  	yyVal = new object[] { yyVals[0+yyTop], GetLocation (yyVals[-1+yyTop]) };
	  }
  break;
case 1025:
  case_1025();
  break;
case 1027:
  case_1027();
  break;
case 1033:
#line 6822 "cs-parser.jay"
  {
	  	current_block = new Linq.QueryBlock (current_block, lexer.Location);
	  }
  break;
case 1034:
  case_1034();
  break;
case 1035:
#line 6841 "cs-parser.jay"
  {
	  	current_block = new Linq.QueryBlock (current_block, lexer.Location);
	  }
  break;
case 1036:
  case_1036();
  break;
case 1037:
  case_1037();
  break;
case 1038:
  case_1038();
  break;
case 1039:
  case_1039();
  break;
case 1040:
  case_1040();
  break;
case 1041:
  case_1041();
  break;
case 1042:
  case_1042();
  break;
case 1043:
  case_1043();
  break;
case 1044:
  case_1044();
  break;
case 1046:
#line 6986 "cs-parser.jay"
  {
		yyVal = yyVals[0+yyTop];
	  }
  break;
case 1047:
#line 6993 "cs-parser.jay"
  {
		current_block = new Linq.QueryBlock (current_block, lexer.Location);
	  }
  break;
case 1048:
  case_1048();
  break;
case 1050:
  case_1050();
  break;
case 1051:
  case_1051();
  break;
case 1053:
  case_1053();
  break;
case 1054:
  case_1054();
  break;
case 1055:
#line 7039 "cs-parser.jay"
  {
		yyVal = new Linq.OrderByAscending ((Linq.QueryBlock) current_block, (Expression)yyVals[0+yyTop]);	
	  }
  break;
case 1056:
  case_1056();
  break;
case 1057:
  case_1057();
  break;
case 1058:
#line 7056 "cs-parser.jay"
  {
		yyVal = new Linq.ThenByAscending ((Linq.QueryBlock) current_block, (Expression)yyVals[0+yyTop]);	
	  }
  break;
case 1059:
  case_1059();
  break;
case 1060:
  case_1060();
  break;
case 1062:
  case_1062();
  break;
case 1063:
  case_1063();
  break;
case 1066:
  case_1066();
  break;
case 1067:
  case_1067();
  break;
case 1075:
#line 7181 "cs-parser.jay"
  {
		module.DocumentationBuilder.ParsedName = (MemberName) yyVals[0+yyTop];
	  }
  break;
case 1076:
#line 7188 "cs-parser.jay"
  {
		module.DocumentationBuilder.ParsedParameters = (List<DocumentationParameter>)yyVals[0+yyTop];
	  }
  break;
case 1077:
  case_1077();
  break;
case 1078:
  case_1078();
  break;
case 1079:
  case_1079();
  break;
case 1080:
#line 7211 "cs-parser.jay"
  {
		yyVal = new MemberName ((MemberName) yyVals[-2+yyTop], MemberCache.IndexerNameAlias, Location.Null);
	  }
  break;
case 1081:
#line 7215 "cs-parser.jay"
  {
		valid_param_mod = ParameterModifierType.Ref | ParameterModifierType.Out;
	  }
  break;
case 1082:
  case_1082();
  break;
case 1083:
  case_1083();
  break;
case 1084:
  case_1084();
  break;
case 1085:
  case_1085();
  break;
case 1087:
#line 7251 "cs-parser.jay"
  {
		yyVal = new MemberName (((MemberName) yyVals[-2+yyTop]), (MemberName) yyVals[0+yyTop]);
	  }
  break;
case 1089:
#line 7259 "cs-parser.jay"
  {
		valid_param_mod = ParameterModifierType.Ref | ParameterModifierType.Out;
	  }
  break;
case 1090:
#line 7263 "cs-parser.jay"
  {
		yyVal = yyVals[-1+yyTop];
	  }
  break;
case 1091:
#line 7270 "cs-parser.jay"
  {
		yyVal = new List<DocumentationParameter> (0);
	  }
  break;
case 1093:
  case_1093();
  break;
case 1094:
  case_1094();
  break;
case 1095:
  case_1095();
  break;
#line default
        }
        yyTop -= yyLen[yyN];
        yyState = yyStates[yyTop];
        int yyM = yyLhs[yyN];
        if (yyState == 0 && yyM == 0) {
//t          if (debug != null) debug.shift(0, yyFinal);
          yyState = yyFinal;
          if (yyToken < 0) {
            yyToken = yyLex.advance() ? yyLex.token() : 0;
//t            if (debug != null)
//t               debug.lex(yyState, yyToken,yyname(yyToken), yyLex.value());
          }
          if (yyToken == 0) {
//t            if (debug != null) debug.accept(yyVal);
            return yyVal;
          }
          goto continue_yyLoop;
        }
        if (((yyN = yyGindex[yyM]) != 0) && ((yyN += yyState) >= 0)
            && (yyN < yyTable.Length) && (yyCheck[yyN] == yyState))
          yyState = yyTable[yyN];
        else
          yyState = yyDgoto[yyM];
//t        if (debug != null) debug.shift(yyStates[yyTop], yyState);
	 goto continue_yyLoop;
      continue_yyDiscarded: ;	// implements the named-loop continue: 'continue yyDiscarded'
      }
    continue_yyLoop: ;		// implements the named-loop continue: 'continue yyLoop'
    }
  }

/*
 All more than 3 lines long rules are wrapped into a method
*/
void case_6()
#line 389 "cs-parser.jay"
{
		if (yyVals[0+yyTop] != null) {
			Attributes attrs = (Attributes) yyVals[0+yyTop];
			report.Error (1730, attrs.Attrs [0].Location,
				"Assembly and module attributes must precede all other elements except using clauses and extern alias declarations");

			current_namespace.UnattachedAttributes = attrs;
		}
	  }

void case_8()
#line 403 "cs-parser.jay"
{
		if (yyToken == Token.EXTERN_ALIAS)
			report.Error (439, lexer.Location, "An extern alias declaration must precede all other elements");
		else
			Error_SyntaxError (yyToken);
	  }

void case_13()
#line 423 "cs-parser.jay"
{
		var lt = (LocatedToken) yyVals[-2+yyTop];
		string s = lt.Value;
		if (s != "alias") {
			syntax_error (lt.Location, "`alias' expected");
		} else {
			if (lang_version == LanguageVersion.ISO_1)
				FeatureIsNotAvailable (lt.Location, "external alias");

			lt = (LocatedToken) yyVals[-1+yyTop];
			if (lt.Value == QualifiedAliasMember.GlobalAlias) {
				RootNamespace.Error_GlobalNamespaceRedefined (report, lt.Location);
			}
			
			var na = new UsingExternAlias (new SimpleMemberName (lt.Value, lt.Location), GetLocation (yyVals[-3+yyTop]));
			current_namespace.AddUsing (na);
			
			lbag.AddLocation (na, GetLocation (yyVals[-2+yyTop]), GetLocation (yyVals[0+yyTop]));
		}
	  }

void case_17()
#line 456 "cs-parser.jay"
{
		if (doc_support)
			Lexer.doc_state = XmlCommentState.Allowed;
	  }

void case_18()
#line 464 "cs-parser.jay"
{
		var un = new UsingNamespace ((ATypeNameExpression) yyVals[-1+yyTop], GetLocation (yyVals[-2+yyTop]));
		current_namespace.AddUsing (un);
		
		lbag.AddLocation (un, GetLocation (yyVals[0+yyTop]));
	  }

void case_19()
#line 471 "cs-parser.jay"
{
		var lt = (LocatedToken) yyVals[-3+yyTop];
		if (lang_version != LanguageVersion.ISO_1 && lt.Value == "global") {
			report.Warning (440, 2, lt.Location,
			 "An alias named `global' will not be used when resolving `global::'. The global namespace will be used instead");
		}

		var un = new UsingAliasNamespace (new SimpleMemberName (lt.Value, lt.Location), (ATypeNameExpression) yyVals[-1+yyTop], GetLocation (yyVals[-4+yyTop]));
		current_namespace.AddUsing (un);
		
		lbag.AddLocation (un, GetLocation (yyVals[-2+yyTop]), GetLocation (yyVals[0+yyTop]));
	  }

void case_20()
#line 484 "cs-parser.jay"
{
		Error_SyntaxError (yyToken);
		yyVal = null;
	 }

void case_21()
#line 497 "cs-parser.jay"
{
		Attributes attrs = (Attributes) yyVals[-2+yyTop];
		var name = (MemberName) yyVals[0+yyTop];
		if (attrs != null) {
			bool valid_global_attrs = true;
			if ((current_namespace.DeclarationFound || current_namespace != file)) {
				valid_global_attrs = false;
			} else {
				foreach (var a in attrs.Attrs) {
					if (a.ExplicitTarget == "assembly" || a.ExplicitTarget == "module")
						continue;
						
					valid_global_attrs = false;
					break;
				}
			}
			
			if (!valid_global_attrs)
				report.Error (1671, name.Location, "A namespace declaration cannot have modifiers or attributes");
		}
	
		module.AddAttributes (attrs, current_namespace);
		
		var ns = new NamespaceContainer (name, current_namespace);
		current_namespace.AddTypeContainer (ns);
		current_container = current_namespace = ns;
	  }

void case_22()
#line 525 "cs-parser.jay"
{
		if (doc_support)
			Lexer.doc_state = XmlCommentState.Allowed;
	  }

void case_23()
#line 530 "cs-parser.jay"
{
		if (yyVals[0+yyTop] != null)
			lbag.AddLocation (current_container, GetLocation (yyVals[-9+yyTop]), GetLocation (yyVals[-6+yyTop]), GetLocation (yyVals[-1+yyTop]), GetLocation (yyVals[0+yyTop]));
		else
			lbag.AddLocation (current_container, GetLocation (yyVals[-9+yyTop]), GetLocation (yyVals[-6+yyTop]), GetLocation (yyVals[-1+yyTop]));
	  
		current_container = current_namespace = current_namespace.Parent;
	  }

void case_24()
#line 539 "cs-parser.jay"
{
		report.Error (1514, lexer.Location, "Unexpected symbol `{0}', expecting `.' or `{{'", GetSymbolName (yyToken));

		var name = (MemberName) yyVals[0+yyTop];		
		var ns = new NamespaceContainer (name, current_namespace);
		lbag.AddLocation (ns, GetLocation (yyVals[-1+yyTop]));
		current_namespace.AddTypeContainer (ns);
	  }

void case_27()
#line 553 "cs-parser.jay"
{
		Error_SyntaxError (yyToken);
	  	yyVal = null;
	  }

void case_28()
#line 561 "cs-parser.jay"
{
		var lt = (LocatedToken) yyVals[0+yyTop];
		yyVal = new MemberName (lt.Value, lt.Location);
	  }

void case_29()
#line 566 "cs-parser.jay"
{
		var lt = (LocatedToken) yyVals[0+yyTop];
		yyVal = new MemberName ((MemberName) yyVals[-2+yyTop], lt.Value, lt.Location);		
		lbag.AddLocation (yyVal, GetLocation (yyVals[-1+yyTop]));
	  }

void case_30()
#line 572 "cs-parser.jay"
{
		Error_SyntaxError (yyToken);
		yyVal = new MemberName ("<invalid>", lexer.Location);
	  }

void case_43()
#line 610 "cs-parser.jay"
{
		if (yyVals[0+yyTop] != null) {
			TypeContainer ds = (TypeContainer)yyVals[0+yyTop];

			if ((ds.ModFlags & (Modifiers.PRIVATE | Modifiers.PROTECTED)) != 0){
				report.Error (1527, ds.Location, 
				"Namespace elements cannot be explicitly declared as private, protected or protected internal");
			}

			/* Here is a trick, for explicit attributes we don't know where they belong to until*/
			/* we parse succeeding declaration hence we parse them as normal and re-attach them*/
			/* when we know whether they are global (assembly:, module:) or local (type:).*/
			if (ds.OptAttributes != null) {
				ds.OptAttributes.ConvertGlobalAttributes (ds, current_namespace, !current_namespace.DeclarationFound && current_namespace == file);
			}
		}
		current_namespace.DeclarationFound = true;
	  }

void case_45()
#line 632 "cs-parser.jay"
{
		current_namespace.UnattachedAttributes = (Attributes) yyVals[-1+yyTop];
		report.Error (1518, lexer.Location, "Attributes must be attached to class, delegate, enum, interface or struct");
		lexer.putback ('}');
	  }

void case_53()
#line 665 "cs-parser.jay"
{
		var sect = (List<Attribute>) yyVals[0+yyTop];
		yyVal = new Attributes (sect);
	  }

void case_54()
#line 670 "cs-parser.jay"
{
		Attributes attrs = yyVals[-1+yyTop] as Attributes;
		var sect = (List<Attribute>) yyVals[0+yyTop];
		if (attrs == null)
			attrs = new Attributes (sect);
		else if (sect != null)
			attrs.AddAttributes (sect);
		yyVal = attrs;
	  }

void case_55()
#line 683 "cs-parser.jay"
{
	  	PushLocation (GetLocation (yyVals[0+yyTop]));
		lexer.parsing_attribute_section = true;
	  }

void case_56()
#line 688 "cs-parser.jay"
{
		lexer.parsing_attribute_section = false;
		yyVal = yyVals[0+yyTop];
	  }

void case_57()
#line 696 "cs-parser.jay"
{
		current_attr_target = (string) yyVals[-1+yyTop];
		if (current_attr_target == "assembly" || current_attr_target == "module") {
			Lexer.check_incorrect_doc_comment ();
		}
	  }

void case_58()
#line 703 "cs-parser.jay"
{
		/* when attribute target is invalid*/
		if (current_attr_target == string.Empty)
			yyVal = new List<Attribute> (0);
		else
			yyVal = yyVals[-2+yyTop];

		lbag.InsertLocation (yyVal, 0, PopLocation ());
		if (yyVals[-1+yyTop] != null) {
			lbag.AddLocation (yyVal, GetLocation (yyVals[-4+yyTop]), GetLocation (yyVals[-1+yyTop]), GetLocation (yyVals[0+yyTop]));
		} else {
			lbag.AddLocation (yyVal, GetLocation (yyVals[-4+yyTop]), GetLocation (yyVals[0+yyTop]));
		}

		current_attr_target = null;
		lexer.parsing_attribute_section = false;
 	  }

void case_59()
#line 721 "cs-parser.jay"
{
		yyVal = yyVals[-2+yyTop];

		lbag.InsertLocation (yyVal, 0, PopLocation ());
		if (yyVals[-1+yyTop] != null) {
			lbag.AddLocation (yyVal, GetLocation(yyVals[-1+yyTop]), GetLocation (yyVals[0+yyTop]));
		} else {
			lbag.AddLocation (yyVal, GetLocation(yyVals[0+yyTop]));
		}
	  }

void case_60()
#line 732 "cs-parser.jay"
{
		Error_SyntaxError (yyToken);

		var lt = (LocatedToken) yyVals[-1+yyTop];
		var tne = new SimpleName (lt.Value, null, lt.Location);

		yyVal = new List<Attribute> () {
			new Attribute (null, tne, null, GetLocation (yyVals[-1+yyTop]), false)
		};
	  }

void case_61()
#line 743 "cs-parser.jay"
{
 		CheckAttributeTarget (yyToken, GetTokenName (yyToken), GetLocation (yyVals[0+yyTop])); 
 		yyVal = null;
	  }

void case_62()
#line 751 "cs-parser.jay"
{
		var lt = (LocatedToken) yyVals[0+yyTop];
		yyVal = CheckAttributeTarget (yyToken, lt.Value, lt.Location);
	  }

void case_66()
#line 765 "cs-parser.jay"
{
		var attrs = (List<Attribute>) yyVals[-2+yyTop];
		if (attrs != null) {
			attrs.Add ((Attribute) yyVals[0+yyTop]);
	  		lbag.AppendTo (attrs, GetLocation (yyVals[-1+yyTop]));
	  	}

		yyVal = attrs;
	  }

void case_68()
#line 782 "cs-parser.jay"
{
		--lexer.parsing_block;
		
		var tne = (ATypeNameExpression) yyVals[-2+yyTop];
		if (tne.HasTypeArguments) {
			report.Error (404, tne.Location, "Attributes cannot be generic");
		}

		yyVal = new Attribute (current_attr_target, tne, (Arguments[]) yyVals[0+yyTop], GetLocation (yyVals[-2+yyTop]), lexer.IsEscapedIdentifier (tne));
	  }

void case_73()
#line 810 "cs-parser.jay"
{
	  	Arguments a = new Arguments (4);
		a.Add ((Argument) yyVals[0+yyTop]);
		yyVal = new Arguments [] { a, null };
	  }

void case_74()
#line 816 "cs-parser.jay"
{
	  	Arguments a = new Arguments (4);
		a.Add ((Argument) yyVals[0+yyTop]);  
		yyVal = new Arguments [] { null, a };
	  }

void case_75()
#line 822 "cs-parser.jay"
{
		Arguments[] o = (Arguments[]) yyVals[-2+yyTop];
		if (o [1] != null) {
			report.Error (1016, ((Argument) yyVals[0+yyTop]).Expr.Location, "Named attribute arguments must appear after the positional arguments");
			o [0] = new Arguments (4);
		}
		
		Arguments args = ((Arguments) o [0]);
		if (args.Count > 0 && !(yyVals[0+yyTop] is NamedArgument) && args [args.Count - 1] is NamedArgument)
			Error_NamedArgumentExpected ((NamedArgument) args [args.Count - 1]);
		
		args.Add ((Argument) yyVals[0+yyTop]);
	  }

void case_76()
#line 836 "cs-parser.jay"
{
		Arguments[] o = (Arguments[]) yyVals[-2+yyTop];
		if (o [1] == null) {
			o [1] = new Arguments (4);
		}

		((Arguments) o [1]).Add ((Argument) yyVals[0+yyTop]);
	  }

void case_79()
#line 853 "cs-parser.jay"
{
		Error_SyntaxError (yyToken);
		yyVal = null;
	  }

void case_81()
#line 865 "cs-parser.jay"
{
	  	--lexer.parsing_block;
	  	var lt = (LocatedToken) yyVals[-3+yyTop];
		yyVal = new NamedArgument (lt.Value, lt.Location, (Expression) yyVals[0+yyTop]);	  
		lbag.AddLocation (yyVal, GetLocation(yyVals[-2+yyTop]));
	  }

void case_82()
#line 875 "cs-parser.jay"
{
		if (lang_version <= LanguageVersion.V_3)
			FeatureIsNotAvailable (GetLocation (yyVals[-3+yyTop]), "named argument");
			
		/* Avoid boxing in common case (no modifier)*/
		var arg_mod = yyVals[-1+yyTop] == null ? Argument.AType.None : (Argument.AType) yyVals[-1+yyTop];
			
		var lt = (LocatedToken) yyVals[-3+yyTop];
		yyVal = new NamedArgument (lt.Value, lt.Location, (Expression) yyVals[0+yyTop], arg_mod);
		lbag.AddLocation (yyVal, GetLocation(yyVals[-2+yyTop]));
	  }

void case_89()
#line 912 "cs-parser.jay"
{
		lexer.parsing_modifiers = true;
		lexer.parsing_block = 0;
	  }

void case_90()
#line 917 "cs-parser.jay"
{
		lexer.parsing_modifiers = true;
		lexer.parsing_block = 0;
	  }

void case_104()
#line 938 "cs-parser.jay"
{
		report.Error (1519, lexer.Location, "Unexpected symbol `{0}' in class, struct, or interface member declaration",
			GetSymbolName (yyToken));
		yyVal = null;
		lexer.parsing_generic_declaration = false;
	  }

void case_105()
#line 948 "cs-parser.jay"
{
		current_local_parameters = current_type.PrimaryConstructorParameters;
		if (current_local_parameters == null) {
			report.Error (9010, GetLocation (yyVals[0+yyTop]), "Primary constructor body is not allowed");
			current_local_parameters = ParametersCompiled.EmptyReadOnlyParameters;
		}

		++lexer.parsing_block;
		start_block (GetLocation (yyVals[0+yyTop]));
	  }

void case_106()
#line 959 "cs-parser.jay"
{
		current_local_parameters = null;
		var t = current_type as ClassOrStruct;
		if (t != null) {
			var b = (ToplevelBlock) yyVals[0+yyTop];
			if (t.PrimaryConstructorBlock != null) {
				report.Error (8041, b.StartLocation, "Primary constructor already has a body");
			} else {
				t.PrimaryConstructorBlock = b;
			}
		}
	  }

void case_108()
#line 981 "cs-parser.jay"
{ 
		lexer.ConstraintsParsing = true;
		valid_param_mod = ParameterModifierType.PrimaryConstructor;
		push_current_container (new Struct (current_container, (MemberName) yyVals[0+yyTop], (Modifiers) yyVals[-4+yyTop], (Attributes) yyVals[-5+yyTop]), yyVals[-3+yyTop]);
	  }

void case_109()
#line 989 "cs-parser.jay"
{
		valid_param_mod = 0;
		lexer.ConstraintsParsing = false;

		if (yyVals[-2+yyTop] != null)
			current_type.PrimaryConstructorParameters = (ParametersCompiled) yyVals[-2+yyTop];

		if (yyVals[0+yyTop] != null)
			current_container.SetConstraints ((List<Constraints>) yyVals[0+yyTop]);

		if (doc_support)
			current_container.PartialContainer.DocComment = Lexer.consume_doc_comment ();

		lbag.AddMember (current_container, mod_locations, GetLocation (yyVals[-6+yyTop]));
		
		lexer.parsing_modifiers = true;
	  }

void case_110()
#line 1007 "cs-parser.jay"
{
		if (doc_support)
			Lexer.doc_state = XmlCommentState.Allowed;
	  }

void case_111()
#line 1012 "cs-parser.jay"
{
		--lexer.parsing_declaration;
		if (doc_support)
			Lexer.doc_state = XmlCommentState.Allowed;
	  }

void case_112()
#line 1018 "cs-parser.jay"
{
		if (yyVals[-1+yyTop] == null) {
			lbag.AppendToMember (current_container, GetLocation (yyVals[-5+yyTop]), GetLocation (yyVals[-2+yyTop]));
		} else {
			lbag.AppendToMember (current_container, GetLocation (yyVals[-5+yyTop]), GetLocation (yyVals[-2+yyTop]), GetLocation (yyVals[0+yyTop]));
		}
		yyVal = pop_current_class ();
	  }

void case_114()
#line 1036 "cs-parser.jay"
{
		var lt = (LocatedToken) yyVals[0+yyTop];
		var mod = (Modifiers) yyVals[-3+yyTop];
		current_field = new Const (current_type, (FullNamedExpression) yyVals[-1+yyTop], mod, new MemberName (lt.Value, lt.Location), (Attributes) yyVals[-4+yyTop]);
		current_type.AddMember (current_field);
		
		if ((mod & Modifiers.STATIC) != 0) {
			report.Error (504, current_field.Location, "The constant `{0}' cannot be marked static", current_field.GetSignatureForError ());
		}
		
		yyVal = current_field;
	  }

void case_115()
#line 1049 "cs-parser.jay"
{
		if (doc_support) {
			current_field.DocComment = Lexer.consume_doc_comment ();
			Lexer.doc_state = XmlCommentState.Allowed;
		}
		
		current_field.Initializer = (ConstInitializer) yyVals[-2+yyTop];
		lbag.AddMember (current_field, mod_locations, GetLocation (yyVals[-6+yyTop]), GetLocation (yyVals[0+yyTop]));
		current_field = null;
	  }

void case_116()
#line 1062 "cs-parser.jay"
{
		Error_SyntaxError (yyToken);

		current_type.AddMember (new Const (current_type, (FullNamedExpression) yyVals[-1+yyTop], (Modifiers) yyVals[-3+yyTop], MemberName.Null, (Attributes) yyVals[-4+yyTop]));
	  }

void case_121()
#line 1087 "cs-parser.jay"
{
		var lt = (LocatedToken) yyVals[-1+yyTop];
	  	yyVal = new FieldDeclarator (new SimpleMemberName (lt.Value, lt.Location), (ConstInitializer) yyVals[0+yyTop]);
	  	lbag.AddLocation (yyVal, GetLocation (yyVals[-2+yyTop]));
	  }

void case_123()
#line 1100 "cs-parser.jay"
{
		--lexer.parsing_block;
		yyVal = new ConstInitializer (current_field, (Expression) yyVals[0+yyTop], GetLocation (yyVals[-2+yyTop]));
	  }

void case_124()
#line 1105 "cs-parser.jay"
{
		report.Error (145, lexer.Location, "A const field requires a value to be provided");
		yyVal = null;
	  }

void case_127()
#line 1120 "cs-parser.jay"
{
	  	lexer.parsing_generic_declaration = false;

		FullNamedExpression type = (FullNamedExpression) yyVals[-1+yyTop];
		if (type.Type != null && type.Type.Kind == MemberKind.Void)
			report.Error (670, GetLocation (yyVals[-1+yyTop]), "Fields cannot have void type");
			
		var lt = (LocatedToken) yyVals[0+yyTop];
		current_field = new Field (current_type, type, (Modifiers) yyVals[-2+yyTop], new MemberName (lt.Value, lt.Location), (Attributes) yyVals[-3+yyTop]);
		current_type.AddField (current_field);
		yyVal = current_field;
	  }

void case_128()
#line 1135 "cs-parser.jay"
{ 
		if (doc_support) {
			current_field.DocComment = Lexer.consume_doc_comment ();
			Lexer.doc_state = XmlCommentState.Allowed;
		}
			
		lbag.AddMember (current_field, mod_locations, GetLocation (yyVals[0+yyTop]));
		yyVal = current_field;
		current_field = null;
	  }

void case_129()
#line 1148 "cs-parser.jay"
{ 
		if (lang_version < LanguageVersion.ISO_2)
			FeatureIsNotAvailable (GetLocation (yyVals[-2+yyTop]), "fixed size buffers");

		var lt = (LocatedToken) yyVals[0+yyTop];
		current_field = new FixedField (current_type, (FullNamedExpression) yyVals[-1+yyTop], (Modifiers) yyVals[-3+yyTop],
			new MemberName (lt.Value, lt.Location), (Attributes) yyVals[-4+yyTop]);
			
		current_type.AddField (current_field);
	  }

void case_130()
#line 1159 "cs-parser.jay"
{
		if (doc_support) {
			current_field.DocComment = Lexer.consume_doc_comment ();
			Lexer.doc_state = XmlCommentState.Allowed;
	    }

		current_field.Initializer = (ConstInitializer) yyVals[-2+yyTop];	    
		lbag.AddMember (current_field, mod_locations, GetLocation (yyVals[0+yyTop]));
		yyVal = current_field;
	    current_field = null;
	  }

void case_133()
#line 1182 "cs-parser.jay"
{
	  	++lexer.parsing_block;
		current_local_parameters = ParametersCompiled.EmptyReadOnlyParameters;
	  	start_block (GetLocation (yyVals[0+yyTop]));
	  }

void case_134()
#line 1188 "cs-parser.jay"
{
	  	--lexer.parsing_block;
		current_field.Initializer = (Expression) yyVals[0+yyTop];
		lbag.AppendToMember (current_field, GetLocation (yyVals[-2+yyTop]));
		end_block (lexer.Location);
		current_local_parameters = null;
	  }

void case_139()
#line 1215 "cs-parser.jay"
{
		var lt = (LocatedToken) yyVals[0+yyTop];
	  	yyVal = new FieldDeclarator (new SimpleMemberName (lt.Value, lt.Location), null);
	  	lbag.AddLocation (yyVal, GetLocation (yyVals[-1+yyTop]));
	  }

void case_141()
#line 1225 "cs-parser.jay"
{
		--lexer.parsing_block;
		var lt = (LocatedToken) yyVals[-3+yyTop];	  
	  	yyVal = new FieldDeclarator (new SimpleMemberName (lt.Value, lt.Location), (Expression) yyVals[0+yyTop]);
	  	lbag.AddLocation (yyVal, GetLocation (yyVals[-4+yyTop]), GetLocation (yyVals[-2+yyTop]));
	  }

void case_146()
#line 1251 "cs-parser.jay"
{
		var lt = (LocatedToken) yyVals[-1+yyTop];	  
		yyVal = new FieldDeclarator (new SimpleMemberName (lt.Value, lt.Location), (ConstInitializer) yyVals[0+yyTop]);
		lbag.AddLocation (yyVal, GetLocation (yyVals[-2+yyTop]));
	  }

void case_148()
#line 1264 "cs-parser.jay"
{
		--lexer.parsing_block;
		yyVal = new ConstInitializer (current_field, (Expression) yyVals[-1+yyTop], GetLocation (yyVals[-3+yyTop]));
		lbag.AddLocation (yyVal, GetLocation (yyVals[0+yyTop]));
	  }

void case_149()
#line 1270 "cs-parser.jay"
{
		report.Error (443, lexer.Location, "Value or constant expected");
		yyVal = null;
	  }

void case_152()
#line 1280 "cs-parser.jay"
{
		/* It has to be here for the parent to safely restore artificial block*/
	  	Error_SyntaxError (yyToken);
	  	yyVal = null;
	  }

void case_153()
#line 1289 "cs-parser.jay"
{
		if (doc_support)
			Lexer.doc_state = XmlCommentState.NotAllowed;

		/* Was added earlier in the case of body being eof for full ast*/
	  }

void case_154()
#line 1296 "cs-parser.jay"
{
		Method method = (Method) yyVals[-2+yyTop];
		method.Block = (ToplevelBlock) yyVals[0+yyTop];
		async_block = false;
		
		if (method.Block == null) {
			method.ParameterInfo.CheckParameters (method);

			if ((method.ModFlags & Modifiers.ASYNC) != 0) {
				report.Error (1994, method.Location, "`{0}': The async modifier can only be used with methods that have a body",
					method.GetSignatureForError ());
			}
		} else {
			if (current_container.Kind == MemberKind.Interface) {
				report.Error (531, method.Location, "`{0}': interface members cannot have a definition",
					method.GetSignatureForError ());
			}
		}

		current_local_parameters = null;

		if (doc_support)
			Lexer.doc_state = XmlCommentState.Allowed;
	  }

void case_156()
#line 1331 "cs-parser.jay"
{
		valid_param_mod = 0;
		MemberName name = (MemberName) yyVals[-4+yyTop];
		current_local_parameters = (ParametersCompiled) yyVals[-1+yyTop];

		var method = Method.Create (current_type, (FullNamedExpression) yyVals[-5+yyTop], (Modifiers) yyVals[-6+yyTop],
				     name, current_local_parameters, (Attributes) yyVals[-7+yyTop]);

		current_type.AddMember (method);

		async_block = (method.ModFlags & Modifiers.ASYNC) != 0;

		if (doc_support)
			method.DocComment = Lexer.consume_doc_comment ();

		lbag.AddMember (method, mod_locations, GetLocation (yyVals[-3+yyTop]), GetLocation (yyVals[0+yyTop]));

		yyVal = method;

		lexer.ConstraintsParsing = true;
	  }

void case_157()
#line 1353 "cs-parser.jay"
{
		lexer.ConstraintsParsing = false;

		if (yyVals[0+yyTop] != null) {
			var method = (Method) yyVals[-1+yyTop];
			method.SetConstraints ((List<Constraints>) yyVals[0+yyTop]);
		}

		yyVal = yyVals[-1+yyTop];
	  }

void case_159()
#line 1372 "cs-parser.jay"
{
		lexer.parsing_generic_declaration = false;
	  	valid_param_mod = ParameterModifierType.All;
	  }

void case_161()
#line 1381 "cs-parser.jay"
{
		lexer.ConstraintsParsing = false;
		valid_param_mod = 0;

		MemberName name = (MemberName) yyVals[-6+yyTop];
		current_local_parameters = (ParametersCompiled) yyVals[-3+yyTop];

		var modifiers = (Modifiers) yyVals[-10+yyTop];
		modifiers |= Modifiers.PARTIAL;

		var method = Method.Create (current_type, new TypeExpression (compiler.BuiltinTypes.Void, GetLocation (yyVals[-8+yyTop])),
				     modifiers, name, current_local_parameters, (Attributes) yyVals[-11+yyTop]);

		current_type.AddMember (method);

		async_block = (method.ModFlags & Modifiers.ASYNC) != 0;

		if (yyVals[0+yyTop] != null)
			method.SetConstraints ((List<Constraints>) yyVals[0+yyTop]);

		if (doc_support)
			method.DocComment = Lexer.consume_doc_comment ();

		StoreModifierLocation (Modifiers.PARTIAL, GetLocation (yyVals[-9+yyTop]));
		lbag.AddMember (method, mod_locations, GetLocation (yyVals[-5+yyTop]), GetLocation (yyVals[-2+yyTop]));
		yyVal = method;
	  }

void case_162()
#line 1412 "cs-parser.jay"
{
		MemberName name = (MemberName) yyVals[-3+yyTop];
		report.Error (1585, name.Location, 
			"Member modifier `{0}' must precede the member type and name", ModifiersExtensions.Name ((Modifiers) yyVals[-4+yyTop]));

		var method = Method.Create (current_type, (FullNamedExpression) yyVals[-5+yyTop],
					    0, name, (ParametersCompiled) yyVals[-1+yyTop], (Attributes) yyVals[-7+yyTop]);

		current_type.AddMember (method);

		current_local_parameters = (ParametersCompiled) yyVals[-1+yyTop];

		if (doc_support)
			method.DocComment = Lexer.consume_doc_comment ();

		yyVal = method;
	  }

void case_163()
#line 1433 "cs-parser.jay"
{
		Error_SyntaxError (yyToken);
		current_local_parameters = ParametersCompiled.Undefined;

		MemberName name = (MemberName) yyVals[-1+yyTop];
		var method = Method.Create (current_type, (FullNamedExpression) yyVals[-2+yyTop], (Modifiers) yyVals[-3+yyTop],
									name, current_local_parameters, (Attributes) yyVals[-4+yyTop]);

		current_type.AddMember (method);

		if (doc_support)
			method.DocComment = Lexer.consume_doc_comment ();

		yyVal = method;
	  }

void case_168()
#line 1462 "cs-parser.jay"
{
		if (lang_version < LanguageVersion.V_6) {
			FeatureIsNotAvailable (GetLocation (yyVals[0+yyTop]), "expression bodied members");
		}

		++lexer.parsing_block;
		start_block (GetLocation (yyVals[0+yyTop]));
	 }

void case_169()
#line 1471 "cs-parser.jay"
{
		lexer.parsing_block = 0;
		current_block.AddStatement (new ContextualReturn ((Expression) yyVals[-1+yyTop]));
		var b = end_block (GetLocation (yyVals[0+yyTop]));
		b.IsCompilerGenerated = true;
		yyVal = b;
	 }

void case_172()
#line 1487 "cs-parser.jay"
{
		var pars_list = (List<Parameter>) yyVals[0+yyTop];
	  	yyVal = new ParametersCompiled (pars_list.ToArray ());
	  }

void case_173()
#line 1492 "cs-parser.jay"
{
		var pars_list = (List<Parameter>) yyVals[-2+yyTop];
		pars_list.Add ((Parameter) yyVals[0+yyTop]);

		yyVal = new ParametersCompiled (pars_list.ToArray ()); 
	  }

void case_174()
#line 1499 "cs-parser.jay"
{
		var pars_list = (List<Parameter>) yyVals[-2+yyTop];
		pars_list.Add (new ArglistParameter (GetLocation (yyVals[0+yyTop])));
		yyVal = new ParametersCompiled (pars_list.ToArray (), true);
	  }

void case_175()
#line 1505 "cs-parser.jay"
{
		if (yyVals[-2+yyTop] != null)
			report.Error (231, ((Parameter) yyVals[-2+yyTop]).Location, "A params parameter must be the last parameter in a formal parameter list");

		yyVal = new ParametersCompiled (new Parameter[] { (Parameter) yyVals[-2+yyTop] } );			
	  }

void case_176()
#line 1512 "cs-parser.jay"
{
		if (yyVals[-2+yyTop] != null)
			report.Error (231, ((Parameter) yyVals[-2+yyTop]).Location, "A params parameter must be the last parameter in a formal parameter list");

		var pars_list = (List<Parameter>) yyVals[-4+yyTop];
		pars_list.Add (new ArglistParameter (GetLocation (yyVals[-2+yyTop])));

		yyVal = new ParametersCompiled (pars_list.ToArray (), true);
	  }

void case_177()
#line 1522 "cs-parser.jay"
{
		report.Error (257, GetLocation (yyVals[-2+yyTop]), "An __arglist parameter must be the last parameter in a formal parameter list");

		yyVal = new ParametersCompiled (new Parameter [] { new ArglistParameter (GetLocation (yyVals[-2+yyTop])) }, true);
	  }

void case_178()
#line 1528 "cs-parser.jay"
{
		report.Error (257, GetLocation (yyVals[-2+yyTop]), "An __arglist parameter must be the last parameter in a formal parameter list");

		var pars_list = (List<Parameter>) yyVals[-4+yyTop];
		pars_list.Add (new ArglistParameter (GetLocation (yyVals[-2+yyTop])));

		yyVal = new ParametersCompiled (pars_list.ToArray (), true);
	  }

void case_181()
#line 1545 "cs-parser.jay"
{
		Error_SyntaxError (yyToken);
		yyVal = ParametersCompiled.EmptyReadOnlyParameters;
	  }

void case_182()
#line 1553 "cs-parser.jay"
{
		parameters_bucket.Clear ();
		Parameter p = (Parameter) yyVals[0+yyTop];
		parameters_bucket.Add (p);
		
		default_parameter_used = p.HasDefaultValue;
		yyVal = parameters_bucket;
	  }

void case_183()
#line 1562 "cs-parser.jay"
{
		var pars = (List<Parameter>) yyVals[-2+yyTop];
		Parameter p = (Parameter) yyVals[0+yyTop];
		if (p != null) {
			if (p.HasExtensionMethodModifier)
				report.Error (1100, p.Location, "The parameter modifier `this' can only be used on the first parameter");
			else if (!p.HasDefaultValue && default_parameter_used)
				report.Error (1737, p.Location, "Optional parameter cannot precede required parameters");

			default_parameter_used |= p.HasDefaultValue;
			pars.Add (p);
			
			lbag.AddLocation (p, GetLocation (yyVals[-1+yyTop]));
		}
		
		yyVal = yyVals[-2+yyTop];
	  }

void case_184()
#line 1586 "cs-parser.jay"
{
		var lt = (LocatedToken) yyVals[0+yyTop];
		yyVal = new Parameter ((FullNamedExpression) yyVals[-1+yyTop], lt.Value, (Parameter.Modifier) yyVals[-2+yyTop], (Attributes) yyVals[-3+yyTop], lt.Location);
	  }

void case_185()
#line 1594 "cs-parser.jay"
{
		var lt = (LocatedToken) yyVals[-2+yyTop];
		report.Error (1552, lt.Location, "Array type specifier, [], must appear before parameter name");
		yyVal = new Parameter ((FullNamedExpression) yyVals[-3+yyTop], lt.Value, (Parameter.Modifier) yyVals[-4+yyTop], (Attributes) yyVals[-5+yyTop], lt.Location);
	  }

void case_186()
#line 1600 "cs-parser.jay"
{
		Error_SyntaxError (yyToken);
	  	Location l = GetLocation (yyVals[0+yyTop]);
		yyVal = new Parameter (null, null, Parameter.Modifier.NONE, (Attributes) yyVals[-1+yyTop], l);
	  }

void case_187()
#line 1609 "cs-parser.jay"
{
		Error_SyntaxError (yyToken);
	  	Location l = GetLocation (yyVals[0+yyTop]);
		yyVal = new Parameter ((FullNamedExpression) yyVals[-1+yyTop], null, (Parameter.Modifier) yyVals[-2+yyTop], (Attributes) yyVals[-3+yyTop], l);
	  }

void case_189()
#line 1623 "cs-parser.jay"
{
	  	--lexer.parsing_block;
		if (lang_version <= LanguageVersion.V_3) {
			FeatureIsNotAvailable (GetLocation (yyVals[-2+yyTop]), "optional parameter");
		}
		
		Parameter.Modifier mod = (Parameter.Modifier) yyVals[-5+yyTop];
		if (mod != Parameter.Modifier.NONE) {
			switch (mod) {
			case Parameter.Modifier.REF:
			case Parameter.Modifier.OUT:
				report.Error (1741, GetLocation (yyVals[-5+yyTop]), "Cannot specify a default value for the `{0}' parameter",
					Parameter.GetModifierSignature (mod));
				break;
				
			case Parameter.Modifier.This:
				report.Error (1743, GetLocation (yyVals[-5+yyTop]), "Cannot specify a default value for the `{0}' parameter",
					Parameter.GetModifierSignature (mod));
				break;
			default:
				throw new NotImplementedException (mod.ToString ());
			}
				
			mod = Parameter.Modifier.NONE;
		}
		
		if ((valid_param_mod & ParameterModifierType.DefaultValue) == 0)
			report.Error (1065, GetLocation (yyVals[-2+yyTop]), "Optional parameter is not valid in this context");
		
		var lt = (LocatedToken) yyVals[-3+yyTop];
		yyVal = new Parameter ((FullNamedExpression) yyVals[-4+yyTop], lt.Value, mod, (Attributes) yyVals[-6+yyTop], lt.Location);
		lbag.AddLocation (yyVal, GetLocation (yyVals[-2+yyTop]));
		
		if (yyVals[0+yyTop] != null)
			((Parameter) yyVal).DefaultValue = new DefaultParameterValueExpression ((Expression) yyVals[0+yyTop]);
	  }

void case_193()
#line 1672 "cs-parser.jay"
{
		Parameter.Modifier p2 = (Parameter.Modifier)yyVals[0+yyTop];
  		Parameter.Modifier mod = (Parameter.Modifier)yyVals[-1+yyTop] | p2;
  		if (((Parameter.Modifier)yyVals[-1+yyTop] & p2) == p2) {
  			Error_DuplicateParameterModifier (lexer.Location, p2);
  		} else {
	  		switch (mod & ~Parameter.Modifier.This) {
  				case Parameter.Modifier.REF:
					report.Error (1101, lexer.Location, "The parameter modifiers `this' and `ref' cannot be used altogether");
  					break;
   				case Parameter.Modifier.OUT:
					report.Error (1102, lexer.Location, "The parameter modifiers `this' and `out' cannot be used altogether");
  					break;
  				default:
 					report.Error (1108, lexer.Location, "A parameter cannot have specified more than one modifier");
 					break;
 			}
  		}
  		yyVal = mod;
	  }

void case_194()
#line 1696 "cs-parser.jay"
{
	  	if ((valid_param_mod & ParameterModifierType.Ref) == 0)
	  		Error_ParameterModifierNotValid ("ref", GetLocation (yyVals[0+yyTop]));
	  		
	  	yyVal = Parameter.Modifier.REF;
	  }

void case_195()
#line 1703 "cs-parser.jay"
{
	  	if ((valid_param_mod & ParameterModifierType.Out) == 0)
	  		Error_ParameterModifierNotValid ("out", GetLocation (yyVals[0+yyTop]));
	  
	  	yyVal = Parameter.Modifier.OUT;
	  }

void case_196()
#line 1710 "cs-parser.jay"
{
		if ((valid_param_mod & ParameterModifierType.This) == 0)
	  		Error_ParameterModifierNotValid ("this", GetLocation (yyVals[0+yyTop]));

	  	if (lang_version <= LanguageVersion.ISO_2)
	  		FeatureIsNotAvailable (GetLocation (yyVals[0+yyTop]), "extension methods");
	  			
		yyVal = Parameter.Modifier.This;
	  }

void case_197()
#line 1723 "cs-parser.jay"
{
		var lt = (LocatedToken) yyVals[0+yyTop];
		yyVal = new ParamsParameter ((FullNamedExpression) yyVals[-1+yyTop], lt.Value, (Attributes) yyVals[-3+yyTop], lt.Location);
	  }

void case_198()
#line 1728 "cs-parser.jay"
{
		report.Error (1751, GetLocation (yyVals[-4+yyTop]), "Cannot specify a default value for a parameter array");
		
		var lt = (LocatedToken) yyVals[-2+yyTop];
		yyVal = new ParamsParameter ((FullNamedExpression) yyVals[-3+yyTop], lt.Value, (Attributes) yyVals[-5+yyTop], lt.Location);		
	  }

void case_199()
#line 1735 "cs-parser.jay"
{
		Error_SyntaxError (yyToken);

		yyVal = new ParamsParameter ((FullNamedExpression) yyVals[-1+yyTop], null, (Attributes) yyVals[-3+yyTop], Location.Null);
	  }

void case_200()
#line 1744 "cs-parser.jay"
{
		if ((valid_param_mod & ParameterModifierType.Params) == 0)
			report.Error (1670, (GetLocation (yyVals[0+yyTop])), "The `params' modifier is not allowed in current context");
	  }

void case_201()
#line 1749 "cs-parser.jay"
{
		Parameter.Modifier mod = (Parameter.Modifier)yyVals[0+yyTop];
		if ((mod & Parameter.Modifier.This) != 0) {
			report.Error (1104, GetLocation (yyVals[-1+yyTop]), "The parameter modifiers `this' and `params' cannot be used altogether");
		} else {
			report.Error (1611, GetLocation (yyVals[-1+yyTop]), "The params parameter cannot be declared as ref or out");
		}	  
	  }

void case_203()
#line 1765 "cs-parser.jay"
{
	  	if ((valid_param_mod & ParameterModifierType.Arglist) == 0)
	  		report.Error (1669, GetLocation (yyVals[0+yyTop]), "__arglist is not valid in this context");
	  }

void case_204()
#line 1776 "cs-parser.jay"
{
		lexer.parsing_generic_declaration = false;
		if (doc_support)
			tmpComment = Lexer.consume_doc_comment ();
	  }

void case_205()
#line 1782 "cs-parser.jay"
{
		var type = (FullNamedExpression) yyVals[-3+yyTop];
		current_property = new Property (current_type, type, (Modifiers) yyVals[-4+yyTop],
			(MemberName) yyVals[-2+yyTop], (Attributes) yyVals[-5+yyTop]);
			
		if (type.Type != null && type.Type.Kind == MemberKind.Void)
			report.Error (547, GetLocation (yyVals[-3+yyTop]), "`{0}': property or indexer cannot have void type", current_property.GetSignatureForError ());					
			
		current_type.AddMember (current_property);
		lbag.AddMember (current_property, mod_locations, GetLocation (yyVals[0+yyTop]));
		
		lexer.PropertyParsing = true;
	  }

void case_206()
#line 1796 "cs-parser.jay"
{
		lexer.PropertyParsing = false;
		
		if (doc_support)
			current_property.DocComment = ConsumeStoredComment ();				
	  }

void case_207()
#line 1803 "cs-parser.jay"
{
		lbag.AppendToMember (current_property, GetLocation (yyVals[0+yyTop]));
		lexer.parsing_modifiers = true;
	  }

void case_209()
#line 1815 "cs-parser.jay"
{
		lexer.parsing_generic_declaration = false;
		if (doc_support)
			tmpComment = Lexer.consume_doc_comment ();
		current_local_parameters = ParametersCompiled.EmptyReadOnlyParameters;
	  }

void case_210()
#line 1822 "cs-parser.jay"
{
		var type = (FullNamedExpression) yyVals[-3+yyTop];
		var property = new Property (current_type, type, (Modifiers) yyVals[-4+yyTop],
			(MemberName) yyVals[-2+yyTop], (Attributes) yyVals[-5+yyTop]);

		property.Get = new Property.GetMethod (property, Modifiers.COMPILER_GENERATED, null, property.Location);
		property.Get.Block = (ToplevelBlock) yyVals[0+yyTop];

		if (current_container.Kind == MemberKind.Interface) {
			report.Error (531, property.Get.Block.StartLocation,
				"`{0}': interface members cannot have a definition", property.GetSignatureForError ());
		}

		if (type.Type != null && type.Type.Kind == MemberKind.Void)
			report.Error (547, GetLocation (yyVals[-3+yyTop]), "`{0}': property or indexer cannot have void type", property.GetSignatureForError ());

		current_type.AddMember (property);

		current_local_parameters = null;
	  }

void case_212()
#line 1847 "cs-parser.jay"
{
		++lexer.parsing_block;
		current_local_parameters = ParametersCompiled.EmptyReadOnlyParameters;
		start_block (GetLocation (yyVals[0+yyTop]));
	  }

void case_213()
#line 1853 "cs-parser.jay"
{
		--lexer.parsing_block;
		((Property)current_property).Initializer = (Expression) yyVals[-1+yyTop];
		lbag.AppendToMember (current_property, GetLocation (yyVals[-3+yyTop]), GetLocation (yyVals[0+yyTop]));
		end_block (GetLocation (yyVals[0+yyTop]));
		current_local_parameters = null;
	  }

void case_217()
#line 1874 "cs-parser.jay"
{
		valid_param_mod = 0;
		var type = (FullNamedExpression) yyVals[-5+yyTop];
		Indexer indexer = new Indexer (current_type, type, (MemberName) yyVals[-4+yyTop], (Modifiers) yyVals[-6+yyTop], (ParametersCompiled) yyVals[-1+yyTop], (Attributes) yyVals[-7+yyTop]);
			
		current_property = indexer;

  		current_type.AddIndexer (indexer);
		lbag.AddMember (current_property, mod_locations, GetLocation (yyVals[-3+yyTop]), GetLocation (yyVals[0+yyTop]));
  		
		if (type.Type != null && type.Type.Kind == MemberKind.Void)
			report.Error (620, GetLocation (yyVals[-5+yyTop]), "`{0}': indexer return type cannot be `void'", indexer.GetSignatureForError ());  		

		if (indexer.ParameterInfo.IsEmpty) {
			report.Error (1551, GetLocation (yyVals[-3+yyTop]), "Indexers must have at least one parameter");
		}

		if (doc_support) {
			tmpComment = Lexer.consume_doc_comment ();
			Lexer.doc_state = XmlCommentState.Allowed;
		}

		lexer.PropertyParsing = true;
		current_local_parameters = (ParametersCompiled) yyVals[-1+yyTop];
	  }

void case_218()
#line 1900 "cs-parser.jay"
{
 		lexer.PropertyParsing = false;
		current_local_parameters = null;

		if (current_property.AccessorFirst != null && current_property.AccessorFirst.Block == null)
			((Indexer) current_property).ParameterInfo.CheckParameters (current_property);
	  
		if (doc_support)
			current_property.DocComment = ConsumeStoredComment ();
			
		current_property = null;		
	  }

void case_220()
#line 1920 "cs-parser.jay"
{
		current_property.Get = new Indexer.GetIndexerMethod (current_property, Modifiers.COMPILER_GENERATED, current_local_parameters, null, current_property.Location);
		current_property.Get.Block = (ToplevelBlock) yyVals[0+yyTop];
	  }

void case_225()
#line 1932 "cs-parser.jay"
{
	  	if (yyToken == Token.CLOSE_BRACE) {
	  		report.Error (548, lexer.Location, "`{0}': property or indexer must have at least one accessor", current_property.GetSignatureForError ());
		} else {
			if (yyToken == Token.SEMICOLON)
				report.Error (1597, lexer.Location, "Semicolon after method or accessor block is not valid");
			else
				report.Error (1014, GetLocation (yyVals[0+yyTop]), "A get or set accessor expected");
		}
	  }

void case_226()
#line 1946 "cs-parser.jay"
{
		if (yyVals[-1+yyTop] != ModifierNone && lang_version == LanguageVersion.ISO_1) {
			FeatureIsNotAvailable (GetLocation (yyVals[-1+yyTop]), "access modifiers on properties");
		}
	  
		if (current_property.Get != null) {
			report.Error (1007, GetLocation (yyVals[0+yyTop]), "Property accessor already defined");
		}
		
		if (current_property is Indexer) {
			current_property.Get = new Indexer.GetIndexerMethod (current_property, (Modifiers) yyVals[-1+yyTop], ((Indexer)current_property).ParameterInfo.Clone (),
				(Attributes) yyVals[-2+yyTop], GetLocation (yyVals[0+yyTop]));
		} else {
			current_property.Get = new Property.GetMethod (current_property,
				(Modifiers) yyVals[-1+yyTop], (Attributes) yyVals[-2+yyTop], GetLocation (yyVals[0+yyTop]));
		}	
	  
		current_local_parameters = current_property.Get.ParameterInfo;	  
		lbag.AddMember (current_property.Get, mod_locations);
		lexer.PropertyParsing = false;
	  }

void case_227()
#line 1968 "cs-parser.jay"
{
	  	if (yyVals[0+yyTop] != null) {
	  		current_property.Get.Block = (ToplevelBlock) yyVals[0+yyTop];			
	  	
			if (current_container.Kind == MemberKind.Interface) {
				report.Error (531, current_property.Get.Block.StartLocation,
					"`{0}': interface members cannot have a definition", current_property.Get.GetSignatureForError ());
			}		
		}
	  
		current_local_parameters = null;
		lexer.PropertyParsing = true;

		if (doc_support)
			if (Lexer.doc_state == XmlCommentState.Error)
				Lexer.doc_state = XmlCommentState.NotAllowed;
	  }

void case_228()
#line 1989 "cs-parser.jay"
{
		if (yyVals[-1+yyTop] != ModifierNone && lang_version == LanguageVersion.ISO_1) {
			FeatureIsNotAvailable (GetLocation (yyVals[-1+yyTop]), "access modifiers on properties");
		}
		
		if (current_property.Set != null) {
			report.Error (1007, GetLocation (yyVals[0+yyTop]), "Property accessor already defined");
		}
	  
		if (current_property is Indexer) {
			current_property.Set = new Indexer.SetIndexerMethod (current_property, (Modifiers) yyVals[-1+yyTop],
				ParametersCompiled.MergeGenerated (compiler,
				((Indexer)current_property).ParameterInfo, true, new Parameter (
					current_property.TypeExpression, "value", Parameter.Modifier.NONE, null, GetLocation (yyVals[0+yyTop])),
					null),
				(Attributes) yyVals[-2+yyTop], GetLocation (yyVals[0+yyTop]));
		} else {
			current_property.Set = new Property.SetMethod (current_property, (Modifiers) yyVals[-1+yyTop], 
				ParametersCompiled.CreateImplicitParameter (current_property.TypeExpression, GetLocation (yyVals[0+yyTop])),
				(Attributes) yyVals[-2+yyTop], GetLocation (yyVals[0+yyTop]));
		}
		
		current_local_parameters = current_property.Set.ParameterInfo;	
		lbag.AddMember (current_property.Set, mod_locations);
		lexer.PropertyParsing = false;
	  }

void case_229()
#line 2016 "cs-parser.jay"
{
		if (yyVals[0+yyTop] != null) {		
			current_property.Set.Block = (ToplevelBlock) yyVals[0+yyTop];
		
			if (current_container.Kind == MemberKind.Interface) {
				report.Error (531, current_property.Set.Block.StartLocation,
					"`{0}': interface members cannot have a definition", current_property.Set.GetSignatureForError ());
			}
		}
		
		current_local_parameters = null;
		lexer.PropertyParsing = true;

		if (doc_support
			&& Lexer.doc_state == XmlCommentState.Error)
			Lexer.doc_state = XmlCommentState.NotAllowed;
	  }

void case_231()
#line 2038 "cs-parser.jay"
{
		/* TODO: lbag*/
	  	yyVal = null;
	  }

void case_232()
#line 2043 "cs-parser.jay"
{
	  	Error_SyntaxError (1043, yyToken, "Invalid accessor body");
	  	yyVal = null;
	  }

void case_234()
#line 2057 "cs-parser.jay"
{
		lexer.ConstraintsParsing = true;
		push_current_container (new Interface (current_container, (MemberName) yyVals[0+yyTop], (Modifiers) yyVals[-4+yyTop], (Attributes) yyVals[-5+yyTop]), yyVals[-3+yyTop]);
		lbag.AddMember (current_container, mod_locations, GetLocation (yyVals[-2+yyTop]));		
	  }

void case_235()
#line 2064 "cs-parser.jay"
{
		lexer.ConstraintsParsing = false;

		if (yyVals[0+yyTop] != null)
			current_container.SetConstraints ((List<Constraints>) yyVals[0+yyTop]);

		if (doc_support) {
			current_container.PartialContainer.DocComment = Lexer.consume_doc_comment ();
			Lexer.doc_state = XmlCommentState.Allowed;
		}
		
		lexer.parsing_modifiers = true;
	  }

void case_236()
#line 2078 "cs-parser.jay"
{
		--lexer.parsing_declaration;	  
		if (doc_support)
			Lexer.doc_state = XmlCommentState.Allowed;
	  }

void case_237()
#line 2084 "cs-parser.jay"
{
		if (yyVals[0+yyTop] == null) {
			lbag.AppendToMember (current_container, GetLocation (yyVals[-4+yyTop]), GetLocation (yyVals[-2+yyTop]));
		} else {
			lbag.AppendToMember (current_container, GetLocation (yyVals[-4+yyTop]), GetLocation (yyVals[-2+yyTop]), GetLocation (yyVals[0+yyTop]));
		}
		yyVal = pop_current_class ();
	  }

void case_241()
#line 2105 "cs-parser.jay"
{
		lexer.parsing_modifiers = true;
		lexer.parsing_block = 0;
	  }

void case_242()
#line 2110 "cs-parser.jay"
{
		lexer.parsing_modifiers = true;
		lexer.parsing_block = 0;
	  }

void case_253()
#line 2148 "cs-parser.jay"
{
		OperatorDeclaration decl = (OperatorDeclaration) yyVals[-2+yyTop];
		if (decl != null) {
			Operator op = new Operator (
				current_type, decl.optype, decl.ret_type, (Modifiers) yyVals[-3+yyTop], 
				current_local_parameters,
				(ToplevelBlock) yyVals[0+yyTop], (Attributes) yyVals[-4+yyTop], decl.location);
				
			if (op.Block == null)
				op.ParameterInfo.CheckParameters (op);

			if (doc_support) {
				op.DocComment = tmpComment;
				Lexer.doc_state = XmlCommentState.Allowed;
			}

			/* Note again, checking is done in semantic analysis*/
			current_type.AddOperator (op);

			lbag.AddMember (op, mod_locations, lbag.GetLocations (decl));
		}
		
		current_local_parameters = null;
	  }

void case_255()
#line 2177 "cs-parser.jay"
{
		report.Error (590, GetLocation (yyVals[0+yyTop]), "User-defined operators cannot return void");
		yyVal = new TypeExpression (compiler.BuiltinTypes.Void, GetLocation (yyVals[0+yyTop]));
	  }

void case_256()
#line 2185 "cs-parser.jay"
{
		valid_param_mod = ParameterModifierType.DefaultValue;
		if ((Operator.OpType) yyVals[-1+yyTop] == Operator.OpType.Is)
			valid_param_mod |= ParameterModifierType.Out;
	  }

void case_257()
#line 2191 "cs-parser.jay"
{
		valid_param_mod = 0;

		Location loc = GetLocation (yyVals[-5+yyTop]);
		Operator.OpType op = (Operator.OpType) yyVals[-4+yyTop];
		current_local_parameters = (ParametersCompiled)yyVals[-1+yyTop];
		
		int p_count = current_local_parameters.Count;
		if (p_count == 1) {
			if (op == Operator.OpType.Addition)
				op = Operator.OpType.UnaryPlus;
			else if (op == Operator.OpType.Subtraction)
				op = Operator.OpType.UnaryNegation;
		}
		
		if (IsUnaryOperator (op)) {
			if (p_count == 2) {
				report.Error (1020, loc, "Overloadable binary operator expected");
			} else if (p_count != 1) {
				report.Error (1535, loc, "Overloaded unary operator `{0}' takes one parameter",
					Operator.GetName (op));
			}
		} else if (op == Operator.OpType.Is) {
			/* TODO: Special checks for is operator*/
		} else {
			if (p_count == 1) {
				report.Error (1019, loc, "Overloadable unary operator expected");
			} else if (p_count != 2) {
				report.Error (1534, loc, "Overloaded binary operator `{0}' takes two parameters",
					Operator.GetName (op));
			}
		}
		
		if (doc_support) {
			tmpComment = Lexer.consume_doc_comment ();
			Lexer.doc_state = XmlCommentState.NotAllowed;
		}

		yyVal = new OperatorDeclaration (op, (FullNamedExpression) yyVals[-6+yyTop], loc);
		lbag.AddLocation (yyVal, GetLocation (yyVals[-5+yyTop]), GetLocation (yyVals[-4+yyTop]), GetLocation (yyVals[-3+yyTop]), GetLocation (yyVals[0+yyTop]));
	  }

void case_281()
#line 2262 "cs-parser.jay"
{
		if (lang_version != LanguageVersion.Experimental)
			FeatureIsNotAvailable (GetLocation (yyVals[0+yyTop]), "is user operator");

		yyVal = Operator.OpType.Is;
	  }

void case_283()
#line 2276 "cs-parser.jay"
{
		valid_param_mod = 0;

		Location loc = GetLocation (yyVals[-5+yyTop]);
		current_local_parameters = (ParametersCompiled)yyVals[-1+yyTop];  

		if (current_local_parameters.Count != 1) {
			report.Error (1535, loc, "Overloaded unary operator `implicit' takes one parameter");
		}

		if (doc_support) {
			tmpComment = Lexer.consume_doc_comment ();
			Lexer.doc_state = XmlCommentState.NotAllowed;
		}

		yyVal = new OperatorDeclaration (Operator.OpType.Implicit, (FullNamedExpression) yyVals[-4+yyTop], loc);
		lbag.AddLocation (yyVal, GetLocation (yyVals[-6+yyTop]), GetLocation (yyVals[-5+yyTop]), GetLocation (yyVals[-3+yyTop]), GetLocation (yyVals[0+yyTop]));
	  }

void case_285()
#line 2299 "cs-parser.jay"
{
		valid_param_mod = 0;
		
		Location loc = GetLocation (yyVals[-5+yyTop]);
		current_local_parameters = (ParametersCompiled)yyVals[-1+yyTop];  

		if (current_local_parameters.Count != 1) {
			report.Error (1535, loc, "Overloaded unary operator `explicit' takes one parameter");
		}

		if (doc_support) {
			tmpComment = Lexer.consume_doc_comment ();
			Lexer.doc_state = XmlCommentState.NotAllowed;
		}

		yyVal = new OperatorDeclaration (Operator.OpType.Explicit, (FullNamedExpression) yyVals[-4+yyTop], loc);
		lbag.AddLocation (yyVal, GetLocation (yyVals[-6+yyTop]), GetLocation (yyVals[-5+yyTop]), GetLocation (yyVals[-3+yyTop]), GetLocation (yyVals[0+yyTop]));
	  }

void case_286()
#line 2318 "cs-parser.jay"
{
	  	Error_SyntaxError (yyToken);
		current_local_parameters = ParametersCompiled.EmptyReadOnlyParameters;
		yyVal = new OperatorDeclaration (Operator.OpType.Implicit, null, GetLocation (yyVals[-1+yyTop]));
	  }

void case_287()
#line 2324 "cs-parser.jay"
{
	  	Error_SyntaxError (yyToken);
		current_local_parameters = ParametersCompiled.EmptyReadOnlyParameters;
	  	yyVal = new OperatorDeclaration (Operator.OpType.Explicit, null, GetLocation (yyVals[-1+yyTop]));
	  }

void case_288()
#line 2334 "cs-parser.jay"
{ 
		Constructor c = (Constructor) yyVals[-1+yyTop];
		c.Block = (ToplevelBlock) yyVals[0+yyTop];
		
		if (doc_support)
			c.DocComment = ConsumeStoredComment ();

		current_local_parameters = null;
		if (doc_support)
			Lexer.doc_state = XmlCommentState.Allowed;
	  }

void case_289()
#line 2351 "cs-parser.jay"
{
		if (doc_support) {
			tmpComment = Lexer.consume_doc_comment ();
			Lexer.doc_state = XmlCommentState.Allowed;
		}
		
		valid_param_mod = ParameterModifierType.All;
	  }

void case_290()
#line 2360 "cs-parser.jay"
{
		valid_param_mod = 0;
		current_local_parameters = (ParametersCompiled) yyVals[-1+yyTop];
		
		var lt = (LocatedToken) yyVals[-4+yyTop];
		var mods = (Modifiers) yyVals[-5+yyTop];
		var c = new Constructor (current_type, lt.Value, mods, (Attributes) yyVals[-6+yyTop], current_local_parameters, lt.Location);

		if (lt.Value != current_container.MemberName.Name) {
			report.Error (1520, c.Location, "Class, struct, or interface method must have a return type");
		} else if ((mods & Modifiers.STATIC) != 0) {
			if (!current_local_parameters.IsEmpty) {
				report.Error (132, c.Location, "`{0}': The static constructor must be parameterless",
					c.GetSignatureForError ());
			}

			if ((mods & Modifiers.AccessibilityMask) != 0){
				report.Error (515, c.Location,
					"`{0}': static constructor cannot have an access modifier",
					c.GetSignatureForError ());
			}
		} else {
			if (current_type.Kind == MemberKind.Struct && current_local_parameters.IsEmpty) {
				if (lang_version < LanguageVersion.V_6)
					FeatureIsNotAvailable (GetLocation (yyVals[-4+yyTop]), "struct parameterless instance constructor");

				if ((mods & Modifiers.PUBLIC) == 0) {
					report.Error (8075, c.Location, "`{0}': Structs parameterless instance constructor must be public", c.GetSignatureForError ());
				}
			}
		}

		current_type.AddConstructor (c);
		lbag.AddMember (c, mod_locations, GetLocation (yyVals[-2+yyTop]), GetLocation (yyVals[0+yyTop]));
		yyVal = c;

		/**/
		/* start block here, so possible anonymous methods inside*/
		/* constructor initializer can get correct parent block*/
		/**/
	  	start_block (lexer.Location);
	  }

void case_291()
#line 2403 "cs-parser.jay"
{
		if (yyVals[0+yyTop] != null) {
			var c = (Constructor) yyVals[-1+yyTop];
			c.Initializer = (ConstructorInitializer) yyVals[0+yyTop];
			
			if (c.IsStatic) {
				report.Error (514, c.Location,
					"`{0}': static constructor cannot have an explicit `this' or `base' constructor call",
					c.GetSignatureForError ());
			}
		}

		yyVal = yyVals[-1+yyTop];
	  }

void case_297()
#line 2435 "cs-parser.jay"
{
	  	--lexer.parsing_block;
		yyVal = new ConstructorBaseInitializer ((Arguments) yyVals[-1+yyTop], GetLocation (yyVals[-4+yyTop]));
		lbag.AddLocation (yyVal, GetLocation (yyVals[-5+yyTop]), GetLocation (yyVals[-3+yyTop]), GetLocation (yyVals[0+yyTop]));
	  }

void case_299()
#line 2445 "cs-parser.jay"
{
	  	--lexer.parsing_block;
		yyVal = new ConstructorThisInitializer ((Arguments) yyVals[-1+yyTop], GetLocation (yyVals[-4+yyTop]));
		lbag.AddLocation (yyVal, GetLocation (yyVals[-5+yyTop]), GetLocation (yyVals[-3+yyTop]), GetLocation (yyVals[0+yyTop]));
	  }

void case_300()
#line 2451 "cs-parser.jay"
{
		Error_SyntaxError (yyToken);	  
		yyVal = new ConstructorThisInitializer (null, GetLocation (yyVals[0+yyTop]));
		lbag.AddLocation (yyVal, GetLocation (yyVals[-1+yyTop]));
	  }

void case_301()
#line 2457 "cs-parser.jay"
{
		Error_SyntaxError (yyToken);
		yyVal = null;
	  }

void case_302()
#line 2465 "cs-parser.jay"
{
		if (doc_support) {
			tmpComment = Lexer.consume_doc_comment ();
			Lexer.doc_state = XmlCommentState.NotAllowed;
		}
		
		current_local_parameters = ParametersCompiled.EmptyReadOnlyParameters;
	  }

void case_303()
#line 2474 "cs-parser.jay"
{
		var lt = (LocatedToken) yyVals[-3+yyTop];
		if (lt.Value != current_container.MemberName.Name){
			report.Error (574, lt.Location, "Name of destructor must match name of class");
		} else if (current_container.Kind != MemberKind.Class){
			report.Error (575, lt.Location, "Only class types can contain destructor");
		}
		
		Destructor d = new Destructor (current_type, (Modifiers) yyVals[-6+yyTop],
			ParametersCompiled.EmptyReadOnlyParameters, (Attributes) yyVals[-7+yyTop], lt.Location);
		if (doc_support)
			d.DocComment = ConsumeStoredComment ();
		  
		d.Block = (ToplevelBlock) yyVals[0+yyTop];
		current_type.AddMember (d);
		lbag.AddMember (d, mod_locations, GetLocation (yyVals[-5+yyTop]), GetLocation (yyVals[-2+yyTop]), GetLocation (yyVals[-1+yyTop]));

		current_local_parameters = null;
	  }

void case_304()
#line 2499 "cs-parser.jay"
{
		current_event_field = new EventField (current_type, (FullNamedExpression) yyVals[-1+yyTop], (Modifiers) yyVals[-3+yyTop], (MemberName) yyVals[0+yyTop], (Attributes) yyVals[-4+yyTop]);
		current_type.AddMember (current_event_field);
		
		if (current_event_field.MemberName.ExplicitInterface != null) {
			report.Error (71, current_event_field.Location, "`{0}': An explicit interface implementation of an event must use property syntax",
			current_event_field.GetSignatureForError ());
		}
		
		yyVal = current_event_field;
	  }

void case_305()
#line 2513 "cs-parser.jay"
{
		if (doc_support) {
			current_event_field.DocComment = Lexer.consume_doc_comment ();
			Lexer.doc_state = XmlCommentState.Allowed;
		}
		
		lbag.AddMember (current_event_field, mod_locations, GetLocation (yyVals[-6+yyTop]), GetLocation (yyVals[0+yyTop]));
		current_event_field = null;
	  }

void case_306()
#line 2526 "cs-parser.jay"
{
		current_event = new EventProperty (current_type, (FullNamedExpression) yyVals[-2+yyTop], (Modifiers) yyVals[-4+yyTop], (MemberName) yyVals[-1+yyTop], (Attributes) yyVals[-5+yyTop]);
		current_type.AddMember (current_event);
		lbag.AddMember (current_event, mod_locations, GetLocation (yyVals[-3+yyTop]), GetLocation (yyVals[0+yyTop]));
		
		lexer.EventParsing = true;
	  }

void case_307()
#line 2534 "cs-parser.jay"
{
		if (current_container.Kind == MemberKind.Interface)
			report.Error (69, GetLocation (yyVals[-2+yyTop]), "Event in interface cannot have add or remove accessors");
	  
		lexer.EventParsing = false;
	  }

void case_308()
#line 2541 "cs-parser.jay"
{
		if (doc_support) {
			current_event.DocComment = Lexer.consume_doc_comment ();
			Lexer.doc_state = XmlCommentState.Allowed;
		}
		
		lbag.AppendToMember (current_event, GetLocation (yyVals[-1+yyTop]));
	  	current_event = null;	
		current_local_parameters = null;
	  }

void case_309()
#line 2554 "cs-parser.jay"
{
		Error_SyntaxError (yyToken);

		current_type.AddMember (new EventField (current_type, (FullNamedExpression) yyVals[-1+yyTop], (Modifiers) yyVals[-3+yyTop], MemberName.Null, (Attributes) yyVals[-4+yyTop]));
	  }

void case_312()
#line 2568 "cs-parser.jay"
{
	  	--lexer.parsing_block;
		current_event_field.Initializer = (Expression) yyVals[0+yyTop];
	  }

void case_317()
#line 2592 "cs-parser.jay"
{
		var lt = (LocatedToken) yyVals[0+yyTop];
	  	yyVal = new FieldDeclarator (new SimpleMemberName (lt.Value, lt.Location), null);
	  	lbag.AddLocation (yyVal, GetLocation (yyVals[-1+yyTop]));
	  }

void case_319()
#line 2602 "cs-parser.jay"
{
		--lexer.parsing_block;
		var lt = (LocatedToken) yyVals[-3+yyTop];	  
	  	yyVal = new FieldDeclarator (new SimpleMemberName (lt.Value, lt.Location), (Expression) yyVals[0+yyTop]);
	  	lbag.AddLocation (yyVal, GetLocation (yyVals[-4+yyTop]), GetLocation (yyVals[-2+yyTop]));
	  }

void case_320()
#line 2611 "cs-parser.jay"
{
		if (current_container.Kind == MemberKind.Interface) {
			report.Error (68, lexer.Location, "`{0}': event in interface cannot have an initializer",
				current_event_field.GetSignatureForError ());
		}
		
	  	if ((current_event_field.ModFlags & Modifiers.ABSTRACT) != 0) {
			report.Error (74, lexer.Location, "`{0}': abstract event cannot have an initializer",
				current_event_field.GetSignatureForError ());
	  	}		
	  }

void case_324()
#line 2632 "cs-parser.jay"
{
		report.Error (65, lexer.Location, "`{0}': event property must have both add and remove accessors",
			current_event.GetSignatureForError ());
	  }

void case_325()
#line 2637 "cs-parser.jay"
{
		report.Error (65, lexer.Location, "`{0}': event property must have both add and remove accessors",
			current_event.GetSignatureForError ());
	  }

void case_326()
#line 2642 "cs-parser.jay"
{ 
		report.Error (1055, GetLocation (yyVals[0+yyTop]), "An add or remove accessor expected");
		yyVal = null;
	  }

void case_327()
#line 2650 "cs-parser.jay"
{
	  	if (yyVals[-1+yyTop] != ModifierNone) {
			report.Error (1609, GetLocation (yyVals[-1+yyTop]), "Modifiers cannot be placed on event accessor declarations");
	  	}
	  	
	  	current_event.Add = new EventProperty.AddDelegateMethod (current_event, (Attributes) yyVals[-2+yyTop], GetLocation (yyVals[0+yyTop]));
		current_local_parameters = current_event.Add.ParameterInfo;
		
		lbag.AddMember (current_event.Add, mod_locations);
		lexer.EventParsing = false;		
	  }

void case_328()
#line 2662 "cs-parser.jay"
{
		lexer.EventParsing = true;
	  
	  	current_event.Add.Block = (ToplevelBlock) yyVals[0+yyTop];
		
		if (current_container.Kind == MemberKind.Interface) {
			report.Error (531, current_event.Add.Block.StartLocation,
				"`{0}': interface members cannot have a definition", current_event.Add.GetSignatureForError ());
		}
		
		current_local_parameters = null;
	  }

void case_329()
#line 2678 "cs-parser.jay"
{
	  	if (yyVals[-1+yyTop] != ModifierNone) {
			report.Error (1609, GetLocation (yyVals[-1+yyTop]), "Modifiers cannot be placed on event accessor declarations");
	  	}
	  	
	  	current_event.Remove = new EventProperty.RemoveDelegateMethod (current_event, (Attributes) yyVals[-2+yyTop], GetLocation (yyVals[0+yyTop]));
		current_local_parameters = current_event.Remove.ParameterInfo;

		lbag.AddMember (current_event.Remove, mod_locations);
		lexer.EventParsing = false;		
	  }

void case_330()
#line 2690 "cs-parser.jay"
{
		lexer.EventParsing = true;
	  
	  	current_event.Remove.Block = (ToplevelBlock) yyVals[0+yyTop];
		
		if (current_container.Kind == MemberKind.Interface) {
			report.Error (531, current_event.Remove.Block.StartLocation,
				"`{0}': interface members cannot have a definition", current_event.Remove.GetSignatureForError ());
		}
		
		current_local_parameters = null;
	  }

void case_331()
#line 2706 "cs-parser.jay"
{
		report.Error (73, lexer.Location, "An add or remove accessor must have a body");
		yyVal = null;
	  }

void case_333()
#line 2715 "cs-parser.jay"
{
		current_type.UnattachedAttributes = (Attributes) yyVals[-1+yyTop];
		report.Error (1519, GetLocation (yyVals[-1+yyTop]), "An attribute is missing member declaration");
		lexer.putback ('}');
	  }

void case_334()
#line 2726 "cs-parser.jay"
{
		report.Error (1519, lexer.Location, "Unexpected symbol `}' in class, struct, or interface member declaration");
 
		lexer.putback ('}');

	  	lexer.parsing_generic_declaration = false;
		FullNamedExpression type = (FullNamedExpression) yyVals[-1+yyTop];
		current_field = new Field (current_type, type, (Modifiers) yyVals[-2+yyTop], MemberName.Null, (Attributes) yyVals[-3+yyTop]);
		current_type.AddField (current_field);
		yyVal = current_field;
	  }

void case_335()
#line 2744 "cs-parser.jay"
{
		if (doc_support)
			enumTypeComment = Lexer.consume_doc_comment ();
	  }

void case_336()
#line 2749 "cs-parser.jay"
{
		if (doc_support)
			Lexer.doc_state = XmlCommentState.Allowed;

		MemberName name = (MemberName) yyVals[-3+yyTop];
		if (name.IsGeneric) {
			report.Error (1675, name.Location, "Enums cannot have type parameters");
		}
		
		push_current_container (new Enum (current_container, (FullNamedExpression) yyVals[-2+yyTop], (Modifiers) yyVals[-5+yyTop], name, (Attributes) yyVals[-6+yyTop]), null);
	  }

void case_337()
#line 2761 "cs-parser.jay"
{
		lexer.parsing_modifiers = true;
	  
	  	/* here will be evaluated after CLOSE_BLACE is consumed.*/
		if (doc_support)
			Lexer.doc_state = XmlCommentState.Allowed;
	  }

void case_338()
#line 2769 "cs-parser.jay"
{
		if (doc_support)
			current_container.DocComment = enumTypeComment;
			
		--lexer.parsing_declaration;

/*			if (doc_support)*/
/*				em.DocComment = ev.DocComment;*/

		lbag.AddMember (current_container, mod_locations, GetLocation (yyVals[-9+yyTop]), GetLocation (yyVals[-5+yyTop]), GetLocation (yyVals[-1+yyTop]));
		yyVal = pop_current_class ();
	  }

void case_341()
#line 2790 "cs-parser.jay"
{
	 	Error_TypeExpected (GetLocation (yyVals[-1+yyTop]));
		yyVal = null;
	 }

void case_346()
#line 2808 "cs-parser.jay"
{
	  	lbag.AddLocation (yyVals[-2+yyTop], GetLocation (yyVals[-1+yyTop]));
	  	yyVal = yyVals[0+yyTop];
	  }

void case_347()
#line 2816 "cs-parser.jay"
{
		var lt = (LocatedToken) yyVals[0+yyTop];
	  	var em = new EnumMember ((Enum) current_type, new MemberName (lt.Value, lt.Location), (Attributes) yyVals[-1+yyTop]);
	  	((Enum) current_type).AddEnumMember (em);

		if (doc_support) {
			em.DocComment = Lexer.consume_doc_comment ();
			Lexer.doc_state = XmlCommentState.Allowed;
		}

		yyVal = em;
	  }

void case_348()
#line 2829 "cs-parser.jay"
{
	  	++lexer.parsing_block;
		if (doc_support) {
			tmpComment = Lexer.consume_doc_comment ();
			Lexer.doc_state = XmlCommentState.NotAllowed;
		}
	  }

void case_349()
#line 2837 "cs-parser.jay"
{ 
		--lexer.parsing_block;
		
		var lt = (LocatedToken) yyVals[-3+yyTop];
	  	var em = new EnumMember ((Enum) current_type, new MemberName (lt.Value, lt.Location), (Attributes) yyVals[-4+yyTop]);
	  	em.Initializer = new ConstInitializer (em, (Expression) yyVals[0+yyTop], GetLocation (yyVals[-1+yyTop]));
	  	((Enum) current_type).AddEnumMember (em);
		
		if (doc_support)
			em.DocComment = ConsumeStoredComment ();

		yyVal = em;
	  }

void case_350()
#line 2851 "cs-parser.jay"
{
		Error_SyntaxError (yyToken);
	  
		var lt = (LocatedToken) yyVals[-1+yyTop];
	  	var em = new EnumMember ((Enum) current_type, new MemberName (lt.Value, lt.Location), (Attributes) yyVals[-2+yyTop]);
	  	((Enum) current_type).AddEnumMember (em);

		if (doc_support) {
			em.DocComment = Lexer.consume_doc_comment ();
			Lexer.doc_state = XmlCommentState.Allowed;
		}

		yyVal = em;
	  }

void case_353()
#line 2878 "cs-parser.jay"
{
		valid_param_mod = 0;

		ParametersCompiled p = (ParametersCompiled) yyVals[-1+yyTop];

		Delegate del = new Delegate (current_container, (FullNamedExpression) yyVals[-5+yyTop], (Modifiers) yyVals[-7+yyTop], (MemberName) yyVals[-4+yyTop], p, (Attributes) yyVals[-8+yyTop]);

		p.CheckParameters (del);

		current_container.AddTypeContainer (del);

		current_delegate = del;
		lexer.ConstraintsParsing = true;
	  }

void case_355()
#line 2897 "cs-parser.jay"
{
		if (doc_support) {
			current_delegate.DocComment = Lexer.consume_doc_comment ();
			Lexer.doc_state = XmlCommentState.Allowed;
		}
	  
		if (yyVals[-2+yyTop] != null)
			current_delegate.SetConstraints ((List<Constraints>) yyVals[-2+yyTop]);
		lbag.AddMember (current_delegate, mod_locations, GetLocation (yyVals[-10+yyTop]), GetLocation (yyVals[-7+yyTop]), GetLocation (yyVals[-4+yyTop]), GetLocation (yyVals[0+yyTop]));

		yyVal = current_delegate;

		current_delegate = null;
	  }

void case_357()
#line 2916 "cs-parser.jay"
{
		if (lang_version < LanguageVersion.ISO_2)
			FeatureIsNotAvailable (GetLocation (yyVals[0+yyTop]), "nullable types");
	  
	  	yyVal = ComposedTypeSpecifier.CreateNullable (GetLocation (yyVals[0+yyTop]));
	  }

void case_359()
#line 2927 "cs-parser.jay"
{
		var lt1 = (LocatedToken) yyVals[-2+yyTop];
		var lt2 = (LocatedToken) yyVals[-1+yyTop];
		
		yyVal = new QualifiedAliasMember (lt1.Value, lt2.Value, (TypeArguments) yyVals[0+yyTop], lt1.Location);
		lbag.AddLocation (yyVal, GetLocation (yyVals[-1+yyTop]));
	  }

void case_360()
#line 2935 "cs-parser.jay"
{
		var lt1 = (LocatedToken) yyVals[-2+yyTop];
		var lt2 = (LocatedToken) yyVals[-1+yyTop];

		yyVal = new QualifiedAliasMember (lt1.Value, lt2.Value, (int) yyVals[0+yyTop], lt1.Location);
		lbag.AddLocation (yyVal, GetLocation (yyVals[-1+yyTop]));
	  }

void case_362()
#line 2947 "cs-parser.jay"
{
		var lt = (LocatedToken) yyVals[-1+yyTop];
		yyVal = new MemberAccess ((Expression) yyVals[-3+yyTop], lt.Value, (TypeArguments) yyVals[0+yyTop], lt.Location);
		lbag.AddLocation (yyVal, GetLocation (yyVals[-2+yyTop]));
	  }

void case_363()
#line 2953 "cs-parser.jay"
{
		var lt = (LocatedToken) yyVals[-1+yyTop];
		yyVal = new MemberAccess ((Expression) yyVals[-3+yyTop], lt.Value, (int) yyVals[0+yyTop], lt.Location);
		lbag.AddLocation (yyVal, GetLocation (yyVals[-2+yyTop]));
	  }

void case_364()
#line 2962 "cs-parser.jay"
{
		var lt = (LocatedToken) yyVals[-1+yyTop];
		yyVal = new SimpleName (lt.Value, (TypeArguments)yyVals[0+yyTop], lt.Location);
	  }

void case_365()
#line 2967 "cs-parser.jay"
{  
		var lt = (LocatedToken) yyVals[-1+yyTop];
		yyVal = new SimpleName (lt.Value, (int) yyVals[0+yyTop], lt.Location);
	  }

void case_367()
#line 2979 "cs-parser.jay"
{
		if (lang_version < LanguageVersion.ISO_2)
			FeatureIsNotAvailable (GetLocation (yyVals[-2+yyTop]), "generics");	  
	  
		yyVal = yyVals[-1+yyTop];
	  }

void case_368()
#line 2986 "cs-parser.jay"
{
		Error_TypeExpected (lexer.Location);
		yyVal = new TypeArguments ();
	  }

void case_369()
#line 2994 "cs-parser.jay"
{
		TypeArguments type_args = new TypeArguments ();
		type_args.Add ((FullNamedExpression) yyVals[0+yyTop]);
		yyVal = type_args;
	  }

void case_370()
#line 3000 "cs-parser.jay"
{
		TypeArguments type_args = (TypeArguments) yyVals[-2+yyTop];
		type_args.Add ((FullNamedExpression) yyVals[0+yyTop]);
		yyVal = type_args;
	  }

void case_372()
#line 3016 "cs-parser.jay"
{
		lexer.parsing_generic_declaration = false;
		var lt = (LocatedToken) yyVals[-2+yyTop];
		yyVal = new MemberName (lt.Value, (TypeParameters)yyVals[0+yyTop], lt.Location);
	  }

void case_373()
#line 3025 "cs-parser.jay"
{
	  	MemberName mn = (MemberName)yyVals[0+yyTop];
	  	if (mn.TypeParameters != null)
	  		syntax_error (mn.Location, string.Format ("Member `{0}' cannot declare type arguments",
	  			mn.GetSignatureForError ()));
	  }

void case_375()
#line 3036 "cs-parser.jay"
{
		lexer.parsing_generic_declaration = false;	  
		var lt = (LocatedToken) yyVals[-1+yyTop];
		yyVal = new MemberName (lt.Value, (TypeParameters) yyVals[0+yyTop], (ATypeNameExpression) yyVals[-2+yyTop], lt.Location);
	  }

void case_376()
#line 3045 "cs-parser.jay"
{
		lexer.parsing_generic_declaration = false;	  
		yyVal = new MemberName (TypeDefinition.DefaultIndexerName, GetLocation (yyVals[0+yyTop]));
	  }

void case_377()
#line 3050 "cs-parser.jay"
{
		lexer.parsing_generic_declaration = false;
		yyVal = new MemberName (TypeDefinition.DefaultIndexerName, null, (ATypeNameExpression) yyVals[-1+yyTop], GetLocation (yyVals[0+yyTop]));
	  }

void case_378()
#line 3058 "cs-parser.jay"
{
		var lt = (LocatedToken) yyVals[-2+yyTop];
		yyVal = new SimpleName (lt.Value, (TypeArguments) yyVals[-1+yyTop], lt.Location);
		lbag.AddLocation (yyVal, GetLocation (yyVals[0+yyTop]));
	  }

void case_379()
#line 3064 "cs-parser.jay"
{
		var lt1 = (LocatedToken) yyVals[-3+yyTop];
		var lt2 = (LocatedToken) yyVals[-2+yyTop];

		yyVal = new QualifiedAliasMember (lt1.Value, lt2.Value, (TypeArguments) yyVals[-1+yyTop], lt1.Location);
		lbag.AddLocation (yyVal, GetLocation (yyVals[0+yyTop]));
	  }

void case_380()
#line 3072 "cs-parser.jay"
{
		var lt = (LocatedToken) yyVals[-2+yyTop];
		yyVal = new MemberAccess ((ATypeNameExpression) yyVals[-3+yyTop], lt.Value, (TypeArguments) yyVals[-1+yyTop], lt.Location);
		lbag.AddLocation (yyVal, GetLocation (yyVals[0+yyTop]));
	  }

void case_382()
#line 3082 "cs-parser.jay"
{
		if (lang_version < LanguageVersion.ISO_2)
			FeatureIsNotAvailable (GetLocation (yyVals[-2+yyTop]), "generics");
	  
		yyVal = yyVals[-1+yyTop];
		lbag.AddLocation (yyVal, GetLocation (yyVals[-2+yyTop]), GetLocation (yyVals[0+yyTop]));
	  }

void case_383()
#line 3093 "cs-parser.jay"
{
		var tparams = new TypeParameters ();
		tparams.Add ((TypeParameter)yyVals[0+yyTop]);
		yyVal = tparams;
	  }

void case_384()
#line 3099 "cs-parser.jay"
{
		var tparams = (TypeParameters) yyVals[-2+yyTop];
		tparams.Add ((TypeParameter)yyVals[0+yyTop]);
		yyVal = tparams;
		lbag.AddLocation (yyVals[0+yyTop], GetLocation (yyVals[0+yyTop]));
	  }

void case_385()
#line 3109 "cs-parser.jay"
{
		var lt = (LocatedToken)yyVals[0+yyTop];
		yyVal = new TypeParameter (new MemberName (lt.Value, lt.Location), (Attributes)yyVals[-2+yyTop], (VarianceDecl) yyVals[-1+yyTop]);
  	  }

void case_386()
#line 3114 "cs-parser.jay"
{
  	  	if (GetTokenName (yyToken) == "type")
			report.Error (81, GetLocation (yyVals[0+yyTop]), "Type parameter declaration must be an identifier not a type");
		else
			Error_SyntaxError (yyToken);
			
  	  	yyVal = new TypeParameter (MemberName.Null, null, null);
  	  }

void case_395()
#line 3158 "cs-parser.jay"
{
		report.Error (1536, GetLocation (yyVals[0+yyTop]), "Invalid parameter type `void'");
		yyVal = new TypeExpression (compiler.BuiltinTypes.Void, GetLocation (yyVals[0+yyTop]));
	  }

void case_398()
#line 3174 "cs-parser.jay"
{
		if (yyVals[0+yyTop] != null) {
			yyVal = new ComposedCast ((ATypeNameExpression) yyVals[-1+yyTop], (ComposedTypeSpecifier) yyVals[0+yyTop]);
		} else {
			var sn = yyVals[-1+yyTop] as SimpleName;
			if (sn != null && sn.Name == "var")
				yyVal = new VarExpr (sn.Location);
			else
				yyVal = yyVals[-1+yyTop];
		}
	  }

void case_401()
#line 3194 "cs-parser.jay"
{
		Expression.Error_VoidInvalidInTheContext (GetLocation (yyVals[0+yyTop]), report);
		yyVal = new TypeExpression (compiler.BuiltinTypes.Void, GetLocation (yyVals[0+yyTop]));
	  }

void case_402()
#line 3202 "cs-parser.jay"
{
		if (yyVals[0+yyTop] != null)
			yyVal = new ComposedCast ((FullNamedExpression) yyVals[-1+yyTop], (ComposedTypeSpecifier) yyVals[0+yyTop]);
	  }

void case_405()
#line 3218 "cs-parser.jay"
{
		var types = new List<FullNamedExpression> (2);
		types.Add ((FullNamedExpression) yyVals[0+yyTop]);
		yyVal = types;
	  }

void case_406()
#line 3224 "cs-parser.jay"
{
		var types = (List<FullNamedExpression>) yyVals[-2+yyTop];
		types.Add ((FullNamedExpression) yyVals[0+yyTop]);
		yyVal = types;
	  }

void case_407()
#line 3233 "cs-parser.jay"
{
		if (yyVals[0+yyTop] is ComposedCast) {
			report.Error (1521, GetLocation (yyVals[0+yyTop]), "Invalid base type `{0}'", ((ComposedCast)yyVals[0+yyTop]).GetSignatureForError ());
		}
		yyVal = yyVals[0+yyTop];
	  }

void case_445()
#line 3297 "cs-parser.jay"
{
		var lt = (LocatedToken) yyVals[-1+yyTop];
	       yyVal = new CompletionSimpleName (MemberName.MakeName (lt.Value, null), lt.Location);
	  }

void case_456()
#line 3338 "cs-parser.jay"
{
		yyVal = new ParenthesizedExpression ((Expression) yyVals[-1+yyTop], GetLocation (yyVals[-2+yyTop]));
		lbag.AddLocation (yyVal, GetLocation (yyVals[-2+yyTop]), GetLocation (yyVals[0+yyTop]));
	  }

void case_458()
#line 3350 "cs-parser.jay"
{
		var lt = (LocatedToken) yyVals[-1+yyTop];
		yyVal = new MemberAccess ((Expression) yyVals[-3+yyTop], lt.Value, (TypeArguments) yyVals[0+yyTop], lt.Location);
		lbag.AddLocation (yyVal, GetLocation (yyVals[-2+yyTop]));
	  }

void case_459()
#line 3356 "cs-parser.jay"
{
		var lt = (LocatedToken) yyVals[-1+yyTop];
		yyVal = new MemberAccess ((Expression) yyVals[-3+yyTop], lt.Value, (int) yyVals[0+yyTop], lt.Location);
		lbag.AddLocation (yyVal, GetLocation (yyVals[-2+yyTop]));
	  }

void case_460()
#line 3362 "cs-parser.jay"
{
		if (lang_version < LanguageVersion.V_6)
			FeatureIsNotAvailable (GetLocation (yyVals[-3+yyTop]), "null propagating operator");

		var lt = (LocatedToken) yyVals[-1+yyTop];
		yyVal = new ConditionalMemberAccess ((Expression) yyVals[-4+yyTop], lt.Value, (TypeArguments) yyVals[0+yyTop], lt.Location);
		lbag.AddLocation (yyVal, GetLocation (yyVals[-3+yyTop]), GetLocation (yyVals[-2+yyTop]));
	  }

void case_461()
#line 3371 "cs-parser.jay"
{
		var lt = (LocatedToken) yyVals[-1+yyTop];
		yyVal = new MemberAccess ((Expression) yyVals[-3+yyTop], lt.Value, (TypeArguments) yyVals[0+yyTop], lt.Location);
		lbag.AddLocation (yyVal, GetLocation (yyVals[-2+yyTop]));
	  }

void case_462()
#line 3377 "cs-parser.jay"
{
		var lt = (LocatedToken) yyVals[-1+yyTop];
		yyVal = new MemberAccess (new BaseThis (GetLocation (yyVals[-3+yyTop])), lt.Value, (TypeArguments) yyVals[0+yyTop], lt.Location);
		lbag.AddLocation (yyVal, GetLocation (yyVals[-2+yyTop]));
	  }

void case_463()
#line 3383 "cs-parser.jay"
{
		var lt = (LocatedToken) yyVals[-1+yyTop];
		yyVal = new MemberAccess (new SimpleName ("await", ((LocatedToken) yyVals[-3+yyTop]).Location), lt.Value, (TypeArguments) yyVals[0+yyTop], lt.Location);
		lbag.AddLocation (yyVal, GetLocation (yyVals[-2+yyTop]));
	  }

void case_464()
#line 3389 "cs-parser.jay"
{
		var lt1 = (LocatedToken) yyVals[-2+yyTop];
		var lt2 = (LocatedToken) yyVals[-1+yyTop];

		yyVal = new QualifiedAliasMember (lt1.Value, lt2.Value, (TypeArguments) yyVals[0+yyTop], lt1.Location);
		lbag.AddLocation (yyVal, GetLocation (yyVals[-1+yyTop]));
	  }

void case_465()
#line 3397 "cs-parser.jay"
{
		var lt1 = (LocatedToken) yyVals[-2+yyTop];
		var lt2 = (LocatedToken) yyVals[-1+yyTop];

		yyVal = new QualifiedAliasMember (lt1.Value, lt2.Value, (int) yyVals[0+yyTop], lt1.Location);
		lbag.AddLocation (yyVal, GetLocation (yyVals[-1+yyTop]));
	  }

void case_467()
#line 3407 "cs-parser.jay"
{
		var lt = (LocatedToken) yyVals[-1+yyTop];
		yyVal = new CompletionMemberAccess ((Expression) yyVals[-3+yyTop], lt.Value, lt.Location);
	  }

void case_469()
#line 3415 "cs-parser.jay"
{
		var lt = (LocatedToken) yyVals[-1+yyTop];
		yyVal = new CompletionMemberAccess ((Expression) yyVals[-3+yyTop], lt.Value, lt.Location);
 	  }

void case_470()
#line 3423 "cs-parser.jay"
{
		yyVal = new Invocation ((Expression) yyVals[-3+yyTop], (Arguments) yyVals[-1+yyTop]);
		lbag.AddLocation (yyVal, GetLocation (yyVals[-2+yyTop]), GetLocation (yyVals[0+yyTop]));
	  }

void case_471()
#line 3428 "cs-parser.jay"
{
		Error_SyntaxError (yyToken);

		yyVal = new Invocation ((Expression) yyVals[-3+yyTop], (Arguments) yyVals[-1+yyTop]);
		lbag.AddLocation (yyVal, GetLocation (yyVals[-2+yyTop]));
	  }

void case_472()
#line 3435 "cs-parser.jay"
{
		Error_SyntaxError (yyToken);

		yyVal = new Invocation ((Expression) yyVals[-2+yyTop], null);
		lbag.AddLocation (yyVal, GetLocation (yyVals[-1+yyTop]));
	  }

void case_475()
#line 3450 "cs-parser.jay"
{
	  	if (yyVals[-1+yyTop] == null) {
	  		yyVal = new CollectionOrObjectInitializers (GetLocation (yyVals[-2+yyTop]));
	  	} else {
	  		yyVal = new CollectionOrObjectInitializers ((List<Expression>) yyVals[-1+yyTop], GetLocation (yyVals[-2+yyTop]));
	  	}
 		lbag.AddLocation (yyVal, GetLocation (yyVals[0+yyTop]));
	  }

void case_476()
#line 3459 "cs-parser.jay"
{
	  	yyVal = new CollectionOrObjectInitializers ((List<Expression>) yyVals[-2+yyTop], GetLocation (yyVals[-3+yyTop]));
	  	lbag.AddLocation (yyVal, GetLocation (yyVals[-1+yyTop]), GetLocation (yyVals[0+yyTop]));
	  }

void case_479()
#line 3475 "cs-parser.jay"
{
	  	var a = new List<Expression> ();
	  	a.Add ((Expression) yyVals[0+yyTop]);
	  	yyVal = a;
	  }

void case_480()
#line 3481 "cs-parser.jay"
{
	  	var a = (List<Expression>)yyVals[-2+yyTop];
	  	a.Add ((Expression) yyVals[0+yyTop]);
	  	yyVal = a;
	  }

void case_481()
#line 3486 "cs-parser.jay"
{
	  	Error_SyntaxError (yyToken);
		yyVal = yyVals[-1+yyTop];
	  }

void case_482()
#line 3494 "cs-parser.jay"
{
	  	var lt = (LocatedToken) yyVals[-2+yyTop];
	  	yyVal = new ElementInitializer (lt.Value, (Expression)yyVals[0+yyTop], lt.Location);
	  	lbag.AddLocation (yyVal, GetLocation (yyVals[-1+yyTop]));
	  }

void case_483()
#line 3500 "cs-parser.jay"
{
	  	var lt = (LocatedToken) Error_AwaitAsIdentifier (yyVals[-2+yyTop]);
	  	yyVal = new ElementInitializer (lt.Value, (Expression)yyVals[0+yyTop], lt.Location);
	  	lbag.AddLocation (yyVal, GetLocation (yyVals[-1+yyTop]));
	  }

void case_485()
#line 3509 "cs-parser.jay"
{
		CompletionSimpleName csn = yyVals[-1+yyTop] as CompletionSimpleName;
		if (csn == null)
			yyVal = new CollectionElementInitializer ((Expression)yyVals[-1+yyTop]);
		else
			yyVal = new CompletionElementInitializer (csn.Prefix, csn.Location);
	  }

void case_486()
#line 3517 "cs-parser.jay"
{
		if (yyVals[-1+yyTop] == null)
			yyVal = new CollectionElementInitializer (GetLocation (yyVals[-2+yyTop]));
		else
	  		yyVal = new CollectionElementInitializer ((List<Expression>)yyVals[-1+yyTop], GetLocation (yyVals[-2+yyTop]));

		lbag.AddLocation (yyVal, GetLocation (yyVals[0+yyTop]));
	  }

void case_487()
#line 3526 "cs-parser.jay"
{
		if (lang_version < LanguageVersion.V_6)
			FeatureIsNotAvailable (GetLocation (yyVals[-4+yyTop]), "dictionary initializer");

		yyVal = new DictionaryElementInitializer ((List<Expression>)yyVals[-3+yyTop], (Expression) yyVals[0+yyTop], GetLocation (yyVals[-4+yyTop]));
		lbag.AddLocation (yyVal, GetLocation (yyVals[-2+yyTop]), GetLocation (yyVals[-1+yyTop]));
	  }

void case_488()
#line 3534 "cs-parser.jay"
{
	  	report.Error (1920, GetLocation (yyVals[-1+yyTop]), "An element initializer cannot be empty");
		yyVal = new CollectionElementInitializer (GetLocation (yyVals[-1+yyTop]));
		lbag.AddLocation (yyVal, GetLocation (yyVals[0+yyTop]));
	  }

void case_493()
#line 3553 "cs-parser.jay"
{ 
		Arguments list = new Arguments (4);
		list.Add ((Argument) yyVals[0+yyTop]);
		yyVal = list;
	  }

void case_494()
#line 3559 "cs-parser.jay"
{
		Arguments list = (Arguments) yyVals[-2+yyTop];
		if (list [list.Count - 1] is NamedArgument)
			Error_NamedArgumentExpected ((NamedArgument) list [list.Count - 1]);
		
		list.Add ((Argument) yyVals[0+yyTop]);
		yyVal = list;
	  }

void case_495()
#line 3568 "cs-parser.jay"
{
		Arguments list = (Arguments) yyVals[-2+yyTop];
		NamedArgument a = (NamedArgument) yyVals[0+yyTop];
		for (int i = 0; i < list.Count; ++i) {
			NamedArgument na = list [i] as NamedArgument;
			if (na != null && na.Name == a.Name)
				report.Error (1740, na.Location, "Named argument `{0}' specified multiple times",
					na.Name);
		}
		
		list.Add (a);
		yyVal = list;
	  }

void case_496()
#line 3582 "cs-parser.jay"
{
	  	if (lexer.putback_char == -1)
	  		lexer.putback (')'); /* TODO: Wrong but what can I do*/
		Error_SyntaxError (yyToken);
		yyVal = yyVals[-2+yyTop];
	  }

void case_497()
#line 3589 "cs-parser.jay"
{
	  	report.Error (839, GetLocation (yyVals[-1+yyTop]), "An argument is missing");
	  	yyVal = null;
	  }

void case_502()
#line 3610 "cs-parser.jay"
{ 
		yyVal = new Argument ((Expression) yyVals[0+yyTop], Argument.AType.Ref);
		lbag.AddLocation (yyVal, GetLocation (yyVals[-1+yyTop]));
	  }

void case_504()
#line 3619 "cs-parser.jay"
{ 
		yyVal = new Argument ((Expression) yyVals[0+yyTop], Argument.AType.Out);
		lbag.AddLocation (yyVal, GetLocation (yyVals[-1+yyTop]));
	  }

void case_506()
#line 3628 "cs-parser.jay"
{
		yyVal = new Argument (new Arglist ((Arguments) yyVals[-1+yyTop], GetLocation (yyVals[-3+yyTop])));
		lbag.AddLocation (yyVal, GetLocation (yyVals[-2+yyTop]), GetLocation (yyVals[0+yyTop]));
	  }

void case_507()
#line 3633 "cs-parser.jay"
{
		yyVal = new Argument (new Arglist (GetLocation (yyVals[-2+yyTop])));
		lbag.AddLocation (yyVal, GetLocation (yyVals[-1+yyTop]), GetLocation (yyVals[0+yyTop]));
	  }

void case_508()
#line 3641 "cs-parser.jay"
{
		yyVal = new ParenthesizedExpression ((Expression) yyVals[-1+yyTop], GetLocation (yyVals[-2+yyTop]));
		lbag.AddLocation (yyVal, GetLocation (yyVals[-2+yyTop]), GetLocation (yyVals[0+yyTop]));
	  }

void case_509()
#line 3658 "cs-parser.jay"
{
		if (lang_version < LanguageVersion.V_6)
			FeatureIsNotAvailable (GetLocation (yyVals[-1+yyTop]), "declaration expression");

		var lt = (LocatedToken) yyVals[0+yyTop];
		var lv = new LocalVariable (current_block, lt.Value, lt.Location);
		current_block.AddLocalName (lv);
		yyVal = new DeclarationExpression ((FullNamedExpression) yyVals[-1+yyTop], lv);
	  }

void case_510()
#line 3668 "cs-parser.jay"
{
		if (lang_version < LanguageVersion.V_6)
			FeatureIsNotAvailable (GetLocation (yyVals[-3+yyTop]), "declaration expression");

		var lt = (LocatedToken) yyVals[-2+yyTop];
		var lv = new LocalVariable (current_block, lt.Value, lt.Location);
		current_block.AddLocalName (lv);
		yyVal = new DeclarationExpression ((FullNamedExpression) yyVals[-3+yyTop], lv) {
			Initializer = (Expression) yyVals[0+yyTop]
		};
	  }

void case_512()
#line 3687 "cs-parser.jay"
{
		yyVal = new ElementAccess ((Expression) yyVals[-3+yyTop], (Arguments) yyVals[-1+yyTop], GetLocation (yyVals[-2+yyTop]));
		lbag.AddLocation (yyVal, GetLocation (yyVals[0+yyTop]));
	  }

void case_513()
#line 3692 "cs-parser.jay"
{
		if (lang_version < LanguageVersion.V_6)
			FeatureIsNotAvailable (GetLocation (yyVals[-3+yyTop]), "null propagating operator");

		yyVal = new ElementAccess ((Expression) yyVals[-4+yyTop], (Arguments) yyVals[-1+yyTop], GetLocation (yyVals[-2+yyTop])) {
			ConditionalAccess = true
		};

		lbag.AddLocation (yyVal, GetLocation (yyVals[-3+yyTop]), GetLocation (yyVals[0+yyTop]));
	  }

void case_514()
#line 3703 "cs-parser.jay"
{
		Error_SyntaxError (yyToken);
		yyVal = new ElementAccess ((Expression) yyVals[-3+yyTop], (Arguments) yyVals[-1+yyTop], GetLocation (yyVals[-2+yyTop]));
	  }

void case_515()
#line 3708 "cs-parser.jay"
{
		Error_SyntaxError (yyToken);
		yyVal = new ElementAccess ((Expression) yyVals[-2+yyTop], null, GetLocation (yyVals[-1+yyTop]));
	  }

void case_516()
#line 3716 "cs-parser.jay"
{
		var list = new List<Expression> (4);
		list.Add ((Expression) yyVals[0+yyTop]);
		yyVal = list;
	  }

void case_517()
#line 3722 "cs-parser.jay"
{
		var list = (List<Expression>) yyVals[-2+yyTop];
		list.Add ((Expression) yyVals[0+yyTop]);
		yyVal = list;
	  }

void case_518()
#line 3731 "cs-parser.jay"
{
		Arguments args = new Arguments (4);
		args.Add ((Argument) yyVals[0+yyTop]);
		yyVal = args;
	  }

void case_519()
#line 3737 "cs-parser.jay"
{
		Arguments args = (Arguments) yyVals[-2+yyTop];
		if (args [args.Count - 1] is NamedArgument && !(yyVals[0+yyTop] is NamedArgument))
			Error_NamedArgumentExpected ((NamedArgument) args [args.Count - 1]);
	  
		args.Add ((Argument) yyVals[0+yyTop]);
		yyVal = args;	  
	  }

void case_523()
#line 3764 "cs-parser.jay"
{
	  	yyVal = new ElementAccess (new BaseThis (GetLocation (yyVals[-3+yyTop])), (Arguments) yyVals[-1+yyTop], GetLocation (yyVals[-2+yyTop]));
		lbag.AddLocation (yyVal, GetLocation (yyVals[0+yyTop]));
	  }

void case_524()
#line 3769 "cs-parser.jay"
{
	  	Error_SyntaxError (yyToken);
		yyVal = new ElementAccess (null, null, GetLocation (yyVals[-1+yyTop]));
	  }

void case_527()
#line 3791 "cs-parser.jay"
{
		if (yyVals[0+yyTop] != null) {
			if (lang_version <= LanguageVersion.ISO_2)
				FeatureIsNotAvailable (GetLocation (yyVals[-5+yyTop]), "object initializers");
				
			yyVal = new NewInitialize ((FullNamedExpression) yyVals[-4+yyTop], (Arguments) yyVals[-2+yyTop], (CollectionOrObjectInitializers) yyVals[0+yyTop], GetLocation (yyVals[-5+yyTop]));
		} else {
			yyVal = new New ((FullNamedExpression) yyVals[-4+yyTop], (Arguments) yyVals[-2+yyTop], GetLocation (yyVals[-5+yyTop]));
		}
		
		lbag.AddLocation (yyVal, GetLocation (yyVals[-3+yyTop]), GetLocation (yyVals[-1+yyTop]));
	  }

void case_528()
#line 3804 "cs-parser.jay"
{
		if (lang_version <= LanguageVersion.ISO_2)
			FeatureIsNotAvailable (GetLocation (yyVals[-2+yyTop]), "collection initializers");
	  
		yyVal = new NewInitialize ((FullNamedExpression) yyVals[-1+yyTop], null, (CollectionOrObjectInitializers) yyVals[0+yyTop], GetLocation (yyVals[-2+yyTop]));
	  }

void case_529()
#line 3816 "cs-parser.jay"
{
		yyVal = new ArrayCreation ((FullNamedExpression) yyVals[-5+yyTop], (List<Expression>) yyVals[-3+yyTop],
				new ComposedTypeSpecifier (((List<Expression>) yyVals[-3+yyTop]).Count, GetLocation (yyVals[-4+yyTop])) {
	  				Next = (ComposedTypeSpecifier) yyVals[-1+yyTop]
			  	}, (ArrayInitializer) yyVals[0+yyTop], GetLocation (yyVals[-6+yyTop]));
		lbag.AddLocation (yyVal, GetLocation (yyVals[-4+yyTop]), GetLocation (yyVals[-2+yyTop]));
	  }

void case_530()
#line 3824 "cs-parser.jay"
{
	  	if (yyVals[0+yyTop] == null)
	  		report.Error (1586, GetLocation (yyVals[-3+yyTop]), "Array creation must have array size or array initializer");

		yyVal = new ArrayCreation ((FullNamedExpression) yyVals[-2+yyTop], (ComposedTypeSpecifier) yyVals[-1+yyTop], (ArrayInitializer) yyVals[0+yyTop], GetLocation (yyVals[-3+yyTop]));
	  }

void case_531()
#line 3831 "cs-parser.jay"
{
		if (lang_version <= LanguageVersion.ISO_2)
			FeatureIsNotAvailable (GetLocation (yyVals[-2+yyTop]), "implicitly typed arrays");
	  
		yyVal = new ImplicitlyTypedArrayCreation ((ComposedTypeSpecifier) yyVals[-1+yyTop], (ArrayInitializer) yyVals[0+yyTop], GetLocation (yyVals[-2+yyTop]));
	  }

void case_532()
#line 3838 "cs-parser.jay"
{
		report.Error (178, GetLocation (yyVals[-1+yyTop]), "Invalid rank specifier, expecting `,' or `]'");
		yyVal = new ArrayCreation ((FullNamedExpression) yyVals[-5+yyTop], null, GetLocation (yyVals[-6+yyTop]));
	  }

void case_533()
#line 3843 "cs-parser.jay"
{
		Error_SyntaxError (yyToken);
		/* It can be any of new expression, create the most common one*/
		yyVal = new New ((FullNamedExpression) yyVals[-1+yyTop], null, GetLocation (yyVals[-2+yyTop]));
	  }

void case_535()
#line 3855 "cs-parser.jay"
{
		--lexer.parsing_type;
		yyVal = yyVals[0+yyTop];
	  }

void case_536()
#line 3863 "cs-parser.jay"
{
		if (lang_version <= LanguageVersion.ISO_2)
	  		FeatureIsNotAvailable (GetLocation (yyVals[-3+yyTop]), "anonymous types");

		yyVal = new NewAnonymousType ((List<AnonymousTypeParameter>) yyVals[-1+yyTop], current_container, GetLocation (yyVals[-3+yyTop]));
		
		/* TODO: lbag comma location*/
		lbag.AddLocation (yyVal, GetLocation (yyVals[-2+yyTop]), GetLocation (yyVals[0+yyTop]));
	  }

void case_542()
#line 3890 "cs-parser.jay"
{
	  	var a = new List<AnonymousTypeParameter> (4);
	  	a.Add ((AnonymousTypeParameter) yyVals[0+yyTop]);
	  	yyVal = a;
	  }

void case_543()
#line 3896 "cs-parser.jay"
{
	  	var a = (List<AnonymousTypeParameter>) yyVals[-2+yyTop];
	  	a.Add ((AnonymousTypeParameter) yyVals[0+yyTop]);
	  	yyVal = a;
	  }

void case_546()
#line 3913 "cs-parser.jay"
{
		var lt = (LocatedToken)yyVals[-2+yyTop];
	  	yyVal = new AnonymousTypeParameter ((Expression)yyVals[0+yyTop], lt.Value, lt.Location);
	  	lbag.AddLocation (yyVal, GetLocation (yyVals[-1+yyTop]));
	  }

void case_547()
#line 3919 "cs-parser.jay"
{
		var lt = (LocatedToken)yyVals[0+yyTop];
	  	yyVal = new AnonymousTypeParameter (new SimpleName (lt.Value, lt.Location),
	  		lt.Value, lt.Location);
	  }

void case_548()
#line 3925 "cs-parser.jay"
{
	  	MemberAccess ma = (MemberAccess) yyVals[0+yyTop];
	  	yyVal = new AnonymousTypeParameter (ma, ma.Name, ma.Location);
	  }

void case_549()
#line 3930 "cs-parser.jay"
{
		report.Error (746, lexer.Location,
			"Invalid anonymous type member declarator. Anonymous type members must be a member assignment, simple name or member access expression");
		yyVal = null;
	  }

void case_553()
#line 3945 "cs-parser.jay"
{
	  	((ComposedTypeSpecifier) yyVals[-1+yyTop]).Next = (ComposedTypeSpecifier) yyVals[0+yyTop];
	  	yyVal = yyVals[-1+yyTop];
	  }

void case_554()
#line 3953 "cs-parser.jay"
{
		yyVal = ComposedTypeSpecifier.CreateArrayDimension (1, GetLocation (yyVals[-1+yyTop]));
		lbag.AddLocation (yyVal, GetLocation (yyVals[0+yyTop]));
	  }

void case_555()
#line 3958 "cs-parser.jay"
{
		yyVal = ComposedTypeSpecifier.CreateArrayDimension ((int)yyVals[-1+yyTop], GetLocation (yyVals[-2+yyTop]));
		lbag.AddLocation (yyVal, GetLocation (yyVals[0+yyTop]));
	  }

void case_560()
#line 3988 "cs-parser.jay"
{
		var ai = new ArrayInitializer (0, GetLocation (yyVals[-1+yyTop]));
		ai.VariableDeclaration = current_variable;
		lbag.AddLocation (ai, GetLocation (yyVals[0+yyTop]));
		yyVal = ai;
	  }

void case_561()
#line 3995 "cs-parser.jay"
{
		var ai = new ArrayInitializer ((List<Expression>) yyVals[-2+yyTop], GetLocation (yyVals[-3+yyTop]));
		ai.VariableDeclaration = current_variable;
		if (yyVals[-1+yyTop] != null) {
			lbag.AddLocation (ai, GetLocation (yyVals[-1+yyTop]), GetLocation (yyVals[0+yyTop]));
		} else {
			lbag.AddLocation (ai, GetLocation (yyVals[0+yyTop]));
		}
		yyVal = ai;
	  }

void case_562()
#line 4009 "cs-parser.jay"
{
		var list = new List<Expression> (4);
		list.Add ((Expression) yyVals[0+yyTop]);
		yyVal = list;
	  }

void case_563()
#line 4015 "cs-parser.jay"
{
		var list = (List<Expression>) yyVals[-2+yyTop];
		list.Add ((Expression) yyVals[0+yyTop]);
		yyVal = list;
	  }

void case_564()
#line 4024 "cs-parser.jay"
{
		yyVal = new TypeOf ((FullNamedExpression) yyVals[-1+yyTop], GetLocation (yyVals[-3+yyTop]));
		lbag.AddLocation (yyVal, GetLocation (yyVals[-2+yyTop]), GetLocation (yyVals[0+yyTop]));
	  }

void case_566()
#line 4033 "cs-parser.jay"
{
	 	Error_TypeExpected (lexer.Location);
	 	yyVal = null;
	 }

void case_567()
#line 4041 "cs-parser.jay"
{
		if (lang_version < LanguageVersion.ISO_2)
			FeatureIsNotAvailable (GetLocation (yyVals[0+yyTop]), "generics");

		yyVal = yyVals[0+yyTop];
	  }

void case_568()
#line 4051 "cs-parser.jay"
{
		var lt = (LocatedToken) yyVals[-1+yyTop];
		if (lang_version == LanguageVersion.ISO_1)
			FeatureIsNotAvailable (lt.Location, "namespace alias qualifier");

		yyVal = lt;		
	  }

void case_569()
#line 4062 "cs-parser.jay"
{ 
		yyVal = new SizeOf ((Expression) yyVals[-1+yyTop], GetLocation (yyVals[-3+yyTop]));
		lbag.AddLocation (yyVal, GetLocation (yyVals[-2+yyTop]), GetLocation (yyVals[0+yyTop]));
	  }

void case_570()
#line 4067 "cs-parser.jay"
{
		Error_SyntaxError (yyToken);

		yyVal = new SizeOf ((Expression) yyVals[-1+yyTop], GetLocation (yyVals[-3+yyTop]));
		lbag.AddLocation (yyVal, GetLocation (yyVals[-2+yyTop]));
	  }

void case_571()
#line 4077 "cs-parser.jay"
{
		yyVal = new CheckedExpr ((Expression) yyVals[-1+yyTop], GetLocation (yyVals[-3+yyTop]));
		lbag.AddLocation (yyVal, GetLocation (yyVals[-2+yyTop]), GetLocation (yyVals[0+yyTop]));
	  }

void case_572()
#line 4082 "cs-parser.jay"
{
		Error_SyntaxError (yyToken);

		yyVal = new CheckedExpr (null, GetLocation (yyVals[-1+yyTop]));
	  }

void case_573()
#line 4091 "cs-parser.jay"
{
		yyVal = new UnCheckedExpr ((Expression) yyVals[-1+yyTop], GetLocation (yyVals[-3+yyTop]));
		lbag.AddLocation (yyVal, GetLocation (yyVals[-2+yyTop]), GetLocation (yyVals[0+yyTop]));
	  }

void case_574()
#line 4096 "cs-parser.jay"
{
		Error_SyntaxError (yyToken);

		yyVal = new UnCheckedExpr (null, GetLocation (yyVals[-1+yyTop]));
	  }

void case_575()
#line 4105 "cs-parser.jay"
{
		var lt = (LocatedToken) yyVals[-1+yyTop];
		yyVal = new MemberAccess (new Indirection ((Expression) yyVals[-3+yyTop], GetLocation (yyVals[-2+yyTop])), lt.Value, (TypeArguments) yyVals[0+yyTop], lt.Location);
	  }

void case_583()
#line 4144 "cs-parser.jay"
{
		valid_param_mod = 0;
	  	yyVal = yyVals[-1+yyTop];
	  }

void case_584()
#line 4152 "cs-parser.jay"
{
		if (lang_version < LanguageVersion.ISO_2)
			FeatureIsNotAvailable (GetLocation (yyVals[-3+yyTop]), "default value expression");

		yyVal = new DefaultValueExpression ((Expression) yyVals[-1+yyTop], GetLocation (yyVals[-3+yyTop]));
		lbag.AddLocation (yyVal, GetLocation (yyVals[-2+yyTop]), GetLocation (yyVals[0+yyTop]));
	  }

void case_588()
#line 4172 "cs-parser.jay"
{
		yyVal = new Cast ((FullNamedExpression) yyVals[-2+yyTop], (Expression) yyVals[0+yyTop], GetLocation (yyVals[-3+yyTop]));
		lbag.AddLocation (yyVal, GetLocation (yyVals[-1+yyTop]));
	  }

void case_589()
#line 4177 "cs-parser.jay"
{
		if (!async_block) {
			 if (current_anonymous_method is LambdaExpression) {
				report.Error (4034, GetLocation (yyVals[-1+yyTop]),
					"The `await' operator can only be used when its containing lambda expression is marked with the `async' modifier");
			} else if (current_anonymous_method != null) {
				report.Error (4035, GetLocation (yyVals[-1+yyTop]),
					"The `await' operator can only be used when its containing anonymous method is marked with the `async' modifier");
			} else if (interactive_async != null) {
				current_block.Explicit.RegisterAsyncAwait ();
				interactive_async = true;
			} else {
				report.Error (4033, GetLocation (yyVals[-1+yyTop]),
					"The `await' operator can only be used when its containing method is marked with the `async' modifier");
			}
		} else {
			current_block.Explicit.RegisterAsyncAwait ();
		}
		
		yyVal = new Await ((Expression) yyVals[0+yyTop], GetLocation (yyVals[-1+yyTop]));
	  }

void case_590()
#line 4199 "cs-parser.jay"
{
		Error_SyntaxError (yyToken);

		yyVal = new Unary (Unary.Operator.LogicalNot, null, GetLocation (yyVals[-1+yyTop]));
	  }

void case_591()
#line 4205 "cs-parser.jay"
{
		Error_SyntaxError (yyToken);

		yyVal = new Unary (Unary.Operator.OnesComplement, null, GetLocation (yyVals[-1+yyTop]));
	  }

void case_592()
#line 4211 "cs-parser.jay"
{
		Error_SyntaxError (yyToken);

		yyVal = new Cast ((FullNamedExpression) yyVals[-2+yyTop], null, GetLocation (yyVals[-3+yyTop]));
		lbag.AddLocation (yyVal, GetLocation (yyVals[-1+yyTop]));
	  }

void case_593()
#line 4218 "cs-parser.jay"
{
		Error_SyntaxError (yyToken);

		yyVal = new Await (null, GetLocation (yyVals[-1+yyTop]));
	  }

void case_601()
#line 4256 "cs-parser.jay"
{ 
		Error_SyntaxError (yyToken);

	  	yyVal = new Unary (Unary.Operator.UnaryPlus, null, GetLocation (yyVals[-1+yyTop]));
	  }

void case_602()
#line 4262 "cs-parser.jay"
{ 
		Error_SyntaxError (yyToken);

		yyVal = new Unary (Unary.Operator.UnaryNegation, null, GetLocation (yyVals[-1+yyTop]));
	  }

void case_603()
#line 4268 "cs-parser.jay"
{
		Error_SyntaxError (yyToken);

		yyVal = new UnaryMutator (UnaryMutator.Mode.PreIncrement, null, GetLocation (yyVals[-1+yyTop]));
	  }

void case_604()
#line 4274 "cs-parser.jay"
{
		Error_SyntaxError (yyToken);

		yyVal = new UnaryMutator (UnaryMutator.Mode.PreDecrement, null, GetLocation (yyVals[-1+yyTop]));
	  }

void case_605()
#line 4280 "cs-parser.jay"
{
		Error_SyntaxError (yyToken);

		yyVal = new Indirection (null, GetLocation (yyVals[-1+yyTop]));
	  }

void case_606()
#line 4286 "cs-parser.jay"
{
		Error_SyntaxError (yyToken);

		yyVal = new Unary (Unary.Operator.AddressOf, null, GetLocation (yyVals[-1+yyTop]));
	  }

void case_608()
#line 4296 "cs-parser.jay"
{
		yyVal = new Binary (Binary.Operator.Multiply, (Expression) yyVals[-2+yyTop], (Expression) yyVals[0+yyTop]);
		lbag.AddLocation (yyVal, GetLocation (yyVals[-1+yyTop]));
	  }

void case_609()
#line 4301 "cs-parser.jay"
{
		yyVal = new Binary (Binary.Operator.Division, (Expression) yyVals[-2+yyTop], (Expression) yyVals[0+yyTop]);
		lbag.AddLocation (yyVal, GetLocation (yyVals[-1+yyTop]));
	  }

void case_610()
#line 4306 "cs-parser.jay"
{
		yyVal = new Binary (Binary.Operator.Modulus, (Expression) yyVals[-2+yyTop], (Expression) yyVals[0+yyTop]);
		lbag.AddLocation (yyVal, GetLocation (yyVals[-1+yyTop]));
	  }

void case_611()
#line 4311 "cs-parser.jay"
{
		Error_SyntaxError (yyToken);

		yyVal = new Binary (Binary.Operator.Multiply, (Expression) yyVals[-2+yyTop], null);
		lbag.AddLocation (yyVal, GetLocation (yyVals[-1+yyTop]));
	  }

void case_612()
#line 4318 "cs-parser.jay"
{
		Error_SyntaxError (yyToken);

		yyVal = new Binary (Binary.Operator.Division, (Expression) yyVals[-2+yyTop], null);
		lbag.AddLocation (yyVal, GetLocation (yyVals[-1+yyTop]));
	  }

void case_613()
#line 4325 "cs-parser.jay"
{
		Error_SyntaxError (yyToken);

		yyVal = new Binary (Binary.Operator.Modulus, (Expression) yyVals[-2+yyTop], null);
		lbag.AddLocation (yyVal, GetLocation (yyVals[-1+yyTop]));
	  }

void case_615()
#line 4336 "cs-parser.jay"
{
		yyVal = new Binary (Binary.Operator.Addition, (Expression) yyVals[-2+yyTop], (Expression) yyVals[0+yyTop]);
		lbag.AddLocation (yyVal, GetLocation (yyVals[-1+yyTop]));
	  }

void case_616()
#line 4341 "cs-parser.jay"
{
		yyVal = new Binary (Binary.Operator.Subtraction, (Expression) yyVals[-2+yyTop], (Expression) yyVals[0+yyTop]);
		lbag.AddLocation (yyVal, GetLocation (yyVals[-1+yyTop]));
	  }

void case_617()
#line 4346 "cs-parser.jay"
{
		Error_SyntaxError (yyToken);

		yyVal = new Binary (Binary.Operator.Addition, (Expression) yyVals[-2+yyTop], null);
		lbag.AddLocation (yyVal, GetLocation (yyVals[-1+yyTop]));
	  }

void case_618()
#line 4353 "cs-parser.jay"
{
		Error_SyntaxError (yyToken);

		yyVal = new Binary (Binary.Operator.Subtraction, (Expression) yyVals[-2+yyTop], null);
		lbag.AddLocation (yyVal, GetLocation (yyVals[-1+yyTop]));
	  }

void case_620()
#line 4364 "cs-parser.jay"
{
		var is_expr = new Is ((Expression) yyVals[-3+yyTop], (Expression) yyVals[-1+yyTop], GetLocation (yyVals[-2+yyTop]));
		if (yyVals[0+yyTop] != null) {
			if (lang_version != LanguageVersion.Experimental)
				FeatureIsNotAvailable (GetLocation (yyVals[0+yyTop]), "type pattern matching");

			var lt = (LocatedToken) yyVals[0+yyTop];
			is_expr.Variable = new LocalVariable (current_block, lt.Value, lt.Location);
			current_block.AddLocalName (is_expr.Variable);
		}

		yyVal = is_expr;
	  }

void case_621()
#line 4378 "cs-parser.jay"
{
		var is_expr = new Is ((Expression) yyVals[-2+yyTop], (Expression) yyVals[0+yyTop], GetLocation (yyVals[-1+yyTop]));
		if (lang_version != LanguageVersion.Experimental)
			FeatureIsNotAvailable (GetLocation (yyVals[-1+yyTop]), "pattern matching");

		yyVal = is_expr;
	  }

void case_622()
#line 4386 "cs-parser.jay"
{
		Error_SyntaxError (yyToken);

		yyVal = new As ((Expression) yyVals[-2+yyTop], null, GetLocation (yyVals[-1+yyTop]));
	  }

void case_623()
#line 4392 "cs-parser.jay"
{
		Error_SyntaxError (yyToken);

		yyVal = new Is ((Expression) yyVals[-2+yyTop], null, GetLocation (yyVals[-1+yyTop]));
	  }

void case_624()
#line 4398 "cs-parser.jay"
{
		var lt = (LocatedToken) yyVals[-2+yyTop];
		yyVal = new Is (new SimpleName (lt.Value, lt.Location), (Expression) yyVals[0+yyTop], GetLocation (yyVals[-1+yyTop]));
	  }

void case_625()
#line 4403 "cs-parser.jay"
{
		var lt = (LocatedToken) yyVals[-2+yyTop];
		yyVal = new As (new SimpleName (lt.Value, lt.Location), (Expression) yyVals[0+yyTop], GetLocation (yyVals[-1+yyTop]));
	  }

void case_632()
#line 4426 "cs-parser.jay"
{
		yyVal = new Cast ((FullNamedExpression) yyVals[-2+yyTop], (Expression) yyVals[0+yyTop], GetLocation (yyVals[-3+yyTop]));
		lbag.AddLocation (yyVal, GetLocation (yyVals[-1+yyTop]));
	  }

void case_638()
#line 4454 "cs-parser.jay"
{
		var list = new List<PropertyPatternMember> ();
		list.Add ((PropertyPatternMember) yyVals[0+yyTop]);
		yyVal = list;
	  }

void case_639()
#line 4460 "cs-parser.jay"
{
		var list = (List<PropertyPatternMember>) yyVals[-2+yyTop];
		list.Add ((PropertyPatternMember) yyVals[0+yyTop]);
		yyVal = list;
	  }

void case_640()
#line 4469 "cs-parser.jay"
{
		var lt = (LocatedToken) yyVals[-2+yyTop];
		yyVal = new PropertyPatternMember (lt.Value, (Expression) yyVals[0+yyTop], lt.Location);
	  }

void case_642()
#line 4478 "cs-parser.jay"
{
	  	if (yyVals[0+yyTop] != null) {
			var lt = (LocatedToken) yyVals[0+yyTop];
			var variable = new LocalVariable (current_block, lt.Value, lt.Location);
			current_block.AddLocalName (variable);
		}
	  }

void case_645()
#line 4497 "cs-parser.jay"
{
		Arguments args = new Arguments (4);
		args.Add ((Argument) yyVals[0+yyTop]);
		yyVal = args;
	  }

void case_646()
#line 4503 "cs-parser.jay"
{
		Arguments args = (Arguments) yyVals[-2+yyTop];
		if (args [args.Count - 1] is NamedArgument && !(yyVals[0+yyTop] is NamedArgument))
			Error_NamedArgumentExpected ((NamedArgument) args [args.Count - 1]);

		args.Add ((Argument) yyVals[0+yyTop]);
		yyVal = args;
	  }

void case_648()
#line 4519 "cs-parser.jay"
{
		var lt = (LocatedToken) yyVals[-2+yyTop];
		yyVal = new NamedArgument (lt.Value, lt.Location, (Expression) yyVals[0+yyTop]);
	  }

void case_650()
#line 4528 "cs-parser.jay"
{
		yyVal = new Binary (Binary.Operator.LeftShift, (Expression) yyVals[-2+yyTop], (Expression) yyVals[0+yyTop]);
		lbag.AddLocation (yyVal, GetLocation (yyVals[-1+yyTop]));
	  }

void case_651()
#line 4533 "cs-parser.jay"
{
		yyVal = new Binary (Binary.Operator.RightShift, (Expression) yyVals[-2+yyTop], (Expression) yyVals[0+yyTop]);
		lbag.AddLocation (yyVal, GetLocation (yyVals[-1+yyTop]));
	  }

void case_652()
#line 4538 "cs-parser.jay"
{
		Error_SyntaxError (yyToken);

		yyVal = new Binary (Binary.Operator.LeftShift, (Expression) yyVals[-2+yyTop], null);
		lbag.AddLocation (yyVal, GetLocation (yyVals[-1+yyTop]));
	  }

void case_653()
#line 4545 "cs-parser.jay"
{
		Error_SyntaxError (yyToken);

		yyVal = new Binary (Binary.Operator.RightShift, (Expression) yyVals[-2+yyTop], null);
		lbag.AddLocation (yyVal, GetLocation (yyVals[-1+yyTop]));
	  }

void case_655()
#line 4556 "cs-parser.jay"
{
		yyVal = new Binary (Binary.Operator.LessThan, (Expression) yyVals[-2+yyTop], (Expression) yyVals[0+yyTop]);
		lbag.AddLocation (yyVal, GetLocation (yyVals[-1+yyTop]));
	  }

void case_656()
#line 4561 "cs-parser.jay"
{
		yyVal = new Binary (Binary.Operator.GreaterThan, (Expression) yyVals[-2+yyTop], (Expression) yyVals[0+yyTop]);
		lbag.AddLocation (yyVal, GetLocation (yyVals[-1+yyTop]));
	  }

void case_657()
#line 4566 "cs-parser.jay"
{
		yyVal = new Binary (Binary.Operator.LessThanOrEqual, (Expression) yyVals[-2+yyTop], (Expression) yyVals[0+yyTop]);
		lbag.AddLocation (yyVal, GetLocation (yyVals[-1+yyTop]));
	  }

void case_658()
#line 4571 "cs-parser.jay"
{
		yyVal = new Binary (Binary.Operator.GreaterThanOrEqual, (Expression) yyVals[-2+yyTop], (Expression) yyVals[0+yyTop]);
		lbag.AddLocation (yyVal, GetLocation (yyVals[-1+yyTop]));
	  }

void case_659()
#line 4576 "cs-parser.jay"
{
		Error_SyntaxError (yyToken);

		yyVal = new Binary (Binary.Operator.LessThan, (Expression) yyVals[-2+yyTop], null);
		lbag.AddLocation (yyVal, GetLocation (yyVals[-1+yyTop]));
	  }

void case_660()
#line 4583 "cs-parser.jay"
{
		Error_SyntaxError (yyToken);

		yyVal = new Binary (Binary.Operator.GreaterThan, (Expression) yyVals[-2+yyTop], null);
		lbag.AddLocation (yyVal, GetLocation (yyVals[-1+yyTop]));
	  }

void case_661()
#line 4590 "cs-parser.jay"
{
		Error_SyntaxError (yyToken);

		yyVal = new Binary (Binary.Operator.LessThanOrEqual, (Expression) yyVals[-2+yyTop], null);
		lbag.AddLocation (yyVal, GetLocation (yyVals[-1+yyTop]));
	  }

void case_662()
#line 4597 "cs-parser.jay"
{
		Error_SyntaxError (yyToken);

		yyVal = new Binary (Binary.Operator.GreaterThanOrEqual, (Expression) yyVals[-2+yyTop], null);
		lbag.AddLocation (yyVal, GetLocation (yyVals[-1+yyTop]));
	  }

void case_664()
#line 4608 "cs-parser.jay"
{
		yyVal = new Binary (Binary.Operator.Equality, (Expression) yyVals[-2+yyTop], (Expression) yyVals[0+yyTop]);
		lbag.AddLocation (yyVal, GetLocation (yyVals[-1+yyTop]));
	  }

void case_665()
#line 4613 "cs-parser.jay"
{
		yyVal = new Binary (Binary.Operator.Inequality, (Expression) yyVals[-2+yyTop], (Expression) yyVals[0+yyTop]);
		lbag.AddLocation (yyVal, GetLocation (yyVals[-1+yyTop]));
	  }

void case_666()
#line 4618 "cs-parser.jay"
{
		Error_SyntaxError (yyToken);

		yyVal = new Binary (Binary.Operator.Equality, (Expression) yyVals[-2+yyTop], null);
		lbag.AddLocation (yyVal, GetLocation (yyVals[-1+yyTop]));
	  }

void case_667()
#line 4625 "cs-parser.jay"
{
		Error_SyntaxError (yyToken);

		yyVal = new Binary (Binary.Operator.Inequality, (Expression) yyVals[-2+yyTop], null);
		lbag.AddLocation (yyVal, GetLocation (yyVals[-1+yyTop]));
	  }

void case_669()
#line 4636 "cs-parser.jay"
{
		yyVal = new Binary (Binary.Operator.BitwiseAnd, (Expression) yyVals[-2+yyTop], (Expression) yyVals[0+yyTop]);
		lbag.AddLocation (yyVal, GetLocation (yyVals[-1+yyTop]));
	  }

void case_670()
#line 4641 "cs-parser.jay"
{
		Error_SyntaxError (yyToken);

		yyVal = new Binary (Binary.Operator.BitwiseAnd, (Expression) yyVals[-2+yyTop], null);
		lbag.AddLocation (yyVal, GetLocation (yyVals[-1+yyTop]));
	  }

void case_672()
#line 4652 "cs-parser.jay"
{
		yyVal = new Binary (Binary.Operator.ExclusiveOr, (Expression) yyVals[-2+yyTop], (Expression) yyVals[0+yyTop]);
		lbag.AddLocation (yyVal, GetLocation (yyVals[-1+yyTop]));
	  }

void case_673()
#line 4657 "cs-parser.jay"
{
		Error_SyntaxError (yyToken);

		yyVal = new Binary (Binary.Operator.ExclusiveOr, (Expression) yyVals[-2+yyTop], null);
		lbag.AddLocation (yyVal, GetLocation (yyVals[-1+yyTop]));
	  }

void case_675()
#line 4668 "cs-parser.jay"
{
		yyVal = new Binary (Binary.Operator.BitwiseOr, (Expression) yyVals[-2+yyTop], (Expression) yyVals[0+yyTop]);
		lbag.AddLocation (yyVal, GetLocation (yyVals[-1+yyTop]));
	  }

void case_676()
#line 4673 "cs-parser.jay"
{
		Error_SyntaxError (yyToken);

		yyVal = new Binary (Binary.Operator.BitwiseOr, (Expression) yyVals[-2+yyTop], null);
		lbag.AddLocation (yyVal, GetLocation (yyVals[-1+yyTop]));
	  }

void case_678()
#line 4684 "cs-parser.jay"
{
		yyVal = new Binary (Binary.Operator.LogicalAnd, (Expression) yyVals[-2+yyTop], (Expression) yyVals[0+yyTop]);
		lbag.AddLocation (yyVal, GetLocation (yyVals[-1+yyTop]));
	  }

void case_679()
#line 4689 "cs-parser.jay"
{
		Error_SyntaxError (yyToken);

		yyVal = new Binary (Binary.Operator.LogicalAnd, (Expression) yyVals[-2+yyTop], null);
		lbag.AddLocation (yyVal, GetLocation (yyVals[-1+yyTop]));
	  }

void case_681()
#line 4700 "cs-parser.jay"
{
		yyVal = new Binary (Binary.Operator.LogicalOr, (Expression) yyVals[-2+yyTop], (Expression) yyVals[0+yyTop]);
		lbag.AddLocation (yyVal, GetLocation (yyVals[-1+yyTop]));
	  }

void case_682()
#line 4705 "cs-parser.jay"
{
		Error_SyntaxError (yyToken);

		yyVal = new Binary (Binary.Operator.LogicalOr, (Expression) yyVals[-2+yyTop], null);
		lbag.AddLocation (yyVal, GetLocation (yyVals[-1+yyTop]));
	  }

void case_684()
#line 4716 "cs-parser.jay"
{
		if (lang_version < LanguageVersion.ISO_2)
			FeatureIsNotAvailable (GetLocation (yyVals[-1+yyTop]), "null coalescing operator");
			
		yyVal = new Nullable.NullCoalescingOperator ((Expression) yyVals[-2+yyTop], (Expression) yyVals[0+yyTop]);
		lbag.AddLocation (yyVal, GetLocation (yyVals[-1+yyTop]));
	  }

void case_686()
#line 4728 "cs-parser.jay"
{
		yyVal = new Conditional (new BooleanExpression ((Expression) yyVals[-4+yyTop]), (Expression) yyVals[-2+yyTop], (Expression) yyVals[0+yyTop], GetLocation (yyVals[-3+yyTop]));
		lbag.AddLocation (yyVal, GetLocation (yyVals[-1+yyTop]));
	  }

void case_687()
#line 4733 "cs-parser.jay"
{
		Error_SyntaxError (yyToken);

		yyVal = new Conditional (new BooleanExpression ((Expression) yyVals[-3+yyTop]), (Expression) yyVals[-1+yyTop], null, GetLocation (yyVals[-2+yyTop]));
	  }

void case_688()
#line 4739 "cs-parser.jay"
{
		Error_SyntaxError (yyToken);

		yyVal = new Conditional (new BooleanExpression ((Expression) yyVals[-4+yyTop]), (Expression) yyVals[-2+yyTop], null, GetLocation (yyVals[-3+yyTop]));
		lbag.AddLocation (yyVal, GetLocation (yyVals[-1+yyTop]));
	  }

void case_689()
#line 4746 "cs-parser.jay"
{
		Error_SyntaxError (Token.CLOSE_BRACE);

		yyVal = new Conditional (new BooleanExpression ((Expression) yyVals[-4+yyTop]), (Expression) yyVals[-2+yyTop], null, GetLocation (yyVals[-3+yyTop]));
		lbag.AddLocation (yyVal, GetLocation (yyVals[-1+yyTop]));
		lexer.putback ('}');
	  }

void case_690()
#line 4757 "cs-parser.jay"
{
		yyVal = new SimpleAssign ((Expression) yyVals[-2+yyTop], (Expression) yyVals[0+yyTop]);
		lbag.AddLocation (yyVal, GetLocation (yyVals[-1+yyTop]));
	  }

void case_691()
#line 4762 "cs-parser.jay"
{
		yyVal = new CompoundAssign (Binary.Operator.Multiply, (Expression) yyVals[-2+yyTop], (Expression) yyVals[0+yyTop]);
		lbag.AddLocation (yyVal, GetLocation (yyVals[-1+yyTop]));
	  }

void case_692()
#line 4767 "cs-parser.jay"
{
		yyVal = new CompoundAssign (Binary.Operator.Division, (Expression) yyVals[-2+yyTop], (Expression) yyVals[0+yyTop]);
		lbag.AddLocation (yyVal, GetLocation (yyVals[-1+yyTop]));
	  }

void case_693()
#line 4772 "cs-parser.jay"
{
		yyVal = new CompoundAssign (Binary.Operator.Modulus, (Expression) yyVals[-2+yyTop], (Expression) yyVals[0+yyTop]);
		lbag.AddLocation (yyVal, GetLocation (yyVals[-1+yyTop]));
	  }

void case_694()
#line 4777 "cs-parser.jay"
{
		yyVal = new CompoundAssign (Binary.Operator.Addition, (Expression) yyVals[-2+yyTop], (Expression) yyVals[0+yyTop]);
		lbag.AddLocation (yyVal, GetLocation (yyVals[-1+yyTop]));
	  }

void case_695()
#line 4782 "cs-parser.jay"
{
		yyVal = new CompoundAssign (Binary.Operator.Subtraction, (Expression) yyVals[-2+yyTop], (Expression) yyVals[0+yyTop]);
		lbag.AddLocation (yyVal, GetLocation (yyVals[-1+yyTop]));
	  }

void case_696()
#line 4787 "cs-parser.jay"
{
		yyVal = new CompoundAssign (Binary.Operator.LeftShift, (Expression) yyVals[-2+yyTop], (Expression) yyVals[0+yyTop]);
		lbag.AddLocation (yyVal, GetLocation (yyVals[-1+yyTop]));
	  }

void case_697()
#line 4792 "cs-parser.jay"
{
		yyVal = new CompoundAssign (Binary.Operator.RightShift, (Expression) yyVals[-2+yyTop], (Expression) yyVals[0+yyTop]);
		lbag.AddLocation (yyVal, GetLocation (yyVals[-1+yyTop]));
	  }

void case_698()
#line 4797 "cs-parser.jay"
{
		yyVal = new CompoundAssign (Binary.Operator.BitwiseAnd, (Expression) yyVals[-2+yyTop], (Expression) yyVals[0+yyTop]);
		lbag.AddLocation (yyVal, GetLocation (yyVals[-1+yyTop]));
	  }

void case_699()
#line 4802 "cs-parser.jay"
{
		yyVal = new CompoundAssign (Binary.Operator.BitwiseOr, (Expression) yyVals[-2+yyTop], (Expression) yyVals[0+yyTop]);
		lbag.AddLocation (yyVal, GetLocation (yyVals[-1+yyTop]));
	  }

void case_700()
#line 4807 "cs-parser.jay"
{
		yyVal = new CompoundAssign (Binary.Operator.ExclusiveOr, (Expression) yyVals[-2+yyTop], (Expression) yyVals[0+yyTop]);
		lbag.AddLocation (yyVal, GetLocation (yyVals[-1+yyTop]));
	  }

void case_701()
#line 4815 "cs-parser.jay"
{
		var pars = new List<Parameter> (4);
		pars.Add ((Parameter) yyVals[0+yyTop]);

		yyVal = pars;
	  }

void case_702()
#line 4822 "cs-parser.jay"
{
		var pars = (List<Parameter>) yyVals[-2+yyTop];
		Parameter p = (Parameter)yyVals[0+yyTop];
		if (pars[0].GetType () != p.GetType ()) {
			report.Error (748, p.Location, "All lambda parameters must be typed either explicitly or implicitly");
		}
		
		pars.Add (p);
		yyVal = pars;
	  }

void case_703()
#line 4836 "cs-parser.jay"
{
		var lt = (LocatedToken) yyVals[0+yyTop];

		yyVal = new Parameter ((FullNamedExpression) yyVals[-1+yyTop], lt.Value, (Parameter.Modifier) yyVals[-2+yyTop], null, lt.Location);
	  }

void case_704()
#line 4842 "cs-parser.jay"
{
		var lt = (LocatedToken) yyVals[0+yyTop];

		yyVal = new Parameter ((FullNamedExpression) yyVals[-1+yyTop], lt.Value, Parameter.Modifier.NONE, null, lt.Location);
	  }

void case_705()
#line 4848 "cs-parser.jay"
{
	  	var lt = (LocatedToken) yyVals[0+yyTop];
		yyVal = new ImplicitLambdaParameter (lt.Value, lt.Location);
	  }

void case_706()
#line 4853 "cs-parser.jay"
{
	  	var lt = (LocatedToken) Error_AwaitAsIdentifier (yyVals[0+yyTop]);
		yyVal = new ImplicitLambdaParameter (lt.Value, lt.Location);
	  }

void case_708()
#line 4861 "cs-parser.jay"
{ 
		var pars_list = (List<Parameter>) yyVals[0+yyTop];
		yyVal = new ParametersCompiled (pars_list.ToArray ());
	  }

void case_710()
#line 4872 "cs-parser.jay"
{
		Block b = end_block (Location.Null);
		b.IsCompilerGenerated = true;
		b.AddStatement (new ContextualReturn ((Expression) yyVals[0+yyTop]));
		yyVal = b;
	  }

void case_712()
#line 4880 "cs-parser.jay"
{
	  	/* Handles only cases like foo = x.FirstOrDefault (l => );*/
	  	/* where we must restore current_variable*/
		Block b = end_block (Location.Null);
		b.IsCompilerGenerated = true;

		Error_SyntaxError (yyToken);
		yyVal = null;
	  }

void case_714()
#line 4894 "cs-parser.jay"
{
		Error_SyntaxError (yyToken);
		yyVal = null;
	  }

void case_715()
#line 4902 "cs-parser.jay"
{
		var lt = (LocatedToken) yyVals[-1+yyTop];	
		Parameter p = new ImplicitLambdaParameter (lt.Value, lt.Location);
		start_anonymous (true, new ParametersCompiled (p), false, lt.Location);
	  }

void case_716()
#line 4908 "cs-parser.jay"
{
		yyVal = end_anonymous ((ParametersBlock) yyVals[0+yyTop]);
		lbag.AddLocation (yyVal, GetLocation (yyVals[-2+yyTop]));
	  }

void case_717()
#line 4913 "cs-parser.jay"
{
		var lt = (LocatedToken) Error_AwaitAsIdentifier (yyVals[-1+yyTop]);
		Parameter p = new ImplicitLambdaParameter (lt.Value, lt.Location);
		start_anonymous (true, new ParametersCompiled (p), false, lt.Location);
	  }

void case_718()
#line 4919 "cs-parser.jay"
{
		yyVal = end_anonymous ((ParametersBlock) yyVals[0+yyTop]);
		lbag.AddLocation (yyVal, GetLocation (yyVals[-2+yyTop]));
	  }

void case_719()
#line 4924 "cs-parser.jay"
{
		var lt = (LocatedToken) yyVals[-1+yyTop];
		Parameter p = new ImplicitLambdaParameter (lt.Value, lt.Location);
		start_anonymous (true, new ParametersCompiled (p), true, lt.Location);
	  }

void case_720()
#line 4930 "cs-parser.jay"
{
		yyVal = end_anonymous ((ParametersBlock) yyVals[0+yyTop]);
		lbag.AddLocation (yyVal, GetLocation (yyVals[-4+yyTop]), GetLocation (yyVals[-2+yyTop]));
	  }

void case_722()
#line 4939 "cs-parser.jay"
{
	  	valid_param_mod = 0;
		start_anonymous (true, (ParametersCompiled) yyVals[-2+yyTop], false, GetLocation (yyVals[-4+yyTop]));
	  }

void case_723()
#line 4944 "cs-parser.jay"
{
		yyVal = end_anonymous ((ParametersBlock) yyVals[0+yyTop]);
		lbag.AddLocation (yyVal, GetLocation (yyVals[-6+yyTop]), GetLocation (yyVals[-3+yyTop]), GetLocation (yyVals[-2+yyTop]));
	  }

void case_725()
#line 4953 "cs-parser.jay"
{
	  	valid_param_mod = 0;
		start_anonymous (true, (ParametersCompiled) yyVals[-2+yyTop], true, GetLocation (yyVals[-5+yyTop]));
	  }

void case_726()
#line 4958 "cs-parser.jay"
{
		yyVal = end_anonymous ((ParametersBlock) yyVals[0+yyTop]);
		lbag.AddLocation (yyVal, GetLocation (yyVals[-7+yyTop]), GetLocation (yyVals[-6+yyTop]), GetLocation (yyVals[-3+yyTop]), GetLocation (yyVals[-2+yyTop]));
	  }

void case_733()
#line 4981 "cs-parser.jay"
{
		yyVal = new RefValueExpr ((Expression) yyVals[-3+yyTop], (FullNamedExpression) yyVals[-1+yyTop], GetLocation (yyVals[-5+yyTop]));
		lbag.AddLocation (yyVal, GetLocation (yyVals[-4+yyTop]), GetLocation (yyVals[-2+yyTop]), GetLocation (yyVals[0+yyTop]));
	  }

void case_734()
#line 4986 "cs-parser.jay"
{
		yyVal = new RefTypeExpr ((Expression) yyVals[-1+yyTop], GetLocation (yyVals[-3+yyTop]));
		lbag.AddLocation (yyVal, GetLocation (yyVals[-2+yyTop]), GetLocation (yyVals[0+yyTop]));
	  }

void case_735()
#line 4991 "cs-parser.jay"
{
		yyVal = new MakeRefExpr ((Expression) yyVals[-1+yyTop], GetLocation (yyVals[-3+yyTop]));
		lbag.AddLocation (yyVal, GetLocation (yyVals[-2+yyTop]), GetLocation (yyVals[0+yyTop]));	  
	  }

void case_740()
#line 5018 "cs-parser.jay"
{
		yyVal = yyVals[-1+yyTop];

		/* Cannot use opt_formal_parameter_list because it can be shared instance for empty parameters*/
	  	lbag.AppendToMember (current_container, GetLocation (yyVals[-2+yyTop]), GetLocation (yyVals[0+yyTop]));

		if (lang_version < LanguageVersion.V_6)
			FeatureIsNotAvailable (GetLocation (yyVals[-2+yyTop]), "primary constructor");
	  }

void case_745()
#line 5047 "cs-parser.jay"
{
		++lexer.parsing_block;
		current_type.PrimaryConstructorBaseArgumentsStart = GetLocation (yyVals[0+yyTop]);
	  }

void case_746()
#line 5052 "cs-parser.jay"
{
		lbag.AppendToMember (current_container, GetLocation (yyVals[0+yyTop]));
		current_type.PrimaryConstructorBaseArguments = (Arguments) yyVals[-1+yyTop];
		--lexer.parsing_block;

		yyVal = yyVals[-5+yyTop];
	  }

void case_748()
#line 5072 "cs-parser.jay"
{
		lexer.ConstraintsParsing = true;

		Class c = new Class (current_container, (MemberName) yyVals[0+yyTop], (Modifiers) yyVals[-4+yyTop], (Attributes) yyVals[-5+yyTop]);
		if (((c.ModFlags & Modifiers.STATIC) != 0) && lang_version == LanguageVersion.ISO_1) {
			FeatureIsNotAvailable (c.Location, "static classes");
		}
			
		push_current_container (c, yyVals[-3+yyTop]);
		valid_param_mod = ParameterModifierType.PrimaryConstructor;
	  }

void case_749()
#line 5085 "cs-parser.jay"
{
		valid_param_mod = 0;
		lexer.ConstraintsParsing = false;

		if (yyVals[-1+yyTop] != null)
			current_type.PrimaryConstructorParameters = (ParametersCompiled) yyVals[-1+yyTop];

		if (yyVals[0+yyTop] != null)
			current_container.SetConstraints ((List<Constraints>) yyVals[0+yyTop]);
		lbag.AddMember (current_container, mod_locations, GetLocation (yyVals[-5+yyTop]));

		if (doc_support) {
			current_container.PartialContainer.DocComment = Lexer.consume_doc_comment ();
			Lexer.doc_state = XmlCommentState.Allowed;
		}
		
		lexer.parsing_modifiers = true;
	  }

void case_750()
#line 5104 "cs-parser.jay"
{
		--lexer.parsing_declaration;
		if (doc_support)
			Lexer.doc_state = XmlCommentState.Allowed;
	  }

void case_751()
#line 5110 "cs-parser.jay"
{
		if (yyVals[0+yyTop] == null) {
			lbag.AppendToMember (current_container, GetLocation (yyVals[-4+yyTop]), GetLocation (yyVals[-2+yyTop]));
		} else {
			lbag.AppendToMember (current_container, GetLocation (yyVals[-4+yyTop]), GetLocation (yyVals[-2+yyTop]), GetLocation (yyVals[0+yyTop]));
		}
		yyVal = pop_current_class ();
	  }

void case_754()
#line 5129 "cs-parser.jay"
{
	    mod_locations = null;
		yyVal = ModifierNone;
		lexer.parsing_modifiers = false;
	  }

void case_757()
#line 5143 "cs-parser.jay"
{ 
		var m1 = (Modifiers) yyVals[-1+yyTop];
		var m2 = (Modifiers) yyVals[0+yyTop];

		if ((m1 & m2) != 0) {
			report.Error (1004, lexer.Location - ModifiersExtensions.Name (m2).Length,
				"Duplicate `{0}' modifier", ModifiersExtensions.Name (m2));
		} else if ((m2 & Modifiers.AccessibilityMask) != 0 && (m1 & Modifiers.AccessibilityMask) != 0 &&
			((m2 | m1 & Modifiers.AccessibilityMask) != (Modifiers.PROTECTED | Modifiers.INTERNAL))) {
			report.Error (107, lexer.Location - ModifiersExtensions.Name (m2).Length,
				"More than one protection modifier specified");
		}
		
		yyVal = m1 | m2;
	  }

void case_758()
#line 5162 "cs-parser.jay"
{
		yyVal = Modifiers.NEW;
		StoreModifierLocation (yyVal, GetLocation (yyVals[0+yyTop]));
		
		if (current_container.Kind == MemberKind.Namespace)
			report.Error (1530, GetLocation (yyVals[0+yyTop]), "Keyword `new' is not allowed on namespace elements");
	  }

void case_759()
#line 5170 "cs-parser.jay"
{
		yyVal = Modifiers.PUBLIC;
		StoreModifierLocation (yyVal, GetLocation (yyVals[0+yyTop]));
	  }

void case_760()
#line 5175 "cs-parser.jay"
{
		yyVal = Modifiers.PROTECTED;
		StoreModifierLocation (yyVal, GetLocation (yyVals[0+yyTop]));
	  }

void case_761()
#line 5180 "cs-parser.jay"
{
		yyVal = Modifiers.INTERNAL;
		StoreModifierLocation (yyVal, GetLocation (yyVals[0+yyTop]));
	  }

void case_762()
#line 5185 "cs-parser.jay"
{
		yyVal = Modifiers.PRIVATE;
		StoreModifierLocation (yyVal, GetLocation (yyVals[0+yyTop]));
	  }

void case_763()
#line 5190 "cs-parser.jay"
{
		yyVal = Modifiers.ABSTRACT;
		StoreModifierLocation (yyVal, GetLocation (yyVals[0+yyTop]));
	  }

void case_764()
#line 5195 "cs-parser.jay"
{
		yyVal = Modifiers.SEALED;
		StoreModifierLocation (yyVal, GetLocation (yyVals[0+yyTop]));
	  }

void case_765()
#line 5200 "cs-parser.jay"
{
		yyVal = Modifiers.STATIC;
		StoreModifierLocation (yyVal, GetLocation (yyVals[0+yyTop]));
	  }

void case_766()
#line 5205 "cs-parser.jay"
{
		yyVal = Modifiers.READONLY;
		StoreModifierLocation (yyVal, GetLocation (yyVals[0+yyTop]));
	  }

void case_767()
#line 5210 "cs-parser.jay"
{
		yyVal = Modifiers.VIRTUAL;
		StoreModifierLocation (yyVal, GetLocation (yyVals[0+yyTop]));
	  }

void case_768()
#line 5215 "cs-parser.jay"
{
		yyVal = Modifiers.OVERRIDE;
		StoreModifierLocation (yyVal, GetLocation (yyVals[0+yyTop]));
	  }

void case_769()
#line 5220 "cs-parser.jay"
{
		yyVal = Modifiers.EXTERN;
		StoreModifierLocation (yyVal, GetLocation (yyVals[0+yyTop]));
	  }

void case_770()
#line 5225 "cs-parser.jay"
{
		yyVal = Modifiers.VOLATILE;
		StoreModifierLocation (yyVal, GetLocation (yyVals[0+yyTop]));
	  }

void case_771()
#line 5230 "cs-parser.jay"
{
		yyVal = Modifiers.UNSAFE;
		StoreModifierLocation (yyVal, GetLocation (yyVals[0+yyTop]));
		if (!settings.Unsafe)
			Error_UnsafeCodeNotAllowed (GetLocation (yyVals[0+yyTop]));
	  }

void case_772()
#line 5237 "cs-parser.jay"
{
		yyVal = Modifiers.ASYNC;
		StoreModifierLocation (yyVal, GetLocation (yyVals[0+yyTop]));
	  }

void case_776()
#line 5254 "cs-parser.jay"
{
		Error_SyntaxError (yyToken);

		current_type.SetBaseTypes ((List<FullNamedExpression>) yyVals[-1+yyTop]);
	  }

void case_779()
#line 5271 "cs-parser.jay"
{
		var constraints = new List<Constraints> (1);
		constraints.Add ((Constraints) yyVals[0+yyTop]);
		yyVal = constraints;
	  }

void case_780()
#line 5277 "cs-parser.jay"
{
		var constraints = (List<Constraints>) yyVals[-1+yyTop];
		Constraints new_constraint = (Constraints)yyVals[0+yyTop];

		foreach (Constraints c in constraints) {
			if (new_constraint.TypeParameter.Value == c.TypeParameter.Value) {
				report.Error (409, new_constraint.Location,
					"A constraint clause has already been specified for type parameter `{0}'",
					new_constraint.TypeParameter.Value);
			}
		}

		constraints.Add (new_constraint);
		yyVal = constraints;
	  }

void case_781()
#line 5296 "cs-parser.jay"
{
		var lt = (LocatedToken) yyVals[-2+yyTop];
		yyVal = new Constraints (new SimpleMemberName (lt.Value, lt.Location), (List<FullNamedExpression>) yyVals[0+yyTop], GetLocation (yyVals[-3+yyTop]));
		lbag.AddLocation (yyVal, GetLocation (yyVals[-1+yyTop]));
	  }

void case_782()
#line 5302 "cs-parser.jay"
{
		Error_SyntaxError (yyToken);
	  
		var lt = (LocatedToken) yyVals[-1+yyTop];
		yyVal = new Constraints (new SimpleMemberName (lt.Value, lt.Location), null, GetLocation (yyVals[-2+yyTop]));
	  }

void case_783()
#line 5312 "cs-parser.jay"
{
		var constraints = new List<FullNamedExpression> (1);
		constraints.Add ((FullNamedExpression) yyVals[0+yyTop]);
		yyVal = constraints;
	  }

void case_784()
#line 5318 "cs-parser.jay"
{
		var constraints = (List<FullNamedExpression>) yyVals[-2+yyTop];
		var prev = constraints [constraints.Count - 1] as SpecialContraintExpr;
		if (prev != null && (prev.Constraint & SpecialConstraint.Constructor) != 0) {			
			report.Error (401, GetLocation (yyVals[-1+yyTop]), "The `new()' constraint must be the last constraint specified");
		}
		
		prev = yyVals[0+yyTop] as SpecialContraintExpr;
		if (prev != null) {
			if ((prev.Constraint & (SpecialConstraint.Class | SpecialConstraint.Struct)) != 0) {
				report.Error (449, prev.Location, "The `class' or `struct' constraint must be the first constraint specified");			
			} else {
			 	prev = constraints [0] as SpecialContraintExpr;
			 	if (prev != null && (prev.Constraint & SpecialConstraint.Struct) != 0) {			
					report.Error (451, GetLocation (yyVals[0+yyTop]), "The `new()' constraint cannot be used with the `struct' constraint");
				}
			}
		}

		constraints.Add ((FullNamedExpression) yyVals[0+yyTop]);
		yyVal = constraints;
	  }

void case_785()
#line 5344 "cs-parser.jay"
{
		if (yyVals[0+yyTop] is ComposedCast)
			report.Error (706, GetLocation (yyVals[0+yyTop]), "Invalid constraint type `{0}'", ((ComposedCast)yyVals[0+yyTop]).GetSignatureForError ());
	  
	  	yyVal = yyVals[0+yyTop];
	  }

void case_786()
#line 5351 "cs-parser.jay"
{
		yyVal = new SpecialContraintExpr (SpecialConstraint.Constructor, GetLocation (yyVals[-2+yyTop]));
		lbag.AddLocation (yyVal, GetLocation (yyVals[-1+yyTop]), GetLocation (yyVals[0+yyTop]));
	  }

void case_790()
#line 5371 "cs-parser.jay"
{
		if (lang_version <= LanguageVersion.V_3)
			FeatureIsNotAvailable (lexer.Location, "generic type variance");
		
		yyVal = yyVals[0+yyTop];
	  }

void case_793()
#line 5405 "cs-parser.jay"
{
		++lexer.parsing_block;
		start_block (GetLocation (yyVals[0+yyTop]));
	  }

void case_795()
#line 5417 "cs-parser.jay"
{
	 	--lexer.parsing_block;
		yyVal = end_block (GetLocation (yyVals[0+yyTop]));
	  }

void case_796()
#line 5422 "cs-parser.jay"
{
	 	--lexer.parsing_block;
		yyVal = end_block (lexer.Location);
	  }

void case_797()
#line 5431 "cs-parser.jay"
{
		++lexer.parsing_block;
		current_block.StartLocation = GetLocation (yyVals[0+yyTop]);
	  }

void case_798()
#line 5436 "cs-parser.jay"
{
		--lexer.parsing_block;
		yyVal = end_block (GetLocation (yyVals[0+yyTop]));
	  }

void case_806()
#line 5463 "cs-parser.jay"
{
		Error_SyntaxError (yyToken);
		yyVal = null;
	  }

void case_839()
#line 5527 "cs-parser.jay"
{
		  report.Error (1023, GetLocation (yyVals[0+yyTop]), "An embedded statement may not be a declaration or labeled statement");
		  yyVal = null;
	  }

void case_840()
#line 5532 "cs-parser.jay"
{
		  report.Error (1023, GetLocation (yyVals[0+yyTop]), "An embedded statement may not be a declaration or labeled statement");
		  yyVal = null;
	  }

void case_841()
#line 5537 "cs-parser.jay"
{
		Error_SyntaxError (yyToken);
		yyVal = new EmptyStatement (GetLocation (yyVals[0+yyTop]));
	  }

void case_842()
#line 5545 "cs-parser.jay"
{
		/* Uses lexer.Location because semicolon location is not kept in quick mode*/
		yyVal = new EmptyStatement (lexer.Location);
	  }

void case_843()
#line 5553 "cs-parser.jay"
{
		var lt = (LocatedToken) yyVals[-1+yyTop];
		LabeledStatement labeled = new LabeledStatement (lt.Value, current_block, lt.Location);
		lbag.AddLocation (labeled, GetLocation (yyVals[0+yyTop]));
		current_block.AddLabel (labeled);
		current_block.AddStatement (labeled);
	  }

void case_846()
#line 5566 "cs-parser.jay"
{
		if (yyVals[-1+yyTop] is VarExpr)
			yyVals[-1+yyTop] = new SimpleName ("var", ((VarExpr) yyVals[-1+yyTop]).Location);
	  
		yyVal = new ComposedCast ((FullNamedExpression) yyVals[-1+yyTop], (ComposedTypeSpecifier) yyVals[0+yyTop]);
	  }

void case_847()
#line 5582 "cs-parser.jay"
{ 
		/* Ok, the above "primary_expression" is there to get rid of*/
		/* both reduce/reduce and shift/reduces in the grammar, it should*/
		/* really just be "type_name".  If you use type_name, a reduce/reduce*/
		/* creeps up.  If you use namespace_or_type_name (which is all we need*/
		/* really) two shift/reduces appear.*/
		/* */

		/* So the super-trick is that primary_expression*/
		/* can only be either a SimpleName or a MemberAccess. */
		/* The MemberAccess case arises when you have a fully qualified type-name like :*/
		/* Foo.Bar.Blah i;*/
		/* SimpleName is when you have*/
		/* Blah i;*/
		
		var expr = (ATypeNameExpression) yyVals[-1+yyTop];
		if (yyVals[0+yyTop] == null) {
			if (expr.Name == "var" && expr is SimpleName)
				yyVal = new VarExpr (expr.Location);
			else
				yyVal = yyVals[-1+yyTop];
		} else {
			yyVal = new ComposedCast (expr, (ComposedTypeSpecifier) yyVals[0+yyTop]);
		}
	  }

void case_848()
#line 5608 "cs-parser.jay"
{
		var expr = (ATypeNameExpression) yyVals[-1+yyTop];
		yyVal = new ComposedCast (expr, (ComposedTypeSpecifier) yyVals[0+yyTop]);
	  }

void case_852()
#line 5619 "cs-parser.jay"
{
	  	((ComposedTypeSpecifier) yyVals[-1+yyTop]).Next = (ComposedTypeSpecifier) yyVals[0+yyTop];
	  	yyVal = yyVals[-1+yyTop];
	  }

void case_856()
#line 5642 "cs-parser.jay"
{
		var lt = (LocatedToken) yyVals[0+yyTop];
		var li = new LocalVariable (current_block, lt.Value, lt.Location);
		current_block.AddLocalName (li);
		current_variable = new BlockVariable ((FullNamedExpression) yyVals[-1+yyTop], li);
	  }

void case_857()
#line 5649 "cs-parser.jay"
{
		yyVal = current_variable;
		current_variable = null;
		if (yyVals[-2+yyTop] != null)
			lbag.AddLocation (yyVal, PopLocation (), GetLocation (yyVals[0+yyTop]));
		else
			lbag.AddLocation (yyVal, GetLocation (yyVals[0+yyTop]));
	  }

void case_858()
#line 5658 "cs-parser.jay"
{
		var lt = (LocatedToken) yyVals[0+yyTop];
		var li = new LocalVariable (current_block, lt.Value, LocalVariable.Flags.Constant, lt.Location);
		current_block.AddLocalName (li);
		current_variable = new BlockConstant ((FullNamedExpression) yyVals[-1+yyTop], li);
	  }

void case_859()
#line 5665 "cs-parser.jay"
{
		yyVal = current_variable;
		current_variable = null;
		lbag.AddLocation (yyVal, GetLocation (yyVals[-6+yyTop]), GetLocation (yyVals[0+yyTop]));
	  }

void case_861()
#line 5675 "cs-parser.jay"
{
		current_variable.Initializer = (Expression) yyVals[0+yyTop];
		PushLocation (GetLocation (yyVals[-1+yyTop]));
		yyVal = current_variable;
	  }

void case_862()
#line 5681 "cs-parser.jay"
{
	  	if (yyToken == Token.OPEN_BRACKET_EXPR) {
			report.Error (650, lexer.Location,
				"Syntax error, bad array declarator. To declare a managed array the rank specifier precedes the variable's identifier. To declare a fixed size buffer field, use the fixed keyword before the field type");
		} else {
			Error_SyntaxError (yyToken);
		}
	  }

void case_866()
#line 5699 "cs-parser.jay"
{
		foreach (var d in current_variable.Declarators) {
			if (d.Initializer == null)
				Error_MissingInitializer (d.Variable.Location);
		}
	  }

void case_869()
#line 5714 "cs-parser.jay"
{
		var lt = (LocatedToken) yyVals[0+yyTop];	  
		var li = new LocalVariable (current_variable.Variable, lt.Value, lt.Location);
		var d = new BlockVariableDeclarator (li, null);
		current_variable.AddDeclarator (d);
		current_block.AddLocalName (li);
	  	lbag.AddLocation (d, GetLocation (yyVals[-1+yyTop]));
	  }

void case_870()
#line 5723 "cs-parser.jay"
{
		var lt = (LocatedToken) yyVals[-2+yyTop];	  
		var li = new LocalVariable (current_variable.Variable, lt.Value, lt.Location);
		var d = new BlockVariableDeclarator (li, (Expression) yyVals[0+yyTop]);
		current_variable.AddDeclarator (d);
		current_block.AddLocalName (li);
	  	lbag.AddLocation (d, GetLocation (yyVals[-3+yyTop]), GetLocation (yyVals[-1+yyTop]));
	  }

void case_877()
#line 5756 "cs-parser.jay"
{
		var lt = (LocatedToken) yyVals[-2+yyTop];	  
		var li = new LocalVariable (current_block, lt.Value, LocalVariable.Flags.Constant, lt.Location);
		var d = new BlockVariableDeclarator (li, (Expression) yyVals[0+yyTop]);
		current_variable.AddDeclarator (d);
		current_block.AddLocalName (li);
	  	lbag.AddLocation (d, GetLocation (yyVals[-3+yyTop]), GetLocation (yyVals[-1+yyTop]));
	  }

void case_879()
#line 5769 "cs-parser.jay"
{
		yyVal = new StackAlloc ((Expression) yyVals[-3+yyTop], (Expression) yyVals[-1+yyTop], GetLocation (yyVals[-4+yyTop]));
		lbag.AddLocation (yyVal, GetLocation (yyVals[-2+yyTop]), GetLocation (yyVals[0+yyTop]));
	  }

void case_880()
#line 5774 "cs-parser.jay"
{
		report.Error (1575, GetLocation (yyVals[-1+yyTop]), "A stackalloc expression requires [] after type");
		yyVal = new StackAlloc ((Expression) yyVals[0+yyTop], null, GetLocation (yyVals[-1+yyTop]));		
	  }

void case_881()
#line 5782 "cs-parser.jay"
{
		yyVal = yyVals[-1+yyTop];
		lbag.AddStatement (yyVal, GetLocation (yyVals[0+yyTop]));
	  }

void case_883()
#line 5788 "cs-parser.jay"
{
		yyVal = yyVals[-1+yyTop];
		report.Error (1002, GetLocation (yyVals[0+yyTop]), "; expected");
		lexer.putback ('}');
	  }

void case_886()
#line 5806 "cs-parser.jay"
{
		ExpressionStatement s = yyVals[0+yyTop] as ExpressionStatement;
		if (s == null) {
			var expr = yyVals[0+yyTop] as Expression;
			yyVal = new StatementErrorExpression (expr);
		} else {
			yyVal = new StatementExpression (s);
		}
	  }

void case_887()
#line 5819 "cs-parser.jay"
{
		Expression expr = (Expression) yyVals[0+yyTop];
		yyVal = new StatementExpression (new OptionalAssign (expr, lexer.Location));
	  }

void case_888()
#line 5824 "cs-parser.jay"
{
		Error_SyntaxError (yyToken);
		yyVal = new EmptyStatement (GetLocation (yyVals[0+yyTop]));
	  }

void case_891()
#line 5838 "cs-parser.jay"
{ 
		if (yyVals[0+yyTop] is EmptyStatement)
			Warning_EmptyStatement (GetLocation (yyVals[0+yyTop]));
		
		yyVal = new If ((BooleanExpression) yyVals[-2+yyTop], (Statement) yyVals[0+yyTop], GetLocation (yyVals[-4+yyTop]));
		lbag.AddStatement (yyVal, GetLocation (yyVals[-3+yyTop]), GetLocation (yyVals[-1+yyTop]));
	  }

void case_892()
#line 5847 "cs-parser.jay"
{
		yyVal = new If ((BooleanExpression) yyVals[-4+yyTop], (Statement) yyVals[-2+yyTop], (Statement) yyVals[0+yyTop], GetLocation (yyVals[-6+yyTop]));
		lbag.AddStatement (yyVal, GetLocation (yyVals[-5+yyTop]), GetLocation (yyVals[-3+yyTop]), GetLocation (yyVals[-1+yyTop]));
		
		if (yyVals[-2+yyTop] is EmptyStatement)
			Warning_EmptyStatement (GetLocation (yyVals[-2+yyTop]));
		if (yyVals[0+yyTop] is EmptyStatement)
			Warning_EmptyStatement (GetLocation (yyVals[0+yyTop]));
	  }

void case_893()
#line 5857 "cs-parser.jay"
{
		Error_SyntaxError (yyToken);
		
		yyVal = new If ((BooleanExpression) yyVals[-1+yyTop], null, GetLocation (yyVals[-3+yyTop]));
		lbag.AddStatement (yyVal, GetLocation (yyVals[-2+yyTop]));
	  }

void case_895()
#line 5871 "cs-parser.jay"
{
		yyVal = new Switch ((Expression) yyVals[-5+yyTop], (ExplicitBlock) current_block.Explicit, GetLocation (yyVals[-7+yyTop]));	
		end_block (GetLocation (yyVals[0+yyTop]));
		lbag.AddStatement (yyVal, GetLocation (yyVals[-6+yyTop]), GetLocation (yyVals[-4+yyTop]));
	  }

void case_896()
#line 5877 "cs-parser.jay"
{
		Error_SyntaxError (yyToken);
	  
		yyVal = new Switch ((Expression) yyVals[-1+yyTop], null, GetLocation (yyVals[-3+yyTop]));	
		lbag.AddStatement (yyVal, GetLocation (yyVals[-2+yyTop]));
	  }

void case_903()
#line 5908 "cs-parser.jay"
{
	  	var label = (SwitchLabel) yyVals[0+yyTop];
	  	label.SectionStart = true;
		current_block.AddStatement (label);
	  }

void case_905()
#line 5921 "cs-parser.jay"
{
	 	yyVal = new SwitchLabel ((Expression) yyVals[-1+yyTop], GetLocation (yyVals[-2+yyTop]));
	 	lbag.AddLocation (yyVal, GetLocation (yyVals[0+yyTop]));
	 }

void case_906()
#line 5926 "cs-parser.jay"
{
		Error_SyntaxError (yyToken);
		yyVal = new SwitchLabel ((Expression) yyVals[-1+yyTop], GetLocation (yyVals[-2+yyTop]));
	  }

void case_907()
#line 5931 "cs-parser.jay"
{
		if (lang_version != LanguageVersion.Experimental)
			FeatureIsNotAvailable (GetLocation (yyVals[-1+yyTop]), "pattern matching");

		yyVal = new SwitchLabel ((Expression) yyVals[-1+yyTop], GetLocation (yyVals[-2+yyTop])) {
			PatternMatching = true
		};
		lbag.AddLocation (yyVal, GetLocation (yyVals[0+yyTop]));
	  }

void case_913()
#line 5955 "cs-parser.jay"
{
		if (yyVals[0+yyTop] is EmptyStatement && lexer.peek_token () == Token.OPEN_BRACE)
			Warning_EmptyStatement (GetLocation (yyVals[0+yyTop]));
	  
		yyVal = new While ((BooleanExpression) yyVals[-2+yyTop], (Statement) yyVals[0+yyTop], GetLocation (yyVals[-4+yyTop]));
		lbag.AddStatement (yyVal, GetLocation (yyVals[-3+yyTop]), GetLocation (yyVals[-1+yyTop]));
	  }

void case_914()
#line 5963 "cs-parser.jay"
{
		Error_SyntaxError (yyToken);
		
		yyVal = new While ((BooleanExpression) yyVals[-1+yyTop], null, GetLocation (yyVals[-3+yyTop]));
		lbag.AddStatement (yyVal, GetLocation (yyVals[-2+yyTop]));
	  }

void case_915()
#line 5973 "cs-parser.jay"
{
		yyVal = new Do ((Statement) yyVals[-5+yyTop], (BooleanExpression) yyVals[-2+yyTop], GetLocation (yyVals[-6+yyTop]), GetLocation (yyVals[-4+yyTop]));
		lbag.AddStatement (yyVal, GetLocation (yyVals[-4+yyTop]), GetLocation (yyVals[-3+yyTop]), GetLocation (yyVals[-1+yyTop]), GetLocation (yyVals[0+yyTop]));
	  }

void case_916()
#line 5978 "cs-parser.jay"
{
		Error_SyntaxError (yyToken);
		yyVal = new Do ((Statement) yyVals[-1+yyTop], null, GetLocation (yyVals[-2+yyTop]), Location.Null);
	  }

void case_917()
#line 5983 "cs-parser.jay"
{
		Error_SyntaxError (yyToken);
	  
		yyVal = new Do ((Statement) yyVals[-4+yyTop], (BooleanExpression) yyVals[-1+yyTop], GetLocation (yyVals[-5+yyTop]), GetLocation (yyVals[-3+yyTop]));
		lbag.AddStatement (yyVal, GetLocation (yyVals[-3+yyTop]), GetLocation (yyVals[-2+yyTop]));
	  }

void case_918()
#line 5993 "cs-parser.jay"
{
		start_block (GetLocation (yyVals[0+yyTop]));
		current_block.IsCompilerGenerated = true;
		For f = new For (GetLocation (yyVals[-1+yyTop]));
		current_block.AddStatement (f);
		yyVal = f;
	  }

void case_920()
#line 6009 "cs-parser.jay"
{
		((For) yyVals[-2+yyTop]).Initializer = (Statement) yyVals[-1+yyTop];

		/* Pass the "For" object to the iterator_part4*/
		oob_stack.Push (yyVals[-2+yyTop]);
	  }

void case_921()
#line 6017 "cs-parser.jay"
{
		var locations = (Tuple<Location,Location>) yyVals[-1+yyTop];
		oob_stack.Pop ();
		if (yyVals[0+yyTop] is EmptyStatement && lexer.peek_token () == Token.OPEN_BRACE)
			Warning_EmptyStatement (GetLocation (yyVals[0+yyTop]));
	  
		For f = ((For) yyVals[-5+yyTop]);
		f.Statement = (Statement) yyVals[0+yyTop];
		lbag.AddStatement (f, current_block.StartLocation, GetLocation (yyVals[-3+yyTop]), GetLocation (locations.Item1), GetLocation (locations.Item2));

		yyVal = end_block (GetLocation (yyVals[-3+yyTop]));
	  }

void case_922()
#line 6030 "cs-parser.jay"
{
		Error_SyntaxError (yyToken);
		yyVal = end_block (current_block.StartLocation);
	  }

void case_923()
#line 6038 "cs-parser.jay"
{
		For f = (For) oob_stack.Peek ();
		f.Condition = (BooleanExpression) yyVals[-1+yyTop];
	  }

void case_925()
#line 6048 "cs-parser.jay"
{
		report.Error (1525, GetLocation (yyVals[0+yyTop]), "Unexpected symbol `}'");
		For f = (For) oob_stack.Peek ();
		f.Condition = (BooleanExpression) yyVals[-1+yyTop];
		yyVal = new Tuple<Location,Location> (GetLocation (yyVals[0+yyTop]), GetLocation (yyVals[0+yyTop]));
	  }

void case_926()
#line 6057 "cs-parser.jay"
{
	  	For f = (For) oob_stack.Peek ();
		f.Iterator = (Statement) yyVals[-1+yyTop];
		yyVal = GetLocation (yyVals[0+yyTop]);
	  }

void case_927()
#line 6062 "cs-parser.jay"
{
		report.Error (1525, GetLocation (yyVals[0+yyTop]), "Unexpected symbol expected ')'");
		For f = (For) oob_stack.Peek ();
		f.Iterator = (Statement) yyVals[-1+yyTop];
		yyVal = GetLocation (yyVals[0+yyTop]);
	  }

void case_932()
#line 6082 "cs-parser.jay"
{
		var lt = (LocatedToken) yyVals[0+yyTop];
		var li = new LocalVariable (current_block, lt.Value, lt.Location);
		current_block.AddLocalName (li);
		current_variable = new BlockVariable ((FullNamedExpression) yyVals[-1+yyTop], li);
	  }

void case_933()
#line 6089 "cs-parser.jay"
{
		yyVal = current_variable;
		if (yyVals[-1+yyTop] != null)
			lbag.AddLocation (current_variable, PopLocation ());

		current_variable = null;
	  }

void case_941()
#line 6116 "cs-parser.jay"
{
	  	var sl = yyVals[-2+yyTop] as StatementList;
	  	if (sl == null) {
	  		sl = new StatementList ((Statement) yyVals[-2+yyTop], (Statement) yyVals[0+yyTop]);
			lbag.AddStatement (sl, GetLocation (yyVals[-1+yyTop]));
	  	} else {
	  		sl.Add ((Statement) yyVals[0+yyTop]);
	  		lbag.AppendTo (sl, GetLocation (yyVals[-1+yyTop]));
	  	}
	  		
		yyVal = sl;
	  }

void case_942()
#line 6132 "cs-parser.jay"
{
		report.Error (230, GetLocation (yyVals[-3+yyTop]), "Type and identifier are both required in a foreach statement");

		start_block (GetLocation (yyVals[-2+yyTop]));
		current_block.IsCompilerGenerated = true;
		
		Foreach f = new Foreach ((Expression) yyVals[-1+yyTop], null, null, null, null, GetLocation (yyVals[-3+yyTop]));
		current_block.AddStatement (f);
		
		lbag.AddStatement (f, GetLocation (yyVals[-2+yyTop]));
		yyVal = end_block (GetLocation (yyVals[0+yyTop]));
	  }

void case_943()
#line 6145 "cs-parser.jay"
{
		Error_SyntaxError (yyToken);
	
		start_block (GetLocation (yyVals[-3+yyTop]));
		current_block.IsCompilerGenerated = true;
		
		var lt = (LocatedToken) yyVals[-1+yyTop];
		var li = new LocalVariable (current_block, lt.Value, LocalVariable.Flags.ForeachVariable | LocalVariable.Flags.Used, lt.Location);
		current_block.AddLocalName (li);
		
		Foreach f = new Foreach ((Expression) yyVals[-2+yyTop], li, null, null, null, GetLocation (yyVals[-4+yyTop]));
		current_block.AddStatement (f);
		
		lbag.AddStatement (f, GetLocation (yyVals[-3+yyTop]));
		yyVal = end_block (GetLocation (yyVals[0+yyTop]));
	  }

void case_944()
#line 6162 "cs-parser.jay"
{
		start_block (GetLocation (yyVals[-5+yyTop]));
		current_block.IsCompilerGenerated = true;
		
		var lt = (LocatedToken) yyVals[-3+yyTop];
		var li = new LocalVariable (current_block, lt.Value, LocalVariable.Flags.ForeachVariable | LocalVariable.Flags.Used, lt.Location);
		current_block.AddLocalName (li);
		yyVal = li;
	  }

void case_945()
#line 6172 "cs-parser.jay"
{
		if (yyVals[0+yyTop] is EmptyStatement && lexer.peek_token () == Token.OPEN_BRACE)
			Warning_EmptyStatement (GetLocation (yyVals[0+yyTop]));
		
		Foreach f = new Foreach ((Expression) yyVals[-6+yyTop], (LocalVariable) yyVals[-1+yyTop], (Expression) yyVals[-3+yyTop], (Statement) yyVals[0+yyTop], current_block, GetLocation (yyVals[-8+yyTop]));
		lbag.AddStatement (f, GetLocation (yyVals[-7+yyTop]), GetLocation (yyVals[-4+yyTop]), GetLocation (yyVals[-2+yyTop]));
		end_block (GetLocation (yyVals[-2+yyTop]));
		
		yyVal = f;
	  }

void case_952()
#line 6195 "cs-parser.jay"
{
		yyVal = new Break (GetLocation (yyVals[-1+yyTop]));
		lbag.AddStatement (yyVal, GetLocation (yyVals[0+yyTop]));
	  }

void case_953()
#line 6203 "cs-parser.jay"
{
		yyVal = new Continue (GetLocation (yyVals[-1+yyTop]));
		lbag.AddStatement (yyVal, GetLocation (yyVals[0+yyTop]));
	  }

void case_954()
#line 6208 "cs-parser.jay"
{
		Error_SyntaxError (yyToken);
		yyVal = new Continue (GetLocation (yyVals[-1+yyTop]));
	  }

void case_955()
#line 6216 "cs-parser.jay"
{
		var lt = (LocatedToken) yyVals[-1+yyTop];
		yyVal = new Goto (lt.Value, GetLocation (yyVals[-2+yyTop]));
		lbag.AddStatement (yyVal, GetLocation (yyVals[-1+yyTop]), GetLocation (yyVals[0+yyTop]));
	  }

void case_956()
#line 6222 "cs-parser.jay"
{
		yyVal = new GotoCase ((Expression) yyVals[-1+yyTop], GetLocation (yyVals[-3+yyTop]));
		lbag.AddStatement (yyVal, GetLocation (yyVals[-2+yyTop]), GetLocation (yyVals[0+yyTop]));
	  }

void case_957()
#line 6227 "cs-parser.jay"
{
		yyVal = new GotoDefault (GetLocation (yyVals[-2+yyTop]));
		lbag.AddStatement (yyVal, GetLocation (yyVals[-1+yyTop]), GetLocation (yyVals[0+yyTop]));
	  }

void case_958()
#line 6235 "cs-parser.jay"
{
		yyVal = new Return ((Expression) yyVals[-1+yyTop], GetLocation (yyVals[-2+yyTop]));
		lbag.AddStatement (yyVal, GetLocation (yyVals[0+yyTop]));
	  }

void case_959()
#line 6240 "cs-parser.jay"
{
		Error_SyntaxError (yyToken);
		yyVal = new Return ((Expression) yyVals[-1+yyTop], GetLocation (yyVals[-2+yyTop]));
	  }

void case_960()
#line 6245 "cs-parser.jay"
{
		Error_SyntaxError (yyToken);
		yyVal = new Return (null, GetLocation (yyVals[-1+yyTop]));
	  }

void case_961()
#line 6253 "cs-parser.jay"
{
		yyVal = new Throw ((Expression) yyVals[-1+yyTop], GetLocation (yyVals[-2+yyTop]));
		lbag.AddStatement (yyVal, GetLocation (yyVals[0+yyTop]));
	  }

void case_962()
#line 6258 "cs-parser.jay"
{
		Error_SyntaxError (yyToken);
		yyVal = new Throw ((Expression) yyVals[-1+yyTop], GetLocation (yyVals[-2+yyTop]));
	  }

void case_963()
#line 6263 "cs-parser.jay"
{
		Error_SyntaxError (yyToken);
		yyVal = new Throw (null, GetLocation (yyVals[-1+yyTop]));
	  }

void case_964()
#line 6271 "cs-parser.jay"
{
		var lt = (LocatedToken) yyVals[-3+yyTop];
		string s = lt.Value;
		if (s != "yield"){
			report.Error (1003, lt.Location, "; expected");
		} else if (yyVals[-1+yyTop] == null) {
			report.Error (1627, GetLocation (yyVals[0+yyTop]), "Expression expected after yield return");
		} else if (lang_version == LanguageVersion.ISO_1){
			FeatureIsNotAvailable (lt.Location, "iterators");
		}
		
		current_block.Explicit.RegisterIteratorYield ();
		yyVal = new Yield ((Expression) yyVals[-1+yyTop], lt.Location);
		lbag.AddStatement (yyVal, GetLocation (yyVals[-2+yyTop]), GetLocation (yyVals[0+yyTop]));
	  }

void case_965()
#line 6287 "cs-parser.jay"
{
		Error_SyntaxError (yyToken);

		var lt = (LocatedToken) yyVals[-3+yyTop];
		string s = lt.Value;
		if (s != "yield"){
			report.Error (1003, lt.Location, "; expected");
		} else if (yyVals[-1+yyTop] == null) {
			report.Error (1627, GetLocation (yyVals[0+yyTop]), "Expression expected after yield return");
		} else if (lang_version == LanguageVersion.ISO_1){
			FeatureIsNotAvailable (lt.Location, "iterators");
		}
		
		current_block.Explicit.RegisterIteratorYield ();
		yyVal = new Yield ((Expression) yyVals[-1+yyTop], lt.Location);
		lbag.AddStatement (yyVal, GetLocation (yyVals[-2+yyTop]));
	  }

void case_966()
#line 6305 "cs-parser.jay"
{
		var lt = (LocatedToken) yyVals[-2+yyTop];
		string s = lt.Value;
		if (s != "yield"){
			report.Error (1003, lt.Location, "; expected");
		} else if (lang_version == LanguageVersion.ISO_1){
			FeatureIsNotAvailable (lt.Location, "iterators");
		}
		
		current_block.ParametersBlock.TopBlock.IsIterator = true;
		yyVal = new YieldBreak (lt.Location);
		lbag.AddStatement (yyVal, GetLocation (yyVals[-1+yyTop]), GetLocation (yyVals[0+yyTop]));
	  }

void case_970()
#line 6331 "cs-parser.jay"
{
		yyVal = new TryFinally ((Statement) yyVals[-2+yyTop], (ExplicitBlock) yyVals[0+yyTop], GetLocation (yyVals[-3+yyTop]));
		lbag.AddStatement (yyVal, GetLocation (yyVals[-1+yyTop]));
	  }

void case_971()
#line 6336 "cs-parser.jay"
{
		yyVal = new TryFinally (new TryCatch ((Block) yyVals[-3+yyTop], (List<Catch>) yyVals[-2+yyTop], Location.Null, true), (ExplicitBlock) yyVals[0+yyTop], GetLocation (yyVals[-4+yyTop]));
		lbag.AddStatement (yyVal, GetLocation (yyVals[-1+yyTop]));
	  }

void case_972()
#line 6341 "cs-parser.jay"
{
		Error_SyntaxError (1524, yyToken);
		yyVal = new TryCatch ((Block) yyVals[-1+yyTop], null, GetLocation (yyVals[-2+yyTop]), false);
	  }

void case_973()
#line 6349 "cs-parser.jay"
{
		var l = new List<Catch> (2);

		l.Add ((Catch) yyVals[0+yyTop]);
		yyVal = l;
	  }

void case_974()
#line 6356 "cs-parser.jay"
{
		var l = (List<Catch>) yyVals[-1+yyTop];
		
		Catch c = (Catch) yyVals[0+yyTop];
		var prev_catch = l [l.Count - 1];
		if (prev_catch.IsGeneral && prev_catch.Filter == null) {
			report.Error (1017, c.loc, "Try statement already has an empty catch block");
		}
		
		l.Add (c);
		yyVal = l;
	  }

void case_977()
#line 6377 "cs-parser.jay"
{
	  	var c = new Catch ((ExplicitBlock) yyVals[0+yyTop], GetLocation (yyVals[-2+yyTop]));
	  	c.Filter = (CatchFilterExpression) yyVals[-1+yyTop];
	  	yyVal = c;
	  }

void case_978()
#line 6383 "cs-parser.jay"
{
		start_block (GetLocation (yyVals[-3+yyTop]));
		var c = new Catch ((ExplicitBlock) current_block, GetLocation (yyVals[-4+yyTop]));
		c.TypeExpression = (FullNamedExpression) yyVals[-2+yyTop];

		if (yyVals[-1+yyTop] != null) {
			var lt = (LocatedToken) yyVals[-1+yyTop];
			c.Variable = new LocalVariable (current_block, lt.Value, lt.Location);
			current_block.AddLocalName (c.Variable);
		}
		
		lbag.AddLocation (c, GetLocation (yyVals[-3+yyTop]), GetLocation (yyVals[0+yyTop]));
		yyVal = c;
	  }

void case_979()
#line 6398 "cs-parser.jay"
{
	  	((Catch) yyVals[-2+yyTop]).Filter = (CatchFilterExpression) yyVals[-1+yyTop];
		yyVal = yyVals[-2+yyTop];
	  }

void case_980()
#line 6403 "cs-parser.jay"
{
	  	if (yyToken == Token.CLOSE_PARENS) {
			report.Error (1015, lexer.Location,
				"A type that derives from `System.Exception', `object', or `string' expected");
		} else {
			Error_SyntaxError (yyToken);
		}
		
		yyVal = new Catch (null, GetLocation (yyVals[-2+yyTop]));
	  }

void case_981()
#line 6414 "cs-parser.jay"
{
		Error_SyntaxError (yyToken);

		/* Required otherwise missing block could not be detected because*/
		/* start_block is run early*/
		var c = new Catch (null, GetLocation (yyVals[-5+yyTop]));
		c.TypeExpression = (FullNamedExpression) yyVals[-3+yyTop];

		if (yyVals[-2+yyTop] != null) {
			var lt = (LocatedToken) yyVals[-2+yyTop];
			c.Variable = new LocalVariable (current_block, lt.Value, lt.Location);
		}

		lbag.AddLocation (c, GetLocation (yyVals[-4+yyTop]), GetLocation (yyVals[-1+yyTop]));

		yyVal = c;
	  }

void case_983()
#line 6436 "cs-parser.jay"
{
		if (lang_version <= LanguageVersion.V_5)
			FeatureIsNotAvailable (GetLocation (yyVals[-3+yyTop]), "exception filter");

	  	yyVal = new CatchFilterExpression ((Expression) yyVals[-1+yyTop], GetLocation (yyVals[-3+yyTop]));
	  	lbag.AddLocation (yyVal, GetLocation (yyVals[-2+yyTop]), GetLocation (yyVals[0+yyTop]));
	  }

void case_986()
#line 6461 "cs-parser.jay"
{
		if (!settings.Unsafe)
			Error_UnsafeCodeNotAllowed (GetLocation (yyVals[0+yyTop]));
	  }

void case_988()
#line 6471 "cs-parser.jay"
{
		if (yyVals[0+yyTop] is EmptyStatement && lexer.peek_token () == Token.OPEN_BRACE)
			Warning_EmptyStatement (GetLocation (yyVals[0+yyTop]));
	  
		yyVal = new Lock ((Expression) yyVals[-2+yyTop], (Statement) yyVals[0+yyTop], GetLocation (yyVals[-4+yyTop]));
		lbag.AddStatement (yyVal, GetLocation (yyVals[-3+yyTop]), GetLocation (yyVals[-1+yyTop]));
	  }

void case_989()
#line 6479 "cs-parser.jay"
{
		Error_SyntaxError (yyToken);

		yyVal = new Lock ((Expression) yyVals[-1+yyTop], null, GetLocation (yyVals[-3+yyTop]));
		lbag.AddStatement (yyVal, GetLocation (yyVals[-2+yyTop]));
	  }

void case_990()
#line 6489 "cs-parser.jay"
{
	    start_block (GetLocation (yyVals[-2+yyTop]));
	    
		current_block.IsCompilerGenerated = true;
		var lt = (LocatedToken) yyVals[0+yyTop];
		var li = new LocalVariable (current_block, lt.Value, LocalVariable.Flags.FixedVariable | LocalVariable.Flags.Used, lt.Location);
		current_block.AddLocalName (li);
		current_variable = new Fixed.VariableDeclaration ((FullNamedExpression) yyVals[-1+yyTop], li);
	  }

void case_991()
#line 6499 "cs-parser.jay"
{
		yyVal = current_variable;
		current_variable = null;
	  }

void case_992()
#line 6504 "cs-parser.jay"
{
		if (yyVals[0+yyTop] is EmptyStatement && lexer.peek_token () == Token.OPEN_BRACE)
			Warning_EmptyStatement (GetLocation (yyVals[0+yyTop]));
	  
		Fixed f = new Fixed ((Fixed.VariableDeclaration) yyVals[-1+yyTop], (Statement) yyVals[0+yyTop], GetLocation (yyVals[-9+yyTop]));
		current_block.AddStatement (f);
		lbag.AddStatement (f, GetLocation (yyVals[-8+yyTop]), GetLocation (yyVals[-2+yyTop]));
		yyVal = end_block (GetLocation (yyVals[-2+yyTop]));
	  }

void case_993()
#line 6517 "cs-parser.jay"
{
	    start_block (GetLocation (yyVals[-2+yyTop]));
	    
		current_block.IsCompilerGenerated = true;
		var lt = (LocatedToken) yyVals[0+yyTop];
		var li = new LocalVariable (current_block, lt.Value, LocalVariable.Flags.UsingVariable | LocalVariable.Flags.Used, lt.Location);
		current_block.AddLocalName (li);
		current_variable = new Using.VariableDeclaration ((FullNamedExpression) yyVals[-1+yyTop], li);
	  }

void case_994()
#line 6527 "cs-parser.jay"
{
		yyVal = current_variable;	  
		current_variable = null;
	  }

void case_995()
#line 6532 "cs-parser.jay"
{
		if (yyVals[0+yyTop] is EmptyStatement && lexer.peek_token () == Token.OPEN_BRACE)
			Warning_EmptyStatement (GetLocation (yyVals[0+yyTop]));
	  
		Using u = new Using ((Using.VariableDeclaration) yyVals[-1+yyTop], (Statement) yyVals[0+yyTop], GetLocation (yyVals[-8+yyTop]));
		current_block.AddStatement (u);
		yyVal = end_block (GetLocation (yyVals[-2+yyTop]));
	  }

void case_996()
#line 6541 "cs-parser.jay"
{
		if (yyVals[0+yyTop] is EmptyStatement && lexer.peek_token () == Token.OPEN_BRACE)
			Warning_EmptyStatement (GetLocation (yyVals[0+yyTop]));
	  
		yyVal = new Using ((Expression) yyVals[-2+yyTop], (Statement) yyVals[0+yyTop], GetLocation (yyVals[-4+yyTop]));
		lbag.AddStatement (yyVal, GetLocation (yyVals[-3+yyTop]), GetLocation (yyVals[-1+yyTop]));
	  }

void case_997()
#line 6549 "cs-parser.jay"
{
		Error_SyntaxError (yyToken);
		
		yyVal = new Using ((Expression) yyVals[-1+yyTop], null, GetLocation (yyVals[-3+yyTop]));
		lbag.AddStatement (yyVal, GetLocation (yyVals[-2+yyTop]));
	  }

void case_999()
#line 6560 "cs-parser.jay"
{
		/* It has to be here for the parent to safely restore artificial block*/
	  	Error_SyntaxError (yyToken);
	  }

void case_1001()
#line 6572 "cs-parser.jay"
{
		current_variable.Initializer = (Expression) yyVals[0+yyTop];
		yyVal = current_variable;
	  }

void case_1002()
#line 6583 "cs-parser.jay"
{
		lexer.query_parsing = false;
			
		Linq.AQueryClause from = yyVals[-1+yyTop] as Linq.AQueryClause;
			
		from.Tail.Next = (Linq.AQueryClause)yyVals[0+yyTop];
		yyVal = from;
		
		current_block.SetEndLocation (lexer.Location);
		current_block = current_block.Parent;
	  }

void case_1003()
#line 6595 "cs-parser.jay"
{
		Linq.AQueryClause from = yyVals[-1+yyTop] as Linq.AQueryClause;
			
		from.Tail.Next = (Linq.AQueryClause)yyVals[0+yyTop];
		yyVal = from;
		
		current_block.SetEndLocation (lexer.Location);
		current_block = current_block.Parent;
	  }

void case_1004()
#line 6606 "cs-parser.jay"
{
	        lexer.query_parsing = false;
		yyVal = yyVals[-1+yyTop];

		current_block.SetEndLocation (lexer.Location);
		current_block = current_block.Parent;
	  }

void case_1005()
#line 6613 "cs-parser.jay"
{
	        yyVal = yyVals[-1+yyTop];
		current_block.SetEndLocation (lexer.Location);
		current_block = current_block.Parent;
	  }

void case_1006()
#line 6622 "cs-parser.jay"
{
		current_block = new Linq.QueryBlock (current_block, lexer.Location);
	  
		var lt = (LocatedToken) yyVals[-2+yyTop];
		var rv = new Linq.RangeVariable (lt.Value, lt.Location);
		var clause = new Linq.QueryStartClause ((Linq.QueryBlock)current_block, (Expression)yyVals[0+yyTop], rv, GetLocation (yyVals[-3+yyTop]));
		lbag.AddLocation (clause, GetLocation (yyVals[-1+yyTop]));
		yyVal = new Linq.QueryExpression (clause);
	  }

void case_1007()
#line 6632 "cs-parser.jay"
{
		current_block = new Linq.QueryBlock (current_block, lexer.Location);
	  
		var lt = (LocatedToken) yyVals[-2+yyTop];
		var rv = new Linq.RangeVariable (lt.Value, lt.Location);
		var clause = new Linq.QueryStartClause ((Linq.QueryBlock)current_block, (Expression)yyVals[0+yyTop], rv, GetLocation (yyVals[-4+yyTop])) {
				IdentifierType = (FullNamedExpression)yyVals[-3+yyTop]
		};
		lbag.AddLocation (clause, GetLocation (yyVals[-1+yyTop]));
		yyVal = new Linq.QueryExpression (clause);
	  }

void case_1008()
#line 6647 "cs-parser.jay"
{
		current_block = new Linq.QueryBlock (current_block, lexer.Location);
	  
		var lt = (LocatedToken) yyVals[-2+yyTop];
		var rv = new Linq.RangeVariable (lt.Value, lt.Location);
		var clause = new Linq.QueryStartClause ((Linq.QueryBlock)current_block, (Expression)yyVals[0+yyTop], rv, GetLocation (yyVals[-3+yyTop]));
		lbag.AddLocation (clause, GetLocation (yyVals[-1+yyTop]));
		yyVal = new Linq.QueryExpression (clause);
	  }

void case_1009()
#line 6657 "cs-parser.jay"
{
		current_block = new Linq.QueryBlock (current_block, lexer.Location);
	  
		var lt = (LocatedToken) yyVals[-2+yyTop];
		var rv = new Linq.RangeVariable (lt.Value, lt.Location);
		var clause = new Linq.QueryStartClause ((Linq.QueryBlock)current_block, (Expression)yyVals[0+yyTop], rv, GetLocation (yyVals[-4+yyTop])) {
				IdentifierType = (FullNamedExpression)yyVals[-3+yyTop]
		};
		lbag.AddLocation (clause, GetLocation (yyVals[-1+yyTop]));
		yyVal = new Linq.QueryExpression (clause);
	  }

void case_1011()
#line 6676 "cs-parser.jay"
{
		var lt = (LocatedToken) yyVals[-3+yyTop];
		var sn = new Linq.RangeVariable (lt.Value, lt.Location);
		yyVal = new Linq.SelectMany ((Linq.QueryBlock)current_block, sn, (Expression)yyVals[0+yyTop], GetLocation (yyVals[-4+yyTop]));
		
		current_block.SetEndLocation (lexer.Location);
		current_block = current_block.Parent;
		
		((Linq.QueryBlock)current_block).AddRangeVariable (sn);
		lbag.AddLocation (yyVal, GetLocation (yyVals[-2+yyTop]));
	  }

void case_1013()
#line 6692 "cs-parser.jay"
{
		var lt = (LocatedToken) yyVals[-3+yyTop];
		var sn = new Linq.RangeVariable (lt.Value, lt.Location);

		yyVal = new Linq.SelectMany ((Linq.QueryBlock)current_block, sn, (Expression)yyVals[0+yyTop], GetLocation (yyVals[-5+yyTop])) {
			IdentifierType = (FullNamedExpression)yyVals[-4+yyTop]
		};
		
		current_block.SetEndLocation (lexer.Location);
		current_block = current_block.Parent;
		
		((Linq.QueryBlock)current_block).AddRangeVariable (sn);
		
		lbag.AddLocation (yyVal, GetLocation (yyVals[-2+yyTop]));
	  }

void case_1014()
#line 6711 "cs-parser.jay"
{
	  	Linq.AQueryClause head = (Linq.AQueryClause)yyVals[-1+yyTop];
		
		if (yyVals[0+yyTop] != null)
			head.Next = (Linq.AQueryClause)yyVals[0+yyTop];
				
		if (yyVals[-2+yyTop] != null) {
			Linq.AQueryClause clause = (Linq.AQueryClause)yyVals[-2+yyTop];
			clause.Tail.Next = head;
			head = clause;
		}
		
		yyVal = head;
	  }

void case_1015()
#line 6726 "cs-parser.jay"
{
	  	Linq.AQueryClause head = (Linq.AQueryClause)yyVals[0+yyTop];

		if (yyVals[-1+yyTop] != null) {
			Linq.AQueryClause clause = (Linq.AQueryClause)yyVals[-1+yyTop];
			clause.Tail.Next = head;
			head = clause;
		}
		
		yyVal = head;
	  }

void case_1017()
#line 6739 "cs-parser.jay"
{
		report.Error (742, GetLocation (yyVals[0+yyTop]), "Unexpected symbol `{0}'. A query body must end with select or group clause", GetSymbolName (yyToken));
		yyVal = yyVals[-1+yyTop];
	  }

void case_1018()
#line 6744 "cs-parser.jay"
{
		Error_SyntaxError (yyToken);
		yyVal = null;
	  }

void case_1020()
#line 6756 "cs-parser.jay"
{
		yyVal = new Linq.Select ((Linq.QueryBlock)current_block, (Expression)yyVals[0+yyTop], GetLocation (yyVals[-2+yyTop]));

		current_block.SetEndLocation (lexer.Location);
		current_block = current_block.Parent;
	  }

void case_1021()
#line 6763 "cs-parser.jay"
{
	  	if (linq_clause_blocks == null)
	  		linq_clause_blocks = new Stack<Linq.QueryBlock> ();
	  		
	  	current_block = new Linq.QueryBlock (current_block, lexer.Location);
	  	linq_clause_blocks.Push ((Linq.QueryBlock)current_block);
	  }

void case_1022()
#line 6771 "cs-parser.jay"
{
		current_block.SetEndLocation (lexer.Location);
		current_block = current_block.Parent;
	  
		current_block = new Linq.QueryBlock (current_block, lexer.Location);
	  }

void case_1023()
#line 6778 "cs-parser.jay"
{
		var obj = (object[]) yyVals[0+yyTop];

		yyVal = new Linq.GroupBy ((Linq.QueryBlock)current_block, (Expression)yyVals[-2+yyTop], linq_clause_blocks.Pop (), (Expression)obj[0], GetLocation (yyVals[-4+yyTop]));
		lbag.AddLocation (yyVal, (Location) obj[1]);
		
		current_block.SetEndLocation (lexer.Location);
		current_block = current_block.Parent;
	  }

void case_1025()
#line 6795 "cs-parser.jay"
{
		Error_SyntaxError (yyToken);
		yyVal = new object[2] { null, Location.Null };
	  }

void case_1027()
#line 6804 "cs-parser.jay"
{
		((Linq.AQueryClause)yyVals[-1+yyTop]).Tail.Next = (Linq.AQueryClause)yyVals[0+yyTop];
		yyVal = yyVals[-1+yyTop];
	  }

void case_1034()
#line 6824 "cs-parser.jay"
{
		var lt = (LocatedToken) yyVals[-3+yyTop];
		var sn = new Linq.RangeVariable (lt.Value, lt.Location);
	  	yyVal = new Linq.Let ((Linq.QueryBlock) current_block, sn, (Expression)yyVals[0+yyTop], GetLocation (yyVals[-4+yyTop]));
		lbag.AddLocation (yyVal, GetLocation (yyVals[-2+yyTop]));
	  	
		current_block.SetEndLocation (lexer.Location);
		current_block = current_block.Parent;
		
		((Linq.QueryBlock)current_block).AddRangeVariable (sn);
	  }

void case_1036()
#line 6843 "cs-parser.jay"
{
		yyVal = new Linq.Where ((Linq.QueryBlock)current_block, (Expression)yyVals[0+yyTop], GetLocation (yyVals[-2+yyTop]));

		current_block.SetEndLocation (lexer.Location);
		current_block = current_block.Parent;
	  }

void case_1037()
#line 6853 "cs-parser.jay"
{
		if (linq_clause_blocks == null)
			linq_clause_blocks = new Stack<Linq.QueryBlock> ();
	  		
		current_block = new Linq.QueryBlock (current_block, lexer.Location);
		linq_clause_blocks.Push ((Linq.QueryBlock) current_block);
	  }

void case_1038()
#line 6861 "cs-parser.jay"
{
		current_block.SetEndLocation (lexer.Location);
		current_block = current_block.Parent;

		current_block = new Linq.QueryBlock (current_block, lexer.Location);
		linq_clause_blocks.Push ((Linq.QueryBlock) current_block);
	  }

void case_1039()
#line 6869 "cs-parser.jay"
{
		current_block.AddStatement (new ContextualReturn ((Expression) yyVals[-1+yyTop]));
		current_block.SetEndLocation (lexer.Location);
		current_block = current_block.Parent;

		current_block = new Linq.QueryBlock (current_block, lexer.Location);
	  }

void case_1040()
#line 6877 "cs-parser.jay"
{
		current_block.AddStatement (new ContextualReturn ((Expression) yyVals[-1+yyTop]));
		current_block.SetEndLocation (lexer.Location);
	  
		var outer_selector = linq_clause_blocks.Pop ();
		var block = linq_clause_blocks.Pop ();

		var lt = (LocatedToken) yyVals[-10+yyTop];	
		var sn = new Linq.RangeVariable (lt.Value, lt.Location);
		Linq.RangeVariable into;
		
		if (yyVals[0+yyTop] == null) {
			into = sn;
	  		yyVal = new Linq.Join (block, sn, (Expression)yyVals[-7+yyTop], outer_selector, (Linq.QueryBlock) current_block, GetLocation (yyVals[-11+yyTop]));
			lbag.AddLocation (yyVal, GetLocation (yyVals[-9+yyTop]), GetLocation (yyVals[-6+yyTop]), GetLocation (yyVals[-3+yyTop]));
		} else {
			/**/
			/* Set equals right side parent to beginning of linq query, it is not accessible therefore cannot cause name collisions*/
			/**/
			var parent = block.Parent;
			while (parent is Linq.QueryBlock) {
				parent = parent.Parent;
			}
			current_block.Parent = parent;
			
			((Linq.QueryBlock)current_block).AddRangeVariable (sn);
		
			lt = (LocatedToken) yyVals[0+yyTop];
			into = new Linq.RangeVariable (lt.Value, lt.Location);

			yyVal = new Linq.GroupJoin (block, sn, (Expression)yyVals[-7+yyTop], outer_selector, (Linq.QueryBlock) current_block, into, GetLocation (yyVals[-11+yyTop]));	
			lbag.AddLocation (yyVal, GetLocation (yyVals[-9+yyTop]), GetLocation (yyVals[-6+yyTop]), GetLocation (yyVals[-3+yyTop]), GetLocation (yyVals[0+yyTop]));
		}

		current_block = block.Parent;
		((Linq.QueryBlock)current_block).AddRangeVariable (into);
	  }

void case_1041()
#line 6915 "cs-parser.jay"
{
		if (linq_clause_blocks == null)
			linq_clause_blocks = new Stack<Linq.QueryBlock> ();
	  		
		current_block = new Linq.QueryBlock (current_block, lexer.Location);
		linq_clause_blocks.Push ((Linq.QueryBlock) current_block);
	  }

void case_1042()
#line 6923 "cs-parser.jay"
{
		current_block.SetEndLocation (lexer.Location);
		current_block = current_block.Parent;

		current_block = new Linq.QueryBlock (current_block, lexer.Location);
		linq_clause_blocks.Push ((Linq.QueryBlock) current_block);
	  }

void case_1043()
#line 6931 "cs-parser.jay"
{
		current_block.AddStatement (new ContextualReturn ((Expression) yyVals[-1+yyTop]));
		current_block.SetEndLocation (lexer.Location);
		current_block = current_block.Parent;

		current_block = new Linq.QueryBlock (current_block, lexer.Location);
	  }

void case_1044()
#line 6939 "cs-parser.jay"
{
		current_block.AddStatement (new ContextualReturn ((Expression) yyVals[-1+yyTop]));
		current_block.SetEndLocation (lexer.Location);
	  
		var outer_selector = linq_clause_blocks.Pop ();
		var block = linq_clause_blocks.Pop ();
		
		var lt = (LocatedToken) yyVals[-10+yyTop];
		var sn = new Linq.RangeVariable (lt.Value, lt.Location);
		Linq.RangeVariable into;
		
		if (yyVals[0+yyTop] == null) {
			into = sn;		
	  		yyVal = new Linq.Join (block, sn, (Expression)yyVals[-7+yyTop], outer_selector, (Linq.QueryBlock) current_block, GetLocation (yyVals[-12+yyTop])) {
	  			IdentifierType = (FullNamedExpression)yyVals[-11+yyTop]
	  		};
			lbag.AddLocation (yyVal, GetLocation (yyVals[-10+yyTop]), GetLocation (yyVals[-7+yyTop]), GetLocation (yyVals[-4+yyTop]));
		} else {
			/**/
			/* Set equals right side parent to beginning of linq query, it is not accessible therefore cannot cause name collisions*/
			/**/
			var parent = block.Parent;
			while (parent is Linq.QueryBlock) {
				parent = parent.Parent;
			}
			current_block.Parent = parent;
		
			((Linq.QueryBlock)current_block).AddRangeVariable (sn);
		
			lt = (LocatedToken) yyVals[0+yyTop];
			into = new Linq.RangeVariable (lt.Value, lt.Location); /* TODO:*/
			
			yyVal = new Linq.GroupJoin (block, sn, (Expression)yyVals[-7+yyTop], outer_selector, (Linq.QueryBlock) current_block, into, GetLocation (yyVals[-12+yyTop])) {
	  			IdentifierType = (FullNamedExpression)yyVals[-11+yyTop]
	  		};			
		}
		
		current_block = block.Parent;
		((Linq.QueryBlock)current_block).AddRangeVariable (into);		
	  }

void case_1048()
#line 6995 "cs-parser.jay"
{
		current_block.SetEndLocation (lexer.Location);
		current_block = current_block.Parent;
	  
		yyVal = yyVals[0+yyTop];
	  }

void case_1050()
#line 7006 "cs-parser.jay"
{
		current_block.SetEndLocation (lexer.Location);
		current_block = current_block.Parent;
	  
		current_block = new Linq.QueryBlock (current_block, lexer.Location);
	  }

void case_1051()
#line 7013 "cs-parser.jay"
{
		((Linq.AQueryClause)yyVals[-3+yyTop]).Next = (Linq.AQueryClause)yyVals[0+yyTop];
		yyVal = yyVals[-3+yyTop];
	  }

void case_1053()
#line 7022 "cs-parser.jay"
{
		current_block.SetEndLocation (lexer.Location);
		current_block = current_block.Parent;
	  
		current_block = new Linq.QueryBlock ((Linq.QueryBlock) current_block, lexer.Location);	 
	 }

void case_1054()
#line 7029 "cs-parser.jay"
{
		((Linq.AQueryClause)yyVals[-3+yyTop]).Tail.Next = (Linq.AQueryClause)yyVals[0+yyTop];
		yyVal = yyVals[-3+yyTop];
	 }

void case_1056()
#line 7041 "cs-parser.jay"
{
		yyVal = new Linq.OrderByAscending ((Linq.QueryBlock) current_block, (Expression)yyVals[-1+yyTop]);	
		lbag.AddLocation (yyVal, GetLocation (yyVals[0+yyTop]));
	  }

void case_1057()
#line 7046 "cs-parser.jay"
{
		yyVal = new Linq.OrderByDescending ((Linq.QueryBlock) current_block, (Expression)yyVals[-1+yyTop]);	
		lbag.AddLocation (yyVal, GetLocation (yyVals[0+yyTop]));
	  }

void case_1059()
#line 7058 "cs-parser.jay"
{
		yyVal = new Linq.ThenByAscending ((Linq.QueryBlock) current_block, (Expression)yyVals[-1+yyTop]);	
		lbag.AddLocation (yyVal, GetLocation (yyVals[0+yyTop]));
	  }

void case_1060()
#line 7063 "cs-parser.jay"
{
		yyVal = new Linq.ThenByDescending ((Linq.QueryBlock) current_block, (Expression)yyVals[-1+yyTop]);	
		lbag.AddLocation (yyVal, GetLocation (yyVals[0+yyTop]));
	  }

void case_1062()
#line 7073 "cs-parser.jay"
{
		/* query continuation block is not linked with query block but with block*/
		/* before. This means each query can use same range variable names for*/
		/* different identifiers.*/

		current_block.SetEndLocation (GetLocation (yyVals[-1+yyTop]));
		current_block = current_block.Parent;
	
		current_block = new Linq.QueryBlock (current_block, lexer.Location);
		
		if (linq_clause_blocks == null)
			linq_clause_blocks = new Stack<Linq.QueryBlock> ();
	  		
		linq_clause_blocks.Push ((Linq.QueryBlock) current_block);		
	  }

void case_1063()
#line 7089 "cs-parser.jay"
{
		var current_block = linq_clause_blocks.Pop ();	  
		var lt = (LocatedToken) yyVals[-2+yyTop];
		var rv = new Linq.RangeVariable (lt.Value, lt.Location);
  		yyVal = new Linq.QueryStartClause ((Linq.QueryBlock)current_block, null, rv, GetLocation (yyVals[-3+yyTop])) {
  			next = (Linq.AQueryClause)yyVals[0+yyTop]
  		};
	  }

void case_1066()
#line 7116 "cs-parser.jay"
{ 
		current_container = current_type = new Class (current_container, new MemberName ("<InteractiveExpressionClass>"), Modifiers.PUBLIC, null);

		/* (ref object retval)*/
		Parameter [] mpar = new Parameter [1];
		mpar [0] = new Parameter (new TypeExpression (compiler.BuiltinTypes.Object, Location.Null), "$retval", Parameter.Modifier.REF, null, Location.Null);

		ParametersCompiled pars = new ParametersCompiled (mpar);
		var mods = Modifiers.PUBLIC | Modifiers.STATIC;
		if (settings.Unsafe)
			mods |= Modifiers.UNSAFE;

		current_local_parameters = pars;
		var method = new InteractiveMethod (
			current_type,
			new TypeExpression (compiler.BuiltinTypes.Void, Location.Null),
			mods,
			pars);
			
		current_type.AddMember (method);			
		oob_stack.Push (method);

		interactive_async = false;

		++lexer.parsing_block;
		start_block (lexer.Location);
	  }

void case_1067()
#line 7144 "cs-parser.jay"
{
		--lexer.parsing_block;
		var method = (InteractiveMethod) oob_stack.Pop ();
		method.Block = (ToplevelBlock) end_block(lexer.Location);

		if (interactive_async == true) {
			method.ChangeToAsync ();
		}

		InteractiveResult = (Class) pop_current_class ();
		current_local_parameters = null;
	  }

void case_1077()
#line 7190 "cs-parser.jay"
{
		module.DocumentationBuilder.ParsedBuiltinType = (TypeExpression)yyVals[-1+yyTop];
		module.DocumentationBuilder.ParsedParameters = (List<DocumentationParameter>)yyVals[0+yyTop];
		yyVal = null;
	  }

void case_1078()
#line 7196 "cs-parser.jay"
{
		module.DocumentationBuilder.ParsedBuiltinType = new TypeExpression (compiler.BuiltinTypes.Void, GetLocation (yyVals[-1+yyTop]));
		module.DocumentationBuilder.ParsedParameters = (List<DocumentationParameter>)yyVals[0+yyTop];
		yyVal = null;
	  }

void case_1079()
#line 7202 "cs-parser.jay"
{
		module.DocumentationBuilder.ParsedBuiltinType = (TypeExpression)yyVals[-3+yyTop];
		module.DocumentationBuilder.ParsedParameters = (List<DocumentationParameter>)yyVals[0+yyTop];
		var lt = (LocatedToken) yyVals[-1+yyTop];
		yyVal = new MemberName (lt.Value);
	  }

void case_1082()
#line 7217 "cs-parser.jay"
{
		module.DocumentationBuilder.ParsedParameters = (List<DocumentationParameter>)yyVals[-1+yyTop];
		yyVal = new MemberName ((MemberName) yyVals[-6+yyTop], MemberCache.IndexerNameAlias, Location.Null);
	  }

void case_1083()
#line 7222 "cs-parser.jay"
{
		var p = (List<DocumentationParameter>)yyVals[0+yyTop] ?? new List<DocumentationParameter> (1);
		p.Add (new DocumentationParameter ((FullNamedExpression) yyVals[-1+yyTop]));
		module.DocumentationBuilder.ParsedParameters = p;
		module.DocumentationBuilder.ParsedOperator = Operator.OpType.Explicit;
		yyVal = null;
	  }

void case_1084()
#line 7230 "cs-parser.jay"
{
		var p = (List<DocumentationParameter>)yyVals[0+yyTop] ?? new List<DocumentationParameter> (1);
		p.Add (new DocumentationParameter ((FullNamedExpression) yyVals[-1+yyTop]));
		module.DocumentationBuilder.ParsedParameters = p;
		module.DocumentationBuilder.ParsedOperator = Operator.OpType.Implicit;
		yyVal = null;
	  }

void case_1085()
#line 7238 "cs-parser.jay"
{
		var p = (List<DocumentationParameter>)yyVals[0+yyTop];
		module.DocumentationBuilder.ParsedParameters = p;
		module.DocumentationBuilder.ParsedOperator = (Operator.OpType) yyVals[-1+yyTop];
		yyVal = null;
	  }

void case_1093()
#line 7276 "cs-parser.jay"
{
		var parameters = new List<DocumentationParameter> ();
		parameters.Add ((DocumentationParameter) yyVals[0+yyTop]);
		yyVal = parameters;
	  }

void case_1094()
#line 7282 "cs-parser.jay"
{
		var parameters = yyVals[-2+yyTop] as List<DocumentationParameter>;
		parameters.Add ((DocumentationParameter) yyVals[0+yyTop]);
		yyVal = parameters;
	  }

void case_1095()
#line 7291 "cs-parser.jay"
{
		if (yyVals[-1+yyTop] != null)
			yyVal = new DocumentationParameter ((Parameter.Modifier) yyVals[-1+yyTop], (FullNamedExpression) yyVals[0+yyTop]);
		else
			yyVal = new DocumentationParameter ((FullNamedExpression) yyVals[0+yyTop]);
	  }

#line default
   static readonly short [] yyLhs  = {              -1,
    0,    4,    0,    0,    1,    1,    1,    1,    2,    2,
   11,   11,   12,   12,   13,   13,   14,   15,   15,   15,
   19,   20,   17,   17,   22,   22,   22,   18,   18,   18,
   23,   23,   24,   24,    7,    7,    6,    6,   21,   21,
    8,    8,   25,   25,   25,   26,   26,   26,   26,   26,
    9,    9,   10,   10,   34,   32,   37,   33,   33,   33,
   33,   35,   35,   35,   36,   36,   41,   38,   39,   40,
   40,   42,   42,   42,   42,   42,   43,   43,   43,   47,
   44,   46,   50,   49,   49,   49,   52,   52,   53,   53,
   54,   54,   54,   54,   54,   54,   54,   54,   54,   54,
   54,   54,   54,   54,   68,   63,   73,   75,   78,   79,
   80,   28,   28,   83,   55,   55,   84,   84,   85,   85,
   86,   88,   82,   82,   87,   87,   93,   56,   97,   56,
   56,   92,  100,   92,   94,   94,  101,  101,  102,  103,
  102,   98,   98,  104,  104,  105,  106,   96,   96,   99,
   99,   99,  109,   57,  112,  113,  107,  114,  115,  116,
  107,  107,  107,  108,  108,  118,  118,  121,  119,  111,
  111,  122,  122,  122,  122,  122,  122,  122,  122,  122,
  122,  123,  123,  126,  126,  126,  126,  129,  126,  127,
  127,  130,  130,  131,  131,  131,  124,  124,  124,  132,
  132,  132,  125,  134,  136,  137,  139,   58,  140,   58,
  138,  142,  138,  141,  141,  144,  146,   60,  145,  145,
  135,  135,  135,  135,  135,  150,  147,  151,  148,  149,
  149,  149,  152,  153,  154,  156,   29,   29,  155,  155,
  157,  157,  158,  158,  158,  158,  158,  158,  158,  158,
  158,  160,   61,  161,  161,  164,  159,  159,  163,  163,
  163,  163,  163,  163,  163,  163,  163,  163,  163,  163,
  163,  163,  163,  163,  163,  163,  163,  163,  163,  163,
  163,  166,  165,  167,  165,  165,  165,   62,  170,  172,
  168,  169,  169,  171,  171,  176,  174,  177,  174,  174,
  174,  178,   64,  180,   59,  183,  184,   59,   59,  179,
  186,  179,  181,  181,  187,  187,  188,  189,  188,  190,
  185,  182,  182,  182,  182,  182,  194,  191,  195,  192,
  193,  193,   65,   66,  197,  199,  200,   30,  196,  196,
  196,  198,  198,  198,  201,  201,  202,  203,  202,  202,
  202,  204,  205,  206,   31,  207,  207,   16,   16,   16,
  208,  208,  208,  212,  212,  210,  210,  210,  213,  213,
  215,   72,  133,  110,  110,  143,  143,  216,  216,  216,
  214,  214,  217,  217,  218,  218,  220,  220,   91,   81,
   81,   95,   95,  128,  128,  162,  162,  222,  222,  222,
  221,  225,  225,  225,  227,  227,  228,  226,  226,  226,
  226,  226,  226,  226,  229,  229,  229,  229,  229,  229,
  229,  229,  229,  230,  230,  230,  230,  230,  230,  230,
  230,  230,  230,  230,  230,  230,  230,  230,  230,  230,
  230,  230,  230,  231,  231,  231,  232,  232,  232,  252,
  252,  253,  253,  254,  254,  234,  234,  251,  251,  251,
  251,  251,  251,  251,  251,  251,  251,  251,  251,  236,
  236,  236,  256,  256,  257,  257,  258,  258,  260,  260,
  260,  261,  261,  261,  261,  261,  261,  261,  262,  262,
  175,  175,  255,  255,  255,  255,  255,  267,  267,  266,
  266,  268,  268,  268,  268,  268,  268,  270,  270,  270,
  269,  237,  237,  237,  237,  265,  265,  272,  272,  273,
  273,  238,  239,  239,  240,  241,  242,  242,  233,  233,
  233,  233,  233,  278,  274,  243,  243,  279,  279,  280,
  280,  281,  281,  281,  281,  282,  282,  282,  282,  275,
  275,  223,  223,  277,  277,  283,  283,  276,  276,   90,
   90,  284,  284,  244,  285,  285,  211,  209,  245,  245,
  246,  246,  247,  247,  248,  287,  249,  288,  249,  286,
  286,  290,  289,  235,  291,  291,  291,  291,  291,  291,
  291,  291,  291,  292,  292,  292,  292,  292,  292,  292,
  292,  292,  292,  292,  292,  292,  293,  293,  293,  293,
  293,  293,  293,  294,  294,  294,  294,  294,  294,  294,
  294,  294,  294,  294,  294,  295,  297,  297,  297,  297,
  297,  297,  297,  297,  297,  298,  299,  301,  301,  302,
  303,  303,  300,  300,  304,  304,  305,  305,  306,  306,
  306,  306,  306,  307,  307,  307,  307,  307,  307,  307,
  307,  307,  308,  308,  308,  308,  308,  309,  309,  309,
  310,  310,  310,  311,  311,  311,  312,  312,  312,  313,
  313,  313,  314,  314,  315,  315,  315,  315,  315,  316,
  316,  316,  316,  316,  316,  316,  316,  316,  316,  316,
  317,  317,  318,  318,  318,  318,  319,  319,  321,  320,
  320,  320,   51,   51,  323,  322,  324,  322,  325,  322,
  326,  327,  322,  328,  329,  322,   45,   45,  263,  263,
  263,  263,  250,  250,  250,   89,  331,   74,   74,  332,
  333,  333,  333,  333,  335,  333,  336,  337,  338,  339,
   27,   71,   71,   70,   70,  117,  117,  340,  340,  340,
  340,  340,  340,  340,  340,  340,  340,  340,  340,  340,
  340,  340,   76,   76,  334,  334,   77,   77,  341,  341,
  342,  342,  343,  343,  344,  344,  344,  344,  219,  219,
  345,  345,  346,  120,   69,   69,  347,  173,   67,   67,
  348,  348,  349,  349,  349,  349,  353,  353,  354,  354,
  354,  351,  351,  351,  351,  351,  351,  351,  351,  351,
  351,  351,  351,  351,  355,  355,  355,  355,  355,  355,
  355,  355,  355,  355,  355,  355,  355,  369,  369,  369,
  369,  356,  370,  352,  271,  271,  371,  371,  371,  371,
  224,  224,  372,   48,   48,  374,  350,  377,  350,  373,
  373,  373,  375,  375,  381,  381,  380,  380,  382,  382,
  376,  376,  378,  378,  383,  383,  384,  379,  379,  379,
  357,  357,  357,  368,  368,  385,  386,  386,  358,  358,
  387,  387,  387,  390,  388,  388,  389,  389,  391,  391,
  391,  392,  393,  393,  394,  394,  394,  394,  359,  359,
  359,  359,  395,  395,  396,  396,  396,  400,  397,  403,
  399,  399,  406,  402,  402,  405,  405,  407,  407,  401,
  401,  410,  409,  409,  404,  404,  408,  408,  412,  411,
  411,  398,  398,  413,  398,  360,  360,  360,  360,  360,
  360,  414,  415,  415,  416,  416,  416,  417,  417,  417,
  418,  418,  418,  419,  419,  419,  420,  420,  361,  361,
  361,  361,  421,  421,  296,  296,  422,  424,  422,  422,
  422,  423,  423,  362,  363,  425,  366,  364,  364,  427,
  428,  367,  430,  431,  365,  365,  365,  429,  429,  426,
  426,  330,  330,  330,  330,  432,  432,  434,  434,  436,
  435,  437,  435,  433,  433,  433,  433,  433,  441,  439,
  442,  444,  439,  443,  443,  438,  438,  445,  445,  445,
  445,  445,  450,  446,  451,  447,  452,  453,  454,  448,
  456,  457,  458,  448,  455,  455,  460,  449,  459,  463,
  459,  462,  465,  462,  461,  461,  461,  464,  464,  464,
  440,  466,  440,    3,    3,  467,    3,    3,  468,  468,
  264,  264,  259,  259,    5,  469,  469,  469,  469,  469,
  473,  469,  469,  469,  469,  470,  470,  471,  474,  471,
  472,  472,  475,  475,  476,
  };
   static readonly short [] yyLen = {           2,
    2,    0,    3,    1,    2,    4,    3,    1,    0,    1,
    1,    2,    4,    2,    1,    2,    1,    3,    5,    2,
    0,    0,   11,    3,    0,    1,    1,    1,    3,    1,
    0,    1,    0,    1,    0,    1,    0,    1,    0,    1,
    1,    2,    1,    1,    2,    1,    1,    1,    1,    1,
    0,    1,    1,    2,    0,    3,    0,    6,    3,    2,
    1,    1,    1,    1,    1,    3,    0,    3,    1,    0,
    3,    0,    1,    1,    3,    3,    1,    1,    1,    0,
    4,    4,    1,    0,    1,    1,    0,    1,    1,    2,
    1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
    1,    1,    1,    1,    0,    4,    0,    0,    0,    0,
    0,   17,    5,    0,    9,    5,    0,    1,    1,    2,
    3,    0,    3,    1,    1,    1,    0,    8,    0,    9,
    6,    0,    0,    3,    0,    1,    1,    2,    2,    0,
    5,    0,    1,    1,    2,    3,    0,    4,    2,    1,
    1,    1,    0,    3,    0,    0,   10,    0,    0,    0,
   12,    8,    5,    1,    1,    1,    1,    0,    4,    0,
    1,    1,    3,    3,    3,    5,    3,    5,    1,    1,
    1,    1,    3,    4,    6,    2,    4,    0,    7,    0,
    1,    1,    2,    1,    1,    1,    4,    6,    4,    1,
    2,    2,    1,    0,    0,    0,    0,   12,    0,    6,
    0,    0,    4,    1,    1,    0,    0,   10,    3,    1,
    1,    2,    1,    2,    1,    0,    5,    0,    5,    1,
    1,    1,    0,    0,    0,    0,   15,    5,    0,    1,
    1,    2,    1,    1,    1,    1,    1,    1,    1,    1,
    1,    0,    5,    1,    1,    0,    7,    1,    1,    1,
    1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
    1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
    1,    0,    7,    0,    7,    2,    2,    2,    0,    0,
    9,    1,    1,    0,    1,    0,    6,    0,    6,    2,
    1,    0,    8,    0,    9,    0,    0,   10,    5,    0,
    0,    3,    0,    1,    1,    2,    2,    0,    5,    0,
    2,    2,    2,    1,    1,    1,    0,    5,    0,    5,
    1,    1,    2,    4,    0,    0,    0,   12,    0,    2,
    2,    0,    1,    2,    1,    3,    2,    0,    5,    3,
    1,    0,    0,    0,   13,    0,    1,    1,    3,    3,
    1,    4,    4,    2,    2,    0,    3,    2,    1,    3,
    0,    3,    1,    1,    3,    1,    2,    3,    4,    4,
    0,    3,    1,    3,    3,    1,    1,    1,    1,    1,
    1,    1,    1,    1,    1,    1,    2,    2,    2,    1,
    1,    2,    2,    2,    1,    3,    1,    1,    1,    1,
    1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
    1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
    1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
    1,    1,    1,    1,    2,    1,    1,    1,    1,    1,
    1,    1,    1,    1,    1,    3,    3,    4,    4,    5,
    4,    4,    4,    3,    3,    3,    4,    3,    4,    4,
    4,    3,    0,    1,    3,    4,    0,    1,    1,    3,
    2,    3,    3,    1,    2,    3,    5,    2,    1,    1,
    0,    1,    1,    3,    3,    3,    2,    1,    1,    1,
    1,    2,    2,    2,    2,    4,    3,    3,    2,    4,
    1,    4,    5,    4,    3,    1,    3,    1,    3,    1,
    1,    1,    4,    3,    2,    2,    6,    3,    7,    4,
    3,    7,    3,    0,    2,    4,    3,    1,    2,    0,
    1,    1,    3,    1,    2,    3,    1,    1,    1,    0,
    1,    1,    2,    2,    3,    1,    2,    0,    1,    2,
    4,    1,    3,    4,    1,    1,    1,    2,    4,    4,
    4,    2,    4,    2,    4,    0,    4,    0,    5,    0,
    1,    0,    4,    4,    1,    2,    2,    4,    2,    2,
    2,    4,    2,    1,    2,    2,    2,    2,    2,    2,
    2,    2,    2,    2,    2,    2,    1,    3,    3,    3,
    3,    3,    3,    1,    3,    3,    3,    3,    3,    4,
    3,    3,    3,    3,    3,    1,    1,    2,    2,    1,
    1,    4,    1,    1,    1,    4,    4,    1,    3,    3,
    1,    2,    0,    1,    1,    3,    1,    3,    1,    3,
    3,    3,    3,    1,    3,    3,    3,    3,    3,    3,
    3,    3,    1,    3,    3,    3,    3,    1,    3,    3,
    1,    3,    3,    1,    3,    3,    1,    3,    3,    1,
    3,    3,    1,    3,    1,    5,    4,    5,    5,    3,
    3,    3,    3,    3,    3,    3,    3,    3,    3,    3,
    1,    3,    3,    2,    1,    1,    0,    1,    0,    2,
    1,    1,    1,    1,    0,    4,    0,    4,    0,    5,
    0,    0,    7,    0,    0,    8,    1,    1,    1,    1,
    1,    1,    6,    4,    4,    1,    1,    0,    1,    3,
    0,    1,    1,    2,    0,    6,    0,    0,    0,    0,
   15,    0,    1,    0,    1,    1,    2,    1,    1,    1,
    1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
    1,    1,    0,    1,    2,    3,    0,    1,    1,    2,
    4,    3,    1,    3,    1,    3,    1,    1,    0,    1,
    1,    1,    0,    4,    1,    1,    0,    4,    0,    1,
    1,    2,    1,    1,    1,    1,    1,    2,    1,    1,
    1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
    1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
    1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
    1,    1,    0,    4,    1,    2,    2,    2,    1,    1,
    1,    2,    1,    1,    1,    0,    6,    0,    7,    0,
    2,    1,    0,    1,    0,    1,    1,    2,    2,    4,
    0,    2,    0,    1,    1,    2,    4,    1,    5,    2,
    2,    2,    2,    2,    2,    1,    1,    1,    1,    1,
    5,    7,    4,    0,    8,    4,    0,    1,    1,    2,
    1,    2,    1,    2,    3,    3,    3,    1,    1,    1,
    1,    1,    5,    4,    7,    3,    6,    0,    4,    0,
    5,    1,    0,    4,    2,    2,    2,    1,    1,    0,
    1,    0,    5,    1,    0,    1,    0,    1,    1,    1,
    3,    4,    5,    0,    9,    1,    1,    1,    1,    1,
    1,    2,    2,    2,    3,    4,    3,    3,    3,    2,
    3,    3,    2,    4,    4,    3,    0,    1,    3,    4,
    5,    3,    1,    2,    0,    1,    3,    0,    8,    3,
    6,    0,    4,    2,    2,    0,    3,    5,    4,    0,
    0,   10,    0,    0,    9,    5,    4,    2,    1,    0,
    2,    2,    2,    2,    2,    4,    5,    4,    5,    0,
    5,    0,    6,    3,    2,    2,    2,    1,    0,    3,
    0,    0,    5,    2,    1,    1,    2,    1,    1,    1,
    1,    1,    0,    5,    0,    3,    0,    0,    0,   12,
    0,    0,    0,   13,    0,    2,    0,    3,    1,    0,
    4,    1,    0,    4,    1,    2,    2,    1,    2,    2,
    0,    0,    4,    2,    3,    0,    4,    2,    2,    3,
    0,    1,    1,    1,    2,    2,    2,    2,    4,    3,
    0,    7,    4,    4,    3,    1,    3,    0,    0,    4,
    0,    1,    1,    3,    2,
  };
   static readonly short [] yyDefRed = {            0,
    8,    0,    0,    0,    0,    0,    0,    0,    2,    4,
    0,    0,   11,   14,    0, 1064,    0,    0, 1068,    0,
    0,   15,   17,  410,  416,  423,  411,  413,    0,  412,
    0,  419,  421,  408,    0,  415,  417,  409,  420,  422,
  418,    0,  371, 1086,    0,  414, 1075,    0,   10,    1,
    0,    0,    0,   12,    0,  888,    0,    0,    0,    0,
    0,    0,    0,    0,  451,    0,    0,    0,    0,    0,
    0,    0,  449,    0,    0,    0,  522,    0,  450,    0,
    0,    0,  986,    0,    0,    0,  732,    0,    0,    0,
    0,    0,    0,    0,  793,    0,  842,    0,    0,    0,
    0,    0,    0,    0,    0,  448,    0,  721,    0,  887,
    0,  825,    0,  444,  850,  849,    0,    0,    0,  425,
  426,  427,  428,  429,  430,  431,  432,  433,  434,  435,
  436,  437,  438,  439,  440,  441,  442,  443,  446,  447,
  728,    0,  594,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,  729,  727,  730,  731,  809,
  811,    0,  807,  810,  826,  828,  829,  830,  831,  832,
  833,  834,  835,  836,  837,  827,    0,    0,  889,  890,
  909,  910,  911,  912,  946,  947,  948,  949,  950,  951,
    0,    0,    0,   20,    0,    0,  358,    0,  361, 1072,
   16, 1065,    0,    0,  264,  281,  263,  260,  265,  266,
  259,  278,  277,  270,  271,  267,  269,  268,  272,  261,
  262,  273,  274,  280,  279,  275,  276,    0, 1089, 1078,
    0,    0, 1077,    0, 1076,    3,   55,    0,    0,    0,
   44,   41,   43,   46,   47,   48,   49,   50,   53,   13,
    0,    0,    0,  952,  572,  452,  453,  984,    0,    0,
    0,    0,    0,    0,    0,    0,  954,  953,    0,  582,
  576,  581,  841,  886,  812,  839,  838,  840,  813,  814,
  815,  816,  817,  818,  819,  820,  821,  822,  823,  824,
    0,    0,    0,  918,    0,    0,    0,  855,  854,    0,
    0,    0,    0,    0,    0,    0,    0,  960,    0,    0,
    0,    0,  424,    0,    0,    0,  963,    0,    0,    0,
    0,  574,  985,    0,    0,    0,  853,  404,    0,    0,
    0,    0,    0,    0,  390,  391,    0,  400,    0,    0,
    0,    0,    0,    0,    0,  724,    0,  593,    0,    0,
  717,    0,    0,  589,    0,    0,  591,  587,  601,  595,
  602,  596,  590,  586,  606,  600,  605,  599,  603,  597,
  604,  598,  715,  568,    0,  567,  445,  364,  365,    0,
    0,    0,    0,    0,  843,    0,  357,    0,  402,  403,
    0,    0,  525,  526,    0,    0,    0,  847,  848,  856,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 1067,  808,  846,    0,  884,  885, 1018,
 1035,    0,    0, 1019, 1021,    0, 1047, 1004, 1002, 1028,
    0,    0, 1026, 1029, 1030, 1031, 1032, 1005, 1003,    0,
    0,    0,    0,   18,    0,    0,    0, 1085,    0,    0,
  372,    0,    0, 1087,    0,    0,   42,  763,  769,  761,
    0,  758,  768,  762,  760,  759,  766,  764,  765,  771,
  767,  770,  772,    0,    0,  756,   45,   54,  524,    0,
  520,  521,    0,    0,  518,    0,  858,    0,    0,    0,
  916,    0,  883,  881,  882,    0,    0,    0,  736,    0,
  957,  955,  737,    0,    0,  549,    0,    0,  537,  544,
    0,    0,    0,  538,    0,    0,  554,  556,    0,  533,
    0,    0,    0,    0,    0,  528,    0,  531,  535,  393,
  392,  959,  958,    0,    0,  962,  961,  972,    0,    0,
    0,  973,  566,    0,  387,  565,    0,    0,  987,    0,
    0,  852,    0,  398,  399,    0,    0,  397,    0,    0,
    0,  607,    0,    0,  578,    0,  719,  625,  624,    0,
    0,  806,    0,    0,  801,  803,  804,  805,  456,  457,
    0,  368,  369,    0,  195,  194,  196,    0,  706,    0,
    0,    0,  394,    0,  701,    0,    0,  966,    0,    0,
    0,  464,  465,    0,  468,    0,    0,    0,    0,  466,
    0,    0,  515,    0,  472,    0,    0,    0,    0,  498,
  501,    0,    0,  493,  500,  499,    0,  690,  691,  692,
  693,  694,  695,  696,  697,  698,  700,  699,  611,  608,
  613,  610,  612,  609,  622,  619,  623,    0,    0,  633,
    0,    0,    0,    0,    0,  626,    0,  621,  634,  635,
  617,    0,  618,    0,  652,    0,    0,  653,    0,  659,
    0,  660,    0,  661,    0,  662,    0,  666,    0,  667,
    0,  670,    0,  673,    0,  676,    0,  679,    0,  682,
    0,  684,    0,  553,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 1017, 1016,    0, 1027,    0, 1015,    0,
    0,  359,  360, 1083, 1084,    0,    0,  192,    0,    0,
 1093,  386,    0,    0,    0,  383, 1079, 1081,   61,   63,
   64,    0,    0,   56,    0,    0,   65,   67,   30,   28,
    0,    0,    0,  753,    0,  757,  462,    0,  523,    0,
  571,    0,  584,  181,  203,    0,    0,    0,  171,    0,
    0,    0,  182,  577,    0,  990,  922,    0,  940,  919,
    0,  931,    0,  942,    0,  956,  893,    0,  989,    0,
    0,  536,    0,  545,  555,  557,    0,    0,    0,    0,
  484,    0,    0,  479,    0,    0,  714,  713,  516,    0,
  559,  530,    0,    0,  152,  560,  150,  151,  562,    0,
  570,  569,  896,    0,    0,    0,    0,  970,    0,  974,
  564,  573,  997,    0,  993,  914,    0, 1008,    0, 1006,
    0,    0,  734,  735,    0,    0,    0,  712,  711,  718,
    0,  463,  795,  796,  794,  802,  716,    0,  367,  704,
    0,    0,    0,  592,  588,  965,  964,  844,  469,  461,
    0,    0,  467,  458,  459,  575,  514,  512,    0,  511,
  504,  505,    0,  502,  503,    0,  497,  454,  455,  470,
  471,    0,  862,    0,    0,  628,  629,    0,    0,    0,
  976,  620,  687,    0, 1036, 1010,    0, 1037,    0, 1020,
 1022, 1033,    0, 1048,    0, 1014, 1062,   19,  362,  363,
 1095,  193, 1090,    0,  792,  791,    0,  790,    0,  382,
    0,   60,   57,    0,    0,    0,    0,    0,    0,  389,
    0,  747,    0,    0,   86,   85,    0,  519,    0,    0,
    0,    0,    0,  186,  583,    0,    0,    0,    0,    0,
  932,  920,    0,  943,    0,    0,  988,  546,  543,    0,
  488,    0,    0,    0, 1073, 1074,  475,  481,    0,  485,
    0,    0,    0,    0,    0,    0,  894,    0,  980,    0,
  977,  971,  996,    0,  913, 1009, 1007,    0,  579,    0,
  720,  710,  370,  703,  702,  722,  460,  513,    0,    0,
  507,    0,  496,  495,  494,    0,  878,  861,    0,    0,
    0,  867,    0,    0,    0,  638,    0,    0,  641,    0,
  647,    0,  645,  688,  689,  686,    0, 1012,    0, 1041,
    0,    0, 1056, 1057, 1050,    0, 1094,  385,  384,    0,
    0,   66,   59,    0,   68,   29,   22,    0,    0,  335,
    0,  238,    0,  113,    0,   82,   83,  872,  125,  126,
    0,    0,    0,  875,  201,  202,    0,    0,    0,    0,
  174,  183,  175,  177,  917,    0,    0,    0,    0,    0,
  941,    0,    0,  489,  490,  483,  486,  482,    0,  476,
  480,    0,  551,    0,  517,  527,  474,  563,  561,    0,
    0,    0,  999,    0,    0,  733,  725,    0,  508,    0,
  506,    0,    0,  857,  868,  632,    0,  637,    0,    0,
  642,  636,    0, 1011,    0,    0,    0, 1025,    0, 1023,
 1034,    0, 1063, 1082,    0,   79,    0,    0,   73,   74,
   77,   78,    0,  352,  341,  340,    0,  748,  234,  108,
    0,  859,  876,  187,    0,  199,    0,    0,    0,  915,
 1001,    0,    0,    0,  936,    0,    0,  944,  892,    0,
  532,  529,  901,    0,  908,    0,    0,  899,    0,  903,
  983,    0,  998,  994,    0,  723,  510,    0,    0,  640,
  639,  648,  646, 1013, 1038,    0, 1024,    0,    0, 1052,
    0,   80,   71,    0,    0,    0,  336,    0,    0,    0,
    0,    0,  188,    0,  178,  176,  991,  933,  921,  929,
  928,  923,  925,    0,  487,    0,    0,    0,  895,  900,
    0,  904,  981,    0,    0,  726,    0,  870,    0, 1042,
 1059, 1060, 1053,   58,    0,   75,   76,    0,    0,    0,
    0,    0,    0,    0,  742,    0,  774,    0,  739,  877,
  185,    0,  198,    0,    0,  945,  906,  905,  907,    0,
  995,  879,    0,    0,    0,   81,    0,    0,  353,    0,
    0,  351,  337,    0,  345,    0,  407,    0,  405,    0,
    0,  749,    0,  779,  235,    0,  189,  992,  924,    0,
    0,  938,  797,  979, 1039,    0, 1054,    0,    0,    0,
  333,    0,    0,  740,  776,    0,  745,    0,    0,  780,
    0,  109,  927,  926,    0,    0, 1043,   27,   26,   23,
  354,  350,    0,    0,  346,  406,    0,  782,    0,    0,
    0,    0,    0,    0,    0,    0,    0,   32,  338,    0,
  787,    0,  788,  785,    0,  783,  104,  105,    0,  101,
    0,    0,   89,   91,   92,   93,   94,   95,   96,   97,
   98,   99,  100,  102,  103,  153,    0,    0,  251,  243,
  244,  245,  246,  247,  248,  249,  250,    0,    0,  241,
  110,  798,    0, 1040,    0,  355,  349,  746,    0,    0,
    0,    0,  750,   90,    0,  293,  288,  292,    0,  236,
  242,    0, 1046, 1044,  786,  784,    0,    0,    0,    0,
    0,    0,    0,    0,  302,    0,    0,  252,    0,    0,
  258,    0,  168,  167,  154,  164,  165,  166,    0,    0,
    0,  106,    0,    0,  287,    0,    0,  286,    0,  158,
    0,    0,  376,  334,    0,  374,    0,    0,    0,    0,
    0,    0,    0,    0,  751,    0,  237,  111,  116,  114,
  309,    0,  373,    0,    0,    0,    0,  129,    0,    0,
    0,    0,    0,    0,  163,  155,    0,    0,    0,  216,
    0,  377,    0,  253,    0,    0,    0,    0,  306,    0,
  284,  131,    0,  282,    0,    0,    0,  133,    0,  378,
    0,    0,  205,  210,    0,    0,    0,  375,  256,  169,
  112,  124,  122,    0,    0,  311,    0,    0,    0,    0,
    0,  159,    0,  290,    0,    0,    0,    0,  137,    0,
    0,    0,    0,  379,  380,    0,    0,    0,    0,    0,
  119,  326,    0,  307,    0,    0,  320,    0,    0,    0,
  315,    0,  149,    0,    0,    0,    0,  144,    0,    0,
  303,    0,  134,    0,  128,  138,  156,  162,  225,    0,
  206,    0,    0,  217,    0,  123,    0,  115,  120,    0,
    0,    0,  322,    0,  323,  312,    0,    0,  305,  316,
  285,    0,    0,  130,  145,  283,    0,  301,    0,  291,
  295,  140,    0,    0,    0,  222,  224,    0,  257,  121,
  327,  329,  308,    0,    0,  321,  318,  148,  146,  160,
  300,    0,    0,    0,  157,  226,  228,  207,    0,  220,
  218,    0,    0,  320,    0,  296,  298,  141,    0,    0,
    0,    0,  331,  332,  328,  330,  319,  161,    0,    0,
  232,  231,  230,  227,  229,  212,  208,  219,    0,    0,
    0,  297,  299,  214,  215,    0,  213,
  };
  protected static readonly short [] yyDgoto  = {             7,
    8,   50,    9,   51,   10,   11,   52,  238,  776,  777,
   12,   13,   53,   22,   23,  332,  241,  761,  948, 1163,
 1298, 1350, 1673,  945,  242,  243,  244,  245,  246,  247,
  248,  249,  754,  475,  755,  756, 1061,  757,  758, 1065,
  946, 1158, 1159, 1160,  274,  641, 1265,  111,  957, 1076,
  819, 1381, 1382, 1383, 1384, 1385, 1386, 1387, 1388, 1389,
 1390, 1391, 1392, 1393, 1394, 1395,  593, 1421,  865,  494,
  765, 1476, 1075, 1278, 1230, 1276, 1312, 1362, 1432, 1517,
 1307, 1544, 1518, 1569, 1570, 1571, 1078, 1567, 1079,  828,
  949, 1529, 1503, 1557,  549, 1550, 1523, 1586, 1027, 1555,
 1558, 1559, 1654, 1587, 1588, 1584, 1396, 1455, 1425, 1477,
  778, 1531, 1633, 1500, 1590, 1665,  495, 1456, 1457,  275,
 1486,  779,  780,  781,  782,  783,  736,  611, 1282,  737,
  738,  963, 1479, 1508, 1601, 1562, 1635, 1687, 1671, 1509,
 1696, 1691, 1480, 1535, 1661, 1638, 1602, 1603, 1684, 1669,
 1670, 1073, 1229, 1341, 1408, 1460, 1409, 1410, 1448, 1483,
 1449,  335,  228, 1566, 1451, 1551, 1548, 1397, 1427, 1472,
 1630, 1592, 1324, 1631,  642, 1679, 1680, 1471, 1547, 1520,
 1579, 1574, 1545, 1611, 1616, 1577, 1580, 1581, 1664, 1617,
 1575, 1576, 1675, 1662, 1663, 1070, 1167, 1303, 1270, 1332,
 1304, 1305, 1353, 1226, 1329, 1366,  389,  197,  113,  378,
  379,  114,  604,  471,  231, 1495,  745,  746,  937,  950,
  115,  337,  436,  328,  338,  312, 1308, 1309,   46,  118,
  313,  120,  121,  122,  123,  124,  125,  126,  127,  128,
  129,  130,  131,  132,  133,  134,  135,  136,  137,  138,
  139,  140,  259,  900,  824, 1116, 1105,  812,  987,  813,
  814, 1106,  141,  202,  820,  644,  645,  646,  891,  892,
  142,  504,  505,  305, 1114,  822,  437,  307,  533,  534,
  535,  536,  539,  830,  567,  271,  510,  855,  272,  509,
  143,  144,  145,  146, 1038,  912, 1039,  679,  680, 1040,
 1035, 1036, 1041, 1042, 1043,  147,  148,  149,  150,  151,
  152,  153,  154,  155,  156,  157,  614,  615,  616,  860,
  861,  158,  601,  590,  857,  380, 1128,  586, 1205,  159,
  524, 1273, 1274, 1277, 1357, 1071, 1228, 1339, 1452,  496,
 1313, 1314, 1375, 1376,  938,  355, 1345,  594,  595,  276,
  277,  278,  162,  163,  164,  279,  280,  281,  282,  283,
  284,  285,  286,  287,  288,  289,  290,  176,  291,  621,
  177,  329,  905,  647, 1030,  960,  772, 1082, 1028, 1031,
 1183, 1032, 1083, 1084,  292,  178,  179,  180, 1196, 1120,
 1197, 1198, 1199, 1200,  181,  182,  183,  184,  790,  517,
  791, 1186, 1100, 1187, 1319, 1285, 1243, 1320,  792, 1099,
  793, 1322, 1244,  185,  186,  187,  188,  189,  190,  314,
  561,  562,  837, 1254,  325, 1098,  970, 1284, 1125, 1004,
 1255,  191,  449,  192,  450, 1047, 1145,  451,  452,  729,
  720,  721, 1150, 1051,  453,  454,  455,  456,  457, 1052,
  715, 1049, 1259, 1346, 1414, 1147, 1294, 1365,  924,  723,
  925, 1219, 1152, 1220, 1295, 1056,   17,   19,   47,   48,
  230,  739,  941,  469,  740,  741,
  };
  protected static readonly short [] yySindex = {         -139,
    0, -188, -138,  -22,   23,19252,    0,  114,    0,    0,
   23,  -22,    0,    0,  186,    0, 8605,   23,    0, -181,
   76,    0,    0,    0,    0,    0,    0,    0,  616,    0,
  632,    0,    0,    0, 8220,    0,    0,    0,    0,    0,
    0,  618,    0,    0,  777,    0,    0,  915,    0,    0,
  114,  670,   23,    0,  678,    0,  401,  693, -144,18698,
 -166, -270,  703, 8763,    0, -270, -270, -270,   61, -270,
 -270,  741,    0,10912, -270, -270,    0,11070,    0,  790,
 -270, -143,    0, -270,  714, -270,    0,19335,19335,  747,
 -270, -270,  -55,11230,    0,17615,    0,12260,12392,12524,
12656,12788,12920,13052,13184,    0,  278,    0, 9847,    0,
  250,    0,  338,    0,    0,    0,  450,  559,  556,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,  338,    0,  889,  867,  292,  236,  800,  810,  776,
  780,  792,  786,  426,  803,    0,    0,    0,    0,    0,
    0, 3930,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,  832,  -32,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
  217,  299,  670,    0,  471,  664,    0,  808,    0,    0,
    0,    0, 9847, 9847,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,  618,    0,    0,
  806,  815,    0, -242,    0,    0,    0,  670, 1287,  891,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
  989,  338,17752,    0,    0,    0,    0,    0,17615, -178,
 -164,  972,  883,  540,  559,  338,    0,    0, 9847,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
  304,  393,18698,    0, 9847,17615,  890,    0,    0,  907,
17615,17615, 1472, -241,   58,  922, 9847,    0,11230,  278,
 1046,  935,    0,  941, 9847,17615,    0, 1056,  953,  701,
 4342,    0,    0,17615,  790,17067,    0,    0,  714,17615,
  564,  652, 1033,  338,    0,    0,  832,    0,  556, 1044,
  338,17615,17615,17615,  703,    0, 1009,    0, 9847, 9847,
    0,12128,  338,    0, 8921, -289,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 4502,    0,    0,    0,    0,19102,
  564,  982,  988,17615,    0,  118,    0,  303,    0,    0,
  165,  347,    0,    0,  960,11362, 9237,    0,    0,    0,
17615,17615,17615,17615,17615,17615,17615,17615,17615,17615,
17615,13316,13448,13580, 4662,15824,13712,13844,13976,14108,
14240,14372,14504,14636,14768,14900,15032,15164,15296,15428,
15560,18163,17615,    0,    0,    0,  832,    0,    0,    0,
    0,19335,19335,    0,    0,  338,    0,    0,    0,    0,
  669, 1048,    0,    0,    0,    0,    0,    0,    0,  670,
  891,  993,  999,    0,  118,  618,  618,    0,  737,  126,
    0,  618, 1061,    0,  -99, 1287,    0,    0,    0,    0,
 -179,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,  230, 3216,    0,    0,    0,    0,  992,
    0,    0, 1038,  143,    0, 1060,    0, 1067,  152,  790,
    0, -270,    0,    0,    0,  338,10146, -174,    0, 1055,
    0,    0,    0, -187,  125,    0,  883,  540,    0,    0,
 1063,    0, 1073,    0, 1068, 1018,    0,    0,  319,    0,
 9511,  359,11520,  922,16930,    0, 9672,    0,    0,    0,
    0,    0,    0,  146,  204,    0,    0,    0, -218,  790,
  719,    0,    0,  714,    0,    0, 1074, 1076,    0,  225,
  338,    0,  229,    0,    0,17615, 1152,    0,17615, 1157,
 1077,    0, 1080, 1082,    0,19102,    0,    0,    0,   74,
  992,    0,  -67, 8921,    0,    0,    0,    0,    0,    0,
   74,    0,    0,  -42,    0,    0,    0,  714,    0,  564,
  338,10005,    0, 1085,    0, 1084,15692,    0, 1214, 1094,
 8921,    0,    0, 1047,    0,  992,  338,17752, 1049,    0,
  118,  992,    0,  182,    0,17204,17204, 1105, 1224,    0,
    0, -280,  103,    0,    0,    0, -149,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,18288,18288,    0,
 9847, -243,    0,    0,    0,    0,  338,    0,    0,    0,
    0,  867,    0,  867,    0,11996,  292,    0,  292,    0,
  236,    0,  236,    0,  236,    0,  236,    0,  800,    0,
  800,    0,  810,    0,  776,    0,  780,    0,  792,    0,
  786,    0,   25,    0,11520, 1187,  338, 1193,  338,11520,
11520, 1107,17615,    0,    0, 1048,    0,  338,    0,  686,
  118,    0,    0,    0,    0,10005,  737,    0, 1117, 1116,
    0,    0,  600,  670,  424,    0,    0,    0,    0,    0,
    0, -176, 1118,    0, 1120, 1119,    0,    0,    0,    0,
 1125,10163, 1081,    0,  508,    0,    0,  621,    0,17752,
    0, 1111,    0,    0,    0,  831,  127, 1131,    0, 1130,
 1132, 1133,    0,    0,17615,    0,    0,  338,    0,    0,
 1129,    0, 1134,    0,  449,    0,    0, 8763,    0, 8763,
10304,    0,16030,    0,    0,    0,10464,10596,  245,11520,
    0,  144,  112,    0, 1087, 1086,    0,    0,    0,  733,
    0,    0, 1145, 1144,    0,    0,    0,    0,    0, 1147,
    0,    0,    0, 1155, -270, 4820,  790,    0,  790,    0,
    0,    0,    0, 8763,    0,    0, 8763,    0,17615,    0,
17615, 9847,    0,    0,  790, 1151,   74,    0,    0,    0,
17615,    0,    0,    0,    0,    0,    0, 9847,    0,    0,
  338,19102, 1190,    0,    0,    0,    0,    0,    0,    0,
  992,  751,    0,    0,    0,    0,    0,    0,17204,    0,
    0,    0,  338,    0,    0,16793,    0,    0,    0,    0,
    0, 9830,    0, 9988, 1160,    0,    0, 1163,  338,18413,
    0,    0,    0,10754,    0,    0, 1236,    0, 1243,    0,
    0,    0,  926,    0, 1164,    0,    0,    0,    0,    0,
    0,    0,    0,  737,    0,    0, 1121,    0,  126,    0,
  737,    0,    0,  993, 1170, 1172, 1123, 1177, 1081,    0,
 1171,    0, 1290, 1293,    0,    0,11520,    0,17341, 1176,
  831,10005, 9847,    0,    0,  269, 1301, 1302,  235, 1178,
    0,    0,17615,    0,17615, 1281,    0,    0,    0,17478,
    0,  533,17478,  763,    0,    0,    0,    0, 9374,    0,
 1306,  832,11520, 1195,10304, 1196,    0,17615,    0,  338,
    0,    0,    0,  -39,    0,    0,    0, 1191,    0, 1223,
    0,    0,    0,    0,    0,    0,    0,    0, 1199, 1186,
    0,  942,    0,    0,    0, 9847,    0,    0,  338, 1198,
 1160,    0,18288, 1275,  645,    0,  546,  338,    0, 1203,
    0, 1202,    0,    0,    0,    0,11520,    0,11520,    0,
  173,11520,    0,    0,    0,  771,    0,    0,    0, 1207,
  993,    0,    0,11678,    0,    0,    0, 1208, 4982,    0,
 1081,    0, 1081,    0, 1081,    0,    0,    0,    0,    0,
  338, 1204, 1176,    0,    0,    0, -165, -151, 1210, 1212,
    0,    0,    0,    0,    0, 1205,10304, 1160, -149,17615,
    0, 1206, 8763,    0,    0,    0,    0,    0, 1201,    0,
    0, 1219,    0,  922,    0,    0,    0,    0,    0, -197,
 1218, 1220,    0, 1160, 1230,    0,    0,   74,    0,17615,
    0, 1168, 1225,    0,    0,    0,18519,    0,  338,18519,
    0,    0,18413,    0,11520, 1259,11520,    0,11520,    0,
    0,17615,    0,    0, 1119,    0,  273,  961,    0,    0,
    0,    0,  -22,    0,    0,    0, 1242,    0,    0,    0,
 1229,    0,    0,    0,  619,    0, 1231, 1358, 1359,    0,
    0, 1160, 1244, 1160,    0, 8763,  708,    0,    0,17478,
    0,    0,    0,17615,    0, 1247, -202,    0, 8445,    0,
    0, 1362,    0,    0,   74,    0,    0,17615, 9988,    0,
    0,    0,    0,    0,    0, 1272,    0, 1017, 1246,    0,
 1253,    0,    0,11678,   23,  152,    0,  620, 1258, 1263,
17341, 1265,    0,17615,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 8763,    0,   55, 1267, 1264,    0,    0,
 8921,    0,    0, 1350, 8763,    0, 1270,    0,11520,    0,
    0,    0,    0,    0,17615,    0,    0,  670, 1271,  670,
  152, 9847, 1258, 1304,    0, 1304,    0, 1258,    0,    0,
    0,17615,    0, 8763,17615,    0,    0,    0,    0, 1278,
    0,    0, 1299,11520,17615,    0,  670, 1279,    0, 1232,
 1029,    0,    0, 1280,    0, 1277,    0,  117,    0, 1282,
 1237,    0, 1304,    0,    0, 1304,    0,    0,    0,  830,
 1134,    0,    0,    0,    0, 1309,    0,  -86, 1304, 1402,
    0, 1292,  670,    0,    0, 9847,    0,   88, 1296,    0,
 1297,    0,    0,    0, 8921,11520,    0,    0,    0,    0,
    0,    0, 1284, 1291,    0,    0,16930,    0,19297,   93,
  670, 1305, 1298, 1313,11520, 1300,17615,    0,    0, 1303,
    0, 1307,    0,    0, 1295,    0,    0,    0, 3216,    0,
 1311,   93,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,  484, 3216,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 1312,  670,    0,
    0,    0,  338,    0, 1313,    0,    0,    0, 1310,19297,
 8921,18902,    0,    0,  530,    0,    0,    0,18970,    0,
    0,   93,    0,    0,    0,    0,  -67, 9847, 9847,  348,
 9847,  408,  714, 1338,    0,  564, 5459,    0, 1373,    0,
    0, 1291,    0,    0,    0,    0,    0,    0, 1827, 1291,
 1315,    0, -147,  -97,    0, 9847,  -95,    0, 9847,    0,
 1266, 1317,    0,    0, -282,    0,   72, 1780,    0, 1320,
 1273,  184,  530, 8220,    0,17615,    0,    0,    0,    0,
    0, -282,    0, 1314, 1288, 1319, 1331,    0, 1322, 1294,
 1339,  152, 1328, 1340,    0,    0, 1343, 1351, 1374,    0,
  992,    0,  985,    0, 1355, 1344, 1291,   75,    0, 1341,
    0,    0, 1364,    0, 1357, 1365, 1368,    0, 1360,    0,
  152,  152,    0,    0,  152, 1370, 1371,    0,    0,    0,
    0,    0,    0, 1375,  177,    0, 1377,  152, 1476, 1378,
  152,    0,  623,    0,10304, 1327, 1379, 1360,    0, 1381,
 1385,  247, 1388,    0,    0,  152,17341, 1342, 1387, 1375,
    0,    0, 3216,    0,  670,  670,    0, 1347, 1392, 1377,
    0, 1389,    0,17615, 1352, 1393, 1378,    0, 1400,  152,
    0,   99,    0, 1396,    0,    0,    0,    0,    0, 3216,
    0,  247,  247,    0, 1408,    0,   75,    0,    0,  266,
 1413, 3216,    0, 3216,    0,    0,10304, 1401,    0,    0,
    0, 1414, 1364,    0,    0,    0, 1411,    0,  336,    0,
    0,    0, 1304, 1037, 1417,    0,    0,  507,    0,    0,
    0,    0,    0, 1474, 1524,    0,    0,    0,    0,    0,
    0, 1420, 1421,10304,    0,    0,    0,    0,  247,    0,
    0,  656,  656,    0, 1304,    0,    0,    0,   50,   50,
 1416, 1427,    0,    0,    0,    0,    0,    0,16930,16930,
    0,    0,    0,    0,    0,    0,    0,    0, 1434, 1441,
17341,    0,    0,    0,    0, 1438,    0,
  };
  protected static readonly short [] yyRindex = {         1716,
    0,    0, 9079, 1716,    0,    0,    0, 1816,    0,    0,
 3596, 3346,    0,    0,    0,    0,    0, 3596,    0,    0,
   46,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 1817,    0,    0, 1817,    0,    0, 1817,    0,    0,
 1816, 2015, 3471,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 1450,    0,    0,    0,    0,    0,    0,    0,
    0,10771,    0, 1442,    0,    0,    0, 1442,    0,    0,
    0,    0,    0,    0, 3214,    0,    0,    0,    0,    0,
    0,    0,    0,  316,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 5461,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 6175, 5555, 5771,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 6280, 6457, 4109, 6860, 7285, 7492, 7630,
 7768, 7906, 8044, 1349, 8182,    0,    0,    0,    0,    0,
    0,   46,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 6352,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 3639,    0,  819,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 1817,    0,    0,
  169,    0,    0,    0,    0,    0,    0, 3702,  444, 3745,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 4341,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 1454,    0,    0,    0,    0,    0,    0, 4501,
 1448,    0,    0,    0,    0,    0,    0, 1448,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 2811,    0,
  382, 2942,    0,    0,    0,    0, 3089,    0, 2942,    0,
    0,    0,    0,    0, 1450,    0,    0,    0,    0,    0,
    0,    0,    0,    0,  216,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 1453,
 1884,    0,    0, 1442,    0, 4341,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,  233,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 2194,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 8217,    0,    0,    0,    0,    0,    0,    0, 3808,
 3851,    0,    0,    0, 2661, 1817, 1817,    0, 9254,   29,
    0, 1817, 1828,    0,    0,  160,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,  534,18630,    0,    0,    0,    0, 4341,
    0,    0,    0,    0,    0,    0,    0,    0,19002,    0,
    0,    0,    0,    0,    0,    0, 1459,    0,    0,    0,
    0,    0,    0,    0,    0,    0,  367, 1090,    0,    0,
  447, 1436,    0,    0, 1458,  785,    0,    0,    0,    0,
  243,    0,    0, 4981, 1465,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 1471,    0,
 2363,    0,    0,  432,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 1453,    0,    0,    0,17889,
 4341,    0,    0,  267,    0,    0,    0,    0,    0,    0,
17889,    0,    0,    0,    0,    0,    0,  113,    0,  439,
    0,    0,    0, 1467,    0,    0,    0,    0, 1448,    0,
    0,    0,    0, 4181,    0, 4341,    0,    0, 4020,    0,
 4341, 5141,    0,    0,    0,    0,    0, -183,    0,    0,
    0,    0,  255,    0,    0,    0,  962,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 5712, 5860, 5965, 6070,    0, 6529,    0,    0,    0,
    0, 6634,    0, 6706,    0,    0, 5640,    0, 6791,    0,
 6929,    0, 6998,    0, 7067,    0, 7136,    0, 7354,    0,
 7423,    0, 7561,    0, 7699,    0, 7837,    0, 7975,    0,
 8113,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 8217,    0,    0,    0,    0,
 2661,    0,    0,    0,    0,    0,10929,    0,    0,  971,
    0,    0, 1423,16233,    0,    0,    0,    0,    0,    0,
    0,  932,  844,    0,    0, 1475,    0,    0,    0,    0,
 3389,    0,    0,    0,    0,    0,    0,11836,    0,    0,
    0,  968,    0,    0,    0,11537,19175,    0,    0,  980,
  987,  990,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 1468,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 1478,    0,    0,    0,    0,    0, 7198,    0,
    0,    0,  285,    0,   83, 4661,    0,    0,    0,    0,
    0,    0,    0, 1477,    0,    0,    0,    0,    0, 1479,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,17889,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 4341,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 1481,    0,    0,    0,    0, 1482,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0, -168,    0,  715,    0,    0,    0,    0,    0,
    0,    0,    0,11537,    0,    0,    0,    0,   29,    0,
 9689,    0,    0, 1480,    0,  833,    0,    0,    0,    0,
 1484,    0, 1440, 1446,    0,    0,    0,    0,    0, 1491,
11695,    0,    0,    0,    0,19207,    0,    0,    0,  995,
    0,    0,    0,    0,    0, 2531,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 4821,    0, 5301, 1492,    0,    0,    0,    0, 1503,
    0,    0,    0,  995,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0, -155,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 1500,    0,    0,    0,    0,    0, 1499,  784,    0,    0,
    0, 1505,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 1002,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 1502,    0,    0,    0,    0,    0, 1005, 1013,
    0,    0,    0,    0,    0,    0,    0, 1508,  962,  729,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 4981,    0,    0,    0,    0,    0, 1513,
    0,    0,    0, 1508,    0,    0,    0,17889,    0,    0,
    0,  789,  895,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 1475,    0,16078,    0,    0,    0,
    0,    0,19398,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,  950,    0,  951,    0,    0,    0,
    0, 1510,    0, 1481,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 1516,    0,    0,    0,
    0,  159,    0,    0,17889,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0, -140,  836,    0,
    0,    0,    0,    0,19441,19002,    0,  520,  578,  480,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 8284,    0,    0,    0,
  -71,    0,    0, 1471,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,19585,    0,  308,
19002,    0,  606, 1519,    0, 1519,    0,  578,    0,    0,
    0,    0,    0,    0,  855,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,19628,    0,    0,    0,
16507,    0,    0, 1520,    0,    0,    0,  542,    0,  626,
    0,    0,  558,    0,    0, 1519,    0,    0,    0,    0,
  860,    0,    0,    0,    0,    0,    0, 3514, 1511, -254,
    0,    0,  322,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 1525,    0,    0,    0,    0,    0,
    0,    0,    0, 2115,    0,    0, 1465,    0,    0,16339,
16591,    0,    0,  886,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,  528,    0,    0,    0,18802,    0,
    0,16423,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,18870,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,16675,    0,
    0,    0,    0,    0,  886,    0,    0,    0,    0,    0,
  216,  534,    0,    0,    0,    0,    0,    0,  534,    0,
    0,16339,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 5139,  543,    0,16717,    0,    0,    0, 5299,
    0, 2115,    0,    0,    0,    0,    0,    0,    0, 2115,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,  674,    0,  586,    0,  634,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,  797,    0,  730,    0,    0,    0,    0,    0,    0,
    0,19002, 1012,    0,    0,    0,    0,    0,    0,    0,
 1522,    0,  869,    0,    0,    0, 2115,    0,    0, 1015,
    0,    0,    0,    0,    0,    0,    0,    0, 1521,    0,
19002,19002,    0,    0,19070,    0,    0,    0,    0,    0,
    0,    0,    0, 1523,15855,    0, 1526,19002,18026, 1527,
19002,    0,    0,    0,    0,    0,    0, 1529,    0,    0,
    0,19781,    0,    0,    0,19002,    0,    0,    0, 1530,
    0,    0,  310,    0,19711,19751,    0,    0,    0, 1531,
    0,    0,    0,    0,    0,    0, 1532,    0,    0,19002,
    0,  710,    0, 1016,    0,    0,    0,    0,    0, 1045,
    0,18444,19671,    0,    0,    0,    0,    0,    0,    0,
    0, 1582,    0, 1636,    0,    0,    0, 1019,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,  596,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,19781,    0,
    0,19485,19515,    0,  596,    0,    0,    0,    0,    0,
15951,    0,    0,    0,    0,    0,    0,    0, 1465, 1465,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,
  };
  protected static readonly short [] yyGindex = {            0,
    0, 1859,    0,    0,    0,    4,  -14, -192,  -50,  -47,
    0, 1901, 1910,  489,    0,   -8,    0,    0,    0,    0,
    0,    0,-1294, -789, -223, -720,    0,    0,    0,    0,
    0, -229,    0,    0,    0,  856,    0,  975,    0,    0,
    0,    0,  692,  697,  -17, -221,    0,    7,    0,    0,
 -548,  490,    0,  544, -686, -644, -624, -606, -601, -574,
 -573, -567,    0,    0,-1171,    0, -850,    0,  487,-1317,
    0,   18,    0,    0,    0,  651,-1218,    0,    0,    0,
  505,  323,    0,    0,    0,  362,-1189,    0, -283, -303,
 -385,    0,    0,    0, -986,  311,    0,    0, -521,    0,
    0,  375,    0,    0,  349,    0,    0,  452,    0, -596,
 -852,    0,    0,    0,    0,    0, -431,  384,-1442,  -10,
    0,    0,    0,  973,  996,  997, 1162, -557,    0,    0,
 -337, 1000,  500,    0, -813,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,  295,    0,
    0,    0,    0,    0,    0,    0,    0,  557,    0,    0,
    0, -293,  483,    0,    0,    0,    0,    0,    0,    0,
    0,    0,  571,    0, -539,    0,    0,    0,    0,    0,
    0,    0,    0,    0,  305,    0,    0,  390,    0,    0,
  396,  400,  313,    0,    0,    0,    0,    0,    0,    0,
    0,  646,    0,    0,    0,    0,  -88,    0,   -3, -361,
 -297, 1059,    0,  465,    0, -390,    0, 1041,    0, 1660,
   47, -298, -272,  -65,  965,  569,    0,  647,    0,  -40,
  933, -393,    0,    0, -389,    0,    0,    0,    0,    0,
    0,    0,    0,    0, -379,    0,    0,    0,    0,    0,
 -267,    0,   56,    0, -358,    0, -275,    0,    0,    0,
  998, -932, -307, -128,  587,    0, 1083,    0, 1345, -584,
  142, -332, 1216,    0,    0,  870, 1920,    0,    0,    0,
    0, 1192,    0,    0,    0, 1649,    0,    0,    0,    0,
    0, 1528, 1007, 1008, 1580, -217, 1583,  807,    0,    0,
    0,  863, -142,    0,  861,  852, 1004, 1578, 1581, 1577,
 1584, 1585,    0, 1579,    0,    0,    0, 1148, 1433, -563,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 -308,  794,    0,-1120,    0,    0,    0,    0,    0, -474,
    0,  716,    0,  610,    0,    0,    0,  835, -565,    1,
 -339,    2,    0, 1869,    0,   86,    0,  138,  163,  170,
  176,  178,  188,  244,  252,  259,  284,    0, -750,    0,
    0,    0,  943,    0,  848,    0,    0,    0,  834, -511,
  916, -968,    0,  963, -503,    0,    0,    0,    0,    0,
    0,  847,    0,  846,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
  762,    0,    0,    0,    0,    0,    0,    0,    0,  -33,
    0, 1487,  796,    0,    0, 1051,    0,    0,    0,    0,
    0,    0, -182,    0,    0,    0,    0,    0, 1602, 1330,
    0,    0,    0,    0, 1606,    0,    0,    0,    0,    0,
    0,    0,    0,    0,  644,    0,    0,    0,    0,    0,
    0,    0,    0,  765,    0,    0,    0,    0,    0,    0,
   -1, 1122,    0,    0,    0, 1127,
  };
  protected static readonly short [] yyTable = {           110,
  460,  239,  548,  193,  240,  823,  112,   18,  551,  459,
  498,  196,  520,  789,  477,  597,  198,  160,  161,  265,
  766,  573,  673,   44,  622,  829,  674,  565,  866,  546,
  398,  502,  544,  434,  581,  532,  675,  867,  643, 1132,
  996, 1280,  612,  233,  319, 1071,  235,  976,  258,  977,
 1108,  390,  895,  399,  871,  878,  311, 1315, 1193, 1369,
  318, 1422, 1135,  634,  578, 1194, 1534,   14,  797,  320,
 1194,  323,  732,  835,  194,  300,  759,  255,  356,  942,
 1429,  794,  473,  599,  198,  198,  613, 1055,  623,  267,
 1174,  322,  898, 1003,  333,  340, 1005, 1342, 1302,  347,
  509,  256,  165,  732, 1176,  198,  903, 1275, 1489,  374,
 1351,  255,  322,  387,  347, 1058,    1,  269,   16,  386,
  347,  293,  294,  295,  909,  301,  302,  348,  910,  537,
  315,  316,  375,  538,  336,  336,  321,  324,  767,  326,
  600,  330,  239,  327,  110,  461,  343,  344,  400,  899,
  257,  112, 1310,  256,  166,  336,  749, 1485, 1491,    6,
 1497, 1302,  160,  161,  714, 1487,  915,  733,  381, 1348,
 1055,  920,  921,  397,  502, 1055,   43, 1055,  931,  167,
 1055, 1055,  750, 1055, 1055,  798,  168,  476,  298,  732,
  461,  732,  169,  256,  170, 1660,  902,  298, 1058,  198,
  198,  266,  257, 1058,  171, 1058, 1055,  256, 1058, 1058,
  268, 1058, 1058, 1135,  751,  374, 1123,  509,    2,  509,
 1195,  345, 1541,   95,   95, 1195,  468,  256,  256,  862,
   15,  498,  904,  815, 1058,  501,  477,  195,  375,  760,
  821,  506,  257,  574,  299,  376,  732,  165,  612,  336,
  336,  474,  265,  299,  597, 1610,  257, 1245,  500,  503,
  172, 1055,  265,  572,  880,  198,  575, 1177,  173,  884,
  886, 1490,  507,  390,  509,  174,  257,  257,  519,  978,
  913,  597, 1634,  523,  525,    3,    4,    5,    6, 1058,
 1349,  198,  613, 1011, 1644,  882, 1645,  902,  555,  166,
  175,  863, 1071,  198, 1019, 1681,  568,  298,  570,  531,
 1287,  198,  523,  540,  569,  336,  324,  198,  613,  752,
  397, 1492,   51, 1498,  167,  583,  584, 1505,  296,  858,
 1542,  168,  868,  885,   51,    2,  297,  169, 1071,  170,
  577,  336, 1097, 1358,  438,  198,  198,  580, 1377,  171,
  620,  902, 1189,  550, 1628,  596,  598,   20,  901,  591,
  545,  336,  864,  299,  346, 1221,  619,  988,  395,  929,
   49,  198, 1335, 1269,  869,  265,  198, 1606,  501,  640,
  799,  742,  964,  648,  649,  650,  651,  652,  653,  654,
  655,  656,  657,  658,  626,  336,  336,  439,  631,  932,
  914,  831,  503,  503, 1087,  172,  502,  774, 1077,  239,
   20,  198,  461,  173, 1655,  713,    6,   95, 1306,  743,
  174,  336,  744,  298,  381,  541, 1682,  542, 1148,  256,
 1288,  754, 1572,  930,  516, 1239,  754,  887,  198,  198,
  754,   95,  613, 1506, 1115,  175, 1678,   51,  716,  718,
  978, 1071,  722,  730, 1467,  754, 1543, 1071,  198,  833,
 1378,  336,  237, 1359,  734,  735,  753,  571,  565, 1101,
  747,  198,  440, 1118, 1629,  395,  969,  902,  257,  299,
  843,  543,  754,  395,  846,  395,  989,  395,  336,  336,
 1095, 1336,  775, 1286, 1363,  237,  237,  800, 1144,  784,
 1146,  754, 1599, 1151, 1291,  200,  762,  381, 1512,  201,
  763,  381,  985,  769,  498,  383,  673,  770,  832, 1017,
  674,  237,  786, 1149,  795,  818,  978,  640, 1641,  827,
  675,  395,  375, 1318,  612,  532,  381, 1022,  627,  376,
  381,  201,  381,  381,  381,  381,  237,  498,  502,  838,
  381,  503,  888,  415,  440,  441,  770,  676,  848,  511,
  442,  850,  443,  384, 1206,  444,  445,  785,  446,  447,
 1437,  764,  754,  986,   45, 1181,  834,  845,  613,  859,
 1642,  855,  198,  398,  799,  117, 1182,  373,  628,  416,
  859, 1651,  334,  341,  596,  598, 1214,  844, 1216, 1652,
 1217,  847, 1513, 1465,   55,  491,  399, 1096,  198, 1089,
  501,  477, 1182,  382,  836,  373,  237,  870,  890,  890,
  373,  596,  598, 1085,  754,  385,  983,  492,  117,  855,
  419,  420,  117,  881,  503,  800,  374,  441,  237, 1380,
 1399, 1256,  442,  512,  443,  799,  448,  444,  445, 1527,
  446,  447, 1466,  478, 1222, 1080,  339,  339,  788,  375,
 1653, 1380,  491, 1468,  374,  298,  376,  198,  613,  374,
  417,  418,  477,  377, 1400,  854,  342,  339, 1560, 1561,
 1024,  815, 1563,  911,  492,  866,  990,  375, 1399,  805,
  344,  855,  375,  806,  376, 1582,  800,  818, 1589,  376,
  298,  377,  818,  818,  974,  923,  377,  466,  467,  298,
 1293, 1380, 1469, 1605,  478,  754, 1401,  336, 1117, 1113,
  754,  624, 1400,  917,  754,  919,   51,  551,  458,  816,
  117,  625,  198,  538,  927,  855, 1402, 1627,  366,  754,
   51,  855,  975,  673,  366, 1326,  673,  674,  855,  673,
  674,  366,  501,  674, 1403,  366,  299,  675,  198, 1404,
  675,  513,  265,  675, 1401,  629,  754,  523,  366,  514,
  251,  339,  339,  508,  252,  630,  503,  893,  893,  952,
  951,  789, 1122,  827, 1402,  754, 1405, 1406, 1636, 1637,
  818, 1185,  818, 1407,  971,  366,  855, 1364,  939,  518,
  366,  366, 1403,  953,  388,  752,  387, 1404,  366,  531,
  821,  705,  366,  705,  753,  547, 1415, 1370,  738,  554,
 1141,  547,  515,  388,  253,  366, 1001,  431, 1002,  752,
  954, 1006,  198, 1007, 1405, 1406,  327,  339,  753,  432,
  940, 1407, 1162, 1012, 1009, 1672,  859,  738,  198, 1453,
  388, 1323,  462,  588,  589,  738,  752,  366,  741,  597,
 1426,  117,  374,  339,  198,  753,  781, 1493,  198,  265,
  781,  356, 1453, 1153, 1659,  339,  547, 1014,  640,  603,
  775, 1507,  336,  339,  640,  375,  827,  741,  743,  339,
  998,  744,  376,  935,  117,  781, 1046,   95,  336, 1020,
  778, 1107,  503, 1525,  781,  936, 1454,  993,  503,  775,
 1246,  597,  387,  775,  336, 1034,  773,  339,  339,  666,
  203, 1140,  391,  117,  724,  778,  955, 1080,  373,  371,
  256,  374,  392,  956,  778,  753,  204,  374,  777,  818,
  198,  519,  327,  339,  743,  773,  717,  719,  339,  119,
 1283,  393,  394,  373,  375,  374,  558, 1102,  198,  198,
  375,  376, 1104,  777,  744, 1104, 1068,  376,  377,  559,
 1049,  395,  777,  743,  377,  818,  209,  827,  375,  257,
 1121,  116,  396,  339,  117,  376,  560,  559, 1232,  229,
   95, 1271,  119,  744, 1210, 1272,  119, 1212, 1317, 1454,
 1233,  204, 1162,  766,  839,  597,  911,  441,  387,  336,
  339,  339,  442, 1138,  443, 1478,  371,  444,  445, 1139,
  446,  447,  198,   95,  116,  463,  440, 1478,  116,  818,
  893,  818, 1368, 1593,  818, 1133, 1447,  463,  327,  237,
  464,  371,  605, 1459,  911,  371, 1161,  366,  127,  606,
  127,  676,  753, 1049,  250,  127, 1482,  198, 1049,  463,
 1049,  607,  928, 1049, 1049,  198, 1049, 1049, 1482,  254,
  503,  498,  550,  477,  270, 1297, 1240,  294,  199,  827,
 1241,  597,  523, 1417, 1242,  117,  294, 1171, 1168,  371,
 1169, 1051, 1170, 1175,  119, 1646,  265,  935,  725,  265,
  327,  935,  265,  992,  304,  935,  304,  993,  303,  441,
  304,  304, 1207, 1504,  442,  336,  443,  859,  342,  444,
  445, 1018,  446,  447,  381,  770,  116,  818, 1450,  818,
 1504,  818, 1668, 1109, 1218, 1450,  605,  993,  961, 1689,
 1690, 1045,  551,  606, 1049, 1034,  199,  199,  229, 1536,
  232, 1537,  975,  542,  339,  607,  975,   95,  975,  542,
  427,  880,  117,  880,  371,  880, 1225,  199,  371,  428,
  366,  371, 1104,  371, 1051,  908,  519,  429,  371, 1051,
  339, 1051,  421,  422, 1051, 1051,  430, 1051, 1051,  117,
 1257,  827,  366,  433,  859,  366,  423,  424, 1343,  596,
  598,  304, 1344,   70,  117,  117, 1161,   70,  425,  426,
 1268,  381,  371,  519,   69,   69,  519,  239,   69, 1300,
  461,  470, 1301,  937, 1045,  119,  465,  937,  939, 1045,
  503, 1045,  939,  472, 1045, 1045,  381, 1045, 1045,  339,
  381,  818,  366,  381,  499,  381,  239, 1296,  345,  461,
  381,  596,  598,  412,  413,  414,  353,  116,  119,  497,
  237,  199,  199, 1080,  519, 1051,  521,  869,  198,  869,
  401,  869,  691,  693,  695,  697,  818, 1218,  676, 1053,
 1054,  676, 1300,  522,  676, 1301,  229,  119,  234,  547,
  116,  402,  403,  404,  405,  406,  407,  408,  409,  410,
  411,  552,  366,  366,  339,  366,  366,   62,  388, 1379,
 1398,  556, 1301,  744, 1131, 1045,  902,  553,  336,  116,
  184,  197,  184,  197,  184,  197,  576,  199,  818,  557,
  339, 1379,  198, 1223, 1301, 1224,  860,  579,  860,  640,
 1000, 1092,  871, 1092,  871,  596,  598,  818,  672,  519,
  172,  587,  172,  199,  617,  198, 1008,  179, 1398,  179,
  180,  744,  180,  503,  618,  199,  117, 1000,  117, 1000,
 1261, 1262, 1013,  199,   72,  203,   72,  203,  632,  199,
  116, 1379,  336,  173, 1301,  173,  132, 1695,  132,  310,
  139,  310,  139,  317,  982,  317,  984, 1331,  237,  375,
  470, 1656, 1657,  728,  339,  336,  375,  199,  199,  754,
  754,  381,  117,  768, 1458,  117,  198,  731,  198, 1433,
  339,  596,  598,  682,  684,  198,  687,  689,  699,  701,
  748,  796,  771,  199,  198,  198,  339,  198,  199,  773,
  339,  802,  803, 1481,  801,  849,  841,  804,  842,  119,
  851,  852,  853,  366,  854, 1481,  873,  117,  854,  872,
 1481,  366,  198,  366,  854,  198,  336, 1088, 1516,  876,
  877,  854, 1458,  199, 1481,  879,  896,  883,  117,  897,
  916,  116,  366,  366,  336,  336,  918,  550,  922,  933,
  934,  463,  959,  944, 1573,  943, 1481,  744,  947,   43,
  199,  199,  366,  965,  966,  972,  967,  968,  973,  991,
  366, 1600,  336,  366,  744,  336,  200,  994,  902,  854,
  199,  995,  997, 1010, 1612, 1614,  119,  744,  744, 1048,
  339,  339, 1016,  199, 1029, 1033, 1050,  827, 1055, 1058,
 1063, 1066, 1458, 1064, 1067, 1072, 1069,  478, 1074,  519,
 1081, 1600, 1600,  119,  744,  744, 1093, 1094,  116, 1097,
 1103, 1112,  541, 1126, 1119, 1127, 1622, 1130,  119,  119,
  479, 1129, 1137, 1166, 1134, 1142, 1143, 1154, 1188, 1164,
 1172, 1180, 1190,  480, 1178,  116, 1179,  481,  482, 1191,
 1201, 1208, 1202,  483,  339,  484,  485,  486,  487,  827,
  116,  116, 1204,  488,  683, 1215, 1209,  489, 1600, 1227,
 1231,  744, 1234, 1235, 1236, 1249, 1237, 1253, 1260,  490,
 1263,  354,  491, 1264,  492,  358,  360,  362,  364,  366,
  368,  370,  372, 1272, 1271, 1281,  827,  339,  910, 1289,
 1292,  835, 1311, 1299,  199, 1323, 1325, 1328,  493, 1334,
 1330, 1674, 1674, 1337, 1333, 1338, 1347, 1352, 1683, 1683,
 1354,  640,  640, 1360, 1361, 1367, 1412, 1368, 1413, 1420,
  199,  117, 1411, 1694, 1470, 1418, 1416, 1484, 1419, 1423,
 1430, 1519, 1435, 1488, 1501,  503,  503,  683, 1502, 1510,
 1521, 1511,  683, 1524,  683,  683,  683,  683,  683,  683,
  683,  683,  683,  683,  683,  117, 1513, 1522,  117, 1528,
 1526,  117, 1492, 1530, 1532,   37, 1453,  683, 1533,  683,
 1540,  683, 1546,  683,  683,  683, 1539,  526, 1552,  199,
  119, 1583,  119, 1549, 1556,   57,   24, 1553,   25,  683,
 1554,   26,  260, 1564, 1565, 1594,   27,   62,   63, 1568,
   28, 1578, 1585, 1597,  117, 1595,   65, 1598, 1604,   30,
 1607, 1621,  116, 1608,  116, 1618,   32,  117, 1619, 1624,
 1623,   33, 1626,   72,   73,   34,  119, 1632,  683,  119,
 1639, 1643, 1647, 1650, 1648, 1658, 1641,   36, 1642,   37,
   75, 1666, 1667,   38,  199, 1688,   77, 1686,   79,  446,
   81,   39,   40,  261,  548,   41, 1692,  446,  116,  446,
  548,  116,  117, 1693, 1697,    9, 1088,  580,  967,  117,
  199,  119,  540,  117,  968,  707,  541, 1080,  446,  446,
   90,   91,   92,  262,  527,  930,  354,  491,  982,  708,
  339,  789,  672,   96,  934,   33,  539,   33,  446,  492,
   34,  339,  117,  116,  643,  366,  446,  863,  233,  446,
   34,  366,  366, 1374,  107,  548,  366,  873,  366,  582,
  366,  366,  366,  366,  116,  975,  864,  644,  874,  354,
  865,  897,  866,  366,  898,  366,  777,  777,  343,  106,
  528,  366,  366,  799,  199,  366,  754,  135,  754,  117,
  529,  530,  313,  142,  339,  136,  118,  314,  143,  236,
  199,  366,   54,  117,   21, 1266, 1155,  366, 1062,  366,
 1267, 1461,  366, 1462, 1374, 1424,  199,  339, 1316, 1640,
  199, 1609, 1596, 1649, 1514, 1625, 1591,  962, 1090,  660,
  662,  664, 1463, 1464,  582,  582,  582,  582,  582,  582,
  582,  582,  582,  582,  582,  582,  582,  582,  582,  582,
 1086, 1091, 1092, 1494, 1685, 1431, 1515, 1428, 1677, 1620,
 1496, 1615,   37, 1499, 1613, 1676,   37, 1538, 1355, 1059,
  566,  894, 1356, 1192, 1025,  958, 1111,   37,  339,  117,
  339,  306,   37,  585,  979,  677,   37,  339,  678,   37,
 1248, 1211,  199, 1213,  703,  707,  339,  339,  705,  339,
  712,   37,   37,  709,    5,  711,   37,   37,  856, 1015,
  199,  199,   37, 1279,   37,   37,   37,   37, 1340, 1436,
  435, 1238,   37, 1251,  339,  119,   37,  339,   37, 1203,
  478, 1184, 1258, 1250, 1252, 1173, 1321,  840,   37, 1290,
   37,   37,  726,   37, 1124,  926,  727,   37, 1434, 1327,
 1057,    0, 1060,  479,    0,    0,    0,  116,  582,  672,
    0,    0,  672,    0,    0,  672,  480,   37,    0,    0,
    0,  482,    0,    0,  199,   37,  483,  478,  484,  485,
  486,  487,    0,    0,    0,    0,  488,    0,    0,    0,
  489,  116,    0,    0,  116,    0,    0,  116,    0,    0,
  479,    0,  490,    0,   31,  491,    0,  492,  119,  199,
    0,    0,    0,  480,    0,    0, 1247,  199,  482,    0,
    0,  119,    0,  483,    0,  484,  485,  486,  487,  366,
    0,  493,    0,  488,  875,  366,    0,  489,    0,    0,
  116, 1473,    0,    0,    0,    0,    0,    0,    0,  490,
    0,    0,  491,  116,  492,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,  119,    0,    0,    0,
    0,  366,    0,  119,    0,    0,    0,  119,  493,    0,
    0,    0,    0,  552,    0,  906,  907,    0, 1492,    0,
    0,    0,    0,    0,    0,    0,    0,    0,  116,    0,
    0,    0,    0,  354,    0,  116,  119,    0,    0,  116,
    0,    0,  366,    0,    0,    0,  366,  366,    0,  366,
  366,  366,  366,  366,  366,  366,  366,  366,  366,  366,
  366,    0,    0,    0,    0, 1475,  366,    0,  116,    0,
    0,  366,  366,  366,  366,  366,  366,  366,  366,  366,
  366,    0,  366,  366,    0,    0,  366,  366,  366,  366,
  366,    5,    0,  366,  366,   51,    0,  119,  366,  366,
  366,  366,  366,  366,  366,  366,   51,    0,    0,    0,
    0,   51,    0,    0,    0,   51,    0,  366,   51,    0,
  366,    0,  366,    0,  366,    0,    0,  366,    0,  116,
   51,   51,    0,  366,    0,   51,   51,    0,    0,    0,
    0,   51,    0,   51,   51,   51,   51,    0,    0,    0,
  199,   51,    0,    0,  354,   51,    0,   51,    0,    0,
    0,    0,    0,    0,    0,    0,    0,   51,    0,    0,
   51,    0,   51,  119,    0,    0,   51,    0,    0,    0,
    0,    0,  969,    0,    0,    0,    0,    0,    0,    0,
   31,   31,    0,    0,    0,   31,   51,    0,    0,   31,
    0,   31,    0,    0,   31,  116,   31,   31,    0,   31,
    0,   31,    0,   31,  199,   31,   31,   31,   31,    0,
    0,   31,   31,    0,    0,    0,    0,   31,    0,   31,
   31,   31,    0,    0,   31,   31,   31,  199,   31,    0,
    0,   31,    0,   31,   31,   31,   31,    0,    0,    0,
   31,   31,   31,    0,    0,   31,   31,   31,    0,    0,
    0,    0,    0,    0,   31,   31,    0,   31,   31,  552,
   31,   31,   31,    0,  552,  552,   31,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,   31,  552,  199,    0,
  199,    0,   31,   31,   31,    0,    0,  199,    0,    0,
  552,  552,   31,    0,    0,  552,  199,  199,  552,  199,
  552,    0,  552,  552,  552,  552,    0,    0,    0,    0,
  552,    0,    0,    0,  552,    0,  582,    0,  552,    0,
    0,    0,    0,    0,  199,    0,  552,  199,    0,  552,
  891,  552,  552,   31,    0,    0,  552,  552,    0,  552,
  552,  552,  552,  552,  552,  552,  552,  552,  552,  552,
    0,    0,    0,    0,    0,  552,  552,  552,    0,    0,
 1136,  552,  552,    0,  552,  552,  552,  552,  552,  552,
  552,    0,  552,  552,    0,  552,  552,  552,  552,  552,
  552,  552,  552,  552,  552,    0,  552,  552,  552,  552,
  552,  552,  552,  552,  552,  552,  552,  552,  552,  552,
  552,  552,  552,  552,  552,  552,  552,  552,    0,    0,
  552,    0,  552,    0,  552,    0,    0,  552,  969,  969,
    0,    0,    0,  552,    0,    0,  969,  969,  969,  969,
  969,    0,  969,  969,    0,  969,  969,  969,  969,  969,
  969,  969,  969,    0,    0,    0,    0,  969,    0,  969,
  969,  969,  969,  969,  969,    0,    0,  969,    0,    0,
  366,  969,  969,    0,  969,  969,  969,    0,    0,    0,
    0,    0,    0,    0,    0,    0,  969,    0,  969,    0,
  969,  969,    0,    0,  969,    0,  969,  969,  969,  969,
  969,  969,  969,  969,  969,  969,  969,  969,    0,  969,
    0,    0,  969,  969,    0,    0,  969,  969,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,  969,  969,  969,  969,  969,    0,    0,    0,    0,
  969,  969,    0,    0,  969,    0,    0,    0,    0,  969,
  969,  969,  969,  969,    0,    0,    0,  969,    0,  969,
    0,    0,    0,    0,    0,  969,  969,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
  969,  969,  969,  969,    0,  969,  891,  891,    0,    0,
    0,    0,  969,    0,  891,  891,  891,  891,  891,    0,
  891,  891,    0,  891,  891,  891,  891,  891,  891,  891,
  851,    0,    0,    0,    0,  891,    0,  891,  891,  891,
  891,  891,  891,    0,    0,  891,    0,    0,    0,  891,
  891,    0,  891,  891,  891,    0,    0,    0,    0,    0,
    0,    0,    0,    0,  891,    0,  891,    0,  891,  891,
    0,    0,  891,    0,  891,  891,  891,  891,  891,  891,
  891,  891,  891,  891,  891,  891,    0,  891,    0,    0,
  891,  891,    0,    0,  891,  891,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,  891,
  891,  891,  891,  891,    0,    0,    0,    0,  891,  891,
    0,    0,  891,    0,    0,    0,    0,  891,  891,  891,
  891,  891,    0,    0,    0,  891,  366,  891,    0,    0,
    0,  366,  366,  891,  891,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,  356,    0,    0,  366,    0,    0,    0,  891,  891,
  891,  891,    0,  891,    0,    0,    0,  366,  366,    0,
  891,    0,  366,    0,    0,  366,    0,  366,    0,  366,
  366,  366,  366,    0,    0,    0,    0,  366,    0,    0,
    0,  366,    0,    0,    0,  366,    0,    0,    0,    0,
    0,    0,    0,  366,    0,    0,  366,    0,  366,  366,
    0,    0,    0,  366,  366,    0,  366,  366,  366,  366,
  366,  366,  366,  366,  366,  366,  366,  366,    0,    0,
    0,    0,  366,  366,    0,    0,    0,    0,  366,  366,
  366,  366,  366,  366,  366,  366,  366,  366,    0,  366,
  366,    0,    0,  366,  366,  366,  366,  366,    0,    0,
  366,  366,    0,    0,    0,  366,  366,  366,  366,  366,
  366,  366,  366,    0,    0,    0,  851,    0,    0,    0,
    0,  851,  851,    0,  366,    0,    0,  366,    0,  366,
    0,  366,    0,    0,  366,    0,    0,    0,  396,    0,
  366,    0,    0,    0,  851,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,  851,  851,    0,
    0,    0,  851,    0,    0,  851,    0,  851,    0,  851,
  851,  851,  851,    0,    0,    0,    0,  851,    0,    0,
    0,  851,    0,    0,    0,  851,    0,    0,    0,    0,
    0,    0,    0,  851,    0,    0,  851,    0,  851,  851,
    0,    0,    0,  851,  851,    0,  851,  851,  851,  851,
  851,  851,  851,  851,  851,  851,  851,    0,    0,    0,
    0,    0,  851,  851,    0,    0,    0,    0,  851,  851,
  851,  851,  851,  851,    0,  851,  851,  851,    0,  851,
  851,    0,    0,  851,  851,  851,  851,  356,    0,    0,
  851,  851,  356,  356,    0,  851,  851,  851,  851,  851,
  851,  851,  851,  401,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,  851,  356,    0,  851,    0,  851,
    0,  851,    0,    0,  851,    0,    0,    0,  356,  356,
  851,    0,    0,  356,    0,    0,  356,    0,  356,    0,
  356,  356,  356,  356,    0,    0,    0,    0,  356,    0,
    0,    0,  356,    0,    0,    0,  356,    0,    0,    0,
    0,    0,    0,    0,  356,    0,    0,  356,    0,  356,
  356,    0,    0,    0,  356,  356,    0,  356,  356,  356,
  356,  356,  356,  356,  356,  356,  356,  356,    0,    0,
    0,    0,    0,  356,  356,    0,    0,    0,    0,  356,
  356,  356,  356,  356,  356,    0,  356,  356,  356,    0,
  356,  356,    0,    0,  356,  356,  356,  356,    0,    0,
    0,  356,  356,    0,    0,    0,  356,  356,  356,  356,
  356,  356,  356,  356,  396,   38,    0,    0,    0,  396,
  396,    0,    0,    0,    0,  356,    0,    0,  356,    0,
  356,    0,  356,    0,    0,  356,    0,    0,    0,    0,
    0,  356,  396,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,  396,  396,    0,   24,    0,
  396,    0,    0,  396,    0,  396,    0,  396,  396,  396,
  396,    0,    0,    0,    0,  396,    0,    0,    0,  396,
    0,    0,    0,  396,    0,    0,    0,    0,    0,    0,
    0,  396,    0,    0,  396,    0,  396,  396,    0,    0,
    0,  396,  396,    0,  396,  396,  396,  396,  396,  396,
  396,  396,  396,  396,  396,    0,    0,    0,    0,    0,
  396,  396,    0,    0,    0,    0,  396,  396,    0,  396,
  396,  396,    0,  396,  396,  396,    0,  396,  396,  401,
   36,  396,  396,  396,  396,  401,  478,    0,  396,  396,
    0,    0,    0,  396,  396,  396,  396,  396,  396,  396,
  396,    0,    0,    0,    0,    0,    0,    0,    0,  479,
    0,    0,  396,    0,    0,  396,    0,  396,    0,    0,
    0,  401,  480,   25,    0,    0,    0,  482,  396,    0,
    0,    0,  483,    0,  484,  485,  486,  487,    0,    0,
    0,    0,  488,    0,    0,    0,  489,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,  490,    0,
    0,  491,  401,  492,    0,    0,  401,  401,    0,  401,
  401,  401,  401,  401,  401,  401,  401,  401,  401,  401,
    0,    0,    0,    0,    0,    0,  401,  493,    0,    0,
    0,  401,  401,  401,  401,  401,  401,    0,  401,  401,
  401,    0,  401,  401,    0,   35,  401,  401,  401,  401,
    0,    0,   38,  401,  401,    0,   38,    0,  401,  401,
  401,  401,  401,  401,  401,  401,    0,   38,    0,    0,
    0,    0,   38,    0,    0,    0,   38,  401,    0,   38,
  401,    0,  401,    0,  401,    0,    0,  401, 1069,    0,
    0,   38,   38,  401,    0,   24,   38,   38,    0,   24,
    0,    0,   38,    0,   38,   38,   38,   38,    0,    0,
   24,    0,   38,    0,    0,   24,   38,    0,   38,   24,
    0,    0,   24,    0,    0,    0,    0,    0,   38,    0,
   38,   38,    0,   38,   24,   24,    0,   38,    0,   24,
   24,    0,    0,    0,    0,   24,    0,   24,   24,   24,
   24,   51,    0,    0,    0,   24,    0,   38,    0,   24,
    0,   24,    0,    0,   38,   38,    0,    0,    0,    0,
    0,   24,    0,    0,   24,    0,   24,   36,    0,    0,
   24,   36,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,   36,    0,    7,    0,    0,   36,    0,    0,
   24,   36,    0,    0,   36,    0,   21,   24,   24,    0,
    0,    0,    0,    0,    0,    0,   36,   36,    0,    0,
   25,   36,   36,    0,   25,    0,    0,   36,    0,   36,
   36,   36,   36,    0,    0,   25,    0,   36,    0,    0,
   25,   36,    0,   36,   25,    0,    0,   25,    0,    0,
    0,    0,    0,   36,    0,    0,   36, 1070,   36,   25,
   25,    0,   36,    0,   25,   25,    0,    0,    0,    0,
   25,    0,   25,   25,   25,   25,    0,    0,    0,    0,
   25,    0,   36,    0,   25,    0,   25,    0,    0,   36,
   36,    0,    0,    0,    0,    0,   25,    0,    0,   25,
   52,   25,   35,    0,    0,   25,   35,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,   35,    0,    0,
    0,    0,   35,    0,    0,   25,   35,    0,    0,   35,
    0,    0,   25,   25,    0,    0,    0,    0,    0,    0,
    0,   35,   35,    0,    0, 1069,   35,   35,    0,   51,
    0,    0,   35,    0,   35,   35,   35,   35,    0,    0,
   51,    0,   35,    0,    0,   51,   35,    0,   35,   51,
    0,    0,   51,    0,    0,    0,    0,    0,   35,    0,
    0,   35,    0,   35,   51,   51,    0,   35,    0,   51,
   51,    0,    0,    0,    0,   51,    0,   51,   51,   51,
   51,    0,    0,    0,    0,   51,    0,   35,   51,   51,
    0,   51,   51,    0,    0,   35,    0,    0,    0,    0,
    0,   51,    0,   51,   51,    0,   51,    0,   51,    0,
   51,    0,   51,    0,    0,   51,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,   51,   51,    0,
   51,    7,   51,   51,    0,   52,    0,    0,   51,    0,
   51,   51,   51,   51,    0,    0,   52,    0,   51,    0,
    0,   52,   51,    0,   51,   52,    0,    0,   52,    0,
    0,    0,    0,    0,   51,    0,    0,   51,    0,   51,
   52,   52,    0,   51,    0,   52,   52,    0,    0,    0,
    0,   52,    0,   52,   52,   52,   52,    0,    0,    0,
    0,   52,    0,   51, 1070,   52,    0,   52,   51,    0,
    0,    0,    0,    0,    0,    0,    0,   52,    0,   51,
   52,    0,   52,    0,   51,    0,   52,    0,   51,    0,
    0,   51,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,   51,   51,    0,   52,   52,   51,   51,
    0,   52,    0,    0,   51,    0,   51,   51,   51,   51,
    0,    0,   52,    0,   51,    0,    0,   52,   51,    0,
   51,   52,    0,    0,   52,    0,    0,    0,    0,    0,
   51,    0,    0,   51,    0,   51,   52,   52,    0,   51,
    0,   52,   52,    0,    0,    0,    0,   52,    0,   52,
   52,   52,   52,    0,    0,    0,    0,   52,    0,   51,
    0,   52,    0,   52,    0,    0,    0,    0,    0,    0,
    0,    0,    0,   52,    0,   56,   52,    0,   52,    0,
    0,    0,   52,   57,   24,   58,   25,    0,    0,   26,
   59,    0,   60,   61,   27,   62,   63,   64,   28,    0,
    0,    0,   52,    0,   65,    0,   66,   30,   67,   68,
   69,   70,    0,    0,   32,    0,    0,    0,   71,   33,
    0,   72,   73,   34,    0,    0,    0,    0,    0,    0,
    0,    0,    0,   74,    0,   36,    0,   37,   75,    0,
    0,   38,    0,   76,   77,   78,   79,   80,   81,   39,
   40,   82,   83,   41,   84,    0,   85,    0,    0,   86,
   87,    0,    0,   88,   89,  854,    0,    0,    0,    0,
    0,  854,    0,    0,    0,    0,    0,    0,   90,   91,
   92,   93,   94,    0,    0,    0,    0,   95,    0,    0,
    0,   96,    0,    0,    0,    0,   97,   98,   99,  100,
  101,    0,    0,    0,  102,    0,  103,  854,    0,    0,
    0,    0,  104,  105,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,  106,  107,  108,
  109,    0,    0,    0,    0,    0,    0,    0,  854,  200,
    0,    0,    0,  854,  649,  854,  854,  854,  854,  854,
  854,  854,  854,  854,  854,  854,  854,    0,    0,    0,
    0,    0,  854,  854,    0,    0,    0,  854,  854,  854,
  854,  854,  854,  854,  854,  854,  854,    0,  854,  854,
    0,  854,  854,  854,  854,  854,  854,  854,  854,  854,
  854,    0,  854,  854,  854,  854,  854,  854,  854,  854,
  854,  854,  854,  854,  854,  854,  854,  854,  854,  854,
  854,  854,  854,  854,  854,    0,  854,    0,  854,    0,
  854,  854,  854,  854,    0,    0,    0,  649,    0,  854,
    0,    0,  649,    0,  649,  649,  649,  649,  649,  649,
  649,  649,  649,  649,  649,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,  649,  854,  649,
    0,  649,    0,  649,  649,  649,    0,    0,    0,    0,
    0,  649,  649,  649,  649,    0,    0,    0,  649,  649,
    0,    0,    0,  649,  649,  649,  649,  649,  649,  649,
  649,    0,    0,    0,    0,    0,    0,    0,    0,  854,
    0,    0,  649,    0,  854,    0,  854,  854,  854,  854,
  854,  854,  854,  854,  854,  854,  854,  854,  649,    0,
    0,    0,    0,  854,  854,    0,    0,    0,  854,  854,
  854,  854,  854,  854,  854,  854,  854,  854,    0,  854,
  854,    0,  854,  854,  854,  854,  854,  854,  854,  854,
  854,  854,    0,  854,  854,  854,  854,  854,  854,  854,
  854,  854,  854,  854,  854,  854,  854,  854,  854,  854,
  854,  854,  854,  854,  854,  854,  366,  563,    0,  854,
    0,  854,  366,    0,  854,    0,   24,    0,   25,    0,
  854,   26,    0,    0,    0,    0,   27,    0,    0,    0,
   28,    0,    0,    0,    0,    0,    0,    0,    0,   30,
    0,    0,    0,    0,    0,    0,   32,    0,  366,    0,
    0,   33,    0,    0,    0,   34,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,   36,    0,   37,
    0,    0,    0,   38,    0,    0,    0,    0,    0,    0,
    0,   39,   40,    0,    0,   41,    0,    0,  564,  366,
    0,    0,    0,    0,  366,    0,  366,  366,  366,  366,
  366,  366,  366,  366,  366,  366,  366,  366,    0,    0,
    0,    0,    0,  366,  366,    0,    0,    0,  366,  366,
  366,  366,  366,  366,  366,  366,  366,  366,    0,  366,
  366,    0,  366,  366,  366,  366,  366,  366,  366,  366,
  366,  366,    0,  366,  366,  366,  366,  366,  366,  366,
  366,  366,  366,  366,  366,  366,  366,  366,  366,  366,
  366,  366,  366,  366,  366,    0,  366,  602,    0,  366,
  381,  366,  366,    0,  366,    0,   24,    0,   25,    0,
  366,   26,    0,    0,    0,    0,   27,    0,    0,    0,
   28,    0,    0,    0,    0,    0,    0,    0,    0,   30,
    0,    0,    0,    0,    0,    0,   32,    0,  366,    0,
    0,   33,    0,    0,    0,   34,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,   36,    0,   37,
    0,    0,    0,   38,    0,    0,    0,    0,    0,    0,
    0,   39,   40,    0,    0,   41,    0,    0,   85,  366,
    0,    0,    0,    0,  366,    0,  366,  366,  366,  366,
  366,  366,  366,  366,  366,  366,  366,  366,    0,    0,
    0,    0,    0,  366,  366,    0,    0,    0,    0,  366,
  366,  366,  366,  366,  366,  366,  366,  366,    0,  366,
  366,    0,  366,  366,  366,  366,  366,  366,  366,  366,
  366,  366,    0,  366,  366,  366,  366,  366,  366,  366,
  366,  366,  366,  366,  366,  366,  366,  366,  366,  366,
  366,  366,  366,  366,  366,    0,  554,  665,    0,  366,
  381,  366,  554,    0,  366,    0,   24,    0,   25,    0,
  366,   26,    0,    0,    0,    0,   27,    0,    0,    0,
   28,    0,    0,    0,    0,    0,    0,    0,    0,   30,
    0,    0,    0,    0,    0,    0,   32,    0,  554,    0,
    0,   33,    0,    0,    0,   34,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,   36,    0,   37,
    0,    0,    0,   38,    0,    0,    0,    0,    0,    0,
    0,   39,   40,    0,    0,   41,    0,    0,   85,  554,
    0,    0,    0,    0,  554,    0,  554,  554,  554,  554,
  554,  554,  554,  554,  554,  554,  554,    0,    0,    0,
    0,    0,    0,    0,  554,    0,    0,    0,  554,  554,
  554,  554,  554,  554,  554,  554,  554,  554,    0,  554,
  554,    0,  554,  554,  554,  554,  554,  554,  554,  554,
  554,  554,    0,  554,  554,  554,  554,  554,  554,  554,
  554,  554,  554,  554,  554,  554,  554,  554,  554,  554,
  554,  554,  554,  554,  554,  999,  550,    0,    0,    0,
  381,  554,  550,    0,   24,    0,   25,    0,    0,   26,
  554,    0,    0,    0,   27,    0,    0,    0,   28,    0,
    0,    0,    0,    0,    0,    0,    0,   30,    0,    0,
    0,    0,    0,    0,   32,    0,    0,    0,  550,   33,
    0,    0,    0,   34,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,   36,    0,   37,    0,    0,
    0,   38,    0,    0,    0,    0,    0,    0,    0,   39,
   40,    0,    0,   41,    0,    0,   85,    0,    0,  550,
    0,    0,    0,    0,  550,    0,  550,  550,  550,  550,
  550,  550,  550,  550,  550,  550,  550,    0,    0,    0,
    0,    0,    0,    0,  550,    0,    0,    0,  550,  550,
    0,  550,  550,  550,  550,  550,  550,  550,    0,  550,
  550,    0,  550,  550,  550,  550,  550,  550,  550,  550,
  550,  550,    0,  550,  550,  550,  550,  550,  550,  550,
  550,  550,  550,  550,  550,  550,  550,  550,  550,  550,
  550,  550,  550,  550,  550,    0,  558, 1165,  381,    0,
    0,  550,  558,    0,  550,    0,   24,    0,   25,    0,
  550,   26,    0,    0,    0,    0,   27,    0,    0,    0,
   28,    0,    0,    0,    0,    0,    0,    0,    0,   30,
    0,    0,    0,    0,    0,    0,   32,    0,  558,    0,
    0,   33,    0,    0,    0,   34,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,   36,    0,   37,
    0,    0,    0,   38,    0,    0,    0,    0,    0,    0,
    0,   39,   40,    0,    0,   41,    0,    0,   85,  558,
    0,    0,    0,    0,  558,    0,  558,  558,  558,  558,
  558,  558,  558,  558,  558,  558,  558,    0,    0,    0,
    0,    0,    0,    0,  558,    0,    0,    0,    0,  558,
    0,  558,  558,  558,  558,  558,  558,  558,    0,  558,
  558,    0,  558,  558,  558,  558,  558,  558,  558,  558,
  558,  558,    0,  558,  558,  558,  558,  558,  558,  558,
  558,  558,  558,  558,  558,  558,  558,  558,  558,  558,
  558,  558,  558,  558,  558,    0,  366,    0,    0,  388,
  381,  558,  366,    0,  558,    0,    0,    0,    0,    0,
  558,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,  388,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,  388,    0,    0,  366,    0,
  388,    0,    0,  255,    0,  388,    0,  388,  388,  388,
  388,    0,    0,    0,    0,  388,    0,    0,    0,  388,
    0,    0,    0,  388,    0,    0,    0,    0,    0,    0,
    0,  388,    0,    0,  388,    0,  388,    0,    0,  366,
    0,    0,    0,    0,  366,    0,  366,  366,  366,  366,
  366,  366,  366,  366,  366,  366,  366,    0,    0,    0,
  388,    0,    0,    0,  366,    0,    0,  388,    0,  366,
    0,  366,  366,  366,  366,  366,  366,  366,    0,  366,
  366,    0,  366,  366,  366,  366,  366,  366,  366,  366,
  366,  366,    0,  366,  366,  366,  366,  366,  366,  366,
  366,  366,  366,  366,  366,  366,  366,  366,  366,  366,
  366,  366,  366,  366,  366,    0,  473,  388,    0,  387,
    0,  366,  473,    0,  366,    0,    0,    0,    0,    0,
  366,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,  387,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,  387,    0,    0,  473,    0,
  387,    0,    0,  254,    0,  387,    0,  387,  387,  387,
  387,    0,    0,    0,    0,  387,    0,    0,    0,  387,
    0,    0,    0,  387,    0,    0,    0,    0,    0,    0,
    0,  387,    0,    0,  387,    0,  387,    0,    0,  473,
    0,    0,    0,    0,  473,    0,  473,  473,  473,  473,
  473,  473,  473,  473,  473,  473,  473,    0,    0,    0,
  387,    0,    0,    0,  473,    0,    0,  387,    0,  473,
    0,  473,  473,  473,  473,  473,  473,  473,    0,  473,
  473,    0,  473,  473,  473,  473,  473,  473,  473,  473,
  473,  473,    0,  473,  473,  473,  473,  473,  473,  473,
  473,  473,  473,  473,  473,  473,  473,  473,  473,  473,
  473,  473,  473,  473,  473,    0,  366,  387,    0,  478,
    0,  473,  366,    0,  473,    0,  854,    0,    0,    0,
  473,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,  479,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,  480,    0,    0,  366,    0,
  482,    0,    0,    0,    0,  483,    0,  484,  485,  486,
  487,    0,    0,    0,  854,  488,    0,    0,    0,  489,
    0,    0,    0, 1473,    0,    0,    0,    0,    0,    0,
    0,  490,    0,    0,  491,    0,  492,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
  585,    0,    0,    0,    0,    0,  585,  366,    0,    0,
  493,    0,    0,  366,  366,    0,    0, 1474,    0,  366,
  366,  366,  366,  366,  366,  366,  854,  366,    0,  366,
  366,    0,  366,  366,  366,  366,  366,  366,  366,  366,
  366,  366,  585,  366,  366,  366,  366,  366,  366,  366,
  366,  366,  366,  366,  366,  366,  366,  366,  366,  366,
  366,  366,  366,  366,  366,    0,    0, 1475,    0,  366,
    0,  366,    0,    0,  366,    0,    0,    0,    0,    0,
  366,    0,    0,  585,    0,  650,    0,    0,  585,    0,
  585,  585,  585,  585,  585,  585,  585,  585,  585,  585,
  585,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,  585,    0,  585,    0,  585,    0,  585,
  585,  585,    0,  585,  585,    0,  585,  585,  585,  585,
  585,  585,  585,  585,  585,  585,    0,    0,    0,  585,
  585,  585,  585,  585,  585,  585,  585,  585,  585,  585,
  585,  585,  585,  585,  585,  585,  585,  356,  585,    0,
    0,    0,    0,  356,    0,    0,    0,    0,  650,    0,
    0,    0,    0,  650,  585,  650,  650,  650,  650,  650,
  650,  650,  650,  650,  650,  650,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,  650,  356,
  650,    0,  650,    0,  650,  650,  650,    0,    0,    0,
    0,    0,  650,  650,  650,  650,  424,    0,    0,  650,
  650,    0,  424,    0,  650,  650,  650,  650,  650,  650,
  650,  650,    0,    0,    0,    0,    0,    0,    0,    0,
  356,    0,    0,  650,    0,  356,    0,  356,  356,  356,
  356,  356,  356,  356,  356,  356,  356,  356,  424,  650,
    0,    0,    0,    0,  356,  424,    0,    0,    0,    0,
  356,  356,  356,    0,  356,  424,  356,  356,  356,    0,
  356,  356,    0,    0,  356,  356,  356,  356,    0,    0,
    0,  356,  356,    0,  424,  424,  356,  356,  356,  356,
  356,  356,  356,  356,    0,  627,    0,    0,    0,    0,
    0,  627,    0,    0,  424,  356,    0,    0,    0,    0,
  356,    0,  424,  356,  424,  424,    0,    0,    0,  424,
  356,  356,  424,  424,  424,  424,    0,  424,    0,  424,
  424,    0,  424,  424,  424,  424,  424,  627,  424,  424,
  424,  424,    0,  424,  424,  424,  424,  424,  424,  424,
  424,  424,  424,  424,  424,  424,  424,  424,  424,  424,
  424,  424,  424,  424,  424,    0,    0,    0,    0,  356,
    0,  424,    0,    0,  424,    0,    0,    0,  627,    0,
  424,    0,    0,  627,    0,  627,  627,  627,  627,  627,
  627,  627,  627,  627,  627,  627,    0,    0,    0,    0,
  631,    0,    0,  425,    0,    0,  631,    0,  627,    0,
  627,  425,  627,  425,  627,  627,  627,    0,  627,  627,
    0,    0,  627,  627,  627,  627,    0,    0,    0,  627,
  627,    0,  425,  425,  627,  627,  627,  627,  627,  627,
  627,  627,  631,    0,    0,    0,    0,    0,    0,    0,
    0,    0,  425,  627,    0,    0,    0,    0,    0,    0,
  425,    0,    0,  425,    0,    0,    0,    0,    0,  627,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,  631,    0,    0,    0,    0,  631,    0,
  631,  631,  631,  631,  631,  631,  631,  631,  631,  631,
  631,    0,    0,    0,    0,  630,    0,    0,  428,    0,
    0,  630,    0,  631,    0,  631,  428,  631,  428,  631,
  631,  631,    0,  631,  631,    0,    0,  631,  631,  631,
  631,    0,    0,    0,  631,  631,    0,  428,  428,  631,
  631,  631,  631,  631,  631,  631,  631,  630,    0,    0,
    0,    0,    0,    0,    0,    0,    0,  428,  631,    0,
    0,    0,    0,    0,    0,  428,    0,    0,  428,    0,
    0,    0,    0,    0,  631,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,  630,    0,
    0,    0,    0,  630,    0,  630,  630,  630,  630,  630,
  630,  630,  630,  630,  630,  630,    0,    0,    0,    0,
  356,    0,    0,  438,    0,    0,  356,    0,  630,    0,
  630,  438,  630,  438,  630,  630,  630,    0,  630,  630,
    0,    0,  630,  630,  630,  630,    0,    0,    0,  630,
  630,    0,  438,  438,  630,  630,  630,  630,  630,  630,
  630,  630,  356,    0,    0,    0,    0,    0,    0,    0,
    0,    0,  438,  630,    0,    0,    0,    0,    0,    0,
  438,    0,    0,  438,    0,    0,    0,    0,    0,  630,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,  356,    0,    0,    0,    0,  356,    0,
  356,  356,  356,  356,  356,  356,  356,  356,  356,  356,
  356,    0,    0,    0,    0,  607,    0,  356,    0,    0,
    0,  607,    0,  356,  356,  356,    0,  356,    0,  356,
  356,  356,    0,  356,  356,    0,    0,  356,  356,  356,
  356,    0,    0,    0,  356,  356,    0,    0,    0,  356,
  356,  356,  356,  356,  356,  356,  356,  607,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,  356,    0,
    0,    0,    0,  356,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,  356,    0,    0,  845,    0,    0,
    0,    0,    0,  845,    0,    0,    0,    0,  607,    0,
    0,    0,    0,  607,    0,  607,  607,  607,  607,  607,
  607,  607,  607,  607,  607,  607,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,  607,  845,
  607,    0,  607,    0,  607,  607,  607,    0,  607,  607,
    0,    0,  607,  607,  607,  607,  607,  607,  607,  607,
  607,    0,    0,    0,  607,  607,  607,  607,  607,  607,
  607,  607,    0,    0,    0,    0,    0,    0,    0,    0,
  845,    0,    0,  607,    0,  845,    0,  845,  845,  845,
  845,  845,  845,  845,  845,  845,  845,  845,    0,  607,
    0,    0,  614,    0,  845,    0,    0,    0,  614,    0,
  845,    0,  845,    0,  845,    0,  845,  845,  845,    0,
  845,  845,    0,    0,  845,  845,  845,  845,    0,    0,
    0,  845,  845,    0,    0,    0,  845,  845,  845,  845,
  845,  845,  845,  845,  614,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,  845,    0,    0,    0,    0,
  845,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,  845,    0,    0,  975,    0,    0,    0,    0,    0,
  975,    0,    0,    0,    0,  614,    0,    0,    0,    0,
  614,    0,  614,  614,  614,  614,  614,  614,  614,  614,
  614,  614,  614,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,  614,  975,  614,    0,  614,
    0,  614,  614,  614,    0,  614,  614,    0,    0,  614,
  614,  614,  614,    0,    0,    0,  614,  614,    0,    0,
    0,  614,  614,  614,  614,  614,  614,  614,  614,    0,
    0,    0,    0,    0,    0,    0,    0,  975,    0,    0,
  614,    0,  975,    0,  975,  975,  975,  975,  975,  975,
  975,  975,  975,  975,  975,    0,  614,    0,    0,  615,
    0,    0,    0,    0,    0,  615,    0,  975,    0,  975,
    0,  975,    0,  975,  975,  975,    0,  975,  975,    0,
    0,  975,  975,  975,  975,    0,    0,    0,  975,  975,
    0,    0,    0,  975,  975,  975,  975,  975,  975,  975,
  975,  615,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,  975,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,  975,    0,
    0,  616,    0,    0,    0,    0,    0,  616,    0,    0,
    0,    0,  615,    0,    0,    0,    0,  615,    0,  615,
  615,  615,  615,  615,  615,  615,  615,  615,  615,  615,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,  615,  616,  615,    0,  615,    0,  615,  615,
  615,    0,  615,  615,    0,    0,  615,  615,  615,  615,
    0,    0,    0,  615,  615,    0,    0,    0,  615,  615,
  615,  615,  615,  615,  615,  615,    0,    0,    0,    0,
    0,    0,    0,    0,  616,    0,  651,  615,    0,  616,
    0,  616,  616,  616,  616,  616,  616,  616,  616,  616,
  616,  616,    0,  615,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,  616,    0,  616,    0,  616,    0,
  616,  616,  616,    0,  616,  616,    0,    0,  616,  616,
  616,  616,    0,    0,    0,  616,  616,    0,    0,    0,
  616,  616,  616,  616,  616,  616,  616,  616,    0,    0,
    0,    0,    0,    0,    0,  654,    0,    0,    0,  616,
    0,    0,    0,    0,    0,    0,    0,    0,    0,  651,
    0,    0,    0,    0,  651,  616,  651,  651,  651,  651,
  651,  651,  651,  651,  651,  651,  651,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,  651,
    0,  651,    0,  651,    0,  651,  651,  651,    0,    0,
    0,    0,    0,  651,  651,  651,  651,    0,    0,    0,
  651,  651,    0,    0,  655,  651,  651,  651,  651,  651,
  651,  651,  651,    0,    0,    0,    0,    0,  654,    0,
    0,    0,    0,  654,  651,  654,  654,  654,  654,  654,
  654,  654,  654,  654,  654,  654,    0,    0,    0,    0,
  651,    0,    0,    0,    0,    0,    0,    0,  654,    0,
  654,    0,  654,    0,  654,  654,  654,    0,    0,    0,
    0,    0,  654,  654,  654,  654,    0,    0,    0,  654,
  654,    0,    0,  656,    0,    0,  654,  654,  654,  654,
  654,  654,    0,    0,    0,    0,    0,  655,    0,    0,
    0,    0,  655,  654,  655,  655,  655,  655,  655,  655,
  655,  655,  655,  655,  655,    0,    0,    0,    0,  654,
    0,    0,    0,    0,    0,    0,    0,  655,    0,  655,
    0,  655,    0,  655,  655,  655,    0,    0,    0,    0,
    0,  655,  655,  655,  655,    0,    0,    0,  655,  655,
    0,    0,  657,    0,    0,  655,  655,  655,  655,  655,
  655,    0,    0,    0,    0,    0,  656,    0,    0,    0,
    0,  656,  655,  656,  656,  656,  656,  656,  656,  656,
  656,  656,  656,  656,    0,    0,    0,    0,  655,    0,
    0,    0,    0,    0,    0,    0,  656,    0,  656,    0,
  656,    0,  656,  656,  656,    0,    0,    0,    0,    0,
  656,  656,  656,  656,    0,    0,    0,  656,  656,    0,
    0,  658,    0,    0,  656,  656,  656,  656,  656,  656,
    0,    0,    0,    0,    0,  657,    0,    0,    0,    0,
  657,  656,  657,  657,  657,  657,  657,  657,  657,  657,
  657,  657,  657,    0,    0,    0,    0,  656,    0,    0,
    0,    0,    0,    0,    0,  657,    0,  657,    0,  657,
    0,  657,  657,  657,    0,    0,    0,    0,    0,  657,
  657,  657,  657,  366,    0,    0,  657,  657,    0,  366,
    0,    0,    0,  657,  657,  657,  657,  657,  657,    0,
    0,    0,    0,    0,  658,    0,    0,    0,    0,  658,
  657,  658,  658,  658,  658,  658,  658,  658,  658,  658,
  658,  658,    0,    0,    0,  366,  657,    0,    0,    0,
    0,    0,    0,    0,  658,    0,  658,    0,  658,    0,
  658,  658,  658,    0,    0,    0,    0,    0,  658,  658,
  658,  658,    0,    0,    0,  658,  658,    0,    0,    0,
    0,    0,  658,  658,  658,  658,  658,  658,    0,    0,
  663,    0,    0,    0,    0,    0,    0,    0,    0,  658,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,  366,    0,    0,    0,  658,  366,    0,    0,  366,
    0,  366,  366,    0,    0,    0,  366,  366,    0,    0,
  366,  366,  366,  366,  366,  366,  366,  366,  366,    0,
  366,  366,  366,  366,  366,  366,  366,  366,  366,  366,
    0,    0,    0,    0,    0,    0,    0,    0,    0,  664,
  366,  366,    0,    0,    0,    0,    0,    0,  366,    0,
    0,  366,    0,  663,    0,    0,    0,  366,  663,    0,
  663,  663,  663,  663,  663,  663,  663,  663,  663,  663,
  663,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,  663,    0,  663,    0,  663,    0,  663,
  663,  663,    0,    0,    0,    0,    0,    0,    0,  663,
  663,    0,    0,    0,  663,  663,    0,    0,  665,    0,
    0,    0,    0,  663,  663,  663,  663,    0,    0,    0,
    0,    0,  664,    0,    0,    0,    0,  664,  663,  664,
  664,  664,  664,  664,  664,  664,  664,  664,  664,  664,
    0,    0,    0,    0,  663,    0,    0,    0,    0,    0,
    0,    0,  664,    0,  664,    0,  664,    0,  664,  664,
  664,    0,    0,    0,    0,    0,    0,    0,  664,  664,
    0,    0,    0,  664,  664,    0,    0,  668,    0,    0,
    0,    0,  664,  664,  664,  664,    0,    0,    0,    0,
    0,  665,    0,    0,    0,    0,  665,  664,  665,  665,
  665,  665,  665,  665,  665,  665,  665,  665,  665,    0,
    0,    0,    0,  664,    0,    0,    0,    0,    0,    0,
    0,  665,    0,  665,    0,  665,    0,  665,  665,  665,
    0,    0,    0,    0,    0,    0,    0,  665,  665,    0,
    0,    0,  665,  665,    0,    0,  669,    0,    0,    0,
    0,  665,  665,  665,  665,    0,    0,    0,    0,    0,
  668,    0,    0,    0,    0,  668,  665,  668,  668,  668,
  668,  668,  668,  668,  668,  668,  668,  668,    0,    0,
    0,    0,  665,    0,    0,    0,    0,    0,    0,    0,
  668,    0,  668,    0,  668,    0,  668,  668,  668,    0,
    0,    0,    0,    0,    0,    0,  668,  668,    0,    0,
    0,  668,  668,    0,    0,  671,    0,    0,    0,    0,
    0,    0,  668,  668,    0,    0,    0,    0,    0,  669,
    0,    0,    0,    0,  669,  668,  669,  669,  669,  669,
  669,  669,  669,  669,  669,  669,  669,    0,    0,    0,
    0,  668,    0,    0,    0,    0,    0,    0,    0,  669,
    0,  669,    0,  669,    0,  669,  669,  669,    0,    0,
    0,    0,    0,    0,    0,  669,  669,    0,    0,    0,
  669,  669,    0,    0,  672,    0,    0,    0,    0,    0,
    0,  669,  669,    0,    0,    0,    0,    0,  671,    0,
    0,    0,    0,  671,  669,  671,  671,  671,  671,  671,
  671,  671,  671,  671,  671,  671,    0,    0,    0,    0,
  669,    0,    0,    0,    0,    0,    0,    0,  671,    0,
  671,    0,  671,    0,  671,  671,  671,    0,    0,    0,
    0,    0,    0,    0,    0,  671,    0,    0,    0,  671,
  671,    0,    0,  674,    0,    0,    0,    0,    0,    0,
  671,  671,    0,    0,    0,    0,    0,  672,    0,    0,
    0,    0,  672,  671,  672,  672,  672,  672,  672,  672,
  672,  672,  672,  672,  672,    0,    0,    0,    0,  671,
    0,    0,    0,    0,    0,    0,    0,  672,    0,  672,
    0,  672,    0,  672,  672,  672,    0,    0,    0,    0,
    0,    0,    0,    0,  672,    0,    0,    0,  672,  672,
    0,    0,  675,    0,    0,    0,    0,    0,    0,  672,
  672,    0,    0,    0,    0,    0,  674,    0,    0,    0,
    0,  674,  672,  674,  674,  674,  674,  674,  674,  674,
  674,  674,  674,  674,    0,    0,    0,    0,  672,    0,
    0,    0,    0,    0,    0,    0,  674,    0,  674,    0,
  674,    0,  674,  674,  674,    0,    0,    0,    0,    0,
    0,    0,    0,  674,    0,    0,    0,    0,  674,    0,
    0,  677,    0,    0,    0,    0,    0,    0,  674,  674,
    0,    0,    0,    0,    0,  675,    0,    0,    0,    0,
  675,  674,  675,  675,  675,  675,  675,  675,  675,  675,
  675,  675,  675,    0,    0,    0,    0,  674,    0,    0,
    0,    0,    0,    0,    0,  675,    0,  675,    0,  675,
    0,  675,  675,  675,    0,    0,    0,    0,    0,    0,
    0,    0,  675,    0,    0,    0,    0,  675,    0,    0,
  678,    0,    0,    0,    0,    0,    0,  675,  675,    0,
    0,    0,    0,    0,  677,    0,    0,    0,    0,  677,
  675,  677,  677,  677,  677,  677,  677,  677,  677,  677,
  677,  677,    0,    0,    0,    0,  675,    0,    0,    0,
    0,    0,    0,    0,  677,    0,  677,    0,  677,    0,
  677,  677,  677,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,  677,    0,    0,  680,
    0,    0,    0,    0,    0,    0,  677,  677,    0,    0,
    0,    0,    0,  678,    0,    0,    0,    0,  678,  677,
  678,  678,  678,  678,  678,  678,  678,  678,  678,  678,
  678,    0,    0,    0,    0,  677,    0,    0,    0,    0,
    0,    0,    0,  678,    0,  678,    0,  678,    0,  678,
  678,  678,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,  678,    0,    0,  681,    0,
    0,    0,    0,    0,    0,  678,  678,    0,    0,    0,
    0,    0,  680,    0,    0,    0,    0,  680,  678,  680,
  680,  680,  680,  680,  680,  680,  680,  680,  680,  680,
    0,    0,    0,    0,  678,    0,    0,    0,    0,    0,
    0,    0,  680,    0,  680,    0,  680,    0,  680,  680,
  680,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,  680,    0,    0,  685,    0,    0,
    0,    0,    0,    0,    0,  680,    0,    0,    0,    0,
    0,  681,    0,    0,    0,    0,  681,  680,  681,  681,
  681,  681,  681,  681,  681,  681,  681,  681,  681,    0,
    0,    0, 1061,  680,    0,    0,    0,    0,    0,    0,
    0,  681,    0,  681,    0,  681,    0,  681,  681,  681,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,  681,  205,    0,    0,    0,    0,    0,
    0,    0,    0,    0,  681,    0,    0,  206,    0,    0,
  685,    0,    0,    0,    0,  685,  681,  685,  685,  685,
  685,  685,  685,  685,  685,  685,  685,  685,    0,  424,
    0,    0,  681,    0,    0,  424,  207,    0,    0,    0,
  685,    0,  685,    0,  685, 1061,  685,  685,  685,    0,
 1061,    0, 1061, 1061, 1061, 1061, 1061, 1061, 1061, 1061,
 1061, 1061,    0,    0,    0,    0,    0,    0,    0,    0,
    0,  424,    0,    0,    0, 1061,    0, 1061,    0, 1061,
    0, 1061, 1061, 1061,    0,    0,    0,  208,  209,  210,
  211,    0,  212,  213,  214,  215,  216,  217,  218,  219,
    0,  685,  220,  221,  222,  223,  224,  225,  226,  227,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 1061,  424,    0,    0,
    0,    0,    0,    0,    0,    0,    0,  424,    0,  424,
    0,    0,  424,  424,    0,  424,  424,  424,  424,  424,
  424,  424,  424,  424,  424,    0,  424,  424,  424,  424,
  424,  424,  424,  424,  424,  424,  424,  424,  424,  424,
  424,  424,  424,  424,  424,  424,  424,  424,    0,    0,
  592,    0,    0,    0,  424,    0,    0,  424,   57,   24,
   58,   25, 1194,    0,   26,   59,    0,   60,   61,   27,
   62,   63,   64,   28,    0,    0,    0,    0,    0,   65,
    0,   66,   30,   67,   68,   69,   70,    0,    0,   32,
    0,    0,    0,   71,   33,    0,   72,   73,   34,    0,
    0,    0,    0,    0,    0,    0,    0,    0,   74,    0,
   36,    0,   37,   75,    0,    0,   38,    0,   76,   77,
   78,   79,   80,   81,   39,   40,   82,   83,   41,   84,
    0,   85,    0,    0,   86,   87,    0,    0,   88,   89,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,   90,   91,   92,   93,   94,    0,    0,
    0,    0,   95,    0,    0,    0,   96,    0,    0,    0,
    0,   97,   98,   99,  100,  101,    0,    0,    0,  102,
    0,  103,    0,    0,    0,    0,    0,  104,  105,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
   56,    0,  106,  107,  108,  109,    0, 1195,   57,   24,
   58,   25,    0,    0,   26,   59,    0,   60,   61,   27,
   62,   63,   64,   28,    0,    0,    0,    0,    0,   65,
    0,   66,   30,   67,   68,   69,   70,    0,    0,   32,
    0,    0,    0,   71,   33,    0,   72,   73,   34,    0,
    0,    0,    0,    0,    0,    0,    0,    0,   74,    0,
   36,    0,   37,   75,    0,    0,   38,    0,   76,   77,
   78,   79,   80,   81,   39,   40,   82,   83,   41,   84,
    0,   85,    0,    0,   86,   87,    0,    0,   88,   89,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,   90,   91,   92,   93,   94,    0,    0,
    0,    0,   95,    0,    0,    0,   96,    0,    0,    0,
    0,   97,   98,   99,  100,  101,    0,    0,    0,  102,
    0,  103,    0,    0,    0,    0,    0,  104,  105,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,  273,    0,
    0,    0,  106,  107,  108,  109,   57,   24,   58,   25,
    0,    0,   26,   59,    0,   60,   61,   27,   62,   63,
   64,   28,    0,    0,    0,    0,    0,   65,    0,   66,
   30,   67,   68,   69,   70,    0,    0,   32,    0,    0,
    0,   71,   33,    0,   72,   73,   34,    0,    0,    0,
    0,    0,    0,    0,    0,    0,   74,    0,   36,    0,
   37,   75,    0,    0,   38,    0,   76,   77,   78,   79,
   80,   81,   39,   40,   82,   83,   41,   84,    0,   85,
    0,    0,   86,   87,    0,    0,   88,   89,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,   90,   91,   92,   93,   94,    0,    0,    0,    0,
   95,    0,    0,    0,   96,    0,    0,    0,    0,   97,
   98,   99,  100,  101,    0,    0,    0,  102,    0,  103,
    0,    0,    0,    0,    0,  104,  105,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,  592,    0,    0,    0,
  106,  107,  108,  109,   57,   24,   58,   25,    0,    0,
   26,   59,    0,   60,   61,   27,   62,   63,   64,   28,
    0,    0,    0,    0,    0,   65,    0,   66,   30,   67,
   68,   69,   70,    0,    0,   32,    0,    0,    0,   71,
   33,    0,   72,   73,   34,    0,    0,    0,    0,    0,
    0,    0,    0,    0,   74,    0,   36,    0,   37,   75,
    0,    0,   38,    0,   76,   77,   78,   79,   80,   81,
   39,   40,   82,   83,   41,   84,    0,   85,    0,    0,
   86,   87,    0,    0,   88,   89,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,   90,
   91,   92,   93,   94,    0,    0,    0,    0,   95,    0,
    0,    0,   96,    0,    0,    0,    0,   97,   98,   99,
  100,  101,    0,    0,    0,  102,    0,  103,    0,    0,
    0,    0,    0,  104,  105,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 1066,    0,    0,    0,  106,  107,
  108,  109, 1066, 1066, 1066, 1066,    0,    0, 1066, 1066,
    0, 1066, 1066, 1066, 1066, 1066, 1066, 1066,    0,    0,
    0,    0,    0, 1066,    0, 1066, 1066, 1066, 1066, 1066,
 1066,    0,    0, 1066,    0,    0,    0, 1066, 1066,    0,
 1066, 1066, 1066,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 1066,    0, 1066,    0, 1066, 1066,    0,    0,
 1066,    0, 1066, 1066, 1066, 1066, 1066, 1066, 1066, 1066,
 1066, 1066, 1066, 1066,    0, 1066,    0,    0, 1066, 1066,
    0,    0, 1066, 1066,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 1066, 1066, 1066,
 1066, 1066,    0,    0,    0,    0, 1066,    0,    0,    0,
 1066,    0,    0,    0,    0, 1066, 1066, 1066, 1066, 1066,
    0,    0,    0, 1066,    0, 1066,    0,    0,    0,    0,
    0, 1066, 1066,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,  635,    0,    0,    0, 1066, 1066, 1066, 1066,
   57,   24,    0,   25,    0,    0,   26,  260,    0,    0,
    0,   27,   62,   63,    0,   28,    0,    0,  190,    0,
  190,   65,    0,  190,   30,    0,    0,    0,  190,    0,
    0,   32,  190,    0,    0,    0,   33,    0,   72,   73,
   34,  190,  636,    0,    0,    0,    0,    0,  190,  637,
    0,    0,   36,  190,   37,   75,    0,  190,   38,    0,
    0,   77,    0,   79,    0,   81,   39,   40,  261,  190,
   41,  190,    0,    0,    0,  190,    0,  638,    0,    0,
   88,   89,    0,  190,  190,    0,    0,  190,    0,    0,
  190,    0,    0,    0,    0,   90,   91,   92,   93,   94,
    0,    0,    0,    0,    0,    0,    0,    0,   96,    0,
    0,  639,    0,    0,   98,   99,  100,  101,    0,    0,
    0,  102,    0,  103,    0,    0, 1091,    0,    0,  104,
  105,    0,    0,    0,    0,    0,    0,   57,   24,    0,
   25,    0,    0,   26,  260,    0,    0,    0,   27,   62,
   63,    0,   28,    0,  106,  107,  108,  109,   65,    0,
    0,   30,    0,    0,    0,    0,    0,    0,   32,    0,
    0,    0,  190,   33,    0,   72,   73,   34,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,   36,
    0,   37,   75,    0,    0,   38,    0,    0,   77,    0,
   79,    0,   81,   39,   40,  261,    0,   41,    0,    0,
    0,    0,    0,    0,   87,    0,    0,   88,   89,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,   90,   91,   92,   93,  807,    0,    0,    0,
    0,  808, 1110,    0,    0,   96,    0,    0,    0,    0,
    0,   98,   99,  100,  101,    0,    0,    0,  102,    0,
  103,    0,    0,    0,    0,    0,  104,  105,    0,    0,
    0,    0,    0,    0,   57,   24,    0,   25,    0,    0,
   26,  260,    0,    0,    0,   27,   62,   63,    0,   28,
    0,  106,  809,  108,  109,   65,    0,  810,   30,    0,
    0,    0,  811,    0,    0,   32,    0,    0,    0,    0,
   33,    0,   72,   73,   34,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,   36,    0,   37,   75,
    0,    0,   38,    0,    0,   77,    0,   79,    0,   81,
   39,   40,  261,    0,   41,    0,    0,    0,    0,    0,
    0,   87,    0,    0,   88,   89,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,   90,
   91,   92,   93,  807,    0,    0,    0,    0,  808,    0,
    0,    0,   96,    0,    0,    0,    0,    0,   98,   99,
  100,  101,    0,    0,    0,  102,    0,  103,    0,    0,
    0,    0,    0,  104,  105,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,  825,  106,  809,
  108,  109,    0,    0,  810,   57,   24,    0,   25,  811,
    0,   26,  260,    0,    0,    0,   27,   62,   63,    0,
   28,    0,    0,  190,    0,  190,   65,    0,  190,   30,
    0,    0,    0,  190,    0,    0,   32,  190,    0,    0,
    0,   33,    0,   72,   73,   34,  190,    0,    0,    0,
    0,    0,    0,  190,    0,    0,    0,   36,  190,   37,
   75,    0,  190,   38,    0,    0,   77,    0,   79,    0,
   81,   39,   40,  261,  190,   41,  190,    0,    0,    0,
  190,    0,   87,    0,    0,   88,   89,    0,  190,  190,
    0,    0,  190,    0,    0,  190,    0,    0,    0,    0,
   90,   91,   92,   93,  309,    0,    0,    0,    0,  547,
  826,    0,    0,   96,    0,    0,    0,    0,    0,   98,
   99,  100,  101,    0,    0,    0,  102,    0,  103, 1091,
    0,    0,    0,    0,  104,  105,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 1023,    0,    0,    0,  106,
  310,  108,  109,   57,   24,    0,   25,    0,    0,   26,
  260,    0,    0,    0,   27,   62,   63,  190,   28,    0,
    0,   24,    0,   25,   65,    0,   26,   30,    0,    0,
    0,   27,    0,    0,   32,   28,    0,    0,    0,   33,
    0,   72,   73,   34,   30,  636,    0,    0,    0,    0,
    0,   32,  637,    0,    0,   36,   33,   37,   75,    0,
   34,   38,    0,    0,   77,    0,   79,    0,   81,   39,
   40,  261,   36,   41,   37,    0,    0,    0,   38,    0,
  638,    0,    0,   88,   89,    0,   39,   40,    0,    0,
   41,    0,    0,   85,    0,    0,    0,    0,   90,   91,
   92,   93,   94,    0,    0,    0,    0,    0,    0,    0,
    0,   96,    0,    0,    0,    0,    0,   98,   99,  100,
  101,    0,    0,    0,  102,    0,  103,    0,    0,    0,
    0,    0,  104,  105,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,  825,    0,    0,    0,  106,  107,  108,
  109,   57,   24,    0,   25,    0,    0,   26,  260,    0,
    0,    0,   27,   62,   63,  381,   28,    0,    0,   24,
    0,   25,   65,    0,   26,   30,    0,    0,    0,   27,
    0,    0,   32,   28,    0,    0,    0,   33,    0,   72,
   73,   34,   30,    0,    0,    0,    0,    0,    0,   32,
    0,    0,    0,   36,   33,   37,   75, 1026,   34,   38,
    0,    0,   77,    0,   79,    0,   81,   39,   40,  261,
   36,   41,   37,    0,    0,    0,   38,    0,   87,    0,
    0,   88,   89,    0,   39,   40,    0,    0,   41,    0,
    0,  608,    0,    0,    0,    0,   90,   91,   92,   93,
  309,    0,    0,    0,    0,  547,    0,    0,    0,   96,
    0,    0,    0,    0,    0,   98,   99,  100,  101,    0,
    0,    0,  102,    0,  103,    0,    0,    0,    0,    0,
  104,  105,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,  787,    0,    0,    0,  106,  310,  108,  109,   57,
   24,    0,   25,    0,    0,   26,  260,    0,    0,    0,
   27,   62,   63,  381,   28,    0,    0,   24,    0,   25,
   65,    0,   26,   30,    0,    0,    0,   27,    0,    0,
   32,   28,    0,    0,    0,   33,    0,   72,   73,   34,
   30,    0,    0,    0,    0,    0,    0,   32,    0,    0,
    0,   36,   33,   37,   75,    0,   34,   38,    0,    0,
   77,    0,   79,    0,   81,   39,   40,  261,   36,   41,
   37,    0,   85,    0,   38,    0,   87,    0,    0,   88,
   89,    0,   39,   40,    0,    0,   41,    0,    0,  564,
    0,    0,    0,    0,   90,   91,   92,   93,  309,    0,
    0,    0,    0,    0,    0,    0,    0,   96,    0,    0,
    0,    0,    0,   98,   99,  100,  101,    0,    0,    0,
  102,    0,  103,    0,    0,    0,    0,    0,  104,  105,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,  825,
    0,    0,    0,  106,  310,  108,  109,   57,   24,    0,
   25,    0,    0,   26,  260,    0,    0,    0,   27,   62,
   63,  381,   28,    0,    0,    0,    0,    0,   65,    0,
    0,   30,    0,    0,    0,    0,    0,    0,   32,    0,
    0,    0,    0,   33,    0,   72,   73,   34,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,   36,
    0,   37,   75,    0,    0,   38,    0,    0,   77,    0,
   79,    0,   81,   39,   40,  261,    0,   41,    0,    0,
    0,    0,    0,    0,   87,    0,    0,   88,   89,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,   90,   91,   92,   93,  309,    0,    0,    0,
    0,  547,    0,    0,    0,   96,    0,    0,    0,    0,
    0,   98,   99,  100,  101,    0,    0,    0,  102,    0,
  103,    0,    0,    0,    0,    0,  104,  105,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,  348,
    0,  106,  310,  108,  109,  349,    0,   57,   24,    0,
   25,    0,    0,   26,  260,    0,    0,    0,   27,   62,
   63,    0,   28,    0,    0,    0,    0,    0,   65,    0,
    0,   30,    0,    0,    0,    0,    0,    0,   32,    0,
    0,  350,    0,   33,    0,   72,   73,   34,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,   36,
    0,   37,   75,    0,    0,   38,    0,    0,   77,    0,
   79,    0,   81,   39,   40,  261,    0,   41,    0,    0,
    0,    0,    0,    0,    0,    0,  351,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,   90,   91,   92,  262,  352,    0,    0,    0,
    0,    0,    0,    0,    0,   96,    0,  353,    0,    0,
    0,   98,   99,  100,  101,  980,    0,    0,  102,    0,
  103,  817,    0,    0,    0,    0,  104,  105,    0,   57,
   24,    0,   25,    0,    0,   26,  260,    0,    0,    0,
   27,   62,   63,    0,   28,    0,    0,    0,    0,    0,
   65,  106,  264,   30,  109,    0,    0,    0,    0,    0,
   32,    0,    0,    0,    0,   33,    0,   72,   73,   34,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,   36,    0,   37,   75,    0,    0,   38,    0,    0,
   77,    0,   79,    0,   81,   39,   40,  261,    0,   41,
    0,    0,    0,    0,    0,    0,   87,    0,    0,   88,
   89,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,   90,   91,   92,   93,  309,    0,
    0,    0,    0,    0,  981,    0,    0,   96,    0,    0,
    0,    0,    0,   98,   99,  100,  101,    0,    0,    0,
  102,    0,  103,    0,    0,    0,    0,    0,  104,  105,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 1044,
    0,    0,    0,  106,  310,  108,  109,   57,   24,    0,
   25,    0,    0,   26,  260,    0,    0,    0,   27,   62,
   63,    0,   28,    0,    0,  534,    0,  534,   65,    0,
  534,   30,    0,    0,    0,  534,    0,    0,   32,  534,
    0,    0,    0,   33,    0,   72,   73,   34,  534,    0,
    0,    0,    0,    0,    0,  534,    0,    0,    0,   36,
  534,   37,   75,    0,  534,   38,    0,    0,   77,    0,
   79,    0,   81,   39,   40,  261,  534,   41,  534,    0,
    0,    0,  534,    0,   87,    0,    0,   88,   89,    0,
  534,  534,    0,    0,  534,    0,    0,  534,    0,    0,
    0,    0,   90,   91,   92,   93,  309,    0,    0,    0,
    0,    0, 1045,    0,    0,   96,    0,    0,    0,    0,
    0,   98,   99,  100,  101,    0,    0,    0,  102,    0,
  103,    0,    0,    0,    0,    0,  104,  105,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,  308,    0,    0,
    0,  106,  310,  108,  109,   57,   24,    0,   25,    0,
    0,   26,  260,    0,    0,    0,   27,   62,   63,  534,
   28,    0,    0,  191,    0,  191,   65,    0,  191,   30,
    0,    0,    0,  191,    0,    0,   32,  191,    0,    0,
    0,   33,    0,   72,   73,   34,  191,    0,    0,    0,
    0,    0,    0,  191,    0,    0,    0,   36,  191,   37,
   75,    0,  191,   38,    0,    0,   77,    0,   79,    0,
   81,   39,   40,  261,  191,   41,  191,    0,    0,    0,
  191,    0,   87,    0,    0,   88,   89,    0,  191,  191,
    0,    0,  191,    0,    0,  191,    0,    0,    0,    0,
   90,   91,   92,   93,  309,    0,    0,    0,    0,    0,
    0,    0,    0,   96,    0,    0,    0,    0,    0,   98,
   99,  100,  101,    0,    0,    0,  102,    0,  103,    0,
    0,    0,    0,    0,  104,  105,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,  317,    0,    0,    0,  106,
  310,  108,  109,   57,   24,    0,   25,    0,    0,   26,
  260,    0,    0,    0,   27,   62,   63,  191,   28,    0,
    0,    0,    0,    0,   65,    0,    0,   30,    0,    0,
    0,    0,    0,    0,   32,    0,    0,    0,    0,   33,
    0,   72,   73,   34,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,   36,    0,   37,   75,    0,
    0,   38,    0,    0,   77,    0,   79,    0,   81,   39,
   40,  261,    0,   41,    0,    0,    0,    0,    0,    0,
   87,    0,    0,   88,   89,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,   90,   91,
   92,   93,  309,    0,    0,    0,    0,    0,    0,    0,
    0,   96,    0,    0,    0,    0,    0,   98,   99,  100,
  101,    0,    0,    0,  102,    0,  103,    0,    0,    0,
    0,    0,  104,  105,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,  348,    0,  106,  310,  108,
  109,  349,    0,   57,   24,    0,   25,    0,    0,   26,
  260,    0,    0,    0,   27,   62,   63,    0,   28,    0,
    0,    0,    0,    0,   65,    0,    0,   30,    0,    0,
    0,    0,    0,    0,   32,    0,    0,  350,    0,   33,
    0,   72,   73,   34,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,   36,    0,   37,   75,    0,
    0,   38,    0,    0,   77,    0,   79,    0,   81,   39,
   40,  261,    0,   41,    0,    0,    0,    0,    0,    0,
    0,    0,  351,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,   90,   91,
   92,  262,  352,    0,    0,    0,    0,    0,    0,    0,
    0,   96,    0,  353,    0,    0,    0,   98,   99,  100,
  101,    0,    0,    0,  102,    0,  103,  633,    0,    0,
    0,    0,  104,  105,    0,   57,   24,    0,   25,    0,
    0,   26,  260,    0,    0,    0,   27,   62,   63,    0,
   28,    0,    0,    0,    0,    0,   65,  106,  264,   30,
  109,    0,    0,    0,    0,    0,   32,    0,    0,    0,
    0,   33,    0,   72,   73,   34,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,   36,    0,   37,
   75,    0,    0,   38,    0,    0,   77,    0,   79,    0,
   81,   39,   40,  261,    0,   41,    0,    0,    0,    0,
    0,    0,   87,    0,    0,   88,   89,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
   90,   91,   92,   93,   94,    0,    0,    0,    0,    0,
    0,    0,    0,   96,    0,    0,    0,    0,    0,   98,
   99,  100,  101,    0,    0,    0,  102,    0,  103,    0,
    0,    0,    0,    0,  104,  105,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,  817,    0,    0,    0,  106,
  107,  108,  109,   57,   24,    0,   25,    0,    0,   26,
  260,    0,    0,    0,   27,   62,   63,    0,   28,    0,
    0,  190,    0,  190,   65,    0,  190,   30,    0,    0,
    0,  190,    0,    0,   32,  190,    0,    0,    0,   33,
    0,   72,   73,   34,  190,    0,    0,    0,    0,    0,
    0,  190,    0,    0,    0,   36,  190,   37,   75,    0,
  190,   38,    0,    0,   77,    0,   79,    0,   81,   39,
   40,  261,  190,   41,  190,    0,    0,    0,  190,    0,
   87,    0,    0,   88,   89,    0,  190,  190,    0,    0,
  190,    0,    0,  190,    0,    0,    0,    0,   90,   91,
   92,   93,  309,    0,    0,    0,    0,    0,    0,    0,
    0,   96,    0,    0,    0,    0,    0,   98,   99,  100,
  101,    0,    0,    0,  102,    0,  103,    0,    0,    0,
    0,    0,  104,  105,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 1156,    0,    0,    0,  106,  310,  108,
  109,   57,   24,    0,   25,    0,    0,   26,  260,    0,
    0,    0,   27,   62,   63,  190,   28,    0,    0,  200,
    0,  200,   65,    0,  200,   30,    0,    0,    0,  200,
    0,    0,   32,  200,    0,    0,    0,   33,    0,   72,
   73,   34,  200,    0,    0,    0,    0,    0,    0,  200,
    0,    0,    0,   36,  200,   37,   75,    0,  200,   38,
    0,    0,   77,    0,   79,    0,   81,   39,   40,  261,
  200,   41,  200,    0,    0,    0,  200,    0,   87,    0,
    0,   88,   89,    0,  200,  200,    0,    0,  200,    0,
    0,  200,    0,    0,    0,    0,   90,   91,   92,   93,
   94,    0,    0,    0,    0,    0,    0,    0,    0,   96,
    0,    0,    0,    0,    0,   98,   99,  100,  101,    0,
    0,    0,  102,    0,  103,    0,    0,    0,    0,    0,
  104,  105,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,   84,    0,    0,    0,  106, 1157,  108,  109,   84,
   84,    0,   84,    0,    0,   84,   84,    0,    0,    0,
   84,   84,   84,  200,   84,    0,    0,    0,    0,    0,
   84,    0,    0,   84,    0,    0,    0,    0,    0,    0,
   84,    0,    0,    0,    0,   84,    0,   84,   84,   84,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,   84,    0,   84,   84,    0,    0,   84,    0,    0,
   84,    0,   84,    0,   84,   84,   84,   84,    0,   84,
    0,    0,    0,    0,    0,    0,   84,    0,    0,   84,
   84,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,   84,   84,   84,   84,   84,    0,
    0,    0,    0,    0,    0,    0,    0,   84,    0,    0,
    0,    0,    0,   84,   84,   84,   84,    0,    0,    0,
   84,    0,   84,    0,    0,    0,    0,    0,   84,   84,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,  348,    0,   84,   84,   84,   84,  349,    0,   57,
   24,    0,   25,    0,    0,   26,  260,    0,    0,    0,
   27,   62,   63,    0,   28,    0,    0,    0,    0,    0,
   65,    0,    0,   30,    0,    0,    0,    0,    0,    0,
   32,    0,    0,  350,    0,   33,    0,   72,   73,   34,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,   36,    0,   37,   75,    0,    0,   38,    0,    0,
   77,    0,   79,    0,   81,   39,   40,  261,    0,   41,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,   90,   91,   92,  262,  352,    0,
    0,    0,    0,    0,    0,    0,    0,   96,    0,  353,
    0,    0,    0,   98,   99,  100,  101,    0,    0,    0,
  102,    0,  103,  348,    0,    0,    0,    0,  104,  105,
    0,   57,   24,    0,   25,    0,    0,   26,  260,    0,
    0,    0,   27,   62,   63,    0,   28,    0,    0,    0,
    0,    0,   65,  106,  264,   30,  109,    0,    0,    0,
    0,    0,   32,    0,    0,    0,    0,   33,    0,   72,
   73,   34,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,   36,    0,   37,   75,    0,    0,   38,
    0,    0,   77,    0,   79,    0,   81,   39,   40,  261,
    0,   41,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,   90,   91,   92,  262,
  352,    0,    0,    0,    0,    0,    0,    0,    0,   96,
    0,  353,    0,    0,    0,   98,   99,  100,  101,    0,
    0,    0,  102,    0,  103,  357,    0,    0,    0,    0,
  104,  105,    0,   57,   24,    0,   25,    0,    0,   26,
  260,    0,    0,    0,   27,   62,   63,    0,   28,    0,
    0,    0,    0,    0,   65,  106,  264,   30,  109,    0,
    0,    0,    0,    0,   32,    0,    0,    0,    0,   33,
    0,   72,   73,   34,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,   36,    0,   37,   75,    0,
    0,   38,    0,    0,   77,    0,   79,    0,   81,   39,
   40,  261,    0,   41,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,   90,   91,
   92,  262,  352,    0,    0,    0,    0,    0,    0,    0,
    0,   96,    0,    0,    0,    0,    0,   98,   99,  100,
  101,    0,    0,    0,  102,    0,  103,  359,    0,    0,
    0,    0,  104,  105,    0,   57,   24,    0,   25,    0,
    0,   26,  260,    0,    0,    0,   27,   62,   63,    0,
   28,    0,    0,    0,    0,    0,   65,  106,  264,   30,
  109,    0,    0,    0,    0,    0,   32,    0,    0,    0,
    0,   33,    0,   72,   73,   34,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,   36,    0,   37,
   75,    0,    0,   38,    0,    0,   77,    0,   79,    0,
   81,   39,   40,  261,    0,   41,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
   90,   91,   92,  262,  352,    0,    0,    0,    0,    0,
    0,    0,    0,   96,    0,    0,    0,    0,    0,   98,
   99,  100,  101,    0,    0,    0,  102,    0,  103,  361,
    0,    0,    0,    0,  104,  105,    0,   57,   24,    0,
   25,    0,    0,   26,  260,    0,    0,    0,   27,   62,
   63,    0,   28,    0,    0,    0,    0,    0,   65,  106,
  264,   30,  109,    0,    0,    0,    0,    0,   32,    0,
    0,    0,    0,   33,    0,   72,   73,   34,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,   36,
    0,   37,   75,    0,    0,   38,    0,    0,   77,    0,
   79,    0,   81,   39,   40,  261,    0,   41,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,   90,   91,   92,  262,  352,    0,    0,    0,
    0,    0,    0,    0,    0,   96,    0,    0,    0,    0,
    0,   98,   99,  100,  101,    0,    0,    0,  102,    0,
  103,  363,    0,    0,    0,    0,  104,  105,    0,   57,
   24,    0,   25,    0,    0,   26,  260,    0,    0,    0,
   27,   62,   63,    0,   28,    0,    0,    0,    0,    0,
   65,  106,  264,   30,  109,    0,    0,    0,    0,    0,
   32,    0,    0,    0,    0,   33,    0,   72,   73,   34,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,   36,    0,   37,   75,    0,    0,   38,    0,    0,
   77,    0,   79,    0,   81,   39,   40,  261,    0,   41,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,   90,   91,   92,  262,  352,    0,
    0,    0,    0,    0,    0,    0,    0,   96,    0,    0,
    0,    0,    0,   98,   99,  100,  101,    0,    0,    0,
  102,    0,  103,  365,    0,    0,    0,    0,  104,  105,
    0,   57,   24,    0,   25,    0,    0,   26,  260,    0,
    0,    0,   27,   62,   63,    0,   28,    0,    0,    0,
    0,    0,   65,  106,  264,   30,  109,    0,    0,    0,
    0,    0,   32,    0,    0,    0,    0,   33,    0,   72,
   73,   34,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,   36,    0,   37,   75,    0,    0,   38,
    0,    0,   77,    0,   79,    0,   81,   39,   40,  261,
    0,   41,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,   90,   91,   92,  262,
  352,    0,    0,    0,    0,    0,    0,    0,    0,   96,
    0,    0,    0,    0,    0,   98,   99,  100,  101,    0,
    0,    0,  102,    0,  103,  367,    0,    0,    0,    0,
  104,  105,    0,   57,   24,    0,   25,    0,    0,   26,
  260,    0,    0,    0,   27,   62,   63,    0,   28,    0,
    0,    0,    0,    0,   65,  106,  264,   30,  109,    0,
    0,    0,    0,    0,   32,    0,    0,    0,    0,   33,
    0,   72,   73,   34,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,   36,    0,   37,   75,    0,
    0,   38,    0,    0,   77,    0,   79,    0,   81,   39,
   40,  261,    0,   41,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,   90,   91,
   92,  262,  352,    0,    0,    0,    0,    0,    0,    0,
    0,   96,    0,    0,    0,    0,    0,   98,   99,  100,
  101,    0,    0,    0,  102,    0,  103,  369,    0,    0,
    0,    0,  104,  105,    0,   57,   24,    0,   25,    0,
    0,   26,  260,    0,    0,    0,   27,   62,   63,    0,
   28,    0,    0,    0,    0,    0,   65,  106,  264,   30,
  109,    0,    0,    0,    0,    0,   32,    0,    0,    0,
    0,   33,    0,   72,   73,   34,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,   36,    0,   37,
   75,    0,    0,   38,    0,    0,   77,    0,   79,    0,
   81,   39,   40,  261,    0,   41,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
   90,   91,   92,  262,  352,    0,    0,    0,    0,    0,
    0,    0,    0,   96,    0,    0,    0,    0,    0,   98,
   99,  100,  101,    0,    0,    0,  102,    0,  103,  371,
    0,    0,    0,    0,  104,  105,    0,   57,   24,    0,
   25,    0,    0,   26,  260,    0,    0,    0,   27,   62,
   63,    0,   28,    0,    0,    0,    0,    0,   65,  106,
  264,   30,  109,    0,    0,    0,    0,    0,   32,    0,
    0,    0,    0,   33,    0,   72,   73,   34,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,   36,
    0,   37,   75,    0,    0,   38,    0,    0,   77,    0,
   79,    0,   81,   39,   40,  261,    0,   41,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,   90,   91,   92,  262,  352,    0,    0,    0,
    0,    0,    0,    0,    0,   96,    0,    0,    0,    0,
    0,   98,   99,  100,  101,    0,    0,    0,  102,    0,
  103,  659,    0,    0,    0,    0,  104,  105,    0,   57,
   24,    0,   25,    0,    0,   26,  260,    0,    0,    0,
   27,   62,   63,    0,   28,    0,    0,    0,    0,    0,
   65,  106,  264,   30,  109,    0,    0,    0,    0,    0,
   32,    0,    0,    0,    0,   33,    0,   72,   73,   34,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,   36,    0,   37,   75,    0,    0,   38,    0,    0,
   77,    0,   79,    0,   81,   39,   40,  261,    0,   41,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,   90,   91,   92,  262,  352,    0,
    0,    0,    0,    0,    0,    0,    0,   96,    0,    0,
    0,    0,    0,   98,   99,  100,  101,    0,    0,    0,
  102,    0,  103,  661,    0,    0,    0,    0,  104,  105,
    0,   57,   24,    0,   25,    0,    0,   26,  260,    0,
    0,    0,   27,   62,   63,    0,   28,    0,    0,    0,
    0,    0,   65,  106,  264,   30,  109,    0,    0,    0,
    0,    0,   32,    0,    0,    0,    0,   33,    0,   72,
   73,   34,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,   36,    0,   37,   75,    0,    0,   38,
    0,    0,   77,    0,   79,    0,   81,   39,   40,  261,
    0,   41,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,   90,   91,   92,  262,
  352,    0,    0,    0,    0,    0,    0,    0,    0,   96,
    0,    0,    0,    0,    0,   98,   99,  100,  101,    0,
    0,    0,  102,    0,  103,  663,    0,    0,    0,    0,
  104,  105,    0,   57,   24,    0,   25,    0,    0,   26,
  260,    0,    0,    0,   27,   62,   63,    0,   28,    0,
    0,    0,    0,    0,   65,  106,  264,   30,  109,    0,
    0,    0,    0,    0,   32,    0,    0,    0,    0,   33,
    0,   72,   73,   34,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,   36,    0,   37,   75,    0,
    0,   38,    0,    0,   77,    0,   79,    0,   81,   39,
   40,  261,    0,   41,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,   90,   91,
   92,  262,  352,    0,    0,    0,    0,    0,    0,    0,
    0,   96,    0,    0,    0,    0,    0,   98,   99,  100,
  101,    0,    0,    0,  102,    0,  103,  681,    0,    0,
    0,    0,  104,  105,    0,   57,   24,    0,   25,    0,
    0,   26,  260,    0,    0,    0,   27,   62,   63,    0,
   28,    0,    0,    0,    0,    0,   65,  106,  264,   30,
  109,    0,    0,    0,    0,    0,   32,    0,    0,    0,
    0,   33,    0,   72,   73,   34,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,   36,    0,   37,
   75,    0,    0,   38,    0,    0,   77,    0,   79,    0,
   81,   39,   40,  261,    0,   41,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
   90,   91,   92,  262,  352,    0,    0,    0,    0,    0,
    0,    0,    0,   96,    0,    0,    0,    0,    0,   98,
   99,  100,  101,    0,    0,    0,  102,    0,  103,  683,
    0,    0,    0,    0,  104,  105,    0,   57,   24,    0,
   25,    0,    0,   26,  260,    0,    0,    0,   27,   62,
   63,    0,   28,    0,    0,    0,    0,    0,   65,  106,
  264,   30,  109,    0,    0,    0,    0,    0,   32,    0,
    0,    0,    0,   33,    0,   72,   73,   34,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,   36,
    0,   37,   75,    0,    0,   38,    0,    0,   77,    0,
   79,    0,   81,   39,   40,  261,    0,   41,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,   90,   91,   92,  262,  352,    0,    0,    0,
    0,    0,    0,    0,    0,   96,    0,    0,    0,    0,
    0,   98,   99,  100,  101,    0,    0,    0,  102,    0,
  103,  685,    0,    0,    0,    0,  104,  105,    0,   57,
   24,    0,   25,    0,    0,   26,  260,    0,    0,    0,
   27,   62,   63,    0,   28,    0,    0,    0,    0,    0,
   65,  106,  264,   30,  109,    0,    0,    0,    0,    0,
   32,    0,    0,    0,    0,   33,    0,   72,   73,   34,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,   36,    0,   37,   75,    0,    0,   38,    0,    0,
   77,    0,   79,    0,   81,   39,   40,  261,    0,   41,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,   90,   91,   92,  262,  686,    0,
    0,    0,    0,    0,    0,    0,    0,   96,    0,    0,
    0,    0,    0,   98,   99,  100,  101,    0,    0,    0,
  102,    0,  103,  688,    0,    0,    0,    0,  104,  105,
    0,   57,   24,    0,   25,    0,    0,   26,  260,    0,
    0,    0,   27,   62,   63,    0,   28,    0,    0,    0,
    0,    0,   65,  106,  264,   30,  109,    0,    0,    0,
    0,    0,   32,    0,    0,    0,    0,   33,    0,   72,
   73,   34,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,   36,    0,   37,   75,    0,    0,   38,
    0,    0,   77,    0,   79,    0,   81,   39,   40,  261,
    0,   41,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,   90,   91,   92,  262,
  686,    0,    0,    0,    0,    0,    0,    0,    0,   96,
    0,    0,    0,    0,    0,   98,   99,  100,  101,    0,
    0,    0,  102,    0,  103,  690,    0,    0,    0,    0,
  104,  105,    0,   57,   24,    0,   25,    0,    0,   26,
  260,    0,    0,    0,   27,   62,   63,    0,   28,    0,
    0,    0,    0,    0,   65,  106,  264,   30,  109,    0,
    0,    0,    0,    0,   32,    0,    0,    0,    0,   33,
    0,   72,   73,   34,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,   36,    0,   37,   75,    0,
    0,   38,    0,    0,   77,    0,   79,    0,   81,   39,
   40,  261,    0,   41,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,   90,   91,
   92,  262,  686,    0,    0,    0,    0,    0,    0,    0,
    0,   96,    0,    0,    0,    0,    0,   98,   99,  100,
  101,    0,    0,    0,  102,    0,  103,  692,    0,    0,
    0,    0,  104,  105,    0,   57,   24,    0,   25,    0,
    0,   26,  260,    0,    0,    0,   27,   62,   63,    0,
   28,    0,    0,    0,    0,    0,   65,  106,  264,   30,
  109,    0,    0,    0,    0,    0,   32,    0,    0,    0,
    0,   33,    0,   72,   73,   34,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,   36,    0,   37,
   75,    0,    0,   38,    0,    0,   77,    0,   79,    0,
   81,   39,   40,  261,    0,   41,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
   90,   91,   92,  262,  686,    0,    0,    0,    0,    0,
    0,    0,    0,   96,    0,    0,    0,    0,    0,   98,
   99,  100,  101,    0,    0,    0,  102,    0,  103,  694,
    0,    0,    0,    0,  104,  105,    0,   57,   24,    0,
   25,    0,    0,   26,  260,    0,    0,    0,   27,   62,
   63,    0,   28,    0,    0,    0,    0,    0,   65,  106,
  264,   30,  109,    0,    0,    0,    0,    0,   32,    0,
    0,    0,    0,   33,    0,   72,   73,   34,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,   36,
    0,   37,   75,    0,    0,   38,    0,    0,   77,    0,
   79,    0,   81,   39,   40,  261,    0,   41,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,   90,   91,   92,  262,  686,    0,    0,    0,
    0,    0,    0,    0,    0,   96,    0,    0,    0,    0,
    0,   98,   99,  100,  101,    0,    0,    0,  102,    0,
  103,  696,    0,    0,    0,    0,  104,  105,    0,   57,
   24,    0,   25,    0,    0,   26,  260,    0,    0,    0,
   27,   62,   63,    0,   28,    0,    0,    0,    0,    0,
   65,  106,  264,   30,  109,    0,    0,    0,    0,    0,
   32,    0,    0,    0,    0,   33,    0,   72,   73,   34,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,   36,    0,   37,   75,    0,    0,   38,    0,    0,
   77,    0,   79,    0,   81,   39,   40,  261,    0,   41,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,   90,   91,   92,  262,  686,    0,
    0,    0,    0,    0,    0,    0,    0,   96,    0,    0,
    0,    0,    0,   98,   99,  100,  101,    0,    0,    0,
  102,    0,  103,  698,    0,    0,    0,    0,  104,  105,
    0,   57,   24,    0,   25,    0,    0,   26,  260,    0,
    0,    0,   27,   62,   63,    0,   28,    0,    0,    0,
    0,    0,   65,  106,  264,   30,  109,    0,    0,    0,
    0,    0,   32,    0,    0,    0,    0,   33,    0,   72,
   73,   34,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,   36,    0,   37,   75,    0,    0,   38,
    0,    0,   77,    0,   79,    0,   81,   39,   40,  261,
    0,   41,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,   90,   91,   92,  262,
  686,    0,    0,    0,    0,    0,    0,    0,    0,   96,
    0,    0,    0,    0,    0,   98,   99,  100,  101,    0,
    0,    0,  102,    0,  103,  700,    0,    0,    0,    0,
  104,  105,    0,   57,   24,    0,   25,    0,    0,   26,
  260,    0,    0,    0,   27,   62,   63,    0,   28,    0,
    0,    0,    0,    0,   65,  106,  264,   30,  109,    0,
    0,    0,    0,    0,   32,    0,    0,    0,    0,   33,
    0,   72,   73,   34,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,   36,    0,   37,   75,    0,
    0,   38,    0,    0,   77,    0,   79,    0,   81,   39,
   40,  261,    0,   41,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,   90,   91,
   92,  262,  686,    0,    0,    0,    0,    0,    0,    0,
    0,   96,    0,    0,    0,    0,    0,   98,   99,  100,
  101,    0,    0,    0,  102,    0,  103,  702,    0,    0,
    0,    0,  104,  105,    0,   57,   24,    0,   25,    0,
    0,   26,  260,    0,    0,    0,   27,   62,   63,    0,
   28,    0,    0,    0,    0,    0,   65,  106,  264,   30,
  109,    0,    0,    0,    0,    0,   32,    0,    0,    0,
    0,   33,    0,   72,   73,   34,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,   36,    0,   37,
   75,    0,    0,   38,    0,    0,   77,    0,   79,    0,
   81,   39,   40,  261,    0,   41,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
   90,   91,   92,  262,  686,    0,    0,    0,    0,    0,
    0,    0,    0,   96,    0,    0,    0,    0,    0,   98,
   99,  100,  101,    0,    0,    0,  102,    0,  103,  704,
    0,    0,    0,    0,  104,  105,    0,   57,   24,    0,
   25,    0,    0,   26,  260,    0,    0,    0,   27,   62,
   63,    0,   28,    0,    0,    0,    0,    0,   65,  106,
  264,   30,  109,    0,    0,    0,    0,    0,   32,    0,
    0,    0,    0,   33,    0,   72,   73,   34,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,   36,
    0,   37,   75,    0,    0,   38,    0,    0,   77,    0,
   79,    0,   81,   39,   40,  261,    0,   41,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,   90,   91,   92,  262,  686,    0,    0,    0,
    0,    0,    0,    0,    0,   96,    0,    0,    0,    0,
    0,   98,   99,  100,  101,    0,    0,    0,  102,    0,
  103,  706,    0,    0,    0,    0,  104,  105,    0,   57,
   24,    0,   25,    0,    0,   26,  260,    0,    0,    0,
   27,   62,   63,    0,   28,    0,    0,    0,    0,    0,
   65,  106,  264,   30,  109,    0,    0,    0,    0,    0,
   32,    0,    0,    0,    0,   33,    0,   72,   73,   34,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,   36,    0,   37,   75,    0,    0,   38,    0,    0,
   77,    0,   79,    0,   81,   39,   40,  261,    0,   41,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,   90,   91,   92,  262,  686,    0,
    0,    0,    0,    0,    0,    0,    0,   96,    0,    0,
    0,    0,    0,   98,   99,  100,  101,    0,    0,    0,
  102,    0,  103,  708,    0,    0,    0,    0,  104,  105,
    0,   57,   24,    0,   25,    0,    0,   26,  260,    0,
    0,    0,   27,   62,   63,    0,   28,    0,    0,    0,
    0,    0,   65,  106,  264,   30,  109,    0,    0,    0,
    0,    0,   32,    0,    0,    0,    0,   33,    0,   72,
   73,   34,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,   36,    0,   37,   75,    0,    0,   38,
    0,    0,   77,    0,   79,    0,   81,   39,   40,  261,
    0,   41,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,   90,   91,   92,  262,
  686,    0,    0,    0,    0,    0,    0,    0,    0,   96,
    0,    0,    0,    0,    0,   98,   99,  100,  101,    0,
    0,    0,  102,    0,  103,  710,    0,    0,    0,    0,
  104,  105,    0,   57,   24,    0,   25,    0,    0,   26,
  260,    0,    0,    0,   27,   62,   63,    0,   28,    0,
    0,    0,    0,    0,   65,  106,  264,   30,  109,    0,
    0,    0,    0,    0,   32,    0,    0,    0,    0,   33,
    0,   72,   73,   34,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,   36,    0,   37,   75,    0,
    0,   38,    0,    0,   77,    0,   79,    0,   81,   39,
   40,  261,    0,   41,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,   90,   91,
   92,  262,  686,    0,    0,    0,    0,    0,    0,    0,
    0,   96,    0,    0,    0,    0,    0,   98,   99,  100,
  101,    0,    0,    0,  102,    0,  103,  874,    0,    0,
    0,    0,  104,  105,    0,   57,   24,    0,   25,    0,
    0,   26,  260,    0,    0,    0,   27,   62,   63,    0,
   28,    0,    0,    0,    0,    0,   65,  106,  264,   30,
  109,    0,    0,    0,    0,    0,   32,    0,    0,    0,
    0,   33,    0,   72,   73,   34,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,   36,    0,   37,
   75,    0,    0,   38,    0,    0,   77,    0,   79,    0,
   81,   39,   40,  261,    0,   41,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
   90,   91,   92,  262,  352,    0,    0,    0,    0,    0,
    0,    0,    0,   96,    0,    0,    0,    0,    0,   98,
   99,  100,  101,    0,    0,    0,  102,    0,  103,  667,
    0,    0,    0,    0,  104,  105,    0,   57,   24,    0,
   25,    0,    0,   26,  260,    0,    0,    0,   27,   62,
   63,    0,   28,    0,    0,    0,    0,    0,   65,  106,
  264,   30,  109,    0,    0,   51,    0,   51,   32,    0,
    0,    0,    0,   33,    0,   72,   73,   34,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,   51,   36,
    0,   37,   75,    0,    0,   38,    0,    0,   77,    0,
   79,   51,   81,   39,   40,  261,   51,   41,    0,    0,
   85,   51,    0,   51,   51,   51,   51,    0,    0,   51,
    0,   51,    0,    0,    0,   51,    0,    0,    0,    0,
    0,    0,   90,   91,   92,  262,  263,   51,    0,    0,
   51,    0,   51,    0,    0,   96,    0,    0,    0,    0,
    0,    0,  668,  669,    0,    0,  211,    0,    0,    0,
  670,  211,    0,    0,    0,  211,   51,  211,    0,    0,
  211,    0,  211,  211,    0,  211,    0,  211,    0,  211,
    0,  211,  211,  211,  211,    0,    0,  211,  211,    0,
    0,  106,  264,  211,  671,  211,  211,  211,    0,    0,
  211,    0,  211,    0,  211,    0,    0,  211,    0,  211,
  211,  211,  211,    0,    0,    0,  211,  211,  211,    0,
    0,  211,  211,  211,    0,    0,    0,    0,    0,    0,
  211,  211,    0,  211,  211,  526,  211,  211,  211,    0,
    0,    0,  211,   57,   24,    0,   25,    0,    0,   26,
  260,    0,    0,    0,   27,   62,   63,    0,   28,    0,
    0,    0,  211,    0,   65,    0,    0,   30,  211,  211,
  211,    0,    0,    0,   32,    0,    0,    0,  211,   33,
    0,   72,   73,   34,    0,    0,    0,    0,    0,  366,
    0,    0,    0,    0,    0,   36,    0,   37,   75,    0,
    0,   38,    0,    0,   77,    0,   79,    0,   81,   39,
   40,  261,    0,   41,    0,    0,    0,    0,    0,  211,
    0,    0,    0,    0,    0,  366,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,   90,   91,
   92,  262,  527,    0,    0,    0,    0,    0,    0,    0,
    0,   96,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,  366,    0,    0,    0,    0,    0,  106,  528,  366,
  366,  366,  366,  854,    0,    0,  366,  366,    0,    0,
  366,  366,  366,  366,  366,  366,  366,  366,  366,    0,
  366,  366,  366,  366,  366,  366,  366,  366,  366,  366,
  366,  366,  366,  366,  366,  366,  366,  366,  366,  366,
  366,  366,    0,   52,    0,   52,    0,   52,  366,   52,
    0,  366,   52,    0,   52,   52,    0,   52,    0,   52,
    0,   52,    0,   52,   52,   52,   52,    0,    0,   52,
   52,    0,    0,    0,    0,   52,   52,   52,   52,   52,
    0,    0,   52,    0,   52,    0,   52,    0,   52,   52,
    0,   52,   52,   52,   52,    0,    0,   52,   52,   52,
   52,    0,    0,   52,   52,   52,    0,    0,    0,    0,
    0,    0,   52,   52,    0,   52,   52,    0,   52,   52,
   52,    0,    0,    0,   52,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,   52,    0,    0,   52,   52,   51,
    0,    0,    0,   51,    0,   51,    0,    0,   51,    0,
   51,   51,    0,   51,    0,   51,    0,   51,    0,   51,
   51,   51,   51,    0,    0,   51,   51,    0,    0,    0,
    0,   51,    0,   51,   51,   51,    0,    0,   51,    0,
   51,    0,   51,    0,    0,   51,    0,   51,   51,   51,
   51,   52,    0,    0,   51,   51,   51,    0,    0,   51,
   51,   51,    0,    0,    0,    0,    0,    0,   51,   51,
    0,   51,   51,    0,   51,   51,   51,    0,    0,    0,
   51,    0,    0,   51,    0,    0,    0,   51,    0,   51,
    0,    0,   51,    0,   51,   51,    0,   51,    0,   51,
   51,   51,    0,   51,   51,   51,   51,   87,    0,   51,
   51,    0,    0,    0,    0,   51,   51,   51,   51,   51,
    0,    0,   51,    0,   51,    0,   51,    0,    0,   51,
    0,   51,   51,   51,   51,    0,    0,    0,   51,   51,
   51,    0,    0,   51,   51,   51,    0,    0,    0,    0,
    0,    0,   51,   51,    0,   51,   51,   51,   51,   51,
   51,    0,    0,    0,   51,    0,    0,   52,    0,    0,
    0,   52,    0,   52,    0,    0,   52,    0,   52,   52,
    0,   52,    0,   52,   51,   52,    0,   52,   52,   52,
   52,   88,    0,   52,   52,    0,    0,    0,    0,   52,
   51,   52,   52,   52,    0,    0,   52,    0,   52,    0,
   52,    0,    0,   52,    0,   52,   52,   52,   52,    0,
    0,    0,   52,   52,   52,    0,    0,   52,   52,   52,
    0,    0,    0,    0,    0,    0,   52,   52,    0,   52,
   52,   51,   52,   52,   52,    0,    0,    0,   52,    0,
    0,   51,    0,    0,    0,   51,    0,   51,    0,    0,
   51,    0,   51,   51,    0,   51,    0,   51,   52,   51,
    0,   51,   51,   51,   51,    0,    0,   51,   51,    0,
    0,    0,    0,   51,   52,   51,   51,   51,    0,    0,
   51,    0,   51,    0,   51,    0,    0,   51,    0,   51,
   51,   51,   51,    0,    0,    0,   51,   51,   51,    0,
    0,   51,   51,   51,    0,    0,    0,    0,    0,    0,
   51,   51,    0,   51,   51,   52,   51,   51,   51,    0,
    0,    0,   51,    0,    0,   51,    0,    0,    0,   51,
    0,   51,    0,    0,   51,    0,   51,   51,    0,   51,
    0,   51,   51,   51,    0,   51,   51,   51,   51,  239,
    0,   51,   51,    0,    0,    0,    0,   51,    0,   51,
   51,   51,    0,    0,   51,    0,   51,  366,   51,    0,
    0,   51,    0,   51,   51,   51,   51,    0,    0,    0,
   51,   51,   51,    0,    0,   51,   51,   51,    0,    0,
  366,    0,    0,    0,   51,   51,    0,   51,   51,   51,
   51,   51,   51,  366,    0,    0,   51,    0,  366,    0,
    0,  366,    0,  366,    0,  366,  366,  366,  366,    0,
    0,    0,    0,  366,    0,    0,   51,  366,    0,    0,
    0,  366,    0,  240,    0,    0,    0,    0,    0,  366,
    0,    0,  366,    0,  366,    0,   57,   24,    0,   25,
    0,    0,   26,  260,    0,    0,    0,   27,   62,   63,
    0,   28,    0,  366,    0,    0,    0,   65,  366,    0,
   30,    0,    0,    0,    0,  366,  366,   32,  289,    0,
  366,    0,   33,   51,   72,   73,   34,    0,  636,    0,
    0,    0,    0,  366,    0,  637,    0,    0,   36,    0,
   37,   75,    0,    0,   38,    0,    0,   77,    0,   79,
    0,   81,   39,   40,  261,    0,   41,    0,    0,    0,
    0,    0,    0,  638,    0,  366,   88,   89,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,   90,   91,   92,   93,   94,    0,    0,    0,    0,
    0,    0,    0,    0,   96, 1021,    0,  639,    0,    0,
   98,   99,  100,  101,    0,    0,    0,  102,    0,  103,
    0,    0,    0,    0,    0,  104,  105,    0,    0,    0,
    0,    0,    0,   57,   24,    0,   25,    0,    0,   26,
  260,    0,    0,    0,   27,   62,   63,    0,   28,    0,
  106,  107,  108,  109,   65,    0,    0,   30,    0,    0,
    0,    0,    0,    0,   32,    0,    0,    0,    0,   33,
    0,   72,   73,   34,    0,  636,    0,    0,    0,    0,
    0,    0,  637,    0,    0,   36,    0,   37,   75,    0,
    0,   38,    0,    0,   77,    0,   79,    0,   81,   39,
   40,  261,    0,   41,    0,    0,    0,    0,    0,    0,
  638,    0,    0,   88,   89,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,   90,   91,
   92,   93,   94,    0,    0,    0,    0,    0,    0,    0,
    0,   96,    0,    0,  639,    0,    0,   98,   99,  100,
  101,    0,    0,    0,  102,    0,  103,    0,    0,    0,
    0,    0,  104,  105,    0,    0,    0,    0,    0,    0,
   57,   24,    0,   25,    0,    0,   26,  260,    0,    0,
    0,   27,   62,   63,    0,   28,    0,  106,  107,  108,
  109,   65,    0,    0,   30,    0,    0,    0,    0,    0,
    0,   32,    0,    0,    0,    0,   33,    0,   72,   73,
   34,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,   36,    0,   37,   75,    0,    0,   38,    0,
    0,   77,    0,   79,    0,   81,   39,   40,  261,    0,
   41,    0,    0,   85,    0,    0,    0,   87,    0,    0,
   88,   89,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,   90,   91,   92,   93,  309,
    0,    0,    0,    0,    0,    0,    0,    0,   96,    0,
    0,    0,    0,    0,   98,   99,  100,  101,    0,    0,
    0,  102,    0,  103,    0,    0,    0,    0,    0,  104,
  105,    0,    0,    0,    0,    0,    0,   57,   24,    0,
   25,    0,    0,   26,  260,    0,    0,    0,   27,   62,
   63,    0,   28,    0,  106,  310,  108,  109,   65,    0,
    0,   30,    0,    0,    0,    0,    0,    0,   32,    0,
    0,    0,    0,   33,    0,   72,   73,   34,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,   36,
    0,   37,   75,    0,    0,   38,    0,    0,   77,    0,
   79,    0,   81,   39,   40,  261,    0,   41,    0,    0,
   85,    0,    0,    0,   87,    0,    0,   88,   89,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,   90,   91,   92,   93,  309,    0,    0,    0,
    0,    0,    0,    0,    0,  889,    0,    0,    0,    0,
    0,   98,   99,  100,  101,    0,    0,    0,  102,    0,
  103,    0,    0,    0,    0,    0,  104,  105,    0,    0,
    0,    0,    0,    0,   57,   24,    0,   25,    0,    0,
   26,  260,    0,    0,    0,   27,   62,   63,    0,   28,
    0,  106,  310,  108,  109,   65,    0,    0,   30,    0,
    0,    0,    0,    0,    0,   32,    0,    0,    0,    0,
   33,    0,   72,   73,   34,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,   36,    0,   37,   75,
    0,    0,   38,    0,    0,   77,    0,   79,    0,   81,
   39,   40,  261,    0,   41,    0,    0,    0,    0,    0,
    0,   87,    0,    0,   88,   89,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,   90,
   91,   92,   93,  309,    0,    0,    0,    0,  547,    0,
    0,    0,   96,    0,    0,    0,    0,    0,   98,   99,
  100,  101,    0,    0,    0,  102,    0,  103,    0,    0,
    0,    0,    0,  104,  105,    0,    0,    0,    0,    0,
    0,   57,   24,    0,   25,    0,    0,   26,  260,    0,
    0,    0,   27,   62,   63,    0,   28,    0,  106,  310,
  108,  109,   65,    0,    0,   30,    0,    0,    0,    0,
    0,    0,   32,    0,    0,    0,    0,   33,    0,   72,
   73,   34,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,   36,    0,   37,   75,    0,    0,   38,
    0,    0,   77,    0,   79,    0,   81,   39,   40,  261,
    0,   41,    0,    0,    0,    0,    0,    0,   87,    0,
    0,   88,   89,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,   90,   91,   92,   93,
  309,    0,    0,    0,    0,  541,    0,    0,    0,   96,
    0,    0,    0,    0,    0,   98,   99,  100,  101,    0,
    0,    0,  102,    0,  103,    0,    0,    0,    0,    0,
  104,  105,    0,    0,    0,    0,    0,    0,   57,   24,
    0,   25,    0,    0,   26,  260,    0,    0,    0,   27,
   62,   63,    0,   28,    0,  106,  310,  108,  109,   65,
    0,    0,   30,    0,    0,    0,    0,    0,    0,   32,
    0,    0,    0,    0,   33,    0,   72,   73,   34,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
   36,    0,   37,   75,    0,    0,   38,    0,    0,   77,
    0,   79,    0,   81,   39,   40,  261,    0,   41,    0,
    0,    0,    0,    0,    0,   87,    0,    0,   88,   89,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,   90,   91,   92,   93,  309,    0,    0,
    0,    0,    0,    0,    0,    0,   96,    0,    0,    0,
    0,    0,   98,   99,  100,  101,    0,    0,    0,  102,
    0,  103,    0,    0,    0,    0,    0,  104,  105,    0,
    0,    0,    0,    0,    0,   57,   24,    0,   25,    0,
    0,   26,  260,    0,    0,    0,   27,   62,   63,    0,
   28,    0,  106,  310,  108,  109,   65,    0,    0,   30,
    0,    0,    0,    0,    0,    0,   32,    0,    0,    0,
    0,   33,    0,   72,   73,   34,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,   36,    0,   37,
   75,    0,    0,   38,    0,    0,   77,    0,   79,    0,
   81,   39,   40,  261,    0,   41,    0,    0,    0,    0,
    0,    0,   87,    0,    0,   88,   89,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
   90,   91,   92,   93,   94,    0,    0,    0,    0,    0,
    0,    0,    0,   96,    0,    0,    0,    0,    0,   98,
   99,  100,  101,    0,    0,    0,  102,    0,  103,    0,
    0,    0,    0,    0,  104,  105,    0,    0,    0,    0,
    0,    0,  709,  709,    0,  709,    0,    0,  709,  709,
    0,    0,    0,  709,  709,  709,    0,  709,    0,  106,
  107,  108,  109,  709,    0,    0,  709,    0,    0,    0,
    0,    0,    0,  709,    0,    0,    0,    0,  709,    0,
  709,  709,  709,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,  709,    0,  709,  709,    0,    0,
  709,    0,    0,  709,    0,  709,    0,  709,  709,  709,
  709,    0,  709,    0,    0,    0,    0,    0,    0,  709,
    0,    0,  709,  709,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,  709,  709,  709,
  709,  709,    0,    0,    0,    0,    0,    0,    0,    0,
  709,    0,    0,    0,    0,    0,  709,  709,  709,  709,
    0,    0,    0,  709,    0,  709,    0,    0,    0,    0,
    0,  709,  709,    0,    0,    0,    0,    0,    0,  147,
  147,    0,  147,    0,    0,  147,  147,    0,    0,    0,
  147,  147,  147,    0,  147,    0,  709,  709,  709,  709,
  147,    0,    0,  147,    0,    0,    0,    0,    0,    0,
  147,    0,    0,    0,    0,  147,    0,  147,  147,  147,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,  147,    0,  147,  147,    0,    0,  147,    0,    0,
  147,    0,  147,    0,  147,  147,  147,  147,    0,  147,
    0,    0,    0,    0,    0,    0,  147,    0,    0,  147,
  147,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,  147,  147,  147,  147,  147,    0,
    0,    0,    0,    0,    0,    0,    0,  147,    0,    0,
    0,    0,    0,  147,  147,  147,  147,    0,    0,    0,
  147,    0,  147,    0,    0,    0,    0,    0,  147,  147,
    0,    0,    0,    0,    0,    0,   57,   24,    0,   25,
    0,    0,   26,  260,    0,    0,    0,   27,   62,   63,
    0,   28,    0,  147,  147,  147,  147,   65,    0,    0,
   30,    0,    0,    0,    0,    0,    0,   32,    0,    0,
    0,    0,   33,    0,   72,   73,   34,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,   36,    0,
   37,   75,    0,    0,   38,    0,    0,   77,    0,   79,
    0,   81,   39,   40,  261,    0,   41,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,   90,   91,   92,  262,  686,    0,    0,    0,    0,
    0,    0,    0,    0,   96,    0,    0,    0,    0,    0,
   98,   99,  100,  101,    0,    0,    0,  102,    0,  103,
    0,   57,   24,    0,   25,  104,  105,   26,  260,    0,
    0,    0,   27,   62,   63,    0,   28,    0,    0,    0,
    0,    0,   65,    0,    0,   30,    0,    0,    0,    0,
  106,  264,   32,  109,    0,    0,    0,   33,    0,   72,
   73,   34,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,   36,    0,   37,   75,    0,    0,   38,
    0,    0,   77,    0,   79,    0,   81,   39,   40,  261,
    0,   41,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,   90,   91,   92,  262,
  352,    0,    0,    0,    0,    0,    0,    0,    0,   96,
    0,    0,    0,    0,    0,   98,   99,  100,  101,    0,
    0,    0,  102,    0,  103,    0,   57,   24,    0,   25,
  104,  105,   26,  260,    0,    0,    0,   27,   62,   63,
    0,   28,    0,    0,    0,    0,    0,   65,    0,    0,
   30,    0,    0,    0,   51,  106,  264,   32,  109,    0,
    0,    0,   33,    0,   72,   73,   34,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,   51,   36,    0,
   37,   75,    0,    0,   38,    0,    0,   77,    0,   79,
   51,   81,   39,   40,  261,   51,   41,    0,    0,   85,
   51,    0,   51,   51,   51,   51,    0,    0,    0,    0,
   51,    0,    0,    0,   51,    0,    0,    0,    0,    0,
    0,   90,   91,   92,  262,  263,   51,    0,    0,   51,
    0,   51,   57,   24,   96,   25,    0,    0,   26,  260,
    0,  668,  669,   27,   62,   63,    0,   28,    0,  670,
    0,    0,    0,   65,    0,   51,   30,    0,   51,   51,
    0,    0,  221,   32,    0,    0,    0,    0,   33,    0,
   72,   73,   34,    0,    0,    0,    0,    0,    0,    0,
  106, 1037,    0,  671,   36,    0,   37,   75,    0,    0,
   38,    0,    0,   77,    0,   79,    0,   81,   39,   40,
  261,    0,   41,    0,    0,   85,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,   90,   91,   92,
  262,  263,    0,    0,    0,    0,    0,    0,    0,    0,
   96,    0,  755,    0,  755,    0,  755,  668,  669,  755,
    0,  755,  755,    0,  755,  670,  755,    0,  755,    0,
  755,  755,  755,    0,    0,    0,  755,  755,    0,    0,
    0,    0,  755,    0,  755,  755,    0,    0,    0,  755,
    0,    0,    0,  755,    0,    0,  106,  264,    0,  671,
    0,    0,    0,    0,  755,  755,    0,  755,    0,    0,
    0,  755,  755,    0,    0,    0,    0,    0,    0,  755,
  755,   57,   24,  755,   25,    0,  755,   26,  260,    0,
    0,  755,   27,   62,   63,    0,   28,    0,    0,    0,
    0,    0,   65,    0,    0,   30,    0,    0,    0,    0,
    0,    0,   32,    0,  755,  755,    0,   33,    0,   72,
   73,   34,    0,    0,    0,    0,    0,  755,    0,    0,
    0,    0,    0,   36,    0,   37,   75,    0,    0,   38,
    0,    0,   77,    0,   79,    0,   81,   39,   40,  261,
    0,   41,    0,    0,   85,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,  755,    0,
    0,    0,    0,    0,    0,    0,   90,   91,   92,  262,
  263,    0,    0,    0,    0,    0,  754,    0,  754,   96,
    0,  754,    0,  754,  754,    0,  754,    0,  754,    0,
  754,    0,  754,  754,  754,    0,    0,    0,  754,  754,
    0,    0,    0,    0,  754,    0,  754,  754,    0,    0,
    0,  754,    0,    0,    0,  754,    0,    0,    0,    0,
    0,    0,    0,    0,    0,  106,  264,  754,    0,  754,
    0,    0,    0,  754,  754,    0,    0,    0,    0,    0,
    0,  754,  754,    0,  754,  754,  754,    0,  754,  754,
    0,  754,  754,  754,  754,    0,  754,    0,  754,    0,
  754,  754,  754,    0,    0,    0,  754,  754,    0,    0,
    0,    0,  754,    0,  754,  754,   24,    0,   25,  754,
    0,   26,    0,  754, 1438,    0,   27,    0,  762,  754,
   28,    0,  763, 1439, 1440,  754,    0,  754, 1441,   30,
    0,  754,  754,    0, 1442,    0,   32,    0,    0,  754,
  754,   33,    0,  754,    0,   34,  754,    0,    0,    0,
    0,  754,    0,    0,    0,    0,    0,   36,    0,   37,
  754,    0,    0,   38,    0,    0,    0,    0,    0,    0,
    0,   39,   40,    0,   24,   41,   25,    0, 1443,   26,
    0,    0, 1438, 1444,   27,    0,  762,    0,   28,    0,
  763, 1439, 1440,    0,    0,    0, 1441,   30,    0,    0,
    0,    0, 1442,    0,   32,    0,   51,    0,   51,   33,
    0,   51,    0,   34,    0,    0,   51,    0,    0, 1445,
   51,    0,    0,    0,    0,   36,    0,   37,  754,   51,
    0,   38,    0,    0,    0,    0,   51,    0,    0,   39,
   40,   51,    0,   41,    0,   51, 1443,   51,    0,   51,
    0, 1444,    0,    0,   51,    0,    0,   51,    0,   51,
 1446,    0,    0,   51,    0,    0,   51,    0,    0,    0,
    0,   51,   51,    0,   51,   51,   51,    0,   51,   51,
    0,    0,    0,    0,   51,    0,    0,    0,   51,    0,
    0,    0,    0,    0,    0,    0,    0,   51,    0,    0,
    0,    0,    0,    0,   51,    0,   24,    0,   25,   51,
    0,   26,    0,   51,  170,   51,   27,   51,    0,    0,
   28,    0,   51,    0,    0,   51,    0,   51, 1446,   30,
    0,   51,    0,    0,   51,    0,   32,    0,    0,   51,
   51,   33,    0,   51,    0,   34,   51,  605,    0,    0,
    0,    0,    0,    0,  606,    0,    0,   36,    0,   37,
   51,    0,    0,   38,    0,    0,  607,    0,    0,    0,
    0,   39,   40,    0,    0,   41,    0,    0,  608,   52,
  170,   52,    0,    0,   52,    0,    0,    0,    0,   52,
    0,    0,    0,   52,    0,    0,    0,    0,    0,    0,
    0,    0,   52,    0,  609,    0,    0,    0,    0,   52,
    0,   51,    0,   51,   52,    0,   51,    0,   52,    0,
   52,   51,   52,    0,    0,   51,    0,   52,   51,    0,
   52,    0,   52,    0,   51,    0,   52,    0,    0,   52,
    0,   51,    0,    0,   52,   52,   51,    0,   52,    0,
   51,   52,   51,    0,   51,    0,   24,    0,   25,   51,
  610,   26,   51,    0,   51,    0,   27,    0,   51,    0,
   28,   51,    0,    0,   29,    0,   51,   51,    0,   30,
   51,    0,    0,   51,   31,    0,   32,    0,    0,    0,
    0,   33,    0,    0,    0,   34,   35,    0,    0,    0,
    0,   24,    0,   25,    0,    0,   26,   36, 1371,   37,
    0,   27,    0,   38,    0,   28,    0,    0,    0,    0,
    0,   39,   40,    0,   30,   41,    0,    0,   42,    0,
    0,   32,    0,   52,    0,    0,   33,    0, 1372,   24,
   34,   25,    0,    0,   26,    0,    0,    0,    0,   27,
    0,    0,   36,   28,   37,    0,    0,    0,   38, 1373,
    0,    0,   30,    0,    0,   51,   39,   40,    0,   32,
   41,    0,    0,   85,   33,    0,    0,    0,   34,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
   36,    0,   37,    0,    0,    0,   38,    0,   37,    0,
    0,    0,    0,    0,   39,   40,    0,    0,   41,   37,
   43,   85,    0,    0,   37,    0,    0,    0,   37,    0,
    0,   37,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,   37,   37,    0,    0,  298,   37,   37,
    0,   35,    0,    0,   37,    0,   37,   37,   37,   37,
    0,    0,   35,    0,   37,  381,    0,   35,   37,    0,
   37,   35,    0,    0,   35,    0,    0,    0,    0,    0,
   37,    0,   37,   37,    0,   37,   35,   35,    0,   37,
    0,   35,   35,    0,    0,   31,    0,   35,    0,   35,
   35,   35,   35,  331,    0,    0,    0,   35,    0,   37,
    0,   35,    0,   35,    0,    0,   37,   37,   31,    0,
    0,    0,    0,   35,    0,   31,   35,   31,   35,    0,
    0,   31,   35,    0,    0,    0,   31,    0,    0,    0,
    0,   31,    0,   31,   31,   31,   31,    0,   31,   31,
    0,   31,   35,    0,    0,   31,    0,    0,    0,   35,
   35,   31,    0,    0,    0,    0,   31,   31,    0,    0,
   31,   31,   31,   31,   31,   31,   31,    0,    0,    0,
    0,   31,    0,    0,    0,   31,    0,    0,    0,    0,
    0,    0,    0,    0,    0,   51,   31,   31,    0,    0,
   31,    0,   31,   31,   31,    0,   51,    0,    0,    0,
    0,   51,    0,    0,    0,   51,    0,    0,   51,    0,
    0,    0,    0,    0,    0,    0,   31,    0,    0,    0,
   51,   51,    0,   31,   31,   51,   51,    0,   51,    0,
    0,   51,    0,   51,   51,   51,   51,    0,    0,   51,
    0,   51,    0,    0,   51,   51,    0,   51,   51,    0,
    0,   51,    0,    0,    0,    0,    0,   51,    0,    0,
   51,    0,   51,   51,   51,    0,   51,    0,   51,   51,
    0,   51,    0,    0,   51,    0,   51,   51,   51,   51,
    0,    0,    0,    0,   51,    0,   51,    0,   51,    0,
   51,    0,    0,   39,   51,    0,    0,    0,    0,    0,
   51,    0,    0,   51,    0,   51,    0,   51,    0,   51,
    0,   51,   51,    0,    0,    0,    0,   51,    0,   51,
   51,   51,   51,    0,    0,    0,    0,   51,    0,   51,
    0,   51,    0,    0,   51,    0,   40,    0,    0,    0,
    0,    0,    0,   51,    0,    0,   51,   51,   51,    0,
    0,   51,   51,   51,    0,    0,    0,   51,    0,   51,
   51,   51,   51,    0,    0,   51,    0,   51,    0,    0,
    0,   51,   51,    0,   51,   51,   51,    0,    0,  223,
    0,   51,    0,   51,    0,    0,   51,   51,   51,    0,
    0,    0,   51,    0,    0,    0,    0,   51,    0,   51,
   51,   51,   51,    0,   51,    0,    0,   51,    0,    0,
    0,   51,   51,    0,    0,    0,    0,   51,    0,  324,
    0,    0,   51,   51,    0,    0,   51,   51,   51,   51,
   51,   51,   51,    0,    0,    0,    0,   51,    0,    0,
    0,   51,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,   51,   51,    0,    0,   51,    0,   51,  325,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,   51,    0,    0,   51,   51,
  };
  protected static readonly short [] yyCheck = {            17,
  193,   52,  306,   18,   52,  545,   17,    4,  307,  192,
  240,   20,  296,  517,  238,  355,   20,   17,   17,   60,
  495,  330,  416,    6,  386,  547,  416,  321,  594,  305,
  119,  253,  305,  162,  342,  303,  416,  601,  397, 1026,
  830, 1231,  380,   45,   78,    0,   48,  798,   59,  800,
  983,  117,  637,  119,  612,  621,   74, 1276,  256, 1354,
   78, 1379, 1031,  396,  337,  268, 1509,  256,  256,   80,
  268,   82,  256,  292,  256,   69,  256,  256,   96,  256,
 1398,  256,  325,  373,   88,   89,  380,  256,  386,  256,
  256,  256,  373,  844,   88,   89,  847, 1316, 1270,   93,
  256,  372,   17,  465,  256,  109,  256, 1228,  256,  392,
 1329,  256,  256,  357,  369,  256,  256,   62,  257,  113,
  375,   66,   67,   68,  368,   70,   71,  382,  372,  371,
   75,   76,  415,  375,   88,   89,   81,   82,  500,   84,
  430,   86,  193,  387,  162,  193,   91,   92,  142,  430,
  421,  162, 1273,  372,   17,  109,  256, 1452,  256,    0,
  256, 1333,  162,  162,  437, 1460,  715,  465,    0,  256,
  339,  720,  721,  118,  396,  344,  419,  346,  736,   17,
  349,  350,  282,  352,  353,  373,   17,  238,  363,  373,
  238,  375,   17,  372,   17, 1638,  268,  363,  339,  203,
  204,   60,  421,  344,   17,  346,  375,  372,  349,  350,
  377,  352,  353, 1182,  314,  392,  256,  373,  358,  375,
  423,  277, 1517,  368,  368,  423,  228,  372,  372,  591,
  419,  461,  382,  541,  375,  253,  460,  419,  415,  419,
  544,  259,  421,  332,  419,  422,  430,  162,  586,  203,
  204,  234,  293,  419,  594, 1573,  421, 1190,  252,  253,
   17,  430,  303,  329,  626,  269,  332,  419,   17,  631,
  632,  419,  266,  339,  430,   17,  421,  421,  296,  801,
  256,  621, 1600,  301,  302,  425,  426,  427,  428,  430,
  377,  295,  586,  857, 1612,  628, 1614,  369,  316,  162,
   17,  369,  257,  307,  889,  256,  324,  363,  326,  303,
  256,  315,  330,  256,  325,  269,  261,  321,  612,  419,
  265,  419,  294,  419,  162,  343,  344,  256,  268,  256,
  256,  162,  375,  631,  306,  358,  276,  162,  256,  162,
  334,  295,  382,  256,  377,  349,  350,  341,  256,  162,
  384,  423, 1103,  307,  256,  355,  355,  335,  256,  353,
  305,  315,  430,  419,  420, 1155,  384,  256,  256,  731,
  257,  375,  256, 1226,  417,  416,  380, 1567,  396,  397,
  256,  256,  256,  401,  402,  403,  404,  405,  406,  407,
  408,  409,  410,  411,  388,  349,  350,  430,  392,  737,
  376,  256,  396,  397,  962,  162,  628,  256,  957,  460,
  335,  415,  460,  162, 1633,  433,  257,  368, 1271,  470,
  162,  375,  470,  363,  256,  368,  377,  370,  256,  372,
  376,  272,  256,  731,  293, 1186,  277,  256,  442,  443,
  281,  368,  736,  372,  993,  162, 1665,  419,  442,  443,
  292,  369,  446,  462, 1441,  296,  382,  375,  462,  256,
  368,  415,  370,  376,  466,  467,  475,  326,  762,  973,
  472,  475,  256,  995,  376,  363,  785,  375,  421,  419,
  256,  424,  323,  371,  256,  373,  375,  375,  442,  443,
  256,  375,  341, 1244, 1345,  370,  370,  373, 1047,  510,
 1049,  342,  256, 1052, 1255,  430,  277,  339,  325,   21,
  281,  343,  369,  371,  744,  266,  910,  375,  373,  881,
  910,  370,  516,  351,  518,  543,  368,  545,  263,  547,
  910,  419,  415, 1284,  872,  803,  368,  896,  374,  422,
  372,   53,  374,  375,  376,  377,  370,  777,  770,  560,
  382,  545,  371,  262,  256,  339,  375,  416,  576,  256,
  344,  579,  346,  314, 1128,  349,  350,  512,  352,  353,
 1421,  342,  263,  430,    6, 1097,  373,  571,  872,  590,
  315,  266,  586,  672,  369,   17, 1098,  343,  424,  298,
  601,  256,   88,   89,  594,  594, 1145,  373, 1147,  264,
 1149,  373,  419,  256,  419,  373,  672,  373,  612,  341,
  628,  369, 1124,  109,  559,  343,  370,  611,  636,  637,
  343,  621,  621,  961,  315,  376,  382,  373,   60,  314,
  395,  396,   64,  627,  628,  369,  392,  339,  370, 1360,
 1361, 1205,  344,  340,  346,  430,  430,  349,  350, 1502,
  352,  353,  305,  369,  382,  959,   88,   89,  517,  415,
  325, 1382,  430,  256,  392,  363,  422,  671,  962,  392,
  379,  380,  430,  429, 1361,  294,  369,  109, 1531, 1532,
  902,  989, 1535,  677,  430, 1251,  815,  415, 1409,  371,
  369,  376,  415,  375,  422, 1548,  430,  715, 1551,  422,
  363,  429,  720,  721,  256,  723,  429,  203,  204,  363,
 1259, 1432,  305, 1566,  430,  272, 1361,  671,  994,  992,
  277,  419, 1409,  717,  281,  719,  419, 1026,  430,  371,
  162,  429,  736,  375,  728,  369, 1361, 1590,  357,  296,
  419,  375,  294, 1137,  363, 1294, 1140, 1137,  382, 1143,
 1140,  370,  770, 1143, 1361,  374,  419, 1137,  762, 1361,
 1140,  369,  803, 1143, 1409,  419,  323,  785,  387,  377,
  370,  203,  204,  269,  374,  429,  770,  636,  637,  272,
  763, 1285, 1000,  801, 1409,  342, 1361, 1361, 1602, 1603,
  808, 1100,  810, 1361,  788,  357,  430, 1346,  375,  295,
  419,  363, 1409,  296,  373,  272,  357, 1409,  370,  803,
 1114,  373,  374,  375,  272,  369, 1365, 1357,  339,  315,
 1038,  375,  430,  374,  424,  387,  837,  402,  839,  296,
  323,  849,  836,  851, 1409, 1409,  387,  269,  296,  414,
  417, 1409, 1064,  861,  855, 1659,  857,  368,  852,  343,
  419,  368,  382,  349,  350,  376,  323,  419,  339, 1199,
  377,  293,  392,  295,  868,  323,  339, 1464,  872,  910,
  343,  889,  343, 1056,  368,  307,  430,  871,  896,  375,
  339, 1478,  836,  315,  902,  415,  904,  368,  939,  321,
  835,  939,  422,  294,  326,  368,  914,  368,  852,  893,
  343,  369,  896, 1500,  377,  306,  377,  375,  902,  368,
 1194, 1251,  357,  372,  868,  909,  339,  349,  350,  415,
  305,  376,  364,  355,  256,  368,  306, 1231,  343,  256,
  372,  392,  374,  313,  377,  944,  305,  392,  343,  957,
  944,  959,  387,  375,  339,  368,  442,  443,  380,   17,
 1234,  393,  394,  368,  415,  392,  256,  975,  962,  963,
  415,  422,  980,  368,  339,  983,  949,  422,  429,  269,
  256,  413,  377,  368,  429,  993,  343,  995,  415,  421,
  998,   17,  424,  415,  416,  422,  286,  269,  370,  372,
  368,  372,   60,  368, 1137,  376,   64, 1140, 1282,  377,
  382,  368, 1224, 1478,  286, 1345, 1000,  339,  357,  963,
  442,  443,  344,  369,  346, 1447,  343,  349,  350,  375,
  352,  353, 1026,  368,   60,  374,  256, 1459,   64, 1047,
  889, 1049,  377, 1555, 1052, 1029, 1422,  374,  387,  370,
  377,  368,  306, 1429, 1038,  372, 1064,  374,  375,  313,
  377,  910, 1061,  339,  377,  382, 1447, 1061,  344,  374,
  346,  325,  377,  349,  350, 1069,  352,  353, 1459,  377,
 1064, 1301, 1026, 1297,  372, 1268,  369,  368,   20, 1097,
  373, 1421, 1100, 1367,  377,  517,  377, 1081, 1071,  416,
 1073,  256, 1075, 1087,  162, 1617, 1137,  369,  430, 1140,
  387,  373, 1143,  371,  375,  377,  377,  375,  368,  339,
  370,  382, 1130, 1475,  344, 1069,  346, 1128,  372,  349,
  350,  371,  352,  353,  256,  375,  162, 1145, 1422, 1147,
 1492, 1149, 1654,  371, 1152, 1429,  306,  375,  308, 1679,
 1680,  256, 1441,  313,  430, 1139,   88,   89,  372, 1511,
  374, 1513,  369,  369,  586,  325,  373,  368,  375,  375,
  385,  373,  594,  375,  368,  377, 1163,  109,  372,  390,
  374,  375, 1190,  377,  339,  671, 1194,  386,  382,  344,
  612,  346,  383,  384,  349,  350,  401,  352,  353,  621,
 1208, 1209,  374,  391, 1205,  377,  397,  398,  369, 1199,
 1199,  370,  373,  371,  636,  637, 1224,  375,  399,  400,
 1225,  343,  416, 1231,  371,  372, 1234, 1268,  375, 1270,
 1268,  416, 1270,  369,  339,  293,  419,  373,  369,  344,
 1224,  346,  373,  419,  349,  350,  368,  352,  353,  671,
  372, 1259,  374,  375,  256,  377, 1297, 1265,  277, 1297,
  382, 1251, 1251,  387,  388,  389,  374,  293,  326,  369,
  370,  203,  204, 1567, 1282,  430,  377,  373, 1272,  375,
  382,  377,  421,  422,  423,  424, 1294, 1295, 1137,  354,
  355, 1140, 1333,  377, 1143, 1333,  372,  355,  374,  368,
  326,  403,  404,  405,  406,  407,  408,  409,  410,  411,
  412,  256,  371,  372,  736,  374,  375,  376,  374, 1360,
 1361,  256, 1360, 1361,  373,  430,  375,  377, 1272,  355,
  371,  371,  373,  373,  375,  375,  294,  269, 1346,  377,
  762, 1382, 1336,  373, 1382,  375,  375,  294,  377, 1357,
  836,  371,  375,  373,  377, 1345, 1345, 1365,  416, 1367,
  371,  343,  373,  295,  373, 1359,  852,  371, 1409,  373,
  371, 1409,  373, 1357,  377,  307,  798,  373,  800,  375,
  354,  355,  868,  315,  373,  371,  375,  373,  419,  321,
  416, 1432, 1336,  371, 1432,  373,  375, 1691,  377,  375,
  375,  377,  377,  375,  808,  377,  810,  369,  370,  415,
  416,  365,  366,  356,  836, 1359,  415,  349,  350,  365,
  366,  419,  844,  376, 1425,  847, 1420,  419, 1422, 1413,
  852, 1421, 1421,  417,  418, 1429,  419,  420,  425,  426,
  370,  377,  373,  375, 1438, 1439,  868, 1441,  380,  373,
  872,  369,  375, 1447,  382,  294,  373,  430,  373,  517,
  294,  375,  373,  364,  373, 1459,  373,  889,  369,  375,
 1464,  372, 1466,  374,  375, 1469, 1420,  963, 1486,  256,
  377,  382, 1483,  415, 1478,  429,  372,  429,  910,  256,
  294,  517,  393,  394, 1438, 1439,  294, 1441,  382,  373,
  375,  374,  382,  375, 1545,  376, 1500, 1545,  374,  419,
  442,  443,  413,  373,  375,  377,  375,  375,  375,  424,
  421, 1562, 1466,  424, 1562, 1469,  430,  373,  375,  430,
  462,  375,  368,  373, 1575, 1576,  594, 1575, 1576,  294,
  962,  963,  343,  475,  375,  373,  294, 1555,  375,  419,
  371,  419, 1553,  372,  368,  256,  376,  261,  256, 1567,
  375, 1602, 1603,  621, 1602, 1603,  256,  256,  594,  382,
  280,  256,  368,  373,  369,  343, 1584,  382,  636,  637,
  284,  373,  298, 1069,  377,  373,  375,  371,  373,  372,
  377,  377,  382,  297,  375,  621,  375,  301,  302,  371,
  373,  424,  373,  307, 1026,  309,  310,  311,  312, 1617,
  636,  637,  373,  317,  256,  347,  382,  321, 1659,  368,
  382, 1659,  382,  256,  256,  369,  373,  256,  347,  333,
  375,   94,  336,  371,  338,   98,   99,  100,  101,  102,
  103,  104,  105,  376,  372,  371, 1654, 1069,  372,  376,
  371,  292,  339,  373,  586,  368,  348,  369,  362,  373,
  419, 1662, 1663,  372,  375,  419,  348,  256, 1669, 1670,
  369, 1679, 1680,  368,  368,  382,  369,  377,  356,  375,
  612, 1103,  368, 1691,  337,  373,  377,  305,  372,  369,
  369,  368,  373,  369,  419, 1679, 1680,  339,  372,  370,
  372,  419,  344,  372,  346,  347,  348,  349,  350,  351,
  352,  353,  354,  355,  356, 1137,  419,  377, 1140,  382,
  372, 1143,  419,  374,  372,    0,  343,  369,  368,  371,
  377,  373,  382,  375,  376,  377,  372,  256,  372,  671,
  798,  256,  800,  370,  375,  264,  265,  373,  267,  391,
  373,  270,  271,  374,  374,  419,  275,  276,  277,  375,
  279,  375,  375,  373, 1186,  377,  285,  373,  371,  288,
  419,  373,  798,  377,  800,  419,  295, 1199,  377,  377,
  419,  300,  373,  302,  303,  304,  844,  382,  430,  847,
  373,  369,  382,  373,  371,  369,  263,  316,  315,  318,
  319,  372,  372,  322,  736,  369,  325,  382,  327,  364,
  329,  330,  331,  332,  369,  334,  373,  372,  844,  374,
  375,  847, 1244,  373,  377,    0,    0,  368,  377, 1251,
  762,  889,  369, 1255,  377,  373,  369,    0,  393,  394,
  359,  360,  361,  362,  363,  377,  309,  373,  368,  373,
 1272,  419,  910,  372,  377,  371,  369,  369,  413,  373,
  371,  368, 1284,  889,  373,  357,  421,  377,  419,  424,
  369,  363,  364, 1359,  419,  430,  368,  377,  370,  342,
  372,  373,  374,  375,  910,  373,  377,  373,  377,  352,
  373,  369,  373,    0,  369,  387,  368,  377,  369,  418,
  419,  393,  394,  369,  836,  374,  315,  377,  263,  377,
  429,  430,  377,  377, 1336,  377,  377,  377,  377,   51,
  852,  413,   12, 1345,    5, 1224, 1061,  419,  944,  421,
 1224, 1432,  424, 1437, 1420, 1382,  868, 1359, 1278, 1607,
  872, 1570, 1558, 1623, 1483, 1587, 1553,  776,  966,  412,
  413,  414, 1438, 1439,  417,  418,  419,  420,  421,  422,
  423,  424,  425,  426,  427,  428,  429,  430,  431,  432,
  961,  966,  966, 1464, 1670, 1409, 1484, 1397, 1664, 1580,
 1466, 1576,  257, 1469, 1575, 1663,  261, 1513, 1333,  939,
  321,  637, 1336, 1114,  902,  770,  989,  272, 1420, 1421,
 1422,   72,  277,  345,  803,  416,  281, 1429,  416,  284,
 1194, 1139,  944, 1143,  427,  429, 1438, 1439,  428, 1441,
  432,  296,  297,  430,    0,  431,  301,  302,  586,  872,
  962,  963,  307, 1230,  309,  310,  311,  312, 1313, 1420,
  162, 1184,  317, 1199, 1466, 1103,  321, 1469,  323, 1124,
  261, 1099, 1209, 1197, 1199, 1083, 1285,  561,  333, 1254,
  335,  336,  451,  338, 1004,  726,  451,  342, 1415, 1295,
  934,   -1,  941,  284,   -1,   -1,   -1, 1103,  541, 1137,
   -1,   -1, 1140,   -1,   -1, 1143,  297,  362,   -1,   -1,
   -1,  302,   -1,   -1, 1026,  370,  307,  261,  309,  310,
  311,  312,   -1,   -1,   -1,   -1,  317,   -1,   -1,   -1,
  321, 1137,   -1,   -1, 1140,   -1,   -1, 1143,   -1,   -1,
  284,   -1,  333,   -1,    0,  336,   -1,  338, 1186, 1061,
   -1,   -1,   -1,  297,   -1,   -1, 1194, 1069,  302,   -1,
   -1, 1199,   -1,  307,   -1,  309,  310,  311,  312,  256,
   -1,  362,   -1,  317,  617,  262,   -1,  321,   -1,   -1,
 1186,  325,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  333,
   -1,   -1,  336, 1199,  338,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1, 1244,   -1,   -1,   -1,
   -1,  298,   -1, 1251,   -1,   -1,   -1, 1255,  362,   -1,
   -1,   -1,   -1,    0,   -1,  668,  669,   -1,  419,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 1244,   -1,
   -1,   -1,   -1,  686,   -1, 1251, 1284,   -1,   -1, 1255,
   -1,   -1,  339,   -1,   -1,   -1,  343,  344,   -1,  346,
  347,  348,  349,  350,  351,  352,  353,  354,  355,  356,
  357,   -1,   -1,   -1,   -1,  419,  363,   -1, 1284,   -1,
   -1,  368,  369,  370,  371,  372,  373,  374,  375,  376,
  377,   -1,  379,  380,   -1,   -1,  383,  384,  385,  386,
  387,  257,   -1,  390,  391,  261,   -1, 1345,  395,  396,
  397,  398,  399,  400,  401,  402,  272,   -1,   -1,   -1,
   -1,  277,   -1,   -1,   -1,  281,   -1,  414,  284,   -1,
  417,   -1,  419,   -1,  421,   -1,   -1,  424,   -1, 1345,
  296,  297,   -1,  430,   -1,  301,  302,   -1,   -1,   -1,
   -1,  307,   -1,  309,  310,  311,  312,   -1,   -1,   -1,
 1272,  317,   -1,   -1,  807,  321,   -1,  323,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  333,   -1,   -1,
  336,   -1,  338, 1421,   -1,   -1,  342,   -1,   -1,   -1,
   -1,   -1,    0,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  256,  257,   -1,   -1,   -1,  261,  362,   -1,   -1,  265,
   -1,  267,   -1,   -1,  270, 1421,  272,  273,   -1,  275,
   -1,  277,   -1,  279, 1336,  281,  282,  283,  284,   -1,
   -1,  287,  288,   -1,   -1,   -1,   -1,  293,   -1,  295,
  296,  297,   -1,   -1,  300,  301,  302, 1359,  304,   -1,
   -1,  307,   -1,  309,  310,  311,  312,   -1,   -1,   -1,
  316,  317,  318,   -1,   -1,  321,  322,  323,   -1,   -1,
   -1,   -1,   -1,   -1,  330,  331,   -1,  333,  334,  256,
  336,  337,  338,   -1,  261,  262,  342,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  362,  284, 1420,   -1,
 1422,   -1,  368,  369,  370,   -1,   -1, 1429,   -1,   -1,
  297,  298,  378,   -1,   -1,  302, 1438, 1439,  305, 1441,
  307,   -1,  309,  310,  311,  312,   -1,   -1,   -1,   -1,
  317,   -1,   -1,   -1,  321,   -1,  989,   -1,  325,   -1,
   -1,   -1,   -1,   -1, 1466,   -1,  333, 1469,   -1,  336,
    0,  338,  339,  419,   -1,   -1,  343,  344,   -1,  346,
  347,  348,  349,  350,  351,  352,  353,  354,  355,  356,
   -1,   -1,   -1,   -1,   -1,  362,  363,  364,   -1,   -1,
 1033,  368,  369,   -1,  371,  372,  373,  374,  375,  376,
  377,   -1,  379,  380,   -1,  382,  383,  384,  385,  386,
  387,  388,  389,  390,  391,   -1,  393,  394,  395,  396,
  397,  398,  399,  400,  401,  402,  403,  404,  405,  406,
  407,  408,  409,  410,  411,  412,  413,  414,   -1,   -1,
  417,   -1,  419,   -1,  421,   -1,   -1,  424,  256,  257,
   -1,   -1,   -1,  430,   -1,   -1,  264,  265,  266,  267,
  268,   -1,  270,  271,   -1,  273,  274,  275,  276,  277,
  278,  279,  280,   -1,   -1,   -1,   -1,  285,   -1,  287,
  288,  289,  290,  291,  292,   -1,   -1,  295,   -1,   -1,
    0,  299,  300,   -1,  302,  303,  304,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  314,   -1,  316,   -1,
  318,  319,   -1,   -1,  322,   -1,  324,  325,  326,  327,
  328,  329,  330,  331,  332,  333,  334,  335,   -1,  337,
   -1,   -1,  340,  341,   -1,   -1,  344,  345,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  359,  360,  361,  362,  363,   -1,   -1,   -1,   -1,
  368,  369,   -1,   -1,  372,   -1,   -1,   -1,   -1,  377,
  378,  379,  380,  381,   -1,   -1,   -1,  385,   -1,  387,
   -1,   -1,   -1,   -1,   -1,  393,  394,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  418,  419,  420,  421,   -1,  423,  256,  257,   -1,   -1,
   -1,   -1,  430,   -1,  264,  265,  266,  267,  268,   -1,
  270,  271,   -1,  273,  274,  275,  276,  277,  278,  279,
    0,   -1,   -1,   -1,   -1,  285,   -1,  287,  288,  289,
  290,  291,  292,   -1,   -1,  295,   -1,   -1,   -1,  299,
  300,   -1,  302,  303,  304,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  314,   -1,  316,   -1,  318,  319,
   -1,   -1,  322,   -1,  324,  325,  326,  327,  328,  329,
  330,  331,  332,  333,  334,  335,   -1,  337,   -1,   -1,
  340,  341,   -1,   -1,  344,  345,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  359,
  360,  361,  362,  363,   -1,   -1,   -1,   -1,  368,  369,
   -1,   -1,  372,   -1,   -1,   -1,   -1,  377,  378,  379,
  380,  381,   -1,   -1,   -1,  385,  256,  387,   -1,   -1,
   -1,  261,  262,  393,  394,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,    0,   -1,   -1,  284,   -1,   -1,   -1,  418,  419,
  420,  421,   -1,  423,   -1,   -1,   -1,  297,  298,   -1,
  430,   -1,  302,   -1,   -1,  305,   -1,  307,   -1,  309,
  310,  311,  312,   -1,   -1,   -1,   -1,  317,   -1,   -1,
   -1,  321,   -1,   -1,   -1,  325,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  333,   -1,   -1,  336,   -1,  338,  339,
   -1,   -1,   -1,  343,  344,   -1,  346,  347,  348,  349,
  350,  351,  352,  353,  354,  355,  356,  357,   -1,   -1,
   -1,   -1,  362,  363,   -1,   -1,   -1,   -1,  368,  369,
  370,  371,  372,  373,  374,  375,  376,  377,   -1,  379,
  380,   -1,   -1,  383,  384,  385,  386,  387,   -1,   -1,
  390,  391,   -1,   -1,   -1,  395,  396,  397,  398,  399,
  400,  401,  402,   -1,   -1,   -1,  256,   -1,   -1,   -1,
   -1,  261,  262,   -1,  414,   -1,   -1,  417,   -1,  419,
   -1,  421,   -1,   -1,  424,   -1,   -1,   -1,    0,   -1,
  430,   -1,   -1,   -1,  284,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  297,  298,   -1,
   -1,   -1,  302,   -1,   -1,  305,   -1,  307,   -1,  309,
  310,  311,  312,   -1,   -1,   -1,   -1,  317,   -1,   -1,
   -1,  321,   -1,   -1,   -1,  325,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  333,   -1,   -1,  336,   -1,  338,  339,
   -1,   -1,   -1,  343,  344,   -1,  346,  347,  348,  349,
  350,  351,  352,  353,  354,  355,  356,   -1,   -1,   -1,
   -1,   -1,  362,  363,   -1,   -1,   -1,   -1,  368,  369,
  370,  371,  372,  373,   -1,  375,  376,  377,   -1,  379,
  380,   -1,   -1,  383,  384,  385,  386,  256,   -1,   -1,
  390,  391,  261,  262,   -1,  395,  396,  397,  398,  399,
  400,  401,  402,    0,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  414,  284,   -1,  417,   -1,  419,
   -1,  421,   -1,   -1,  424,   -1,   -1,   -1,  297,  298,
  430,   -1,   -1,  302,   -1,   -1,  305,   -1,  307,   -1,
  309,  310,  311,  312,   -1,   -1,   -1,   -1,  317,   -1,
   -1,   -1,  321,   -1,   -1,   -1,  325,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  333,   -1,   -1,  336,   -1,  338,
  339,   -1,   -1,   -1,  343,  344,   -1,  346,  347,  348,
  349,  350,  351,  352,  353,  354,  355,  356,   -1,   -1,
   -1,   -1,   -1,  362,  363,   -1,   -1,   -1,   -1,  368,
  369,  370,  371,  372,  373,   -1,  375,  376,  377,   -1,
  379,  380,   -1,   -1,  383,  384,  385,  386,   -1,   -1,
   -1,  390,  391,   -1,   -1,   -1,  395,  396,  397,  398,
  399,  400,  401,  402,  256,    0,   -1,   -1,   -1,  261,
  262,   -1,   -1,   -1,   -1,  414,   -1,   -1,  417,   -1,
  419,   -1,  421,   -1,   -1,  424,   -1,   -1,   -1,   -1,
   -1,  430,  284,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  297,  298,   -1,    0,   -1,
  302,   -1,   -1,  305,   -1,  307,   -1,  309,  310,  311,
  312,   -1,   -1,   -1,   -1,  317,   -1,   -1,   -1,  321,
   -1,   -1,   -1,  325,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  333,   -1,   -1,  336,   -1,  338,  339,   -1,   -1,
   -1,  343,  344,   -1,  346,  347,  348,  349,  350,  351,
  352,  353,  354,  355,  356,   -1,   -1,   -1,   -1,   -1,
  362,  363,   -1,   -1,   -1,   -1,  368,  369,   -1,  371,
  372,  373,   -1,  375,  376,  377,   -1,  379,  380,  256,
    0,  383,  384,  385,  386,  262,  261,   -1,  390,  391,
   -1,   -1,   -1,  395,  396,  397,  398,  399,  400,  401,
  402,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  284,
   -1,   -1,  414,   -1,   -1,  417,   -1,  419,   -1,   -1,
   -1,  298,  297,    0,   -1,   -1,   -1,  302,  430,   -1,
   -1,   -1,  307,   -1,  309,  310,  311,  312,   -1,   -1,
   -1,   -1,  317,   -1,   -1,   -1,  321,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  333,   -1,
   -1,  336,  339,  338,   -1,   -1,  343,  344,   -1,  346,
  347,  348,  349,  350,  351,  352,  353,  354,  355,  356,
   -1,   -1,   -1,   -1,   -1,   -1,  363,  362,   -1,   -1,
   -1,  368,  369,  370,  371,  372,  373,   -1,  375,  376,
  377,   -1,  379,  380,   -1,    0,  383,  384,  385,  386,
   -1,   -1,  257,  390,  391,   -1,  261,   -1,  395,  396,
  397,  398,  399,  400,  401,  402,   -1,  272,   -1,   -1,
   -1,   -1,  277,   -1,   -1,   -1,  281,  414,   -1,  284,
  417,   -1,  419,   -1,  421,   -1,   -1,  424,    0,   -1,
   -1,  296,  297,  430,   -1,  257,  301,  302,   -1,  261,
   -1,   -1,  307,   -1,  309,  310,  311,  312,   -1,   -1,
  272,   -1,  317,   -1,   -1,  277,  321,   -1,  323,  281,
   -1,   -1,  284,   -1,   -1,   -1,   -1,   -1,  333,   -1,
  335,  336,   -1,  338,  296,  297,   -1,  342,   -1,  301,
  302,   -1,   -1,   -1,   -1,  307,   -1,  309,  310,  311,
  312,    0,   -1,   -1,   -1,  317,   -1,  362,   -1,  321,
   -1,  323,   -1,   -1,  369,  370,   -1,   -1,   -1,   -1,
   -1,  333,   -1,   -1,  336,   -1,  338,  257,   -1,   -1,
  342,  261,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  272,   -1,    0,   -1,   -1,  277,   -1,   -1,
  362,  281,   -1,   -1,  284,   -1,  368,  369,  370,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  296,  297,   -1,   -1,
  257,  301,  302,   -1,  261,   -1,   -1,  307,   -1,  309,
  310,  311,  312,   -1,   -1,  272,   -1,  317,   -1,   -1,
  277,  321,   -1,  323,  281,   -1,   -1,  284,   -1,   -1,
   -1,   -1,   -1,  333,   -1,   -1,  336,    0,  338,  296,
  297,   -1,  342,   -1,  301,  302,   -1,   -1,   -1,   -1,
  307,   -1,  309,  310,  311,  312,   -1,   -1,   -1,   -1,
  317,   -1,  362,   -1,  321,   -1,  323,   -1,   -1,  369,
  370,   -1,   -1,   -1,   -1,   -1,  333,   -1,   -1,  336,
    0,  338,  257,   -1,   -1,  342,  261,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  272,   -1,   -1,
   -1,   -1,  277,   -1,   -1,  362,  281,   -1,   -1,  284,
   -1,   -1,  369,  370,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  296,  297,   -1,   -1,  257,  301,  302,   -1,  261,
   -1,   -1,  307,   -1,  309,  310,  311,  312,   -1,   -1,
  272,   -1,  317,   -1,   -1,  277,  321,   -1,  323,  281,
   -1,   -1,  284,   -1,   -1,   -1,   -1,   -1,  333,   -1,
   -1,  336,   -1,  338,  296,  297,   -1,  342,   -1,  301,
  302,   -1,   -1,   -1,   -1,  307,   -1,  309,  310,  311,
  312,   -1,   -1,   -1,   -1,  317,   -1,  362,  257,  321,
   -1,  323,  261,   -1,   -1,  370,   -1,   -1,   -1,   -1,
   -1,  333,   -1,  272,  336,   -1,  338,   -1,  277,   -1,
  342,   -1,  281,   -1,   -1,  284,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  296,  297,   -1,
  362,  257,  301,  302,   -1,  261,   -1,   -1,  307,   -1,
  309,  310,  311,  312,   -1,   -1,  272,   -1,  317,   -1,
   -1,  277,  321,   -1,  323,  281,   -1,   -1,  284,   -1,
   -1,   -1,   -1,   -1,  333,   -1,   -1,  336,   -1,  338,
  296,  297,   -1,  342,   -1,  301,  302,   -1,   -1,   -1,
   -1,  307,   -1,  309,  310,  311,  312,   -1,   -1,   -1,
   -1,  317,   -1,  362,  257,  321,   -1,  323,  261,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  333,   -1,  272,
  336,   -1,  338,   -1,  277,   -1,  342,   -1,  281,   -1,
   -1,  284,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  296,  297,   -1,  362,  257,  301,  302,
   -1,  261,   -1,   -1,  307,   -1,  309,  310,  311,  312,
   -1,   -1,  272,   -1,  317,   -1,   -1,  277,  321,   -1,
  323,  281,   -1,   -1,  284,   -1,   -1,   -1,   -1,   -1,
  333,   -1,   -1,  336,   -1,  338,  296,  297,   -1,  342,
   -1,  301,  302,   -1,   -1,   -1,   -1,  307,   -1,  309,
  310,  311,  312,   -1,   -1,   -1,   -1,  317,   -1,  362,
   -1,  321,   -1,  323,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  333,   -1,  256,  336,   -1,  338,   -1,
   -1,   -1,  342,  264,  265,  266,  267,   -1,   -1,  270,
  271,   -1,  273,  274,  275,  276,  277,  278,  279,   -1,
   -1,   -1,  362,   -1,  285,   -1,  287,  288,  289,  290,
  291,  292,   -1,   -1,  295,   -1,   -1,   -1,  299,  300,
   -1,  302,  303,  304,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  314,   -1,  316,   -1,  318,  319,   -1,
   -1,  322,   -1,  324,  325,  326,  327,  328,  329,  330,
  331,  332,  333,  334,  335,   -1,  337,   -1,   -1,  340,
  341,   -1,   -1,  344,  345,  256,   -1,   -1,   -1,   -1,
   -1,  262,   -1,   -1,   -1,   -1,   -1,   -1,  359,  360,
  361,  362,  363,   -1,   -1,   -1,   -1,  368,   -1,   -1,
   -1,  372,   -1,   -1,   -1,   -1,  377,  378,  379,  380,
  381,   -1,   -1,   -1,  385,   -1,  387,  298,   -1,   -1,
   -1,   -1,  393,  394,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  418,  419,  420,
  421,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  339,  430,
   -1,   -1,   -1,  344,  256,  346,  347,  348,  349,  350,
  351,  352,  353,  354,  355,  356,  357,   -1,   -1,   -1,
   -1,   -1,  363,  364,   -1,   -1,   -1,  368,  369,  370,
  371,  372,  373,  374,  375,  376,  377,   -1,  379,  380,
   -1,  382,  383,  384,  385,  386,  387,  388,  389,  390,
  391,   -1,  393,  394,  395,  396,  397,  398,  399,  400,
  401,  402,  403,  404,  405,  406,  407,  408,  409,  410,
  411,  412,  413,  414,  415,   -1,  256,   -1,  419,   -1,
  421,  422,  262,  424,   -1,   -1,   -1,  339,   -1,  430,
   -1,   -1,  344,   -1,  346,  347,  348,  349,  350,  351,
  352,  353,  354,  355,  356,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  369,  298,  371,
   -1,  373,   -1,  375,  376,  377,   -1,   -1,   -1,   -1,
   -1,  383,  384,  385,  386,   -1,   -1,   -1,  390,  391,
   -1,   -1,   -1,  395,  396,  397,  398,  399,  400,  401,
  402,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  339,
   -1,   -1,  414,   -1,  344,   -1,  346,  347,  348,  349,
  350,  351,  352,  353,  354,  355,  356,  357,  430,   -1,
   -1,   -1,   -1,  363,  364,   -1,   -1,   -1,  368,  369,
  370,  371,  372,  373,  374,  375,  376,  377,   -1,  379,
  380,   -1,  382,  383,  384,  385,  386,  387,  388,  389,
  390,  391,   -1,  393,  394,  395,  396,  397,  398,  399,
  400,  401,  402,  403,  404,  405,  406,  407,  408,  409,
  410,  411,  412,  413,  414,  415,  256,  256,   -1,  419,
   -1,  421,  262,   -1,  424,   -1,  265,   -1,  267,   -1,
  430,  270,   -1,   -1,   -1,   -1,  275,   -1,   -1,   -1,
  279,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  288,
   -1,   -1,   -1,   -1,   -1,   -1,  295,   -1,  298,   -1,
   -1,  300,   -1,   -1,   -1,  304,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  316,   -1,  318,
   -1,   -1,   -1,  322,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  330,  331,   -1,   -1,  334,   -1,   -1,  337,  339,
   -1,   -1,   -1,   -1,  344,   -1,  346,  347,  348,  349,
  350,  351,  352,  353,  354,  355,  356,  357,   -1,   -1,
   -1,   -1,   -1,  363,  364,   -1,   -1,   -1,  368,  369,
  370,  371,  372,  373,  374,  375,  376,  377,   -1,  379,
  380,   -1,  382,  383,  384,  385,  386,  387,  388,  389,
  390,  391,   -1,  393,  394,  395,  396,  397,  398,  399,
  400,  401,  402,  403,  404,  405,  406,  407,  408,  409,
  410,  411,  412,  413,  414,   -1,  256,  256,   -1,  419,
  419,  421,  262,   -1,  424,   -1,  265,   -1,  267,   -1,
  430,  270,   -1,   -1,   -1,   -1,  275,   -1,   -1,   -1,
  279,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  288,
   -1,   -1,   -1,   -1,   -1,   -1,  295,   -1,  298,   -1,
   -1,  300,   -1,   -1,   -1,  304,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  316,   -1,  318,
   -1,   -1,   -1,  322,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  330,  331,   -1,   -1,  334,   -1,   -1,  337,  339,
   -1,   -1,   -1,   -1,  344,   -1,  346,  347,  348,  349,
  350,  351,  352,  353,  354,  355,  356,  357,   -1,   -1,
   -1,   -1,   -1,  363,  364,   -1,   -1,   -1,   -1,  369,
  370,  371,  372,  373,  374,  375,  376,  377,   -1,  379,
  380,   -1,  382,  383,  384,  385,  386,  387,  388,  389,
  390,  391,   -1,  393,  394,  395,  396,  397,  398,  399,
  400,  401,  402,  403,  404,  405,  406,  407,  408,  409,
  410,  411,  412,  413,  414,   -1,  256,  256,   -1,  419,
  419,  421,  262,   -1,  424,   -1,  265,   -1,  267,   -1,
  430,  270,   -1,   -1,   -1,   -1,  275,   -1,   -1,   -1,
  279,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  288,
   -1,   -1,   -1,   -1,   -1,   -1,  295,   -1,  298,   -1,
   -1,  300,   -1,   -1,   -1,  304,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  316,   -1,  318,
   -1,   -1,   -1,  322,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  330,  331,   -1,   -1,  334,   -1,   -1,  337,  339,
   -1,   -1,   -1,   -1,  344,   -1,  346,  347,  348,  349,
  350,  351,  352,  353,  354,  355,  356,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  364,   -1,   -1,   -1,  368,  369,
  370,  371,  372,  373,  374,  375,  376,  377,   -1,  379,
  380,   -1,  382,  383,  384,  385,  386,  387,  388,  389,
  390,  391,   -1,  393,  394,  395,  396,  397,  398,  399,
  400,  401,  402,  403,  404,  405,  406,  407,  408,  409,
  410,  411,  412,  413,  414,  256,  256,   -1,   -1,   -1,
  419,  421,  262,   -1,  265,   -1,  267,   -1,   -1,  270,
  430,   -1,   -1,   -1,  275,   -1,   -1,   -1,  279,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  288,   -1,   -1,
   -1,   -1,   -1,   -1,  295,   -1,   -1,   -1,  298,  300,
   -1,   -1,   -1,  304,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  316,   -1,  318,   -1,   -1,
   -1,  322,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  330,
  331,   -1,   -1,  334,   -1,   -1,  337,   -1,   -1,  339,
   -1,   -1,   -1,   -1,  344,   -1,  346,  347,  348,  349,
  350,  351,  352,  353,  354,  355,  356,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  364,   -1,   -1,   -1,  368,  369,
   -1,  371,  372,  373,  374,  375,  376,  377,   -1,  379,
  380,   -1,  382,  383,  384,  385,  386,  387,  388,  389,
  390,  391,   -1,  393,  394,  395,  396,  397,  398,  399,
  400,  401,  402,  403,  404,  405,  406,  407,  408,  409,
  410,  411,  412,  413,  414,   -1,  256,  256,  419,   -1,
   -1,  421,  262,   -1,  424,   -1,  265,   -1,  267,   -1,
  430,  270,   -1,   -1,   -1,   -1,  275,   -1,   -1,   -1,
  279,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  288,
   -1,   -1,   -1,   -1,   -1,   -1,  295,   -1,  298,   -1,
   -1,  300,   -1,   -1,   -1,  304,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  316,   -1,  318,
   -1,   -1,   -1,  322,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  330,  331,   -1,   -1,  334,   -1,   -1,  337,  339,
   -1,   -1,   -1,   -1,  344,   -1,  346,  347,  348,  349,
  350,  351,  352,  353,  354,  355,  356,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  364,   -1,   -1,   -1,   -1,  369,
   -1,  371,  372,  373,  374,  375,  376,  377,   -1,  379,
  380,   -1,  382,  383,  384,  385,  386,  387,  388,  389,
  390,  391,   -1,  393,  394,  395,  396,  397,  398,  399,
  400,  401,  402,  403,  404,  405,  406,  407,  408,  409,
  410,  411,  412,  413,  414,   -1,  256,   -1,   -1,  261,
  419,  421,  262,   -1,  424,   -1,   -1,   -1,   -1,   -1,
  430,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  284,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  297,   -1,   -1,  298,   -1,
  302,   -1,   -1,  305,   -1,  307,   -1,  309,  310,  311,
  312,   -1,   -1,   -1,   -1,  317,   -1,   -1,   -1,  321,
   -1,   -1,   -1,  325,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  333,   -1,   -1,  336,   -1,  338,   -1,   -1,  339,
   -1,   -1,   -1,   -1,  344,   -1,  346,  347,  348,  349,
  350,  351,  352,  353,  354,  355,  356,   -1,   -1,   -1,
  362,   -1,   -1,   -1,  364,   -1,   -1,  369,   -1,  369,
   -1,  371,  372,  373,  374,  375,  376,  377,   -1,  379,
  380,   -1,  382,  383,  384,  385,  386,  387,  388,  389,
  390,  391,   -1,  393,  394,  395,  396,  397,  398,  399,
  400,  401,  402,  403,  404,  405,  406,  407,  408,  409,
  410,  411,  412,  413,  414,   -1,  256,  419,   -1,  261,
   -1,  421,  262,   -1,  424,   -1,   -1,   -1,   -1,   -1,
  430,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  284,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  297,   -1,   -1,  298,   -1,
  302,   -1,   -1,  305,   -1,  307,   -1,  309,  310,  311,
  312,   -1,   -1,   -1,   -1,  317,   -1,   -1,   -1,  321,
   -1,   -1,   -1,  325,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  333,   -1,   -1,  336,   -1,  338,   -1,   -1,  339,
   -1,   -1,   -1,   -1,  344,   -1,  346,  347,  348,  349,
  350,  351,  352,  353,  354,  355,  356,   -1,   -1,   -1,
  362,   -1,   -1,   -1,  364,   -1,   -1,  369,   -1,  369,
   -1,  371,  372,  373,  374,  375,  376,  377,   -1,  379,
  380,   -1,  382,  383,  384,  385,  386,  387,  388,  389,
  390,  391,   -1,  393,  394,  395,  396,  397,  398,  399,
  400,  401,  402,  403,  404,  405,  406,  407,  408,  409,
  410,  411,  412,  413,  414,   -1,  256,  419,   -1,  261,
   -1,  421,  262,   -1,  424,   -1,  266,   -1,   -1,   -1,
  430,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  284,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  297,   -1,   -1,  298,   -1,
  302,   -1,   -1,   -1,   -1,  307,   -1,  309,  310,  311,
  312,   -1,   -1,   -1,  314,  317,   -1,   -1,   -1,  321,
   -1,   -1,   -1,  325,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  333,   -1,   -1,  336,   -1,  338,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  256,   -1,   -1,   -1,   -1,   -1,  262,  357,   -1,   -1,
  362,   -1,   -1,  363,  364,   -1,   -1,  369,   -1,  369,
  370,  371,  372,  373,  374,  375,  376,  377,   -1,  379,
  380,   -1,  382,  383,  384,  385,  386,  387,  388,  389,
  390,  391,  298,  393,  394,  395,  396,  397,  398,  399,
  400,  401,  402,  403,  404,  405,  406,  407,  408,  409,
  410,  411,  412,  413,  414,   -1,   -1,  419,   -1,  419,
   -1,  421,   -1,   -1,  424,   -1,   -1,   -1,   -1,   -1,
  430,   -1,   -1,  339,   -1,  256,   -1,   -1,  344,   -1,
  346,  347,  348,  349,  350,  351,  352,  353,  354,  355,
  356,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  369,   -1,  371,   -1,  373,   -1,  375,
  376,  377,   -1,  379,  380,   -1,  382,  383,  384,  385,
  386,  387,  388,  389,  390,  391,   -1,   -1,   -1,  395,
  396,  397,  398,  399,  400,  401,  402,  403,  404,  405,
  406,  407,  408,  409,  410,  411,  412,  256,  414,   -1,
   -1,   -1,   -1,  262,   -1,   -1,   -1,   -1,  339,   -1,
   -1,   -1,   -1,  344,  430,  346,  347,  348,  349,  350,
  351,  352,  353,  354,  355,  356,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  369,  298,
  371,   -1,  373,   -1,  375,  376,  377,   -1,   -1,   -1,
   -1,   -1,  383,  384,  385,  386,  256,   -1,   -1,  390,
  391,   -1,  262,   -1,  395,  396,  397,  398,  399,  400,
  401,  402,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  339,   -1,   -1,  414,   -1,  344,   -1,  346,  347,  348,
  349,  350,  351,  352,  353,  354,  355,  356,  298,  430,
   -1,   -1,   -1,   -1,  363,  364,   -1,   -1,   -1,   -1,
  369,  370,  371,   -1,  373,  374,  375,  376,  377,   -1,
  379,  380,   -1,   -1,  383,  384,  385,  386,   -1,   -1,
   -1,  390,  391,   -1,  393,  394,  395,  396,  397,  398,
  399,  400,  401,  402,   -1,  256,   -1,   -1,   -1,   -1,
   -1,  262,   -1,   -1,  413,  414,   -1,   -1,   -1,   -1,
  419,   -1,  421,  363,  364,  424,   -1,   -1,   -1,  369,
  370,  430,  372,  373,  374,  375,   -1,  377,   -1,  379,
  380,   -1,  382,  383,  384,  385,  386,  298,  388,  389,
  390,  391,   -1,  393,  394,  395,  396,  397,  398,  399,
  400,  401,  402,  403,  404,  405,  406,  407,  408,  409,
  410,  411,  412,  413,  414,   -1,   -1,   -1,   -1,  419,
   -1,  421,   -1,   -1,  424,   -1,   -1,   -1,  339,   -1,
  430,   -1,   -1,  344,   -1,  346,  347,  348,  349,  350,
  351,  352,  353,  354,  355,  356,   -1,   -1,   -1,   -1,
  256,   -1,   -1,  364,   -1,   -1,  262,   -1,  369,   -1,
  371,  372,  373,  374,  375,  376,  377,   -1,  379,  380,
   -1,   -1,  383,  384,  385,  386,   -1,   -1,   -1,  390,
  391,   -1,  393,  394,  395,  396,  397,  398,  399,  400,
  401,  402,  298,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  413,  414,   -1,   -1,   -1,   -1,   -1,   -1,
  421,   -1,   -1,  424,   -1,   -1,   -1,   -1,   -1,  430,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  339,   -1,   -1,   -1,   -1,  344,   -1,
  346,  347,  348,  349,  350,  351,  352,  353,  354,  355,
  356,   -1,   -1,   -1,   -1,  256,   -1,   -1,  364,   -1,
   -1,  262,   -1,  369,   -1,  371,  372,  373,  374,  375,
  376,  377,   -1,  379,  380,   -1,   -1,  383,  384,  385,
  386,   -1,   -1,   -1,  390,  391,   -1,  393,  394,  395,
  396,  397,  398,  399,  400,  401,  402,  298,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  413,  414,   -1,
   -1,   -1,   -1,   -1,   -1,  421,   -1,   -1,  424,   -1,
   -1,   -1,   -1,   -1,  430,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  339,   -1,
   -1,   -1,   -1,  344,   -1,  346,  347,  348,  349,  350,
  351,  352,  353,  354,  355,  356,   -1,   -1,   -1,   -1,
  256,   -1,   -1,  364,   -1,   -1,  262,   -1,  369,   -1,
  371,  372,  373,  374,  375,  376,  377,   -1,  379,  380,
   -1,   -1,  383,  384,  385,  386,   -1,   -1,   -1,  390,
  391,   -1,  393,  394,  395,  396,  397,  398,  399,  400,
  401,  402,  298,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  413,  414,   -1,   -1,   -1,   -1,   -1,   -1,
  421,   -1,   -1,  424,   -1,   -1,   -1,   -1,   -1,  430,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  339,   -1,   -1,   -1,   -1,  344,   -1,
  346,  347,  348,  349,  350,  351,  352,  353,  354,  355,
  356,   -1,   -1,   -1,   -1,  256,   -1,  363,   -1,   -1,
   -1,  262,   -1,  369,  370,  371,   -1,  373,   -1,  375,
  376,  377,   -1,  379,  380,   -1,   -1,  383,  384,  385,
  386,   -1,   -1,   -1,  390,  391,   -1,   -1,   -1,  395,
  396,  397,  398,  399,  400,  401,  402,  298,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  414,   -1,
   -1,   -1,   -1,  419,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  430,   -1,   -1,  256,   -1,   -1,
   -1,   -1,   -1,  262,   -1,   -1,   -1,   -1,  339,   -1,
   -1,   -1,   -1,  344,   -1,  346,  347,  348,  349,  350,
  351,  352,  353,  354,  355,  356,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  369,  298,
  371,   -1,  373,   -1,  375,  376,  377,   -1,  379,  380,
   -1,   -1,  383,  384,  385,  386,  387,  388,  389,  390,
  391,   -1,   -1,   -1,  395,  396,  397,  398,  399,  400,
  401,  402,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  339,   -1,   -1,  414,   -1,  344,   -1,  346,  347,  348,
  349,  350,  351,  352,  353,  354,  355,  356,   -1,  430,
   -1,   -1,  256,   -1,  363,   -1,   -1,   -1,  262,   -1,
  369,   -1,  371,   -1,  373,   -1,  375,  376,  377,   -1,
  379,  380,   -1,   -1,  383,  384,  385,  386,   -1,   -1,
   -1,  390,  391,   -1,   -1,   -1,  395,  396,  397,  398,
  399,  400,  401,  402,  298,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  414,   -1,   -1,   -1,   -1,
  419,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  430,   -1,   -1,  256,   -1,   -1,   -1,   -1,   -1,
  262,   -1,   -1,   -1,   -1,  339,   -1,   -1,   -1,   -1,
  344,   -1,  346,  347,  348,  349,  350,  351,  352,  353,
  354,  355,  356,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  369,  298,  371,   -1,  373,
   -1,  375,  376,  377,   -1,  379,  380,   -1,   -1,  383,
  384,  385,  386,   -1,   -1,   -1,  390,  391,   -1,   -1,
   -1,  395,  396,  397,  398,  399,  400,  401,  402,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  339,   -1,   -1,
  414,   -1,  344,   -1,  346,  347,  348,  349,  350,  351,
  352,  353,  354,  355,  356,   -1,  430,   -1,   -1,  256,
   -1,   -1,   -1,   -1,   -1,  262,   -1,  369,   -1,  371,
   -1,  373,   -1,  375,  376,  377,   -1,  379,  380,   -1,
   -1,  383,  384,  385,  386,   -1,   -1,   -1,  390,  391,
   -1,   -1,   -1,  395,  396,  397,  398,  399,  400,  401,
  402,  298,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  414,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  430,   -1,
   -1,  256,   -1,   -1,   -1,   -1,   -1,  262,   -1,   -1,
   -1,   -1,  339,   -1,   -1,   -1,   -1,  344,   -1,  346,
  347,  348,  349,  350,  351,  352,  353,  354,  355,  356,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  369,  298,  371,   -1,  373,   -1,  375,  376,
  377,   -1,  379,  380,   -1,   -1,  383,  384,  385,  386,
   -1,   -1,   -1,  390,  391,   -1,   -1,   -1,  395,  396,
  397,  398,  399,  400,  401,  402,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  339,   -1,  256,  414,   -1,  344,
   -1,  346,  347,  348,  349,  350,  351,  352,  353,  354,
  355,  356,   -1,  430,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  369,   -1,  371,   -1,  373,   -1,
  375,  376,  377,   -1,  379,  380,   -1,   -1,  383,  384,
  385,  386,   -1,   -1,   -1,  390,  391,   -1,   -1,   -1,
  395,  396,  397,  398,  399,  400,  401,  402,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  256,   -1,   -1,   -1,  414,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  339,
   -1,   -1,   -1,   -1,  344,  430,  346,  347,  348,  349,
  350,  351,  352,  353,  354,  355,  356,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  369,
   -1,  371,   -1,  373,   -1,  375,  376,  377,   -1,   -1,
   -1,   -1,   -1,  383,  384,  385,  386,   -1,   -1,   -1,
  390,  391,   -1,   -1,  256,  395,  396,  397,  398,  399,
  400,  401,  402,   -1,   -1,   -1,   -1,   -1,  339,   -1,
   -1,   -1,   -1,  344,  414,  346,  347,  348,  349,  350,
  351,  352,  353,  354,  355,  356,   -1,   -1,   -1,   -1,
  430,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  369,   -1,
  371,   -1,  373,   -1,  375,  376,  377,   -1,   -1,   -1,
   -1,   -1,  383,  384,  385,  386,   -1,   -1,   -1,  390,
  391,   -1,   -1,  256,   -1,   -1,  397,  398,  399,  400,
  401,  402,   -1,   -1,   -1,   -1,   -1,  339,   -1,   -1,
   -1,   -1,  344,  414,  346,  347,  348,  349,  350,  351,
  352,  353,  354,  355,  356,   -1,   -1,   -1,   -1,  430,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  369,   -1,  371,
   -1,  373,   -1,  375,  376,  377,   -1,   -1,   -1,   -1,
   -1,  383,  384,  385,  386,   -1,   -1,   -1,  390,  391,
   -1,   -1,  256,   -1,   -1,  397,  398,  399,  400,  401,
  402,   -1,   -1,   -1,   -1,   -1,  339,   -1,   -1,   -1,
   -1,  344,  414,  346,  347,  348,  349,  350,  351,  352,
  353,  354,  355,  356,   -1,   -1,   -1,   -1,  430,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  369,   -1,  371,   -1,
  373,   -1,  375,  376,  377,   -1,   -1,   -1,   -1,   -1,
  383,  384,  385,  386,   -1,   -1,   -1,  390,  391,   -1,
   -1,  256,   -1,   -1,  397,  398,  399,  400,  401,  402,
   -1,   -1,   -1,   -1,   -1,  339,   -1,   -1,   -1,   -1,
  344,  414,  346,  347,  348,  349,  350,  351,  352,  353,
  354,  355,  356,   -1,   -1,   -1,   -1,  430,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  369,   -1,  371,   -1,  373,
   -1,  375,  376,  377,   -1,   -1,   -1,   -1,   -1,  383,
  384,  385,  386,  256,   -1,   -1,  390,  391,   -1,  262,
   -1,   -1,   -1,  397,  398,  399,  400,  401,  402,   -1,
   -1,   -1,   -1,   -1,  339,   -1,   -1,   -1,   -1,  344,
  414,  346,  347,  348,  349,  350,  351,  352,  353,  354,
  355,  356,   -1,   -1,   -1,  298,  430,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  369,   -1,  371,   -1,  373,   -1,
  375,  376,  377,   -1,   -1,   -1,   -1,   -1,  383,  384,
  385,  386,   -1,   -1,   -1,  390,  391,   -1,   -1,   -1,
   -1,   -1,  397,  398,  399,  400,  401,  402,   -1,   -1,
  256,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  414,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  364,   -1,   -1,   -1,  430,  369,   -1,   -1,  372,
   -1,  374,  375,   -1,   -1,   -1,  379,  380,   -1,   -1,
  383,  384,  385,  386,  387,  388,  389,  390,  391,   -1,
  393,  394,  395,  396,  397,  398,  399,  400,  401,  402,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  256,
  413,  414,   -1,   -1,   -1,   -1,   -1,   -1,  421,   -1,
   -1,  424,   -1,  339,   -1,   -1,   -1,  430,  344,   -1,
  346,  347,  348,  349,  350,  351,  352,  353,  354,  355,
  356,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  369,   -1,  371,   -1,  373,   -1,  375,
  376,  377,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  385,
  386,   -1,   -1,   -1,  390,  391,   -1,   -1,  256,   -1,
   -1,   -1,   -1,  399,  400,  401,  402,   -1,   -1,   -1,
   -1,   -1,  339,   -1,   -1,   -1,   -1,  344,  414,  346,
  347,  348,  349,  350,  351,  352,  353,  354,  355,  356,
   -1,   -1,   -1,   -1,  430,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  369,   -1,  371,   -1,  373,   -1,  375,  376,
  377,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  385,  386,
   -1,   -1,   -1,  390,  391,   -1,   -1,  256,   -1,   -1,
   -1,   -1,  399,  400,  401,  402,   -1,   -1,   -1,   -1,
   -1,  339,   -1,   -1,   -1,   -1,  344,  414,  346,  347,
  348,  349,  350,  351,  352,  353,  354,  355,  356,   -1,
   -1,   -1,   -1,  430,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  369,   -1,  371,   -1,  373,   -1,  375,  376,  377,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  385,  386,   -1,
   -1,   -1,  390,  391,   -1,   -1,  256,   -1,   -1,   -1,
   -1,  399,  400,  401,  402,   -1,   -1,   -1,   -1,   -1,
  339,   -1,   -1,   -1,   -1,  344,  414,  346,  347,  348,
  349,  350,  351,  352,  353,  354,  355,  356,   -1,   -1,
   -1,   -1,  430,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  369,   -1,  371,   -1,  373,   -1,  375,  376,  377,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  385,  386,   -1,   -1,
   -1,  390,  391,   -1,   -1,  256,   -1,   -1,   -1,   -1,
   -1,   -1,  401,  402,   -1,   -1,   -1,   -1,   -1,  339,
   -1,   -1,   -1,   -1,  344,  414,  346,  347,  348,  349,
  350,  351,  352,  353,  354,  355,  356,   -1,   -1,   -1,
   -1,  430,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  369,
   -1,  371,   -1,  373,   -1,  375,  376,  377,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  385,  386,   -1,   -1,   -1,
  390,  391,   -1,   -1,  256,   -1,   -1,   -1,   -1,   -1,
   -1,  401,  402,   -1,   -1,   -1,   -1,   -1,  339,   -1,
   -1,   -1,   -1,  344,  414,  346,  347,  348,  349,  350,
  351,  352,  353,  354,  355,  356,   -1,   -1,   -1,   -1,
  430,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  369,   -1,
  371,   -1,  373,   -1,  375,  376,  377,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  386,   -1,   -1,   -1,  390,
  391,   -1,   -1,  256,   -1,   -1,   -1,   -1,   -1,   -1,
  401,  402,   -1,   -1,   -1,   -1,   -1,  339,   -1,   -1,
   -1,   -1,  344,  414,  346,  347,  348,  349,  350,  351,
  352,  353,  354,  355,  356,   -1,   -1,   -1,   -1,  430,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  369,   -1,  371,
   -1,  373,   -1,  375,  376,  377,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  386,   -1,   -1,   -1,  390,  391,
   -1,   -1,  256,   -1,   -1,   -1,   -1,   -1,   -1,  401,
  402,   -1,   -1,   -1,   -1,   -1,  339,   -1,   -1,   -1,
   -1,  344,  414,  346,  347,  348,  349,  350,  351,  352,
  353,  354,  355,  356,   -1,   -1,   -1,   -1,  430,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  369,   -1,  371,   -1,
  373,   -1,  375,  376,  377,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  386,   -1,   -1,   -1,   -1,  391,   -1,
   -1,  256,   -1,   -1,   -1,   -1,   -1,   -1,  401,  402,
   -1,   -1,   -1,   -1,   -1,  339,   -1,   -1,   -1,   -1,
  344,  414,  346,  347,  348,  349,  350,  351,  352,  353,
  354,  355,  356,   -1,   -1,   -1,   -1,  430,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  369,   -1,  371,   -1,  373,
   -1,  375,  376,  377,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  386,   -1,   -1,   -1,   -1,  391,   -1,   -1,
  256,   -1,   -1,   -1,   -1,   -1,   -1,  401,  402,   -1,
   -1,   -1,   -1,   -1,  339,   -1,   -1,   -1,   -1,  344,
  414,  346,  347,  348,  349,  350,  351,  352,  353,  354,
  355,  356,   -1,   -1,   -1,   -1,  430,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  369,   -1,  371,   -1,  373,   -1,
  375,  376,  377,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  391,   -1,   -1,  256,
   -1,   -1,   -1,   -1,   -1,   -1,  401,  402,   -1,   -1,
   -1,   -1,   -1,  339,   -1,   -1,   -1,   -1,  344,  414,
  346,  347,  348,  349,  350,  351,  352,  353,  354,  355,
  356,   -1,   -1,   -1,   -1,  430,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  369,   -1,  371,   -1,  373,   -1,  375,
  376,  377,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  391,   -1,   -1,  256,   -1,
   -1,   -1,   -1,   -1,   -1,  401,  402,   -1,   -1,   -1,
   -1,   -1,  339,   -1,   -1,   -1,   -1,  344,  414,  346,
  347,  348,  349,  350,  351,  352,  353,  354,  355,  356,
   -1,   -1,   -1,   -1,  430,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  369,   -1,  371,   -1,  373,   -1,  375,  376,
  377,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  391,   -1,   -1,  256,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  402,   -1,   -1,   -1,   -1,
   -1,  339,   -1,   -1,   -1,   -1,  344,  414,  346,  347,
  348,  349,  350,  351,  352,  353,  354,  355,  356,   -1,
   -1,   -1,  256,  430,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  369,   -1,  371,   -1,  373,   -1,  375,  376,  377,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  391,  285,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  402,   -1,   -1,  298,   -1,   -1,
  339,   -1,   -1,   -1,   -1,  344,  414,  346,  347,  348,
  349,  350,  351,  352,  353,  354,  355,  356,   -1,  256,
   -1,   -1,  430,   -1,   -1,  262,  327,   -1,   -1,   -1,
  369,   -1,  371,   -1,  373,  339,  375,  376,  377,   -1,
  344,   -1,  346,  347,  348,  349,  350,  351,  352,  353,
  354,  355,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  298,   -1,   -1,   -1,  369,   -1,  371,   -1,  373,
   -1,  375,  376,  377,   -1,   -1,   -1,  378,  379,  380,
  381,   -1,  383,  384,  385,  386,  387,  388,  389,  390,
   -1,  430,  393,  394,  395,  396,  397,  398,  399,  400,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  430,  364,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  374,   -1,  376,
   -1,   -1,  379,  380,   -1,  382,  383,  384,  385,  386,
  387,  388,  389,  390,  391,   -1,  393,  394,  395,  396,
  397,  398,  399,  400,  401,  402,  403,  404,  405,  406,
  407,  408,  409,  410,  411,  412,  413,  414,   -1,   -1,
  256,   -1,   -1,   -1,  421,   -1,   -1,  424,  264,  265,
  266,  267,  268,   -1,  270,  271,   -1,  273,  274,  275,
  276,  277,  278,  279,   -1,   -1,   -1,   -1,   -1,  285,
   -1,  287,  288,  289,  290,  291,  292,   -1,   -1,  295,
   -1,   -1,   -1,  299,  300,   -1,  302,  303,  304,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  314,   -1,
  316,   -1,  318,  319,   -1,   -1,  322,   -1,  324,  325,
  326,  327,  328,  329,  330,  331,  332,  333,  334,  335,
   -1,  337,   -1,   -1,  340,  341,   -1,   -1,  344,  345,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  359,  360,  361,  362,  363,   -1,   -1,
   -1,   -1,  368,   -1,   -1,   -1,  372,   -1,   -1,   -1,
   -1,  377,  378,  379,  380,  381,   -1,   -1,   -1,  385,
   -1,  387,   -1,   -1,   -1,   -1,   -1,  393,  394,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  256,   -1,  418,  419,  420,  421,   -1,  423,  264,  265,
  266,  267,   -1,   -1,  270,  271,   -1,  273,  274,  275,
  276,  277,  278,  279,   -1,   -1,   -1,   -1,   -1,  285,
   -1,  287,  288,  289,  290,  291,  292,   -1,   -1,  295,
   -1,   -1,   -1,  299,  300,   -1,  302,  303,  304,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  314,   -1,
  316,   -1,  318,  319,   -1,   -1,  322,   -1,  324,  325,
  326,  327,  328,  329,  330,  331,  332,  333,  334,  335,
   -1,  337,   -1,   -1,  340,  341,   -1,   -1,  344,  345,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  359,  360,  361,  362,  363,   -1,   -1,
   -1,   -1,  368,   -1,   -1,   -1,  372,   -1,   -1,   -1,
   -1,  377,  378,  379,  380,  381,   -1,   -1,   -1,  385,
   -1,  387,   -1,   -1,   -1,   -1,   -1,  393,  394,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  256,   -1,
   -1,   -1,  418,  419,  420,  421,  264,  265,  266,  267,
   -1,   -1,  270,  271,   -1,  273,  274,  275,  276,  277,
  278,  279,   -1,   -1,   -1,   -1,   -1,  285,   -1,  287,
  288,  289,  290,  291,  292,   -1,   -1,  295,   -1,   -1,
   -1,  299,  300,   -1,  302,  303,  304,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  314,   -1,  316,   -1,
  318,  319,   -1,   -1,  322,   -1,  324,  325,  326,  327,
  328,  329,  330,  331,  332,  333,  334,  335,   -1,  337,
   -1,   -1,  340,  341,   -1,   -1,  344,  345,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  359,  360,  361,  362,  363,   -1,   -1,   -1,   -1,
  368,   -1,   -1,   -1,  372,   -1,   -1,   -1,   -1,  377,
  378,  379,  380,  381,   -1,   -1,   -1,  385,   -1,  387,
   -1,   -1,   -1,   -1,   -1,  393,  394,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  256,   -1,   -1,   -1,
  418,  419,  420,  421,  264,  265,  266,  267,   -1,   -1,
  270,  271,   -1,  273,  274,  275,  276,  277,  278,  279,
   -1,   -1,   -1,   -1,   -1,  285,   -1,  287,  288,  289,
  290,  291,  292,   -1,   -1,  295,   -1,   -1,   -1,  299,
  300,   -1,  302,  303,  304,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  314,   -1,  316,   -1,  318,  319,
   -1,   -1,  322,   -1,  324,  325,  326,  327,  328,  329,
  330,  331,  332,  333,  334,  335,   -1,  337,   -1,   -1,
  340,  341,   -1,   -1,  344,  345,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  359,
  360,  361,  362,  363,   -1,   -1,   -1,   -1,  368,   -1,
   -1,   -1,  372,   -1,   -1,   -1,   -1,  377,  378,  379,
  380,  381,   -1,   -1,   -1,  385,   -1,  387,   -1,   -1,
   -1,   -1,   -1,  393,  394,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  256,   -1,   -1,   -1,  418,  419,
  420,  421,  264,  265,  266,  267,   -1,   -1,  270,  271,
   -1,  273,  274,  275,  276,  277,  278,  279,   -1,   -1,
   -1,   -1,   -1,  285,   -1,  287,  288,  289,  290,  291,
  292,   -1,   -1,  295,   -1,   -1,   -1,  299,  300,   -1,
  302,  303,  304,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  314,   -1,  316,   -1,  318,  319,   -1,   -1,
  322,   -1,  324,  325,  326,  327,  328,  329,  330,  331,
  332,  333,  334,  335,   -1,  337,   -1,   -1,  340,  341,
   -1,   -1,  344,  345,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  359,  360,  361,
  362,  363,   -1,   -1,   -1,   -1,  368,   -1,   -1,   -1,
  372,   -1,   -1,   -1,   -1,  377,  378,  379,  380,  381,
   -1,   -1,   -1,  385,   -1,  387,   -1,   -1,   -1,   -1,
   -1,  393,  394,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  256,   -1,   -1,   -1,  418,  419,  420,  421,
  264,  265,   -1,  267,   -1,   -1,  270,  271,   -1,   -1,
   -1,  275,  276,  277,   -1,  279,   -1,   -1,  265,   -1,
  267,  285,   -1,  270,  288,   -1,   -1,   -1,  275,   -1,
   -1,  295,  279,   -1,   -1,   -1,  300,   -1,  302,  303,
  304,  288,  306,   -1,   -1,   -1,   -1,   -1,  295,  313,
   -1,   -1,  316,  300,  318,  319,   -1,  304,  322,   -1,
   -1,  325,   -1,  327,   -1,  329,  330,  331,  332,  316,
  334,  318,   -1,   -1,   -1,  322,   -1,  341,   -1,   -1,
  344,  345,   -1,  330,  331,   -1,   -1,  334,   -1,   -1,
  337,   -1,   -1,   -1,   -1,  359,  360,  361,  362,  363,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  372,   -1,
   -1,  375,   -1,   -1,  378,  379,  380,  381,   -1,   -1,
   -1,  385,   -1,  387,   -1,   -1,  373,   -1,   -1,  393,
  394,   -1,   -1,   -1,   -1,   -1,   -1,  264,  265,   -1,
  267,   -1,   -1,  270,  271,   -1,   -1,   -1,  275,  276,
  277,   -1,  279,   -1,  418,  419,  420,  421,  285,   -1,
   -1,  288,   -1,   -1,   -1,   -1,   -1,   -1,  295,   -1,
   -1,   -1,  419,  300,   -1,  302,  303,  304,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  316,
   -1,  318,  319,   -1,   -1,  322,   -1,   -1,  325,   -1,
  327,   -1,  329,  330,  331,  332,   -1,  334,   -1,   -1,
   -1,   -1,   -1,   -1,  341,   -1,   -1,  344,  345,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  359,  360,  361,  362,  363,   -1,   -1,   -1,
   -1,  368,  369,   -1,   -1,  372,   -1,   -1,   -1,   -1,
   -1,  378,  379,  380,  381,   -1,   -1,   -1,  385,   -1,
  387,   -1,   -1,   -1,   -1,   -1,  393,  394,   -1,   -1,
   -1,   -1,   -1,   -1,  264,  265,   -1,  267,   -1,   -1,
  270,  271,   -1,   -1,   -1,  275,  276,  277,   -1,  279,
   -1,  418,  419,  420,  421,  285,   -1,  424,  288,   -1,
   -1,   -1,  429,   -1,   -1,  295,   -1,   -1,   -1,   -1,
  300,   -1,  302,  303,  304,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  316,   -1,  318,  319,
   -1,   -1,  322,   -1,   -1,  325,   -1,  327,   -1,  329,
  330,  331,  332,   -1,  334,   -1,   -1,   -1,   -1,   -1,
   -1,  341,   -1,   -1,  344,  345,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  359,
  360,  361,  362,  363,   -1,   -1,   -1,   -1,  368,   -1,
   -1,   -1,  372,   -1,   -1,   -1,   -1,   -1,  378,  379,
  380,  381,   -1,   -1,   -1,  385,   -1,  387,   -1,   -1,
   -1,   -1,   -1,  393,  394,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  256,  418,  419,
  420,  421,   -1,   -1,  424,  264,  265,   -1,  267,  429,
   -1,  270,  271,   -1,   -1,   -1,  275,  276,  277,   -1,
  279,   -1,   -1,  265,   -1,  267,  285,   -1,  270,  288,
   -1,   -1,   -1,  275,   -1,   -1,  295,  279,   -1,   -1,
   -1,  300,   -1,  302,  303,  304,  288,   -1,   -1,   -1,
   -1,   -1,   -1,  295,   -1,   -1,   -1,  316,  300,  318,
  319,   -1,  304,  322,   -1,   -1,  325,   -1,  327,   -1,
  329,  330,  331,  332,  316,  334,  318,   -1,   -1,   -1,
  322,   -1,  341,   -1,   -1,  344,  345,   -1,  330,  331,
   -1,   -1,  334,   -1,   -1,  337,   -1,   -1,   -1,   -1,
  359,  360,  361,  362,  363,   -1,   -1,   -1,   -1,  368,
  369,   -1,   -1,  372,   -1,   -1,   -1,   -1,   -1,  378,
  379,  380,  381,   -1,   -1,   -1,  385,   -1,  387,  371,
   -1,   -1,   -1,   -1,  393,  394,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  256,   -1,   -1,   -1,  418,
  419,  420,  421,  264,  265,   -1,  267,   -1,   -1,  270,
  271,   -1,   -1,   -1,  275,  276,  277,  419,  279,   -1,
   -1,  265,   -1,  267,  285,   -1,  270,  288,   -1,   -1,
   -1,  275,   -1,   -1,  295,  279,   -1,   -1,   -1,  300,
   -1,  302,  303,  304,  288,  306,   -1,   -1,   -1,   -1,
   -1,  295,  313,   -1,   -1,  316,  300,  318,  319,   -1,
  304,  322,   -1,   -1,  325,   -1,  327,   -1,  329,  330,
  331,  332,  316,  334,  318,   -1,   -1,   -1,  322,   -1,
  341,   -1,   -1,  344,  345,   -1,  330,  331,   -1,   -1,
  334,   -1,   -1,  337,   -1,   -1,   -1,   -1,  359,  360,
  361,  362,  363,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  372,   -1,   -1,   -1,   -1,   -1,  378,  379,  380,
  381,   -1,   -1,   -1,  385,   -1,  387,   -1,   -1,   -1,
   -1,   -1,  393,  394,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  256,   -1,   -1,   -1,  418,  419,  420,
  421,  264,  265,   -1,  267,   -1,   -1,  270,  271,   -1,
   -1,   -1,  275,  276,  277,  419,  279,   -1,   -1,  265,
   -1,  267,  285,   -1,  270,  288,   -1,   -1,   -1,  275,
   -1,   -1,  295,  279,   -1,   -1,   -1,  300,   -1,  302,
  303,  304,  288,   -1,   -1,   -1,   -1,   -1,   -1,  295,
   -1,   -1,   -1,  316,  300,  318,  319,  320,  304,  322,
   -1,   -1,  325,   -1,  327,   -1,  329,  330,  331,  332,
  316,  334,  318,   -1,   -1,   -1,  322,   -1,  341,   -1,
   -1,  344,  345,   -1,  330,  331,   -1,   -1,  334,   -1,
   -1,  337,   -1,   -1,   -1,   -1,  359,  360,  361,  362,
  363,   -1,   -1,   -1,   -1,  368,   -1,   -1,   -1,  372,
   -1,   -1,   -1,   -1,   -1,  378,  379,  380,  381,   -1,
   -1,   -1,  385,   -1,  387,   -1,   -1,   -1,   -1,   -1,
  393,  394,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  256,   -1,   -1,   -1,  418,  419,  420,  421,  264,
  265,   -1,  267,   -1,   -1,  270,  271,   -1,   -1,   -1,
  275,  276,  277,  419,  279,   -1,   -1,  265,   -1,  267,
  285,   -1,  270,  288,   -1,   -1,   -1,  275,   -1,   -1,
  295,  279,   -1,   -1,   -1,  300,   -1,  302,  303,  304,
  288,   -1,   -1,   -1,   -1,   -1,   -1,  295,   -1,   -1,
   -1,  316,  300,  318,  319,   -1,  304,  322,   -1,   -1,
  325,   -1,  327,   -1,  329,  330,  331,  332,  316,  334,
  318,   -1,  337,   -1,  322,   -1,  341,   -1,   -1,  344,
  345,   -1,  330,  331,   -1,   -1,  334,   -1,   -1,  337,
   -1,   -1,   -1,   -1,  359,  360,  361,  362,  363,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  372,   -1,   -1,
   -1,   -1,   -1,  378,  379,  380,  381,   -1,   -1,   -1,
  385,   -1,  387,   -1,   -1,   -1,   -1,   -1,  393,  394,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  256,
   -1,   -1,   -1,  418,  419,  420,  421,  264,  265,   -1,
  267,   -1,   -1,  270,  271,   -1,   -1,   -1,  275,  276,
  277,  419,  279,   -1,   -1,   -1,   -1,   -1,  285,   -1,
   -1,  288,   -1,   -1,   -1,   -1,   -1,   -1,  295,   -1,
   -1,   -1,   -1,  300,   -1,  302,  303,  304,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  316,
   -1,  318,  319,   -1,   -1,  322,   -1,   -1,  325,   -1,
  327,   -1,  329,  330,  331,  332,   -1,  334,   -1,   -1,
   -1,   -1,   -1,   -1,  341,   -1,   -1,  344,  345,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  359,  360,  361,  362,  363,   -1,   -1,   -1,
   -1,  368,   -1,   -1,   -1,  372,   -1,   -1,   -1,   -1,
   -1,  378,  379,  380,  381,   -1,   -1,   -1,  385,   -1,
  387,   -1,   -1,   -1,   -1,   -1,  393,  394,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  256,
   -1,  418,  419,  420,  421,  262,   -1,  264,  265,   -1,
  267,   -1,   -1,  270,  271,   -1,   -1,   -1,  275,  276,
  277,   -1,  279,   -1,   -1,   -1,   -1,   -1,  285,   -1,
   -1,  288,   -1,   -1,   -1,   -1,   -1,   -1,  295,   -1,
   -1,  298,   -1,  300,   -1,  302,  303,  304,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  316,
   -1,  318,  319,   -1,   -1,  322,   -1,   -1,  325,   -1,
  327,   -1,  329,  330,  331,  332,   -1,  334,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  343,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  359,  360,  361,  362,  363,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  372,   -1,  374,   -1,   -1,
   -1,  378,  379,  380,  381,  382,   -1,   -1,  385,   -1,
  387,  256,   -1,   -1,   -1,   -1,  393,  394,   -1,  264,
  265,   -1,  267,   -1,   -1,  270,  271,   -1,   -1,   -1,
  275,  276,  277,   -1,  279,   -1,   -1,   -1,   -1,   -1,
  285,  418,  419,  288,  421,   -1,   -1,   -1,   -1,   -1,
  295,   -1,   -1,   -1,   -1,  300,   -1,  302,  303,  304,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  316,   -1,  318,  319,   -1,   -1,  322,   -1,   -1,
  325,   -1,  327,   -1,  329,  330,  331,  332,   -1,  334,
   -1,   -1,   -1,   -1,   -1,   -1,  341,   -1,   -1,  344,
  345,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  359,  360,  361,  362,  363,   -1,
   -1,   -1,   -1,   -1,  369,   -1,   -1,  372,   -1,   -1,
   -1,   -1,   -1,  378,  379,  380,  381,   -1,   -1,   -1,
  385,   -1,  387,   -1,   -1,   -1,   -1,   -1,  393,  394,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  256,
   -1,   -1,   -1,  418,  419,  420,  421,  264,  265,   -1,
  267,   -1,   -1,  270,  271,   -1,   -1,   -1,  275,  276,
  277,   -1,  279,   -1,   -1,  265,   -1,  267,  285,   -1,
  270,  288,   -1,   -1,   -1,  275,   -1,   -1,  295,  279,
   -1,   -1,   -1,  300,   -1,  302,  303,  304,  288,   -1,
   -1,   -1,   -1,   -1,   -1,  295,   -1,   -1,   -1,  316,
  300,  318,  319,   -1,  304,  322,   -1,   -1,  325,   -1,
  327,   -1,  329,  330,  331,  332,  316,  334,  318,   -1,
   -1,   -1,  322,   -1,  341,   -1,   -1,  344,  345,   -1,
  330,  331,   -1,   -1,  334,   -1,   -1,  337,   -1,   -1,
   -1,   -1,  359,  360,  361,  362,  363,   -1,   -1,   -1,
   -1,   -1,  369,   -1,   -1,  372,   -1,   -1,   -1,   -1,
   -1,  378,  379,  380,  381,   -1,   -1,   -1,  385,   -1,
  387,   -1,   -1,   -1,   -1,   -1,  393,  394,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  256,   -1,   -1,
   -1,  418,  419,  420,  421,  264,  265,   -1,  267,   -1,
   -1,  270,  271,   -1,   -1,   -1,  275,  276,  277,  419,
  279,   -1,   -1,  265,   -1,  267,  285,   -1,  270,  288,
   -1,   -1,   -1,  275,   -1,   -1,  295,  279,   -1,   -1,
   -1,  300,   -1,  302,  303,  304,  288,   -1,   -1,   -1,
   -1,   -1,   -1,  295,   -1,   -1,   -1,  316,  300,  318,
  319,   -1,  304,  322,   -1,   -1,  325,   -1,  327,   -1,
  329,  330,  331,  332,  316,  334,  318,   -1,   -1,   -1,
  322,   -1,  341,   -1,   -1,  344,  345,   -1,  330,  331,
   -1,   -1,  334,   -1,   -1,  337,   -1,   -1,   -1,   -1,
  359,  360,  361,  362,  363,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  372,   -1,   -1,   -1,   -1,   -1,  378,
  379,  380,  381,   -1,   -1,   -1,  385,   -1,  387,   -1,
   -1,   -1,   -1,   -1,  393,  394,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  256,   -1,   -1,   -1,  418,
  419,  420,  421,  264,  265,   -1,  267,   -1,   -1,  270,
  271,   -1,   -1,   -1,  275,  276,  277,  419,  279,   -1,
   -1,   -1,   -1,   -1,  285,   -1,   -1,  288,   -1,   -1,
   -1,   -1,   -1,   -1,  295,   -1,   -1,   -1,   -1,  300,
   -1,  302,  303,  304,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  316,   -1,  318,  319,   -1,
   -1,  322,   -1,   -1,  325,   -1,  327,   -1,  329,  330,
  331,  332,   -1,  334,   -1,   -1,   -1,   -1,   -1,   -1,
  341,   -1,   -1,  344,  345,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  359,  360,
  361,  362,  363,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  372,   -1,   -1,   -1,   -1,   -1,  378,  379,  380,
  381,   -1,   -1,   -1,  385,   -1,  387,   -1,   -1,   -1,
   -1,   -1,  393,  394,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  256,   -1,  418,  419,  420,
  421,  262,   -1,  264,  265,   -1,  267,   -1,   -1,  270,
  271,   -1,   -1,   -1,  275,  276,  277,   -1,  279,   -1,
   -1,   -1,   -1,   -1,  285,   -1,   -1,  288,   -1,   -1,
   -1,   -1,   -1,   -1,  295,   -1,   -1,  298,   -1,  300,
   -1,  302,  303,  304,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  316,   -1,  318,  319,   -1,
   -1,  322,   -1,   -1,  325,   -1,  327,   -1,  329,  330,
  331,  332,   -1,  334,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  343,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  359,  360,
  361,  362,  363,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  372,   -1,  374,   -1,   -1,   -1,  378,  379,  380,
  381,   -1,   -1,   -1,  385,   -1,  387,  256,   -1,   -1,
   -1,   -1,  393,  394,   -1,  264,  265,   -1,  267,   -1,
   -1,  270,  271,   -1,   -1,   -1,  275,  276,  277,   -1,
  279,   -1,   -1,   -1,   -1,   -1,  285,  418,  419,  288,
  421,   -1,   -1,   -1,   -1,   -1,  295,   -1,   -1,   -1,
   -1,  300,   -1,  302,  303,  304,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  316,   -1,  318,
  319,   -1,   -1,  322,   -1,   -1,  325,   -1,  327,   -1,
  329,  330,  331,  332,   -1,  334,   -1,   -1,   -1,   -1,
   -1,   -1,  341,   -1,   -1,  344,  345,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  359,  360,  361,  362,  363,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  372,   -1,   -1,   -1,   -1,   -1,  378,
  379,  380,  381,   -1,   -1,   -1,  385,   -1,  387,   -1,
   -1,   -1,   -1,   -1,  393,  394,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  256,   -1,   -1,   -1,  418,
  419,  420,  421,  264,  265,   -1,  267,   -1,   -1,  270,
  271,   -1,   -1,   -1,  275,  276,  277,   -1,  279,   -1,
   -1,  265,   -1,  267,  285,   -1,  270,  288,   -1,   -1,
   -1,  275,   -1,   -1,  295,  279,   -1,   -1,   -1,  300,
   -1,  302,  303,  304,  288,   -1,   -1,   -1,   -1,   -1,
   -1,  295,   -1,   -1,   -1,  316,  300,  318,  319,   -1,
  304,  322,   -1,   -1,  325,   -1,  327,   -1,  329,  330,
  331,  332,  316,  334,  318,   -1,   -1,   -1,  322,   -1,
  341,   -1,   -1,  344,  345,   -1,  330,  331,   -1,   -1,
  334,   -1,   -1,  337,   -1,   -1,   -1,   -1,  359,  360,
  361,  362,  363,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  372,   -1,   -1,   -1,   -1,   -1,  378,  379,  380,
  381,   -1,   -1,   -1,  385,   -1,  387,   -1,   -1,   -1,
   -1,   -1,  393,  394,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  256,   -1,   -1,   -1,  418,  419,  420,
  421,  264,  265,   -1,  267,   -1,   -1,  270,  271,   -1,
   -1,   -1,  275,  276,  277,  419,  279,   -1,   -1,  265,
   -1,  267,  285,   -1,  270,  288,   -1,   -1,   -1,  275,
   -1,   -1,  295,  279,   -1,   -1,   -1,  300,   -1,  302,
  303,  304,  288,   -1,   -1,   -1,   -1,   -1,   -1,  295,
   -1,   -1,   -1,  316,  300,  318,  319,   -1,  304,  322,
   -1,   -1,  325,   -1,  327,   -1,  329,  330,  331,  332,
  316,  334,  318,   -1,   -1,   -1,  322,   -1,  341,   -1,
   -1,  344,  345,   -1,  330,  331,   -1,   -1,  334,   -1,
   -1,  337,   -1,   -1,   -1,   -1,  359,  360,  361,  362,
  363,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  372,
   -1,   -1,   -1,   -1,   -1,  378,  379,  380,  381,   -1,
   -1,   -1,  385,   -1,  387,   -1,   -1,   -1,   -1,   -1,
  393,  394,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  256,   -1,   -1,   -1,  418,  419,  420,  421,  264,
  265,   -1,  267,   -1,   -1,  270,  271,   -1,   -1,   -1,
  275,  276,  277,  419,  279,   -1,   -1,   -1,   -1,   -1,
  285,   -1,   -1,  288,   -1,   -1,   -1,   -1,   -1,   -1,
  295,   -1,   -1,   -1,   -1,  300,   -1,  302,  303,  304,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  316,   -1,  318,  319,   -1,   -1,  322,   -1,   -1,
  325,   -1,  327,   -1,  329,  330,  331,  332,   -1,  334,
   -1,   -1,   -1,   -1,   -1,   -1,  341,   -1,   -1,  344,
  345,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  359,  360,  361,  362,  363,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  372,   -1,   -1,
   -1,   -1,   -1,  378,  379,  380,  381,   -1,   -1,   -1,
  385,   -1,  387,   -1,   -1,   -1,   -1,   -1,  393,  394,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  256,   -1,  418,  419,  420,  421,  262,   -1,  264,
  265,   -1,  267,   -1,   -1,  270,  271,   -1,   -1,   -1,
  275,  276,  277,   -1,  279,   -1,   -1,   -1,   -1,   -1,
  285,   -1,   -1,  288,   -1,   -1,   -1,   -1,   -1,   -1,
  295,   -1,   -1,  298,   -1,  300,   -1,  302,  303,  304,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  316,   -1,  318,  319,   -1,   -1,  322,   -1,   -1,
  325,   -1,  327,   -1,  329,  330,  331,  332,   -1,  334,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  359,  360,  361,  362,  363,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  372,   -1,  374,
   -1,   -1,   -1,  378,  379,  380,  381,   -1,   -1,   -1,
  385,   -1,  387,  256,   -1,   -1,   -1,   -1,  393,  394,
   -1,  264,  265,   -1,  267,   -1,   -1,  270,  271,   -1,
   -1,   -1,  275,  276,  277,   -1,  279,   -1,   -1,   -1,
   -1,   -1,  285,  418,  419,  288,  421,   -1,   -1,   -1,
   -1,   -1,  295,   -1,   -1,   -1,   -1,  300,   -1,  302,
  303,  304,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  316,   -1,  318,  319,   -1,   -1,  322,
   -1,   -1,  325,   -1,  327,   -1,  329,  330,  331,  332,
   -1,  334,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  359,  360,  361,  362,
  363,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  372,
   -1,  374,   -1,   -1,   -1,  378,  379,  380,  381,   -1,
   -1,   -1,  385,   -1,  387,  256,   -1,   -1,   -1,   -1,
  393,  394,   -1,  264,  265,   -1,  267,   -1,   -1,  270,
  271,   -1,   -1,   -1,  275,  276,  277,   -1,  279,   -1,
   -1,   -1,   -1,   -1,  285,  418,  419,  288,  421,   -1,
   -1,   -1,   -1,   -1,  295,   -1,   -1,   -1,   -1,  300,
   -1,  302,  303,  304,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  316,   -1,  318,  319,   -1,
   -1,  322,   -1,   -1,  325,   -1,  327,   -1,  329,  330,
  331,  332,   -1,  334,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  359,  360,
  361,  362,  363,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  372,   -1,   -1,   -1,   -1,   -1,  378,  379,  380,
  381,   -1,   -1,   -1,  385,   -1,  387,  256,   -1,   -1,
   -1,   -1,  393,  394,   -1,  264,  265,   -1,  267,   -1,
   -1,  270,  271,   -1,   -1,   -1,  275,  276,  277,   -1,
  279,   -1,   -1,   -1,   -1,   -1,  285,  418,  419,  288,
  421,   -1,   -1,   -1,   -1,   -1,  295,   -1,   -1,   -1,
   -1,  300,   -1,  302,  303,  304,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  316,   -1,  318,
  319,   -1,   -1,  322,   -1,   -1,  325,   -1,  327,   -1,
  329,  330,  331,  332,   -1,  334,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  359,  360,  361,  362,  363,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  372,   -1,   -1,   -1,   -1,   -1,  378,
  379,  380,  381,   -1,   -1,   -1,  385,   -1,  387,  256,
   -1,   -1,   -1,   -1,  393,  394,   -1,  264,  265,   -1,
  267,   -1,   -1,  270,  271,   -1,   -1,   -1,  275,  276,
  277,   -1,  279,   -1,   -1,   -1,   -1,   -1,  285,  418,
  419,  288,  421,   -1,   -1,   -1,   -1,   -1,  295,   -1,
   -1,   -1,   -1,  300,   -1,  302,  303,  304,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  316,
   -1,  318,  319,   -1,   -1,  322,   -1,   -1,  325,   -1,
  327,   -1,  329,  330,  331,  332,   -1,  334,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  359,  360,  361,  362,  363,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  372,   -1,   -1,   -1,   -1,
   -1,  378,  379,  380,  381,   -1,   -1,   -1,  385,   -1,
  387,  256,   -1,   -1,   -1,   -1,  393,  394,   -1,  264,
  265,   -1,  267,   -1,   -1,  270,  271,   -1,   -1,   -1,
  275,  276,  277,   -1,  279,   -1,   -1,   -1,   -1,   -1,
  285,  418,  419,  288,  421,   -1,   -1,   -1,   -1,   -1,
  295,   -1,   -1,   -1,   -1,  300,   -1,  302,  303,  304,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  316,   -1,  318,  319,   -1,   -1,  322,   -1,   -1,
  325,   -1,  327,   -1,  329,  330,  331,  332,   -1,  334,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  359,  360,  361,  362,  363,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  372,   -1,   -1,
   -1,   -1,   -1,  378,  379,  380,  381,   -1,   -1,   -1,
  385,   -1,  387,  256,   -1,   -1,   -1,   -1,  393,  394,
   -1,  264,  265,   -1,  267,   -1,   -1,  270,  271,   -1,
   -1,   -1,  275,  276,  277,   -1,  279,   -1,   -1,   -1,
   -1,   -1,  285,  418,  419,  288,  421,   -1,   -1,   -1,
   -1,   -1,  295,   -1,   -1,   -1,   -1,  300,   -1,  302,
  303,  304,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  316,   -1,  318,  319,   -1,   -1,  322,
   -1,   -1,  325,   -1,  327,   -1,  329,  330,  331,  332,
   -1,  334,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  359,  360,  361,  362,
  363,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  372,
   -1,   -1,   -1,   -1,   -1,  378,  379,  380,  381,   -1,
   -1,   -1,  385,   -1,  387,  256,   -1,   -1,   -1,   -1,
  393,  394,   -1,  264,  265,   -1,  267,   -1,   -1,  270,
  271,   -1,   -1,   -1,  275,  276,  277,   -1,  279,   -1,
   -1,   -1,   -1,   -1,  285,  418,  419,  288,  421,   -1,
   -1,   -1,   -1,   -1,  295,   -1,   -1,   -1,   -1,  300,
   -1,  302,  303,  304,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  316,   -1,  318,  319,   -1,
   -1,  322,   -1,   -1,  325,   -1,  327,   -1,  329,  330,
  331,  332,   -1,  334,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  359,  360,
  361,  362,  363,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  372,   -1,   -1,   -1,   -1,   -1,  378,  379,  380,
  381,   -1,   -1,   -1,  385,   -1,  387,  256,   -1,   -1,
   -1,   -1,  393,  394,   -1,  264,  265,   -1,  267,   -1,
   -1,  270,  271,   -1,   -1,   -1,  275,  276,  277,   -1,
  279,   -1,   -1,   -1,   -1,   -1,  285,  418,  419,  288,
  421,   -1,   -1,   -1,   -1,   -1,  295,   -1,   -1,   -1,
   -1,  300,   -1,  302,  303,  304,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  316,   -1,  318,
  319,   -1,   -1,  322,   -1,   -1,  325,   -1,  327,   -1,
  329,  330,  331,  332,   -1,  334,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  359,  360,  361,  362,  363,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  372,   -1,   -1,   -1,   -1,   -1,  378,
  379,  380,  381,   -1,   -1,   -1,  385,   -1,  387,  256,
   -1,   -1,   -1,   -1,  393,  394,   -1,  264,  265,   -1,
  267,   -1,   -1,  270,  271,   -1,   -1,   -1,  275,  276,
  277,   -1,  279,   -1,   -1,   -1,   -1,   -1,  285,  418,
  419,  288,  421,   -1,   -1,   -1,   -1,   -1,  295,   -1,
   -1,   -1,   -1,  300,   -1,  302,  303,  304,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  316,
   -1,  318,  319,   -1,   -1,  322,   -1,   -1,  325,   -1,
  327,   -1,  329,  330,  331,  332,   -1,  334,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  359,  360,  361,  362,  363,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  372,   -1,   -1,   -1,   -1,
   -1,  378,  379,  380,  381,   -1,   -1,   -1,  385,   -1,
  387,  256,   -1,   -1,   -1,   -1,  393,  394,   -1,  264,
  265,   -1,  267,   -1,   -1,  270,  271,   -1,   -1,   -1,
  275,  276,  277,   -1,  279,   -1,   -1,   -1,   -1,   -1,
  285,  418,  419,  288,  421,   -1,   -1,   -1,   -1,   -1,
  295,   -1,   -1,   -1,   -1,  300,   -1,  302,  303,  304,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  316,   -1,  318,  319,   -1,   -1,  322,   -1,   -1,
  325,   -1,  327,   -1,  329,  330,  331,  332,   -1,  334,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  359,  360,  361,  362,  363,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  372,   -1,   -1,
   -1,   -1,   -1,  378,  379,  380,  381,   -1,   -1,   -1,
  385,   -1,  387,  256,   -1,   -1,   -1,   -1,  393,  394,
   -1,  264,  265,   -1,  267,   -1,   -1,  270,  271,   -1,
   -1,   -1,  275,  276,  277,   -1,  279,   -1,   -1,   -1,
   -1,   -1,  285,  418,  419,  288,  421,   -1,   -1,   -1,
   -1,   -1,  295,   -1,   -1,   -1,   -1,  300,   -1,  302,
  303,  304,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  316,   -1,  318,  319,   -1,   -1,  322,
   -1,   -1,  325,   -1,  327,   -1,  329,  330,  331,  332,
   -1,  334,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  359,  360,  361,  362,
  363,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  372,
   -1,   -1,   -1,   -1,   -1,  378,  379,  380,  381,   -1,
   -1,   -1,  385,   -1,  387,  256,   -1,   -1,   -1,   -1,
  393,  394,   -1,  264,  265,   -1,  267,   -1,   -1,  270,
  271,   -1,   -1,   -1,  275,  276,  277,   -1,  279,   -1,
   -1,   -1,   -1,   -1,  285,  418,  419,  288,  421,   -1,
   -1,   -1,   -1,   -1,  295,   -1,   -1,   -1,   -1,  300,
   -1,  302,  303,  304,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  316,   -1,  318,  319,   -1,
   -1,  322,   -1,   -1,  325,   -1,  327,   -1,  329,  330,
  331,  332,   -1,  334,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  359,  360,
  361,  362,  363,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  372,   -1,   -1,   -1,   -1,   -1,  378,  379,  380,
  381,   -1,   -1,   -1,  385,   -1,  387,  256,   -1,   -1,
   -1,   -1,  393,  394,   -1,  264,  265,   -1,  267,   -1,
   -1,  270,  271,   -1,   -1,   -1,  275,  276,  277,   -1,
  279,   -1,   -1,   -1,   -1,   -1,  285,  418,  419,  288,
  421,   -1,   -1,   -1,   -1,   -1,  295,   -1,   -1,   -1,
   -1,  300,   -1,  302,  303,  304,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  316,   -1,  318,
  319,   -1,   -1,  322,   -1,   -1,  325,   -1,  327,   -1,
  329,  330,  331,  332,   -1,  334,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  359,  360,  361,  362,  363,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  372,   -1,   -1,   -1,   -1,   -1,  378,
  379,  380,  381,   -1,   -1,   -1,  385,   -1,  387,  256,
   -1,   -1,   -1,   -1,  393,  394,   -1,  264,  265,   -1,
  267,   -1,   -1,  270,  271,   -1,   -1,   -1,  275,  276,
  277,   -1,  279,   -1,   -1,   -1,   -1,   -1,  285,  418,
  419,  288,  421,   -1,   -1,   -1,   -1,   -1,  295,   -1,
   -1,   -1,   -1,  300,   -1,  302,  303,  304,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  316,
   -1,  318,  319,   -1,   -1,  322,   -1,   -1,  325,   -1,
  327,   -1,  329,  330,  331,  332,   -1,  334,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  359,  360,  361,  362,  363,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  372,   -1,   -1,   -1,   -1,
   -1,  378,  379,  380,  381,   -1,   -1,   -1,  385,   -1,
  387,  256,   -1,   -1,   -1,   -1,  393,  394,   -1,  264,
  265,   -1,  267,   -1,   -1,  270,  271,   -1,   -1,   -1,
  275,  276,  277,   -1,  279,   -1,   -1,   -1,   -1,   -1,
  285,  418,  419,  288,  421,   -1,   -1,   -1,   -1,   -1,
  295,   -1,   -1,   -1,   -1,  300,   -1,  302,  303,  304,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  316,   -1,  318,  319,   -1,   -1,  322,   -1,   -1,
  325,   -1,  327,   -1,  329,  330,  331,  332,   -1,  334,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  359,  360,  361,  362,  363,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  372,   -1,   -1,
   -1,   -1,   -1,  378,  379,  380,  381,   -1,   -1,   -1,
  385,   -1,  387,  256,   -1,   -1,   -1,   -1,  393,  394,
   -1,  264,  265,   -1,  267,   -1,   -1,  270,  271,   -1,
   -1,   -1,  275,  276,  277,   -1,  279,   -1,   -1,   -1,
   -1,   -1,  285,  418,  419,  288,  421,   -1,   -1,   -1,
   -1,   -1,  295,   -1,   -1,   -1,   -1,  300,   -1,  302,
  303,  304,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  316,   -1,  318,  319,   -1,   -1,  322,
   -1,   -1,  325,   -1,  327,   -1,  329,  330,  331,  332,
   -1,  334,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  359,  360,  361,  362,
  363,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  372,
   -1,   -1,   -1,   -1,   -1,  378,  379,  380,  381,   -1,
   -1,   -1,  385,   -1,  387,  256,   -1,   -1,   -1,   -1,
  393,  394,   -1,  264,  265,   -1,  267,   -1,   -1,  270,
  271,   -1,   -1,   -1,  275,  276,  277,   -1,  279,   -1,
   -1,   -1,   -1,   -1,  285,  418,  419,  288,  421,   -1,
   -1,   -1,   -1,   -1,  295,   -1,   -1,   -1,   -1,  300,
   -1,  302,  303,  304,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  316,   -1,  318,  319,   -1,
   -1,  322,   -1,   -1,  325,   -1,  327,   -1,  329,  330,
  331,  332,   -1,  334,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  359,  360,
  361,  362,  363,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  372,   -1,   -1,   -1,   -1,   -1,  378,  379,  380,
  381,   -1,   -1,   -1,  385,   -1,  387,  256,   -1,   -1,
   -1,   -1,  393,  394,   -1,  264,  265,   -1,  267,   -1,
   -1,  270,  271,   -1,   -1,   -1,  275,  276,  277,   -1,
  279,   -1,   -1,   -1,   -1,   -1,  285,  418,  419,  288,
  421,   -1,   -1,   -1,   -1,   -1,  295,   -1,   -1,   -1,
   -1,  300,   -1,  302,  303,  304,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  316,   -1,  318,
  319,   -1,   -1,  322,   -1,   -1,  325,   -1,  327,   -1,
  329,  330,  331,  332,   -1,  334,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  359,  360,  361,  362,  363,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  372,   -1,   -1,   -1,   -1,   -1,  378,
  379,  380,  381,   -1,   -1,   -1,  385,   -1,  387,  256,
   -1,   -1,   -1,   -1,  393,  394,   -1,  264,  265,   -1,
  267,   -1,   -1,  270,  271,   -1,   -1,   -1,  275,  276,
  277,   -1,  279,   -1,   -1,   -1,   -1,   -1,  285,  418,
  419,  288,  421,   -1,   -1,   -1,   -1,   -1,  295,   -1,
   -1,   -1,   -1,  300,   -1,  302,  303,  304,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  316,
   -1,  318,  319,   -1,   -1,  322,   -1,   -1,  325,   -1,
  327,   -1,  329,  330,  331,  332,   -1,  334,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  359,  360,  361,  362,  363,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  372,   -1,   -1,   -1,   -1,
   -1,  378,  379,  380,  381,   -1,   -1,   -1,  385,   -1,
  387,  256,   -1,   -1,   -1,   -1,  393,  394,   -1,  264,
  265,   -1,  267,   -1,   -1,  270,  271,   -1,   -1,   -1,
  275,  276,  277,   -1,  279,   -1,   -1,   -1,   -1,   -1,
  285,  418,  419,  288,  421,   -1,   -1,   -1,   -1,   -1,
  295,   -1,   -1,   -1,   -1,  300,   -1,  302,  303,  304,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  316,   -1,  318,  319,   -1,   -1,  322,   -1,   -1,
  325,   -1,  327,   -1,  329,  330,  331,  332,   -1,  334,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  359,  360,  361,  362,  363,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  372,   -1,   -1,
   -1,   -1,   -1,  378,  379,  380,  381,   -1,   -1,   -1,
  385,   -1,  387,  256,   -1,   -1,   -1,   -1,  393,  394,
   -1,  264,  265,   -1,  267,   -1,   -1,  270,  271,   -1,
   -1,   -1,  275,  276,  277,   -1,  279,   -1,   -1,   -1,
   -1,   -1,  285,  418,  419,  288,  421,   -1,   -1,   -1,
   -1,   -1,  295,   -1,   -1,   -1,   -1,  300,   -1,  302,
  303,  304,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  316,   -1,  318,  319,   -1,   -1,  322,
   -1,   -1,  325,   -1,  327,   -1,  329,  330,  331,  332,
   -1,  334,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  359,  360,  361,  362,
  363,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  372,
   -1,   -1,   -1,   -1,   -1,  378,  379,  380,  381,   -1,
   -1,   -1,  385,   -1,  387,  256,   -1,   -1,   -1,   -1,
  393,  394,   -1,  264,  265,   -1,  267,   -1,   -1,  270,
  271,   -1,   -1,   -1,  275,  276,  277,   -1,  279,   -1,
   -1,   -1,   -1,   -1,  285,  418,  419,  288,  421,   -1,
   -1,   -1,   -1,   -1,  295,   -1,   -1,   -1,   -1,  300,
   -1,  302,  303,  304,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  316,   -1,  318,  319,   -1,
   -1,  322,   -1,   -1,  325,   -1,  327,   -1,  329,  330,
  331,  332,   -1,  334,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  359,  360,
  361,  362,  363,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  372,   -1,   -1,   -1,   -1,   -1,  378,  379,  380,
  381,   -1,   -1,   -1,  385,   -1,  387,  256,   -1,   -1,
   -1,   -1,  393,  394,   -1,  264,  265,   -1,  267,   -1,
   -1,  270,  271,   -1,   -1,   -1,  275,  276,  277,   -1,
  279,   -1,   -1,   -1,   -1,   -1,  285,  418,  419,  288,
  421,   -1,   -1,   -1,   -1,   -1,  295,   -1,   -1,   -1,
   -1,  300,   -1,  302,  303,  304,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  316,   -1,  318,
  319,   -1,   -1,  322,   -1,   -1,  325,   -1,  327,   -1,
  329,  330,  331,  332,   -1,  334,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  359,  360,  361,  362,  363,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  372,   -1,   -1,   -1,   -1,   -1,  378,
  379,  380,  381,   -1,   -1,   -1,  385,   -1,  387,  256,
   -1,   -1,   -1,   -1,  393,  394,   -1,  264,  265,   -1,
  267,   -1,   -1,  270,  271,   -1,   -1,   -1,  275,  276,
  277,   -1,  279,   -1,   -1,   -1,   -1,   -1,  285,  418,
  419,  288,  421,   -1,   -1,   -1,   -1,   -1,  295,   -1,
   -1,   -1,   -1,  300,   -1,  302,  303,  304,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  316,
   -1,  318,  319,   -1,   -1,  322,   -1,   -1,  325,   -1,
  327,   -1,  329,  330,  331,  332,   -1,  334,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  359,  360,  361,  362,  363,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  372,   -1,   -1,   -1,   -1,
   -1,  378,  379,  380,  381,   -1,   -1,   -1,  385,   -1,
  387,  256,   -1,   -1,   -1,   -1,  393,  394,   -1,  264,
  265,   -1,  267,   -1,   -1,  270,  271,   -1,   -1,   -1,
  275,  276,  277,   -1,  279,   -1,   -1,   -1,   -1,   -1,
  285,  418,  419,  288,  421,   -1,   -1,   -1,   -1,   -1,
  295,   -1,   -1,   -1,   -1,  300,   -1,  302,  303,  304,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  316,   -1,  318,  319,   -1,   -1,  322,   -1,   -1,
  325,   -1,  327,   -1,  329,  330,  331,  332,   -1,  334,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  359,  360,  361,  362,  363,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  372,   -1,   -1,
   -1,   -1,   -1,  378,  379,  380,  381,   -1,   -1,   -1,
  385,   -1,  387,  256,   -1,   -1,   -1,   -1,  393,  394,
   -1,  264,  265,   -1,  267,   -1,   -1,  270,  271,   -1,
   -1,   -1,  275,  276,  277,   -1,  279,   -1,   -1,   -1,
   -1,   -1,  285,  418,  419,  288,  421,   -1,   -1,   -1,
   -1,   -1,  295,   -1,   -1,   -1,   -1,  300,   -1,  302,
  303,  304,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  316,   -1,  318,  319,   -1,   -1,  322,
   -1,   -1,  325,   -1,  327,   -1,  329,  330,  331,  332,
   -1,  334,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  359,  360,  361,  362,
  363,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  372,
   -1,   -1,   -1,   -1,   -1,  378,  379,  380,  381,   -1,
   -1,   -1,  385,   -1,  387,  256,   -1,   -1,   -1,   -1,
  393,  394,   -1,  264,  265,   -1,  267,   -1,   -1,  270,
  271,   -1,   -1,   -1,  275,  276,  277,   -1,  279,   -1,
   -1,   -1,   -1,   -1,  285,  418,  419,  288,  421,   -1,
   -1,   -1,   -1,   -1,  295,   -1,   -1,   -1,   -1,  300,
   -1,  302,  303,  304,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  316,   -1,  318,  319,   -1,
   -1,  322,   -1,   -1,  325,   -1,  327,   -1,  329,  330,
  331,  332,   -1,  334,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  359,  360,
  361,  362,  363,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  372,   -1,   -1,   -1,   -1,   -1,  378,  379,  380,
  381,   -1,   -1,   -1,  385,   -1,  387,  256,   -1,   -1,
   -1,   -1,  393,  394,   -1,  264,  265,   -1,  267,   -1,
   -1,  270,  271,   -1,   -1,   -1,  275,  276,  277,   -1,
  279,   -1,   -1,   -1,   -1,   -1,  285,  418,  419,  288,
  421,   -1,   -1,   -1,   -1,   -1,  295,   -1,   -1,   -1,
   -1,  300,   -1,  302,  303,  304,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  316,   -1,  318,
  319,   -1,   -1,  322,   -1,   -1,  325,   -1,  327,   -1,
  329,  330,  331,  332,   -1,  334,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  359,  360,  361,  362,  363,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  372,   -1,   -1,   -1,   -1,   -1,  378,
  379,  380,  381,   -1,   -1,   -1,  385,   -1,  387,  256,
   -1,   -1,   -1,   -1,  393,  394,   -1,  264,  265,   -1,
  267,   -1,   -1,  270,  271,   -1,   -1,   -1,  275,  276,
  277,   -1,  279,   -1,   -1,   -1,   -1,   -1,  285,  418,
  419,  288,  421,   -1,   -1,  261,   -1,  263,  295,   -1,
   -1,   -1,   -1,  300,   -1,  302,  303,  304,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  284,  316,
   -1,  318,  319,   -1,   -1,  322,   -1,   -1,  325,   -1,
  327,  297,  329,  330,  331,  332,  302,  334,   -1,   -1,
  337,  307,   -1,  309,  310,  311,  312,   -1,   -1,  315,
   -1,  317,   -1,   -1,   -1,  321,   -1,   -1,   -1,   -1,
   -1,   -1,  359,  360,  361,  362,  363,  333,   -1,   -1,
  336,   -1,  338,   -1,   -1,  372,   -1,   -1,   -1,   -1,
   -1,   -1,  379,  380,   -1,   -1,  256,   -1,   -1,   -1,
  387,  261,   -1,   -1,   -1,  265,  362,  267,   -1,   -1,
  270,   -1,  272,  273,   -1,  275,   -1,  277,   -1,  279,
   -1,  281,  282,  283,  284,   -1,   -1,  287,  288,   -1,
   -1,  418,  419,  293,  421,  295,  296,  297,   -1,   -1,
  300,   -1,  302,   -1,  304,   -1,   -1,  307,   -1,  309,
  310,  311,  312,   -1,   -1,   -1,  316,  317,  318,   -1,
   -1,  321,  322,  323,   -1,   -1,   -1,   -1,   -1,   -1,
  330,  331,   -1,  333,  334,  256,  336,  337,  338,   -1,
   -1,   -1,  342,  264,  265,   -1,  267,   -1,   -1,  270,
  271,   -1,   -1,   -1,  275,  276,  277,   -1,  279,   -1,
   -1,   -1,  362,   -1,  285,   -1,   -1,  288,  368,  369,
  370,   -1,   -1,   -1,  295,   -1,   -1,   -1,  378,  300,
   -1,  302,  303,  304,   -1,   -1,   -1,   -1,   -1,  262,
   -1,   -1,   -1,   -1,   -1,  316,   -1,  318,  319,   -1,
   -1,  322,   -1,   -1,  325,   -1,  327,   -1,  329,  330,
  331,  332,   -1,  334,   -1,   -1,   -1,   -1,   -1,  419,
   -1,   -1,   -1,   -1,   -1,  298,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  359,  360,
  361,  362,  363,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  372,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  364,   -1,   -1,   -1,   -1,   -1,  418,  419,  372,
  373,  374,  375,  376,   -1,   -1,  379,  380,   -1,   -1,
  383,  384,  385,  386,  387,  388,  389,  390,  391,   -1,
  393,  394,  395,  396,  397,  398,  399,  400,  401,  402,
  403,  404,  405,  406,  407,  408,  409,  410,  411,  412,
  413,  414,   -1,  261,   -1,  263,   -1,  265,  421,  267,
   -1,  424,  270,   -1,  272,  273,   -1,  275,   -1,  277,
   -1,  279,   -1,  281,  282,  283,  284,   -1,   -1,  287,
  288,   -1,   -1,   -1,   -1,  293,  294,  295,  296,  297,
   -1,   -1,  300,   -1,  302,   -1,  304,   -1,  306,  307,
   -1,  309,  310,  311,  312,   -1,   -1,  315,  316,  317,
  318,   -1,   -1,  321,  322,  323,   -1,   -1,   -1,   -1,
   -1,   -1,  330,  331,   -1,  333,  334,   -1,  336,  337,
  338,   -1,   -1,   -1,  342,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  362,   -1,   -1,  365,  366,  261,
   -1,   -1,   -1,  265,   -1,  267,   -1,   -1,  270,   -1,
  272,  273,   -1,  275,   -1,  277,   -1,  279,   -1,  281,
  282,  283,  284,   -1,   -1,  287,  288,   -1,   -1,   -1,
   -1,  293,   -1,  295,  296,  297,   -1,   -1,  300,   -1,
  302,   -1,  304,   -1,   -1,  307,   -1,  309,  310,  311,
  312,  419,   -1,   -1,  316,  317,  318,   -1,   -1,  321,
  322,  323,   -1,   -1,   -1,   -1,   -1,   -1,  330,  331,
   -1,  333,  334,   -1,  336,  337,  338,   -1,   -1,   -1,
  342,   -1,   -1,  261,   -1,   -1,   -1,  265,   -1,  267,
   -1,   -1,  270,   -1,  272,  273,   -1,  275,   -1,  277,
  362,  279,   -1,  281,  282,  283,  284,  369,   -1,  287,
  288,   -1,   -1,   -1,   -1,  293,  378,  295,  296,  297,
   -1,   -1,  300,   -1,  302,   -1,  304,   -1,   -1,  307,
   -1,  309,  310,  311,  312,   -1,   -1,   -1,  316,  317,
  318,   -1,   -1,  321,  322,  323,   -1,   -1,   -1,   -1,
   -1,   -1,  330,  331,   -1,  333,  334,  419,  336,  337,
  338,   -1,   -1,   -1,  342,   -1,   -1,  261,   -1,   -1,
   -1,  265,   -1,  267,   -1,   -1,  270,   -1,  272,  273,
   -1,  275,   -1,  277,  362,  279,   -1,  281,  282,  283,
  284,  369,   -1,  287,  288,   -1,   -1,   -1,   -1,  293,
  378,  295,  296,  297,   -1,   -1,  300,   -1,  302,   -1,
  304,   -1,   -1,  307,   -1,  309,  310,  311,  312,   -1,
   -1,   -1,  316,  317,  318,   -1,   -1,  321,  322,  323,
   -1,   -1,   -1,   -1,   -1,   -1,  330,  331,   -1,  333,
  334,  419,  336,  337,  338,   -1,   -1,   -1,  342,   -1,
   -1,  261,   -1,   -1,   -1,  265,   -1,  267,   -1,   -1,
  270,   -1,  272,  273,   -1,  275,   -1,  277,  362,  279,
   -1,  281,  282,  283,  284,   -1,   -1,  287,  288,   -1,
   -1,   -1,   -1,  293,  378,  295,  296,  297,   -1,   -1,
  300,   -1,  302,   -1,  304,   -1,   -1,  307,   -1,  309,
  310,  311,  312,   -1,   -1,   -1,  316,  317,  318,   -1,
   -1,  321,  322,  323,   -1,   -1,   -1,   -1,   -1,   -1,
  330,  331,   -1,  333,  334,  419,  336,  337,  338,   -1,
   -1,   -1,  342,   -1,   -1,  261,   -1,   -1,   -1,  265,
   -1,  267,   -1,   -1,  270,   -1,  272,  273,   -1,  275,
   -1,  277,  362,  279,   -1,  281,  282,  283,  284,  369,
   -1,  287,  288,   -1,   -1,   -1,   -1,  293,   -1,  295,
  296,  297,   -1,   -1,  300,   -1,  302,  261,  304,   -1,
   -1,  307,   -1,  309,  310,  311,  312,   -1,   -1,   -1,
  316,  317,  318,   -1,   -1,  321,  322,  323,   -1,   -1,
  284,   -1,   -1,   -1,  330,  331,   -1,  333,  334,  419,
  336,  337,  338,  297,   -1,   -1,  342,   -1,  302,   -1,
   -1,  305,   -1,  307,   -1,  309,  310,  311,  312,   -1,
   -1,   -1,   -1,  317,   -1,   -1,  362,  321,   -1,   -1,
   -1,  325,   -1,  369,   -1,   -1,   -1,   -1,   -1,  333,
   -1,   -1,  336,   -1,  338,   -1,  264,  265,   -1,  267,
   -1,   -1,  270,  271,   -1,   -1,   -1,  275,  276,  277,
   -1,  279,   -1,  357,   -1,   -1,   -1,  285,  362,   -1,
  288,   -1,   -1,   -1,   -1,  369,  370,  295,  372,   -1,
  374,   -1,  300,  419,  302,  303,  304,   -1,  306,   -1,
   -1,   -1,   -1,  387,   -1,  313,   -1,   -1,  316,   -1,
  318,  319,   -1,   -1,  322,   -1,   -1,  325,   -1,  327,
   -1,  329,  330,  331,  332,   -1,  334,   -1,   -1,   -1,
   -1,   -1,   -1,  341,   -1,  419,  344,  345,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  359,  360,  361,  362,  363,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  372,  373,   -1,  375,   -1,   -1,
  378,  379,  380,  381,   -1,   -1,   -1,  385,   -1,  387,
   -1,   -1,   -1,   -1,   -1,  393,  394,   -1,   -1,   -1,
   -1,   -1,   -1,  264,  265,   -1,  267,   -1,   -1,  270,
  271,   -1,   -1,   -1,  275,  276,  277,   -1,  279,   -1,
  418,  419,  420,  421,  285,   -1,   -1,  288,   -1,   -1,
   -1,   -1,   -1,   -1,  295,   -1,   -1,   -1,   -1,  300,
   -1,  302,  303,  304,   -1,  306,   -1,   -1,   -1,   -1,
   -1,   -1,  313,   -1,   -1,  316,   -1,  318,  319,   -1,
   -1,  322,   -1,   -1,  325,   -1,  327,   -1,  329,  330,
  331,  332,   -1,  334,   -1,   -1,   -1,   -1,   -1,   -1,
  341,   -1,   -1,  344,  345,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  359,  360,
  361,  362,  363,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  372,   -1,   -1,  375,   -1,   -1,  378,  379,  380,
  381,   -1,   -1,   -1,  385,   -1,  387,   -1,   -1,   -1,
   -1,   -1,  393,  394,   -1,   -1,   -1,   -1,   -1,   -1,
  264,  265,   -1,  267,   -1,   -1,  270,  271,   -1,   -1,
   -1,  275,  276,  277,   -1,  279,   -1,  418,  419,  420,
  421,  285,   -1,   -1,  288,   -1,   -1,   -1,   -1,   -1,
   -1,  295,   -1,   -1,   -1,   -1,  300,   -1,  302,  303,
  304,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  316,   -1,  318,  319,   -1,   -1,  322,   -1,
   -1,  325,   -1,  327,   -1,  329,  330,  331,  332,   -1,
  334,   -1,   -1,  337,   -1,   -1,   -1,  341,   -1,   -1,
  344,  345,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  359,  360,  361,  362,  363,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  372,   -1,
   -1,   -1,   -1,   -1,  378,  379,  380,  381,   -1,   -1,
   -1,  385,   -1,  387,   -1,   -1,   -1,   -1,   -1,  393,
  394,   -1,   -1,   -1,   -1,   -1,   -1,  264,  265,   -1,
  267,   -1,   -1,  270,  271,   -1,   -1,   -1,  275,  276,
  277,   -1,  279,   -1,  418,  419,  420,  421,  285,   -1,
   -1,  288,   -1,   -1,   -1,   -1,   -1,   -1,  295,   -1,
   -1,   -1,   -1,  300,   -1,  302,  303,  304,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  316,
   -1,  318,  319,   -1,   -1,  322,   -1,   -1,  325,   -1,
  327,   -1,  329,  330,  331,  332,   -1,  334,   -1,   -1,
  337,   -1,   -1,   -1,  341,   -1,   -1,  344,  345,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  359,  360,  361,  362,  363,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  372,   -1,   -1,   -1,   -1,
   -1,  378,  379,  380,  381,   -1,   -1,   -1,  385,   -1,
  387,   -1,   -1,   -1,   -1,   -1,  393,  394,   -1,   -1,
   -1,   -1,   -1,   -1,  264,  265,   -1,  267,   -1,   -1,
  270,  271,   -1,   -1,   -1,  275,  276,  277,   -1,  279,
   -1,  418,  419,  420,  421,  285,   -1,   -1,  288,   -1,
   -1,   -1,   -1,   -1,   -1,  295,   -1,   -1,   -1,   -1,
  300,   -1,  302,  303,  304,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  316,   -1,  318,  319,
   -1,   -1,  322,   -1,   -1,  325,   -1,  327,   -1,  329,
  330,  331,  332,   -1,  334,   -1,   -1,   -1,   -1,   -1,
   -1,  341,   -1,   -1,  344,  345,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  359,
  360,  361,  362,  363,   -1,   -1,   -1,   -1,  368,   -1,
   -1,   -1,  372,   -1,   -1,   -1,   -1,   -1,  378,  379,
  380,  381,   -1,   -1,   -1,  385,   -1,  387,   -1,   -1,
   -1,   -1,   -1,  393,  394,   -1,   -1,   -1,   -1,   -1,
   -1,  264,  265,   -1,  267,   -1,   -1,  270,  271,   -1,
   -1,   -1,  275,  276,  277,   -1,  279,   -1,  418,  419,
  420,  421,  285,   -1,   -1,  288,   -1,   -1,   -1,   -1,
   -1,   -1,  295,   -1,   -1,   -1,   -1,  300,   -1,  302,
  303,  304,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  316,   -1,  318,  319,   -1,   -1,  322,
   -1,   -1,  325,   -1,  327,   -1,  329,  330,  331,  332,
   -1,  334,   -1,   -1,   -1,   -1,   -1,   -1,  341,   -1,
   -1,  344,  345,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  359,  360,  361,  362,
  363,   -1,   -1,   -1,   -1,  368,   -1,   -1,   -1,  372,
   -1,   -1,   -1,   -1,   -1,  378,  379,  380,  381,   -1,
   -1,   -1,  385,   -1,  387,   -1,   -1,   -1,   -1,   -1,
  393,  394,   -1,   -1,   -1,   -1,   -1,   -1,  264,  265,
   -1,  267,   -1,   -1,  270,  271,   -1,   -1,   -1,  275,
  276,  277,   -1,  279,   -1,  418,  419,  420,  421,  285,
   -1,   -1,  288,   -1,   -1,   -1,   -1,   -1,   -1,  295,
   -1,   -1,   -1,   -1,  300,   -1,  302,  303,  304,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  316,   -1,  318,  319,   -1,   -1,  322,   -1,   -1,  325,
   -1,  327,   -1,  329,  330,  331,  332,   -1,  334,   -1,
   -1,   -1,   -1,   -1,   -1,  341,   -1,   -1,  344,  345,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  359,  360,  361,  362,  363,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  372,   -1,   -1,   -1,
   -1,   -1,  378,  379,  380,  381,   -1,   -1,   -1,  385,
   -1,  387,   -1,   -1,   -1,   -1,   -1,  393,  394,   -1,
   -1,   -1,   -1,   -1,   -1,  264,  265,   -1,  267,   -1,
   -1,  270,  271,   -1,   -1,   -1,  275,  276,  277,   -1,
  279,   -1,  418,  419,  420,  421,  285,   -1,   -1,  288,
   -1,   -1,   -1,   -1,   -1,   -1,  295,   -1,   -1,   -1,
   -1,  300,   -1,  302,  303,  304,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  316,   -1,  318,
  319,   -1,   -1,  322,   -1,   -1,  325,   -1,  327,   -1,
  329,  330,  331,  332,   -1,  334,   -1,   -1,   -1,   -1,
   -1,   -1,  341,   -1,   -1,  344,  345,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  359,  360,  361,  362,  363,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  372,   -1,   -1,   -1,   -1,   -1,  378,
  379,  380,  381,   -1,   -1,   -1,  385,   -1,  387,   -1,
   -1,   -1,   -1,   -1,  393,  394,   -1,   -1,   -1,   -1,
   -1,   -1,  264,  265,   -1,  267,   -1,   -1,  270,  271,
   -1,   -1,   -1,  275,  276,  277,   -1,  279,   -1,  418,
  419,  420,  421,  285,   -1,   -1,  288,   -1,   -1,   -1,
   -1,   -1,   -1,  295,   -1,   -1,   -1,   -1,  300,   -1,
  302,  303,  304,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  316,   -1,  318,  319,   -1,   -1,
  322,   -1,   -1,  325,   -1,  327,   -1,  329,  330,  331,
  332,   -1,  334,   -1,   -1,   -1,   -1,   -1,   -1,  341,
   -1,   -1,  344,  345,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  359,  360,  361,
  362,  363,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  372,   -1,   -1,   -1,   -1,   -1,  378,  379,  380,  381,
   -1,   -1,   -1,  385,   -1,  387,   -1,   -1,   -1,   -1,
   -1,  393,  394,   -1,   -1,   -1,   -1,   -1,   -1,  264,
  265,   -1,  267,   -1,   -1,  270,  271,   -1,   -1,   -1,
  275,  276,  277,   -1,  279,   -1,  418,  419,  420,  421,
  285,   -1,   -1,  288,   -1,   -1,   -1,   -1,   -1,   -1,
  295,   -1,   -1,   -1,   -1,  300,   -1,  302,  303,  304,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  316,   -1,  318,  319,   -1,   -1,  322,   -1,   -1,
  325,   -1,  327,   -1,  329,  330,  331,  332,   -1,  334,
   -1,   -1,   -1,   -1,   -1,   -1,  341,   -1,   -1,  344,
  345,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  359,  360,  361,  362,  363,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  372,   -1,   -1,
   -1,   -1,   -1,  378,  379,  380,  381,   -1,   -1,   -1,
  385,   -1,  387,   -1,   -1,   -1,   -1,   -1,  393,  394,
   -1,   -1,   -1,   -1,   -1,   -1,  264,  265,   -1,  267,
   -1,   -1,  270,  271,   -1,   -1,   -1,  275,  276,  277,
   -1,  279,   -1,  418,  419,  420,  421,  285,   -1,   -1,
  288,   -1,   -1,   -1,   -1,   -1,   -1,  295,   -1,   -1,
   -1,   -1,  300,   -1,  302,  303,  304,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  316,   -1,
  318,  319,   -1,   -1,  322,   -1,   -1,  325,   -1,  327,
   -1,  329,  330,  331,  332,   -1,  334,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  359,  360,  361,  362,  363,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  372,   -1,   -1,   -1,   -1,   -1,
  378,  379,  380,  381,   -1,   -1,   -1,  385,   -1,  387,
   -1,  264,  265,   -1,  267,  393,  394,  270,  271,   -1,
   -1,   -1,  275,  276,  277,   -1,  279,   -1,   -1,   -1,
   -1,   -1,  285,   -1,   -1,  288,   -1,   -1,   -1,   -1,
  418,  419,  295,  421,   -1,   -1,   -1,  300,   -1,  302,
  303,  304,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  316,   -1,  318,  319,   -1,   -1,  322,
   -1,   -1,  325,   -1,  327,   -1,  329,  330,  331,  332,
   -1,  334,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  359,  360,  361,  362,
  363,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  372,
   -1,   -1,   -1,   -1,   -1,  378,  379,  380,  381,   -1,
   -1,   -1,  385,   -1,  387,   -1,  264,  265,   -1,  267,
  393,  394,  270,  271,   -1,   -1,   -1,  275,  276,  277,
   -1,  279,   -1,   -1,   -1,   -1,   -1,  285,   -1,   -1,
  288,   -1,   -1,   -1,  261,  418,  419,  295,  421,   -1,
   -1,   -1,  300,   -1,  302,  303,  304,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  284,  316,   -1,
  318,  319,   -1,   -1,  322,   -1,   -1,  325,   -1,  327,
  297,  329,  330,  331,  332,  302,  334,   -1,   -1,  337,
  307,   -1,  309,  310,  311,  312,   -1,   -1,   -1,   -1,
  317,   -1,   -1,   -1,  321,   -1,   -1,   -1,   -1,   -1,
   -1,  359,  360,  361,  362,  363,  333,   -1,   -1,  336,
   -1,  338,  264,  265,  372,  267,   -1,   -1,  270,  271,
   -1,  379,  380,  275,  276,  277,   -1,  279,   -1,  387,
   -1,   -1,   -1,  285,   -1,  362,  288,   -1,  365,  366,
   -1,   -1,  369,  295,   -1,   -1,   -1,   -1,  300,   -1,
  302,  303,  304,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  418,  419,   -1,  421,  316,   -1,  318,  319,   -1,   -1,
  322,   -1,   -1,  325,   -1,  327,   -1,  329,  330,  331,
  332,   -1,  334,   -1,   -1,  337,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  359,  360,  361,
  362,  363,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  372,   -1,  263,   -1,  265,   -1,  267,  379,  380,  270,
   -1,  272,  273,   -1,  275,  387,  277,   -1,  279,   -1,
  281,  282,  283,   -1,   -1,   -1,  287,  288,   -1,   -1,
   -1,   -1,  293,   -1,  295,  296,   -1,   -1,   -1,  300,
   -1,   -1,   -1,  304,   -1,   -1,  418,  419,   -1,  421,
   -1,   -1,   -1,   -1,  315,  316,   -1,  318,   -1,   -1,
   -1,  322,  323,   -1,   -1,   -1,   -1,   -1,   -1,  330,
  331,  264,  265,  334,  267,   -1,  337,  270,  271,   -1,
   -1,  342,  275,  276,  277,   -1,  279,   -1,   -1,   -1,
   -1,   -1,  285,   -1,   -1,  288,   -1,   -1,   -1,   -1,
   -1,   -1,  295,   -1,  365,  366,   -1,  300,   -1,  302,
  303,  304,   -1,   -1,   -1,   -1,   -1,  378,   -1,   -1,
   -1,   -1,   -1,  316,   -1,  318,  319,   -1,   -1,  322,
   -1,   -1,  325,   -1,  327,   -1,  329,  330,  331,  332,
   -1,  334,   -1,   -1,  337,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  419,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  359,  360,  361,  362,
  363,   -1,   -1,   -1,   -1,   -1,  265,   -1,  267,  372,
   -1,  270,   -1,  272,  273,   -1,  275,   -1,  277,   -1,
  279,   -1,  281,  282,  283,   -1,   -1,   -1,  287,  288,
   -1,   -1,   -1,   -1,  293,   -1,  295,  296,   -1,   -1,
   -1,  300,   -1,   -1,   -1,  304,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  418,  419,  316,   -1,  318,
   -1,   -1,   -1,  322,  323,   -1,   -1,   -1,   -1,   -1,
   -1,  330,  331,   -1,  265,  334,  267,   -1,  337,  270,
   -1,  272,  273,  342,  275,   -1,  277,   -1,  279,   -1,
  281,  282,  283,   -1,   -1,   -1,  287,  288,   -1,   -1,
   -1,   -1,  293,   -1,  295,  296,  265,   -1,  267,  300,
   -1,  270,   -1,  304,  273,   -1,  275,   -1,  277,  378,
  279,   -1,  281,  282,  283,  316,   -1,  318,  287,  288,
   -1,  322,  323,   -1,  293,   -1,  295,   -1,   -1,  330,
  331,  300,   -1,  334,   -1,  304,  337,   -1,   -1,   -1,
   -1,  342,   -1,   -1,   -1,   -1,   -1,  316,   -1,  318,
  419,   -1,   -1,  322,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  330,  331,   -1,  265,  334,  267,   -1,  337,  270,
   -1,   -1,  273,  342,  275,   -1,  277,   -1,  279,   -1,
  281,  282,  283,   -1,   -1,   -1,  287,  288,   -1,   -1,
   -1,   -1,  293,   -1,  295,   -1,  265,   -1,  267,  300,
   -1,  270,   -1,  304,   -1,   -1,  275,   -1,   -1,  378,
  279,   -1,   -1,   -1,   -1,  316,   -1,  318,  419,  288,
   -1,  322,   -1,   -1,   -1,   -1,  295,   -1,   -1,  330,
  331,  300,   -1,  334,   -1,  304,  337,  306,   -1,  308,
   -1,  342,   -1,   -1,  313,   -1,   -1,  316,   -1,  318,
  419,   -1,   -1,  322,   -1,   -1,  325,   -1,   -1,   -1,
   -1,  330,  331,   -1,  265,  334,  267,   -1,  337,  270,
   -1,   -1,   -1,   -1,  275,   -1,   -1,   -1,  279,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  288,   -1,   -1,
   -1,   -1,   -1,   -1,  295,   -1,  265,   -1,  267,  300,
   -1,  270,   -1,  304,  373,  306,  275,  308,   -1,   -1,
  279,   -1,  313,   -1,   -1,  316,   -1,  318,  419,  288,
   -1,  322,   -1,   -1,  325,   -1,  295,   -1,   -1,  330,
  331,  300,   -1,  334,   -1,  304,  337,  306,   -1,   -1,
   -1,   -1,   -1,   -1,  313,   -1,   -1,  316,   -1,  318,
  419,   -1,   -1,  322,   -1,   -1,  325,   -1,   -1,   -1,
   -1,  330,  331,   -1,   -1,  334,   -1,   -1,  337,  265,
  371,  267,   -1,   -1,  270,   -1,   -1,   -1,   -1,  275,
   -1,   -1,   -1,  279,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  288,   -1,  363,   -1,   -1,   -1,   -1,  295,
   -1,  265,   -1,  267,  300,   -1,  270,   -1,  304,   -1,
  306,  275,  308,   -1,   -1,  279,   -1,  313,  419,   -1,
  316,   -1,  318,   -1,  288,   -1,  322,   -1,   -1,  325,
   -1,  295,   -1,   -1,  330,  331,  300,   -1,  334,   -1,
  304,  337,  306,   -1,  308,   -1,  265,   -1,  267,  313,
  419,  270,  316,   -1,  318,   -1,  275,   -1,  322,   -1,
  279,  325,   -1,   -1,  283,   -1,  330,  331,   -1,  288,
  334,   -1,   -1,  337,  293,   -1,  295,   -1,   -1,   -1,
   -1,  300,   -1,   -1,   -1,  304,  305,   -1,   -1,   -1,
   -1,  265,   -1,  267,   -1,   -1,  270,  316,  272,  318,
   -1,  275,   -1,  322,   -1,  279,   -1,   -1,   -1,   -1,
   -1,  330,  331,   -1,  288,  334,   -1,   -1,  337,   -1,
   -1,  295,   -1,  419,   -1,   -1,  300,   -1,  302,  265,
  304,  267,   -1,   -1,  270,   -1,   -1,   -1,   -1,  275,
   -1,   -1,  316,  279,  318,   -1,   -1,   -1,  322,  323,
   -1,   -1,  288,   -1,   -1,  419,  330,  331,   -1,  295,
  334,   -1,   -1,  337,  300,   -1,   -1,   -1,  304,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  316,   -1,  318,   -1,   -1,   -1,  322,   -1,  261,   -1,
   -1,   -1,   -1,   -1,  330,  331,   -1,   -1,  334,  272,
  419,  337,   -1,   -1,  277,   -1,   -1,   -1,  281,   -1,
   -1,  284,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  296,  297,   -1,   -1,  363,  301,  302,
   -1,  261,   -1,   -1,  307,   -1,  309,  310,  311,  312,
   -1,   -1,  272,   -1,  317,  419,   -1,  277,  321,   -1,
  323,  281,   -1,   -1,  284,   -1,   -1,   -1,   -1,   -1,
  333,   -1,  335,  336,   -1,  338,  296,  297,   -1,  342,
   -1,  301,  302,   -1,   -1,  261,   -1,  307,   -1,  309,
  310,  311,  312,  419,   -1,   -1,   -1,  317,   -1,  362,
   -1,  321,   -1,  323,   -1,   -1,  369,  370,  284,   -1,
   -1,   -1,   -1,  333,   -1,  261,  336,  263,  338,   -1,
   -1,  297,  342,   -1,   -1,   -1,  302,   -1,   -1,   -1,
   -1,  307,   -1,  309,  310,  311,  312,   -1,  284,  315,
   -1,  317,  362,   -1,   -1,  321,   -1,   -1,   -1,  369,
  370,  297,   -1,   -1,   -1,   -1,  302,  333,   -1,   -1,
  336,  307,  338,  309,  310,  311,  312,   -1,   -1,   -1,
   -1,  317,   -1,   -1,   -1,  321,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  261,  362,  333,   -1,   -1,
  336,   -1,  338,  369,  370,   -1,  272,   -1,   -1,   -1,
   -1,  277,   -1,   -1,   -1,  281,   -1,   -1,  284,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  362,   -1,   -1,   -1,
  296,  297,   -1,  369,  370,  301,  302,   -1,  261,   -1,
   -1,  307,   -1,  309,  310,  311,  312,   -1,   -1,  272,
   -1,  317,   -1,   -1,  277,  321,   -1,  323,  281,   -1,
   -1,  284,   -1,   -1,   -1,   -1,   -1,  333,   -1,   -1,
  336,   -1,  338,  296,  297,   -1,  342,   -1,  301,  302,
   -1,  261,   -1,   -1,  307,   -1,  309,  310,  311,  312,
   -1,   -1,   -1,   -1,  317,   -1,  362,   -1,  321,   -1,
  323,   -1,   -1,  369,  284,   -1,   -1,   -1,   -1,   -1,
  333,   -1,   -1,  336,   -1,  338,   -1,  297,   -1,  342,
   -1,  261,  302,   -1,   -1,   -1,   -1,  307,   -1,  309,
  310,  311,  312,   -1,   -1,   -1,   -1,  317,   -1,  362,
   -1,  321,   -1,   -1,  284,   -1,  369,   -1,   -1,   -1,
   -1,   -1,   -1,  333,   -1,   -1,  336,  297,  338,   -1,
   -1,  261,  302,  263,   -1,   -1,   -1,  307,   -1,  309,
  310,  311,  312,   -1,   -1,  315,   -1,  317,   -1,   -1,
   -1,  321,  362,   -1,  284,  365,  366,   -1,   -1,  369,
   -1,  261,   -1,  333,   -1,   -1,  336,  297,  338,   -1,
   -1,   -1,  302,   -1,   -1,   -1,   -1,  307,   -1,  309,
  310,  311,  312,   -1,  284,   -1,   -1,  317,   -1,   -1,
   -1,  321,  362,   -1,   -1,   -1,   -1,  297,   -1,  369,
   -1,   -1,  302,  333,   -1,   -1,  336,  307,  338,  309,
  310,  311,  312,   -1,   -1,   -1,   -1,  317,   -1,   -1,
   -1,  321,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  362,  333,   -1,   -1,  336,   -1,  338,  369,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  362,   -1,   -1,  365,  366,
  };

#line 7300 "cs-parser.jay"

// <summary>
//  A class used to hold info about an operator declarator
// </summary>
class OperatorDeclaration {
	public readonly Operator.OpType optype;
	public readonly FullNamedExpression ret_type;
	public readonly Location location;

	public OperatorDeclaration (Operator.OpType op, FullNamedExpression ret_type, Location location)
	{
		optype = op;
		this.ret_type = ret_type;
		this.location = location;
	}
}

void Error_ExpectingTypeName (Expression expr)
{
	if (expr is Invocation){
		report.Error (1002, expr.Location, "Expecting `;'");
	} else {
		expr.Error_InvalidExpressionStatement (report);
	}
}

void Error_ParameterModifierNotValid (string modifier, Location loc)
{
	report.Error (631, loc, "The parameter modifier `{0}' is not valid in this context",
			              modifier);
}

void Error_DuplicateParameterModifier (Location loc, Parameter.Modifier mod)
{
	report.Error (1107, loc, "Duplicate parameter modifier `{0}'",
  		Parameter.GetModifierSignature (mod));
}

void Error_TypeExpected (Location loc)
{
	report.Error (1031, loc, "Type expected");
}

void Error_UnsafeCodeNotAllowed (Location loc)
{
	report.Error (227, loc, "Unsafe code requires the `unsafe' command line option to be specified");
}

void Warning_EmptyStatement (Location loc)
{
	report.Warning (642, 3, loc, "Possible mistaken empty statement");
}

void Error_NamedArgumentExpected (NamedArgument a)
{
	report.Error (1738, a.Location, "Named arguments must appear after the positional arguments");
}

void Error_MissingInitializer (Location loc)
{
	report.Error (210, loc, "You must provide an initializer in a fixed or using statement declaration");
}

object Error_AwaitAsIdentifier (object token)
{
	if (async_block) {
		report.Error (4003, GetLocation (token), "`await' cannot be used as an identifier within an async method or lambda expression");
		return new LocatedToken ("await", GetLocation (token));
	}

	return token;
}

void push_current_container (TypeDefinition tc, object partial_token)
{
	if (module.Evaluator != null){
		tc.Definition.Modifiers = tc.ModFlags = (tc.ModFlags & ~Modifiers.AccessibilityMask) | Modifiers.PUBLIC;
		if (undo == null)
			undo = new Undo ();

		undo.AddTypeContainer (current_container, tc);
	}
	
	if (partial_token != null)
		current_container.AddPartial (tc);
	else
		current_container.AddTypeContainer (tc);
		
	++lexer.parsing_declaration;
	current_container = tc;
	current_type = tc;
}

TypeContainer pop_current_class ()
{
	var retval = current_container;

	current_container = current_container.Parent;
	current_type = current_type.Parent as TypeDefinition;

	return retval;
}

[System.Diagnostics.Conditional ("FULL_AST")]
void StoreModifierLocation (object token, Location loc)
{
	if (lbag == null)
		return;

	if (mod_locations == null)
  		mod_locations = new List<Tuple<Modifiers, Location>> ();

	mod_locations.Add (Tuple.Create ((Modifiers) token, loc));
}

[System.Diagnostics.Conditional ("FULL_AST")]
void PushLocation (Location loc)
{
	if (location_stack == null)
		location_stack = new Stack<Location> ();

	location_stack.Push (loc);
}

Location PopLocation ()
{
	if (location_stack == null)
		return Location.Null;

	return location_stack.Pop ();
}

string CheckAttributeTarget (int token, string a, Location l)
{
	switch (a) {
	case "assembly" : case "module" : case "field" : case "method" : case "param" : case "property" : case "type" :
			return a;
	}

	if (!Tokenizer.IsValidIdentifier (a)) {
		Error_SyntaxError (token);
	} else {
		report.Warning (658, 1, l,
			 "`{0}' is invalid attribute target. All attributes in this attribute section will be ignored", a);
	}

	return string.Empty;
}

static bool IsUnaryOperator (Operator.OpType op)
{
	switch (op) {
		
	case Operator.OpType.LogicalNot: 
	case Operator.OpType.OnesComplement: 
	case Operator.OpType.Increment:
	case Operator.OpType.Decrement:
	case Operator.OpType.True: 
	case Operator.OpType.False: 
	case Operator.OpType.UnaryPlus: 
	case Operator.OpType.UnaryNegation:
		return true;
	}
	return false;
}

void syntax_error (Location l, string msg)
{
	report.Error (1003, l, "Syntax error, " + msg);
}

Tokenizer lexer;

public Tokenizer Lexer {
	get {
		return lexer;
	}
}		   

public CSharpParser (SeekableStreamReader reader, CompilationSourceFile file, ParserSession session)
	: this (reader, file, file.Compiler.Report, session)
{
}

public CSharpParser (SeekableStreamReader reader, CompilationSourceFile file, Report report, ParserSession session)
{
	this.file = file;
	current_container = current_namespace = file;
	
	this.module = file.Module;
	this.compiler = file.Compiler;
	this.settings = compiler.Settings;
	this.report = report;
	
	lang_version = settings.Version;
	yacc_verbose_flag = settings.VerboseParserFlag;
	doc_support = settings.DocumentationFile != null;
	lexer = new Tokenizer (reader, file, session, report);
	oob_stack = new Stack<object> ();
	lbag = session.LocationsBag;
	use_global_stacks = session.UseJayGlobalArrays;
	parameters_bucket = session.ParametersStack;
}

public void parse ()
{
	eof_token = Token.EOF;
	
	try {
		if (yacc_verbose_flag > 1)
			yyparse (lexer, new yydebug.yyDebugSimple ());
		else
			yyparse (lexer);
			
		Tokenizer tokenizer = lexer as Tokenizer;
		tokenizer.cleanup ();		
	} catch (Exception e){
	  	if (e is yyParser.yyUnexpectedEof) {
			Error_SyntaxError (yyToken);
			UnexpectedEOF = true;
			return;
		}
			
		if (e is yyParser.yyException) {
			if (report.Errors == 0)
				report.Error (-25, lexer.Location, "Parsing error");
		} else {
			// Used by compiler-tester to test internal errors
			if (yacc_verbose_flag > 0 || e is FatalException)
				throw;
		
			report.Error (589, lexer.Location, "Internal compiler error during parsing" + e);
		}
	}
}

void CheckToken (int error, int yyToken, string msg, Location loc)
{
	if (yyToken >= Token.FIRST_KEYWORD && yyToken <= Token.LAST_KEYWORD)
		report.Error (error, loc, "{0}: `{1}' is a keyword", msg, GetTokenName (yyToken));
	else
		report.Error (error, loc, msg);
}

string ConsumeStoredComment ()
{
	string s = tmpComment;
	tmpComment = null;
	Lexer.doc_state = XmlCommentState.Allowed;
	return s;
}

void FeatureIsNotAvailable (Location loc, string feature)
{
	report.FeatureIsNotAvailable (compiler, loc, feature);
}

Location GetLocation (object obj)
{
	var lt = obj as LocatedToken;
	if (lt != null)
		return lt.Location;
		
	var mn = obj as MemberName;
	if (mn != null)
		return mn.Location;
		
	var expr = obj as Expression;
	if (expr != null)
		return expr.Location;

	return lexer.Location;
}

void start_block (Location loc)
{
	if (current_block == null) {
		current_block = new ToplevelBlock (compiler, current_local_parameters, loc);
		parsing_anonymous_method = false;
	} else if (parsing_anonymous_method) {
		current_block = new ParametersBlock (current_block, current_local_parameters, loc);
		parsing_anonymous_method = false;
	} else {
		current_block = new ExplicitBlock (current_block, loc, Location.Null);
	}
}

Block
end_block (Location loc)
{
	Block retval = current_block.Explicit;
	retval.SetEndLocation (loc);
	current_block = retval.Parent;
	return retval;
}

void start_anonymous (bool isLambda, ParametersCompiled parameters, bool isAsync, Location loc)
{
	oob_stack.Push (current_anonymous_method);
	oob_stack.Push (current_local_parameters);
	oob_stack.Push (current_variable);
	oob_stack.Push (async_block);

	current_local_parameters = parameters;
	if (isLambda) {
		if (lang_version <= LanguageVersion.ISO_2)
			FeatureIsNotAvailable (loc, "lambda expressions");

		current_anonymous_method = new LambdaExpression (loc);
	} else {
		if (lang_version == LanguageVersion.ISO_1)
			FeatureIsNotAvailable (loc, "anonymous methods");
			
		current_anonymous_method = new AnonymousMethodExpression (loc);
	}

	async_block = isAsync;
	// Force the next block to be created as a ToplevelBlock
	parsing_anonymous_method = true;
}

/*
 * Completes the anonymous method processing, if lambda_expr is null, this
 * means that we have a Statement instead of an Expression embedded 
 */
AnonymousMethodExpression end_anonymous (ParametersBlock anon_block)
{
	AnonymousMethodExpression retval;

	if (async_block)
		anon_block.IsAsync = true;

	current_anonymous_method.Block = anon_block;
	retval = current_anonymous_method;

	async_block = (bool) oob_stack.Pop ();
	current_variable = (BlockVariable) oob_stack.Pop ();
	current_local_parameters = (ParametersCompiled) oob_stack.Pop ();
	current_anonymous_method = (AnonymousMethodExpression) oob_stack.Pop ();

	return retval;
}

void Error_SyntaxError (int token)
{
	Error_SyntaxError (0, token);
}

void Error_SyntaxError (int error_code, int token)
{
	Error_SyntaxError (error_code, token, "Unexpected symbol");
}

void Error_SyntaxError (int error_code, int token, string msg)
{
	Lexer.CompleteOnEOF = false;

	// An error message has been reported by tokenizer
	if (token == Token.ERROR)
		return;
	
	// Avoid duplicit error message after unterminated string literals
	if (token == Token.LITERAL && lexer.Location.Column == 0)
		return;

	string symbol = GetSymbolName (token);
	string expecting = GetExpecting ();
	var loc = lexer.Location - symbol.Length;
	
	if (error_code == 0) {
		if (expecting == "`identifier'") {
			if (token > Token.FIRST_KEYWORD && token < Token.LAST_KEYWORD) {
				report.Error (1041, loc, "Identifier expected, `{0}' is a keyword", symbol);
				return;
			}
			
			error_code = 1001;
			expecting = "identifier";
		} else if (expecting == "`)'") {
			error_code = 1026;
		} else {
			error_code = 1525;
		}
	}
	
	if (string.IsNullOrEmpty (expecting))
		report.Error (error_code, loc, "{1} `{0}'", symbol, msg);
	else
		report.Error (error_code, loc, "{2} `{0}', expecting {1}", symbol, expecting, msg);	  
}

string GetExpecting ()
{
	int [] tokens = yyExpectingTokens (yyExpectingState);
	var names = new List<string> (tokens.Length);
	bool has_type = false;
	bool has_identifier = false;
	for (int i = 0; i < tokens.Length; i++){
		int token = tokens [i];
		has_identifier |= token == Token.IDENTIFIER;
		
		string name = GetTokenName (token);
		if (name == "<internal>")
			continue;
			
		has_type |= name == "type";
		if (names.Contains (name))
			continue;
		
		names.Add (name);
	}

	//
	// Too many tokens to enumerate
	//
	if (names.Count > 8)
		return null;

	if (has_type && has_identifier)
		names.Remove ("identifier");

	if (names.Count == 1)
		return "`" + GetTokenName (tokens [0]) + "'";
	
	StringBuilder sb = new StringBuilder ();
	names.Sort ();
	int count = names.Count;
	for (int i = 0; i < count; i++){
		bool last = i + 1 == count;
		if (last)
			sb.Append ("or ");
		sb.Append ('`');
		sb.Append (names [i]);
		sb.Append (last ? "'" : count < 3 ? "' " : "', ");
	}
	return sb.ToString ();
}


string GetSymbolName (int token)
{
	switch (token){
	case Token.LITERAL:
		return ((Constant)lexer.Value).GetValue ().ToString ();
	case Token.IDENTIFIER:
		return ((LocatedToken)lexer.Value).Value;

	case Token.BOOL:
		return "bool";
	case Token.BYTE:
		return "byte";
	case Token.CHAR:
		return "char";
	case Token.VOID:
		return "void";
	case Token.DECIMAL:
		return "decimal";
	case Token.DOUBLE:
		return "double";
	case Token.FLOAT:
		return "float";
	case Token.INT:
		return "int";
	case Token.LONG:
		return "long";
	case Token.SBYTE:
		return "sbyte";
	case Token.SHORT:
		return "short";
	case Token.STRING:
		return "string";
	case Token.UINT:
		return "uint";
	case Token.ULONG:
		return "ulong";
	case Token.USHORT:
		return "ushort";
	case Token.OBJECT:
		return "object";
		
	case Token.PLUS:
		return "+";
	case Token.UMINUS:
	case Token.MINUS:
		return "-";
	case Token.BANG:
		return "!";
	case Token.BITWISE_AND:
		return "&";
	case Token.BITWISE_OR:
		return "|";
	case Token.STAR:
		return "*";
	case Token.PERCENT:
		return "%";
	case Token.DIV:
		return "/";
	case Token.CARRET:
		return "^";
	case Token.OP_INC:
		return "++";
	case Token.OP_DEC:
		return "--";
	case Token.OP_SHIFT_LEFT:
		return "<<";
	case Token.OP_SHIFT_RIGHT:
		return ">>";
	case Token.OP_LT:
		return "<";
	case Token.OP_GT:
		return ">";
	case Token.OP_LE:
		return "<=";
	case Token.OP_GE:
		return ">=";
	case Token.OP_EQ:
		return "==";
	case Token.OP_NE:
		return "!=";
	case Token.OP_AND:
		return "&&";
	case Token.OP_OR:
		return "||";
	case Token.OP_PTR:
		return "->";
	case Token.OP_COALESCING:	
		return "??";
	case Token.OP_MULT_ASSIGN:
		return "*=";
	case Token.OP_DIV_ASSIGN:
		return "/=";
	case Token.OP_MOD_ASSIGN:
		return "%=";
	case Token.OP_ADD_ASSIGN:
		return "+=";
	case Token.OP_SUB_ASSIGN:
		return "-=";
	case Token.OP_SHIFT_LEFT_ASSIGN:
		return "<<=";
	case Token.OP_SHIFT_RIGHT_ASSIGN:
		return ">>=";
	case Token.OP_AND_ASSIGN:
		return "&=";
	case Token.OP_XOR_ASSIGN:
		return "^=";
	case Token.OP_OR_ASSIGN:
		return "|=";
	}

	return GetTokenName (token);
}

static string GetTokenName (int token)
{
	switch (token){
	case Token.ABSTRACT:
		return "abstract";
	case Token.AS:
		return "as";
	case Token.ADD:
		return "add";
	case Token.ASYNC:
		return "async";
	case Token.BASE:
		return "base";
	case Token.BREAK:
		return "break";
	case Token.CASE:
		return "case";
	case Token.CATCH:
		return "catch";
	case Token.CHECKED:
		return "checked";
	case Token.CLASS:
		return "class";
	case Token.CONST:
		return "const";
	case Token.CONTINUE:
		return "continue";
	case Token.DEFAULT:
		return "default";
	case Token.DELEGATE:
		return "delegate";
	case Token.DO:
		return "do";
	case Token.ELSE:
		return "else";
	case Token.ENUM:
		return "enum";
	case Token.EVENT:
		return "event";
	case Token.EXPLICIT:
		return "explicit";
	case Token.EXTERN:
	case Token.EXTERN_ALIAS:
		return "extern";
	case Token.FALSE:
		return "false";
	case Token.FINALLY:
		return "finally";
	case Token.FIXED:
		return "fixed";
	case Token.FOR:
		return "for";
	case Token.FOREACH:
		return "foreach";
	case Token.GOTO:
		return "goto";
	case Token.IF:
		return "if";
	case Token.IMPLICIT:
		return "implicit";
	case Token.IN:
		return "in";
	case Token.INTERFACE:
		return "interface";
	case Token.INTERNAL:
		return "internal";
	case Token.IS:
		return "is";
	case Token.LOCK:
		return "lock";
	case Token.NAMESPACE:
		return "namespace";
	case Token.NEW:
		return "new";
	case Token.NULL:
		return "null";
	case Token.OPERATOR:
		return "operator";
	case Token.OUT:
		return "out";
	case Token.OVERRIDE:
		return "override";
	case Token.PARAMS:
		return "params";
	case Token.PRIVATE:
		return "private";
	case Token.PROTECTED:
		return "protected";
	case Token.PUBLIC:
		return "public";
	case Token.READONLY:
		return "readonly";
	case Token.REF:
		return "ref";
	case Token.RETURN:
		return "return";
	case Token.REMOVE:
		return "remove";
	case Token.SEALED:
		return "sealed";
	case Token.SIZEOF:
		return "sizeof";
	case Token.STACKALLOC:
		return "stackalloc";
	case Token.STATIC:
		return "static";
	case Token.STRUCT:
		return "struct";
	case Token.SWITCH:
		return "switch";
	case Token.THIS:
		return "this";
	case Token.THROW:
		return "throw";
	case Token.TRUE:
		return "true";
	case Token.TRY:
		return "try";
	case Token.TYPEOF:
		return "typeof";
	case Token.UNCHECKED:
		return "unchecked";
	case Token.UNSAFE:
		return "unsafe";
	case Token.USING:
		return "using";
	case Token.VIRTUAL:
		return "virtual";
	case Token.VOLATILE:
		return "volatile";
	case Token.WHERE:
		return "where";
	case Token.WHILE:
		return "while";
	case Token.ARGLIST:
		return "__arglist";
	case Token.REFVALUE:
		return "__refvalue";
	case Token.REFTYPE:
		return "__reftype";
	case Token.MAKEREF:
		return "__makeref";
	case Token.PARTIAL:
		return "partial";
	case Token.ARROW:
		return "=>";
	case Token.FROM:
	case Token.FROM_FIRST:
		return "from";
	case Token.JOIN:
		return "join";
	case Token.ON:
		return "on";
	case Token.EQUALS:
		return "equals";
	case Token.SELECT:
		return "select";
	case Token.GROUP:
		return "group";
	case Token.BY:
		return "by";
	case Token.LET:
		return "let";
	case Token.ORDERBY:
		return "orderby";
	case Token.ASCENDING:
		return "ascending";
	case Token.DESCENDING:
		return "descending";
	case Token.INTO:
		return "into";
	case Token.GET:
		return "get";
	case Token.SET:
		return "set";
	case Token.OPEN_BRACE:
		return "{";
	case Token.CLOSE_BRACE:
		return "}";
	case Token.OPEN_BRACKET:
	case Token.OPEN_BRACKET_EXPR:
		return "[";
	case Token.CLOSE_BRACKET:
		return "]";
	case Token.OPEN_PARENS_CAST:
	case Token.OPEN_PARENS_LAMBDA:
	case Token.OPEN_PARENS:
		return "(";
	case Token.CLOSE_PARENS:
		return ")";
	case Token.DOT:
		return ".";
	case Token.COMMA:
		return ",";
	case Token.DEFAULT_COLON:
		return "default:";
	case Token.COLON:
		return ":";
	case Token.SEMICOLON:
		return ";";
	case Token.TILDE:
		return "~";
		
	case Token.PLUS:
	case Token.UMINUS:
	case Token.MINUS:
	case Token.BANG:
	case Token.OP_LT:
	case Token.OP_GT:
	case Token.BITWISE_AND:
	case Token.BITWISE_OR:
	case Token.STAR:
	case Token.PERCENT:
	case Token.DIV:
	case Token.CARRET:
	case Token.OP_INC:
	case Token.OP_DEC:
	case Token.OP_SHIFT_LEFT:
	case Token.OP_SHIFT_RIGHT:
	case Token.OP_LE:
	case Token.OP_GE:
	case Token.OP_EQ:
	case Token.OP_NE:
	case Token.OP_AND:
	case Token.OP_OR:
	case Token.OP_PTR:
	case Token.OP_COALESCING:	
	case Token.OP_MULT_ASSIGN:
	case Token.OP_DIV_ASSIGN:
	case Token.OP_MOD_ASSIGN:
	case Token.OP_ADD_ASSIGN:
	case Token.OP_SUB_ASSIGN:
	case Token.OP_SHIFT_LEFT_ASSIGN:
	case Token.OP_SHIFT_RIGHT_ASSIGN:
	case Token.OP_AND_ASSIGN:
	case Token.OP_XOR_ASSIGN:
	case Token.OP_OR_ASSIGN:
	case Token.INTERR_OPERATOR:
		return "<operator>";

	case Token.BOOL:
	case Token.BYTE:
	case Token.CHAR:
	case Token.VOID:
	case Token.DECIMAL:
	case Token.DOUBLE:
	case Token.FLOAT:
	case Token.INT:
	case Token.LONG:
	case Token.SBYTE:
	case Token.SHORT:
	case Token.STRING:
	case Token.UINT:
	case Token.ULONG:
	case Token.USHORT:
	case Token.OBJECT:
		return "type";
	
	case Token.ASSIGN:
		return "=";
	case Token.OP_GENERICS_LT:
	case Token.GENERIC_DIMENSION:
		return "<";
	case Token.OP_GENERICS_GT:
		return ">";
	case Token.INTERR:
	case Token.INTERR_NULLABLE:
		return "?";
	case Token.DOUBLE_COLON:
		return "::";
	case Token.LITERAL:
		return "value";
	case Token.IDENTIFIER:
	case Token.AWAIT:
		return "identifier";

	case Token.EOF:
		return "end-of-file";

		// All of these are internal.
	case Token.NONE:
	case Token.ERROR:
	case Token.FIRST_KEYWORD:
	case Token.EVAL_COMPILATION_UNIT_PARSER:
	case Token.EVAL_USING_DECLARATIONS_UNIT_PARSER:
	case Token.EVAL_STATEMENT_PARSER:
	case Token.LAST_KEYWORD:
	case Token.GENERATE_COMPLETION:
	case Token.COMPLETE_COMPLETION:
		return "<internal>";

		// A bit more robust.
	default:
		return yyNames [token];
        }
}

/* end end end */
}
#line default
namespace yydebug {
        using System;
	 internal interface yyDebug {
		 void push (int state, Object value);
		 void lex (int state, int token, string name, Object value);
		 void shift (int from, int to, int errorFlag);
		 void pop (int state);
		 void discard (int state, int token, string name, Object value);
		 void reduce (int from, int to, int rule, string text, int len);
		 void shift (int from, int to);
		 void accept (Object value);
		 void error (string message);
		 void reject ();
	 }
	 
	 class yyDebugSimple : yyDebug {
		 void println (string s){
			 Console.Error.WriteLine (s);
		 }
		 
		 public void push (int state, Object value) {
			 println ("push\tstate "+state+"\tvalue "+value);
		 }
		 
		 public void lex (int state, int token, string name, Object value) {
			 println("lex\tstate "+state+"\treading "+name+"\tvalue "+value);
		 }
		 
		 public void shift (int from, int to, int errorFlag) {
			 switch (errorFlag) {
			 default:				// normally
				 println("shift\tfrom state "+from+" to "+to);
				 break;
			 case 0: case 1: case 2:		// in error recovery
				 println("shift\tfrom state "+from+" to "+to
					     +"\t"+errorFlag+" left to recover");
				 break;
			 case 3:				// normally
				 println("shift\tfrom state "+from+" to "+to+"\ton error");
				 break;
			 }
		 }
		 
		 public void pop (int state) {
			 println("pop\tstate "+state+"\ton error");
		 }
		 
		 public void discard (int state, int token, string name, Object value) {
			 println("discard\tstate "+state+"\ttoken "+name+"\tvalue "+value);
		 }
		 
		 public void reduce (int from, int to, int rule, string text, int len) {
			 println("reduce\tstate "+from+"\tuncover "+to
				     +"\trule ("+rule+") "+text);
		 }
		 
		 public void shift (int from, int to) {
			 println("goto\tfrom state "+from+" to "+to);
		 }
		 
		 public void accept (Object value) {
			 println("accept\tvalue "+value);
		 }
		 
		 public void error (string message) {
			 println("error\t"+message);
		 }
		 
		 public void reject () {
			 println("reject");
		 }
		 
	 }
}
// %token constants
 class Token {
  public const int EOF = 257;
  public const int NONE = 258;
  public const int ERROR = 259;
  public const int FIRST_KEYWORD = 260;
  public const int ABSTRACT = 261;
  public const int AS = 262;
  public const int ADD = 263;
  public const int BASE = 264;
  public const int BOOL = 265;
  public const int BREAK = 266;
  public const int BYTE = 267;
  public const int CASE = 268;
  public const int CATCH = 269;
  public const int CHAR = 270;
  public const int CHECKED = 271;
  public const int CLASS = 272;
  public const int CONST = 273;
  public const int CONTINUE = 274;
  public const int DECIMAL = 275;
  public const int DEFAULT = 276;
  public const int DELEGATE = 277;
  public const int DO = 278;
  public const int DOUBLE = 279;
  public const int ELSE = 280;
  public const int ENUM = 281;
  public const int EVENT = 282;
  public const int EXPLICIT = 283;
  public const int EXTERN = 284;
  public const int FALSE = 285;
  public const int FINALLY = 286;
  public const int FIXED = 287;
  public const int FLOAT = 288;
  public const int FOR = 289;
  public const int FOREACH = 290;
  public const int GOTO = 291;
  public const int IF = 292;
  public const int IMPLICIT = 293;
  public const int IN = 294;
  public const int INT = 295;
  public const int INTERFACE = 296;
  public const int INTERNAL = 297;
  public const int IS = 298;
  public const int LOCK = 299;
  public const int LONG = 300;
  public const int NAMESPACE = 301;
  public const int NEW = 302;
  public const int NULL = 303;
  public const int OBJECT = 304;
  public const int OPERATOR = 305;
  public const int OUT = 306;
  public const int OVERRIDE = 307;
  public const int PARAMS = 308;
  public const int PRIVATE = 309;
  public const int PROTECTED = 310;
  public const int PUBLIC = 311;
  public const int READONLY = 312;
  public const int REF = 313;
  public const int RETURN = 314;
  public const int REMOVE = 315;
  public const int SBYTE = 316;
  public const int SEALED = 317;
  public const int SHORT = 318;
  public const int SIZEOF = 319;
  public const int STACKALLOC = 320;
  public const int STATIC = 321;
  public const int STRING = 322;
  public const int STRUCT = 323;
  public const int SWITCH = 324;
  public const int THIS = 325;
  public const int THROW = 326;
  public const int TRUE = 327;
  public const int TRY = 328;
  public const int TYPEOF = 329;
  public const int UINT = 330;
  public const int ULONG = 331;
  public const int UNCHECKED = 332;
  public const int UNSAFE = 333;
  public const int USHORT = 334;
  public const int USING = 335;
  public const int VIRTUAL = 336;
  public const int VOID = 337;
  public const int VOLATILE = 338;
  public const int WHERE = 339;
  public const int WHILE = 340;
  public const int ARGLIST = 341;
  public const int PARTIAL = 342;
  public const int ARROW = 343;
  public const int FROM = 344;
  public const int FROM_FIRST = 345;
  public const int JOIN = 346;
  public const int ON = 347;
  public const int EQUALS = 348;
  public const int SELECT = 349;
  public const int GROUP = 350;
  public const int BY = 351;
  public const int LET = 352;
  public const int ORDERBY = 353;
  public const int ASCENDING = 354;
  public const int DESCENDING = 355;
  public const int INTO = 356;
  public const int INTERR_NULLABLE = 357;
  public const int EXTERN_ALIAS = 358;
  public const int REFVALUE = 359;
  public const int REFTYPE = 360;
  public const int MAKEREF = 361;
  public const int ASYNC = 362;
  public const int AWAIT = 363;
  public const int INTERR_OPERATOR = 364;
  public const int GET = 365;
  public const int SET = 366;
  public const int LAST_KEYWORD = 367;
  public const int OPEN_BRACE = 368;
  public const int CLOSE_BRACE = 369;
  public const int OPEN_BRACKET = 370;
  public const int CLOSE_BRACKET = 371;
  public const int OPEN_PARENS = 372;
  public const int CLOSE_PARENS = 373;
  public const int DOT = 374;
  public const int COMMA = 375;
  public const int COLON = 376;
  public const int SEMICOLON = 377;
  public const int TILDE = 378;
  public const int PLUS = 379;
  public const int MINUS = 380;
  public const int BANG = 381;
  public const int ASSIGN = 382;
  public const int OP_LT = 383;
  public const int OP_GT = 384;
  public const int BITWISE_AND = 385;
  public const int BITWISE_OR = 386;
  public const int STAR = 387;
  public const int PERCENT = 388;
  public const int DIV = 389;
  public const int CARRET = 390;
  public const int INTERR = 391;
  public const int DOUBLE_COLON = 392;
  public const int OP_INC = 393;
  public const int OP_DEC = 394;
  public const int OP_SHIFT_LEFT = 395;
  public const int OP_SHIFT_RIGHT = 396;
  public const int OP_LE = 397;
  public const int OP_GE = 398;
  public const int OP_EQ = 399;
  public const int OP_NE = 400;
  public const int OP_AND = 401;
  public const int OP_OR = 402;
  public const int OP_MULT_ASSIGN = 403;
  public const int OP_DIV_ASSIGN = 404;
  public const int OP_MOD_ASSIGN = 405;
  public const int OP_ADD_ASSIGN = 406;
  public const int OP_SUB_ASSIGN = 407;
  public const int OP_SHIFT_LEFT_ASSIGN = 408;
  public const int OP_SHIFT_RIGHT_ASSIGN = 409;
  public const int OP_AND_ASSIGN = 410;
  public const int OP_XOR_ASSIGN = 411;
  public const int OP_OR_ASSIGN = 412;
  public const int OP_PTR = 413;
  public const int OP_COALESCING = 414;
  public const int OP_GENERICS_LT = 415;
  public const int OP_GENERICS_LT_DECL = 416;
  public const int OP_GENERICS_GT = 417;
  public const int LITERAL = 418;
  public const int IDENTIFIER = 419;
  public const int OPEN_PARENS_LAMBDA = 420;
  public const int OPEN_PARENS_CAST = 421;
  public const int GENERIC_DIMENSION = 422;
  public const int DEFAULT_COLON = 423;
  public const int OPEN_BRACKET_EXPR = 424;
  public const int EVAL_STATEMENT_PARSER = 425;
  public const int EVAL_COMPILATION_UNIT_PARSER = 426;
  public const int EVAL_USING_DECLARATIONS_UNIT_PARSER = 427;
  public const int DOC_SEE = 428;
  public const int GENERATE_COMPLETION = 429;
  public const int COMPLETE_COMPLETION = 430;
  public const int UMINUS = 431;
  public const int yyErrorCode = 256;
 }
 namespace yyParser {
  using System;
  /** thrown for irrecoverable syntax errors and stack overflow.
    */
  internal class yyException : System.Exception {
    public yyException (string message) : base (message) {
    }
  }
  internal class yyUnexpectedEof : yyException {
    public yyUnexpectedEof (string message) : base (message) {
    }
    public yyUnexpectedEof () : base ("") {
    }
  }

  /** must be implemented by a scanner object to supply input to the parser.
    */
  internal interface yyInput {
    /** move on to next token.
        @return false if positioned beyond tokens.
        @throws IOException on input error.
      */
    bool advance (); // throws java.io.IOException;
    /** classifies current token.
        Should not be called if advance() returned false.
        @return current %token or single character.
      */
    int token ();
    /** associated with current token.
        Should not be called if advance() returned false.
        @return value for token().
      */
    Object value ();
  }
 }
} // close outermost namespace, that MUST HAVE BEEN opened in the prolog
